(* This is file S_Effects.pas, which is included in S_dat.pas *)

(* Defines Effects for Siena from version 3 *)

(* In the effect names, the symbol # is replaced by the parameter
   during program execution. *)

(* Functions for observations = 1 : *)

function DataSpecification.ChooseOut(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
 result := rchoose(dg.Outdegree(i)-t,par-1);
end;

function DataSpecification.ChooseIn(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
 result := rchoose(dg.Indegree(j)-t,par-1);
end;

function DataSpecification.ChooseBoth(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar:  real): Real;
begin
 result := (rchoose(dg.Outdegree(i)-t,par-1) + rchoose(dg.Indegree(j)-t,par-1))
                    /factorial[par];                  // changed 3.17t (16 apr 09)
end;


function DataSpecification.InvDEmpty(const dg: TValuedDigraph; const i,k,par: integer): Real;
begin
  result := 1.0/par;
end;


function DataSpecification.InvDEmpty2(const dg: TValuedDigraph; const i,k,par: integer): Real;
begin
  result := 1.0/(par*(par+1));
end;

function DataSpecification.InvDegrees(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var di,dj: real;
begin
  di := dg.Outdegree(i)-t+par;
  dj := dg.Outdegree(j)-t+par;
  result := -1.0/(di*(di+1.0)) - 1.0/(dj*(dj+1.0));
end;


function DataSpecification.InvDegree(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var di: real;
begin
  di := dg.Outdegree(i)-t+par;
  result := - 1.0/(di*(di+1.0));
end;


function DataSpecification.InvDegrees2(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var di,dj: real;
begin
  di := dg.Outdegree(i)-t+par;
  dj := dg.Outdegree(j)-t+par;
  result := -2.0/(di*(di+1.0)*(di+2.0)) - 2.0/(dj*(dj+1.0)*(dj+2.0));
end;

function DataSpecification.InvDegree2(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var di: real;
begin
  di := dg.Outdegree(i)-t+par;
  result := - 2.0/(di*(di+1.0)*(di+2.0));
end;

function DataSpecification.DirectIndirect(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it1, it2: TDoublyLinkedDoubleInt;
  h,tp: integer;
  nd2: real;
begin
  if dg.NoTwoPath(i,j,high(integer)) then nd2 := 0.0 else nd2 := 1.0;
  it1 := dg.OutTies[i].First;
  it2 := dg.OutTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         h := it1.Value1;
         if  dg.NoOrOneTwoPath(i,h,tp) then
           if  ((tp <= 0) or (t >= 1))
              then nd2 := nd2 + 1.0;
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)

  it1 := dg.InTies[i].First;
  it2 := dg.InTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         h := it1.Value1;
         if  dg.NoOrOneTwoPath(h,j,tp) then
           if  ((tp <= 0) or (t >= 1))
              then nd2 := nd2 + 1.0;
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)
  result := nd2;
end;

function DataSpecification.Indirect(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it: TDoublyLinkedDoubleInt;
  h,tp: integer;
  nd2: real;
begin
  nd2 := 0.0;
  it := dg.OutTies[j].First;
  while it.hasnext do
    begin
    it := it.next;
    h := it.Value1;
    if dg.NoOrOneTwoPath(i,h,tp) then
      begin
      if (h <> i) and ((tp <= 0) or (t >= 1)) then nd2 := nd2 + 1.0;
      end;
    end;

  it := dg.InTies[i].First;
  while it.hasnext do
    begin
    it := it.next;
    h := it.Value1;
    if dg.NoOrOneTwoPath(h,j,tp) then
      begin
      if (h <> j) and ((tp <= 0) or (t >= 1)) then  nd2 := nd2 + 1.0;
      end;
    end;
  result := nd2;
end;

function DataSpecification.AltStars(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
  result := arraykoutstars[dg.OutDegree(i)-t] +  arraykoutstars[dg.OutDegree(j)-t];
end;

function DataSpecification.AltOutStars(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
  result := arraykoutstars[dg.OutDegree(i)-t];
end;

function DataSpecification.AltInStars(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
  result := arraykinstars[dg.InDegree(j)-t];
end;

function DataSpecification.AltTriangles(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it1, it2: TDoublyLinkedDoubleInt;
  h: integer;
  rd2: real;
begin
  if (par = 1) then result := DirectIndirect(dg,i,j,t,par,rpar)
   else
  begin
  rd2 := par*(1.0 - arrayktri[dg.TwoPaths(i,j)]);
  it1 := dg.OutTies[i].First;
  it2 := dg.OutTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         h := it1.Value1;
         rd2 := rd2 + arrayktri[dg.TwoPaths(i,h)-t];
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)

  it1 := dg.InTies[i].First;
  it2 := dg.InTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         h := it1.Value1;
         rd2 := rd2 + arrayktri[dg.TwoPaths(h,j)-t];
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)
  result := rd2;
  end;
end; (* AltTriangles *)

function DataSpecification.TwoTriangles(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it1, it2: TDoublyLinkedDoubleInt;
  h: integer;
  rd2: real;
begin
  rd2 := dg.TwoPaths(i,j);
  rd2 := 0.5*rd2*(rd2-1);
  it1 := dg.OutTies[i].First;
  it2 := dg.OutTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         h := it1.Value1;
         rd2 := rd2 + (dg.TwoPaths(i,h)-t);
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)

  it1 := dg.InTies[i].First;
  it2 := dg.InTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         h := it1.Value1;
         rd2 := rd2 + (dg.TwoPaths(h,j)-t);
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)
  result := rd2;
end; (* TwoTriangles *)

function DataSpecification.ThreeTriangles(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it1, it2: TDoublyLinkedDoubleInt;
  tp,rd2: real;
begin
  rd2 := dg.TwoPaths(i,j);
  rd2 := rd2*(rd2-1)*(rd2-2)/6.0;
  it1 := dg.OutTies[i].First;
  it2 := dg.OutTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         tp := dg.TwoPaths(i,it1.value1)-t;
         rd2 := rd2 + tp*(tp-1)/2.0;
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)

  it1 := dg.InTies[i].First;
  it2 := dg.InTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         tp := dg.TwoPaths(it1.value1,j)-t;
         rd2 := rd2 + tp*(tp-1)/2.0;
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)
  result := rd2;
end; (* ThreeTriangles *)

function DataSpecification.ThreePaths(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it: TDoublyLinkedDoubleInt;
  h: integer;
  rd: real;
begin
  rd := 0.0;
  it := dg.OutTies[j].First;
  while it.hasnext do
    begin
    it := it.next;
    h := it.value1;
    if (h <> i) then
      begin
      rd := rd + dg.TwoPaths(h,i);
      if (dg.TieValue(j,i)>= 1) and (dg.TieValue(h,j) >= 1) then rd := rd-1.0;
      end;
    end;
  result := rd;
end; (* ThreePaths *)

function DataSpecification.ThreeTwoPaths(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it: TDoublyLinkedDoubleInt;
  h: integer;
  rd2,tp: real;
begin
  rd2 := 0.0;
  it := dg.OutTies[j].First;
  while it.hasnext do
    begin
    it := it.next;
    h := it.Value1;
    if (h <> i) then
      begin
      tp := dg.TwoPaths(i,h)-t;
      rd2 := rd2 + 0.5*tp*(tp-1.0);
      end;
    end;

  it := dg.InTies[i].First;
  while it.hasnext do
    begin
    it := it.next;
    h := it.Value1;
    if (h <> j) then
      begin
      tp := dg.TwoPaths(h,j)-t;
      rd2 := rd2 + 0.5*tp*(tp-1.0);
      end;
    end;
  result := rd2;
end; (* ThreeTwoPaths *)

function DataSpecification.AltParTwoPaths(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it: TDoublyLinkedDoubleInt;
  h: integer;
  rd2: real;
begin
  if (par = 1) then result := Indirect(dg,i,j,t,par,rpar)
   else
  begin
  rd2 := 0.0;
  it := dg.OutTies[j].First;
  while it.hasnext do
    begin
    it := it.next;
    h := it.Value1;
    if (h <> i) then rd2 := rd2 + arrayktwop[dg.TwoPaths(i,h)-t];
    end;

  it := dg.InTies[i].First;
  while it.hasnext do
    begin
    it := it.next;
    h := it.Value1;
    if (h <> j) then rd2 := rd2 + arrayktwop[dg.TwoPaths(h,j)-t];
    end;
  result := rd2;
  end;
end;   (* AltParTwoPaths *)

function DataSpecification.CoatHanger(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it1, it2: TDoublyLinkedDoubleInt;
  h,k,degs: integer;
  rd2: real;
begin
 with dg do
 begin
 if symmetric then
  begin
  degs := Outdegree(i) + Outdegree(j) -4 -2*t;
  rd2 := 0.0;
  it1 := OutTies[i].First;
  while it1.hasnext do
    begin
    it1 := it1.next;
    it2 := OutTies[i].First;
    h := it1.Value1;
    if (h <> j) then while it2.hasnext do
      begin
      it2 := it2.next;
      k := it2.Value1;
      if (k <> j) and (h < k) and (TieValue(h,k) >= 1) then rd2 := rd2 + 1.0;
      end;
    end;
  it1 := OutTies[j].First;
  while it1.hasnext do
    begin
    it1 := it1.next;
    it2 := OutTies[j].First;
    h := it1.Value1;
    if (h <> i) then while it2.hasnext do
      begin
      it2 := it2.next;
      k := it2.Value1;
      if ( k <> i) and (h < k) and (TieValue(h,k) >= 1) then rd2 := rd2 + 1.0;
      end;
    end;
  it1 := InTies[i].First;
  it2 := InTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         rd2 := rd2 + (Outdegree(it1.value1) + degs);
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)
  end;
  end;

{  end (* if symmetric *)
 else
  begin
  rd2 := 0.0;
  it1 := OutTies[i].First;
  while it1.hasnext do
    begin
    it1 := it1.next;
    h := it1.Value1;
    if (h <> j) then rd2 := rd2 + TwoPaths(i,h);
    end;
  rd2 := rd2 + (OutDegree(i)-1-2*t)*InTwoStars(i,j);

  it1 := InTies[i].First;
  it2 := InTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
          it1 := it1.next;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         rd2 := rd2 + (Outdegree(it1.value1) - 2);
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
        end;
   end; (* case *)
  rd2 := rd2 + (outdegree(i)-1-t)*TwoPaths(i,j);
  end;
  end; (* with *)
}
  result := rd2;
end;

function DataSpecification.ThreeLine(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
var it1: TDoublyLinkedDoubleInt;
  h,rt: integer;
  rd: real;
begin
 with dg do
  begin
  rd := 0.0;
  it1 := OutTies[j].First;
  while it1.hasnext do
    begin
    it1 := it1.next;
    h := it1.Value1;
    if (h <> i) then rd := rd + Outdegree(h);
    end;

  if symmetric then rt := t else
    begin
    if (TieValue(j,i) >= 1) then rt := 1 else rt := 0;
    end;
  rd := rd - TwoPaths(j,i) - ReciDegree(j) + 2*t*rt
          + (Indegree(i) - rt)*(Outdegree(j) - rt) - TwoPaths(j,i);

  it1 := InTies[i].First;
  while it1.hasnext do
    begin
    it1 := it1.next;
    h := it1.Value1;
    if (h <> j) then rd := rd + Indegree(h);
    end;
  rd := rd - Recidegree(i) - TwoPaths(j,i);
  result := rd;
  end;
end; (* ThreeLine *)

function DataSpecification.Covari(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
  if symmetric then
    result := z[i,par] + z[j,par]
   else result := z[i,par];
end;

function DataSpecification.Covarj(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
  result := z[j,par];
end;

function DataSpecification.Simcovar(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
   result := similarity(z[i,par],z[j,par],par);
end;

function DataSpecification.Samecovar(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
   if (abs(z[i,par] - z[j,par]) < epsi) then result := 1.0 else result := 0.0;
end;

function DataSpecification.DyCovar(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
  result := zz[i,j,par] - zzmean[par];
end;


function DataSpecification.CliqueChange(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
  result := ExtraClique[par];
end;

function DataSpecification.Red1_CliqueChange(const dg: TValuedDigraph;
                              const i,j: integer; const t: shortint;
                              const par: integer; const rpar: real): Real;
begin
  result := ExtraRed1_Clique[par];
end;

(* Functions for observations >= 2 : *)

function DataSpecification.CorNDist2(const dg: TValuedDigraph;
                                     const i,k,par: integer): real;
(* This must be redefined here, it seems. *)
begin
 result := CorrectionNDist2(dg,i,k);
end;


function DataSpecification.CorBetween(const dg: TValuedDigraph;
                                     const i,k,par: integer): real;
(* This must be redefined here, it seems. *)
begin
 result := CorrectionBetween(dg,i,k);
end;

function DataSpecification.BalanceSimilarity(const dg: TValuedDigraph;
                                           const i,j,k,par: integer): real;
(* Refers to function defined in Eight because it uses missing data indicators.*)
begin
  Result := fbalsim(dg,i,j,k,balmean);
end;

function DataSpecification.Dist2Contr(const dg: TValuedDigraph;
                                      const i,j,k,par: integer): real;
(* Contribution of tie (i,j) to number of distances 2:
   sum_(i,j) yn(i,j) Dist2contr(i,j) is the number of directed distances 2.
   There are many functions Dist2contr(i,j) with this property.
   The function used here was derived as follows:
   Dist2contr(i,j) is the change in number of distances 2
   from node i to any other node,
   when the arc i => j is added,
   given that all arcs i => h for h > j are considered absent.
   In other words, the arcs in row i are added one by one,
   and for each j, Dist2contr(i,j) is the change in distances 2
   emanating from node i. *)
var it1,it2: TDoublyLinkedDoubleInt;
 h: integer;
begin
  Result := 0;
  if (not dg.NoTwoPath(i,j,j)) then Result := -1;
  it1 := dg.outties[i].first;
  it2 := dg.outties[j].first;
  h := 0;
  while (it1.HasNext and it2.HasNext and (h < j)) do
    begin
    case sign(it1.Next.Value1 - it2.Next.Value1) of
     -1: it1 := it1.Next;
      0:  begin
          it1 := it1.Next;
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h and from i to h *)
          if (j <= h) and dg.NoTwoPath(i,h,j) then Result := Result+1;
          end;
      1:  begin
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h but not from i to h *)
          if (i <> h) and dg.NoTwoPath(i,h,j) then Result := Result+1;
          end;
      end; (* case *)
    end;
  while it2.HasNext do
    begin
    it2 := it2.Next;
    h := it2.Value1; (* there is a tie from j to h but not from i to h *)
    if (i <> h) and dg.NoTwoPath(i,h,j) then Result := Result+1;
    end;
end; (* Dist2Contr *)

function DataSpecification.HalfDist2Contr(const dg: TValuedDigraph;
                                      const i,j,k,par: integer): real;
begin
  HalfDist2Contr := 0.5*Dist2Contr(dg,i,j,k,par);
end;

function DataSpecification.HalfCorNDist2(const dg: TValuedDigraph;
                                     const i,k,par: integer): real;
begin
 result := 0.5*CorNDist2(dg,i,k,par);
end;

function DataSpecification.Dist2ActorContr(const dg: TValuedDigraph;
                          const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* Distance 2 contribution to actor utility function. *)
var it1, it2: TDoublyLinkedDoubleInt;
    h: integer;
begin
  if dg.NoTwoPath(i,j,high(integer)) then Result := 0 else Result := -1;
  it1 := dg.OutTies[j].First;
  it2 := dg.OutTies[i].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
        it1 := it1.next;
        h := it1.value1;    (* there is a tie from j to h,
                               and no tie from i to h *)
        if (i <> h) and dg.NoTwoPath(i,h,high(integer)) then Result := Result+1;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
       end; (* case 0 *)
    1: it2 := it2.next;
   end; (* case *)
  while it1.HasNext do
    begin
    it1 := it1.next;
    h := it1.value1;    (* there is a tie from j to h *)
    if (i <> h) and dg.NoTwoPath(i,h,high(integer)) then Result := Result+1;
    end; (* while *)
end; (* Dist2ActorContr *)

function DataSpecification.TwoDist2ActorContr(const dg: TValuedDigraph;
                          const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* Two-Distance 2 contribution to actor utility function. *)
var it1, it2: TDoublyLinkedDoubleInt;
    h: integer;
begin
  if (dg.TwoPathsTrunc2(i,j) <= 1)
     then Result := 0 else Result := -1;
  it1 := dg.OutTies[j].First;
  it2 := dg.OutTies[i].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
        it1 := it1.next;
        h := it1.value1;    (* there is a tie from j to h,
                               and no tie from i to h *)
        if (i <> h) and (dg.TwoPathsTrunc2(i,h) <= 1) then Result := Result+1;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
       end; (* case 0 *)
    1: it2 := it2.next;
   end; (* case *)
  while it1.HasNext do
    begin
    it1 := it1.next;
    h := it1.value1;    (* there is a tie from j to h *)
    if (i <> h) and (dg.TwoPathsTrunc2(i,h) <= 1) then Result := Result+1;
    end; (* while *)
end; (* TwoDist2ActorContr *)

function DataSpecification.TwoDist2Contr(const dg: TValuedDigraph;
                  const i,j,k,par: integer): Real;
(* Contribution of tie (i,j) to number of TWO-distances 2:
   A two-distance 2 is (i,j) where there are at least two h with i >= h >= j.
   sum_(i,j) yn(i,j) TwoDist2contr(i,j) is the number of directed two-distances 2.
   There are many functions with this property.
   The function used here was derived as follows:
   TwoDist2contr(i,j) is the change in number of two-distances 2
   from node i to any other node,
   when the arc i => j is added,
   given that all arcs i => h for h > j are considered absent.
   In other words, the arcs in row i are added one by one,
   and for each j, TwoDist2contr(i,j) is the change in two-distances 2
   emanating from node i. *)
var it1,it2: TDoublyLinkedDoubleInt;
 h: integer;
begin
  if dg.NoOrOneTwoPathMax(i,j,j) then Result := 0 else result := -1;
  it1 := dg.outties[i].first;
  it2 := dg.outties[j].first;
  h := 0;
  while (it1.HasNext and it2.HasNext) and (h < j) do
    begin
    case sign(it1.Next.Value1 - it2.Next.Value1) of
     -1:  it1 := it1.Next;
      0:  begin
          it1 := it1.Next;
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h and from i to h *)
          if (j < h) and dg.OneTwoPath(i,h,j) then Result := Result+1;
          end;
      1:  begin
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h but not from i to h *)
          if (i <> h) and dg.OneTwoPath(i,h,j) then Result := Result+1;
          end;
      end; (* case *)
    end;
  while it2.HasNext do
    begin
    it2 := it2.Next;
    h := it2.Value1; (* there is a tie from j to h but not from i to h *)
    if (i <> h) and dg.OneTwoPath(i,h,j) then Result := Result+1;
    end;
end; (* TwoDist2Contr *)

function DataSpecification.HalfTwoDist2Contr(const dg: TValuedDigraph;
                                             const i,j,k,par: integer): Real;
begin
  result := 0.5*TwoDist2Contr(dg,i,j,k,par);
end;

function DataSpecification.TwoCorNDist2(const dg: TValuedDigraph; const i,k,par: integer): real;
(* This must be redefined here, it seems. *)
begin
 result := CorrectionNTwoDist2(dg,i,k);
end;

function DataSpecification.HalfTwoCorNDist2(const dg: TValuedDigraph;
                                     const i,k,par: integer): real;
begin
  result := 0.5*TwoCorNDist2(dg,i,k,par);
end;


function DataSpecification.DifSqrtInt
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* square root transformation of indegree alter: *)
var ind: integer;
begin
  ind := dg.InDegree(j);
// corrected 3.17v
  if (tv >= 1) then
    Result := sqrt(ind)
   else
    Result := sqrt(ind+1);
{was from 3.15 to 3.17t:
  if (tv >= 1) then
    Result := (sqrt(ind)*(ind)) - ((ind-1)*sqrt(ind-1))
   else
    Result := (sqrt(ind+1)*(ind+1)) - (ind*sqrt(ind));
was until 3.14:  if (tv >= 1) then
    Result := sqrt(ind)
   else Result := sqrt(ind);
}
{was t/m 3bet14
  if (tv >= 1) then Result := rootn[dg.InDegree(j)+1]/rootn[n]
   else Result := rootn[dg.InDegree(j)]/rootn[n];
}
{pure tie-oriented: }
{  ind  := dg.InDegree(j);
  if (tv >= 1) then
    Result := ((ind)*rootn[ind] - (ind-1)*rootn[ind-1])
   else Result := ((ind+1)*rootn[ind+1] - ind*rootn[ind]);
}
end;

function DataSpecification.DifSqrtIn
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
(* square root transformation of indegree alter: *)
begin
  Result := sqrt(dg.InDegree(j));
end;


function DataSpecification.SqrtInTot
           (const dg: TValuedDigraph; const i,k,par: integer): real;
begin
  result := 0;
  result := dg.InDegree(i)*root(dg.InDegree(i));
end;

function DataSpecification.DifSqrtOutt
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* square root transformation of outdegree alter: *)
begin
  Result := root(dg.OutDegree(j));
end;

function DataSpecification.DifSqrtOut
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
(* square root transformation of outdegree alter: *)
begin
  Result := root(dg.OutDegree(j));
end;

function DataSpecification.SqrtOutTot
           (const dg: TValuedDigraph; const i,k,par: integer): real;
begin
  result := dg.InDegree(i)*root(dg.OutDegree(i));
end;

function DataSpecification.DifIndirectt
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
var it1, it2: TDoublyLinkedDoubleInt;
begin
  if dg.NoTwoPath(i,j,high(integer)) then Result := 0 else Result := 1;
  it1 := dg.OutTies[i].First;
  it2 := dg.OutTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: it1 := it1.next;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
         (* there are ties from i and from j to it1.value1 *)
         if (dg.TwoPathsTrunc2(i,it1.value1) <= tv) then Result := Result+1;
         (* because then the tie i -> j makes the difference between
            no or at least one twopath from i to it1.value1. *)
       end; (* case 0 *)
    1: it2 := it2.next;
   end; (* case *)
end;

function DataSpecification.DifIndirect
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
  if dg.NoTwoPath(i,j,high(integer)) then Result := 0.0 else Result := 1.0;
end;


function DataSpecification.Interactiont_var_trans
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* for evaluation function *)
var p1,p3,resu,varval,jj,kk,code: integer;
    it1, it2, it3, it4: TDoublyLinkedDoubleInt;
    dya, depvar, s1, s2, s3, s4, found: boolean;
function sameval(it: TDoublyLinkedDoubleInt): boolean;
begin
  if misz(it1.value1,p3) then
    sameval := false
   else
    begin
    if depvar then
      sameval := (abs(varval - ya(it.value1,p3)) <= 0)
     else
      sameval := (abs(varval - round(z[it.value1,p3])) <= 0);
    end;
end;
begin (* Interactiont_var_trans *)
(*
See Siena manual, "additional interaction effects".
The following interactions are defined in Interactiont_rec_trans in Digraph:
1002003, 2002003: reciprocity x transitive triplets
The following interactions are defined here:
deze codes oud, weg:
1003def, 2003def, 3003def:
  (similarity on actor variable # def) x transitive triplets
4003def, 5003def, 6003def:
  transitive triplets, 3-cycles, transitive ties
  only within same values of actor variable # def
11003def, 12003def, 13003def:
  (similarity on dyadic variable # def) x transitive triplets
14003def, 15003def, 16003def:
  transitive triplets, 3-cycles, transitive ties
  only within values zz[i,j]=zz[j,h]=zz[i,h]=1 of dyadic variable # def.

1003def, 2003def, 3003def, 4003def:
  (similarity on actor variable # def) x transitive triplets
8003def, 8005def, 8006def:
  transitive triplets, 3-cycles, transitive ties
  only within same values of actor variable # def
11003def, 12003def, 13003def, 14003def:
  (similarity on dyadic variable # def) x transitive triplets
18003def, 18005def, 18006def:
  transitive triplets, 3-cycles, transitive ties
  only within values zz[i,j]=zz[j,h]=zz[i,h]=1 of dyadic variable # def.

Names defined in Digraph.SpecifyInteractionEffect.
*)

  Resu := 0;
  result := 0.0;
  p1 := (par div 1000);
  dya := (p1 >= 10001); (* dya = "variable is dyadic " *)
  if dya then
    begin
    p1 := p1-10000;
    end;
  p3 := (par mod 1000);
  case p1 of
    1003: code := 1;
    2003: code := 2;
    3003: code := 3;
    4003: code := 4;
    8003: code := 5;
    8005: code := 6;
    8006: code := 7;
    else code := 0;
    end; (* case p1 *)
(*1003def, 2003def, 3003def, 4003def:
  (similarity on actor variable # def) x transitive triplets
8003def, 8005def, 8006def:
  transitive triplets, 3-cycles, transitive ties
  only within same values of actor variable # def
11003def, 12003def, 13003def, 14003def:
  (similarity on dyadic variable # def) x transitive triplets
18003def, 18005def, 18006def:     *)

  if (code <= 0) then exit;
  if dya then
    begin
    if miszz(i,j,p3) then exit;
    end
   else
    begin
    if misz(i,p3) or misz(j,p3) then exit;
    end;

  depvar := (not dya) and (p3 > nz) and (p3 <= nz+nza); (* p3 refers to one of the dependent behavior variables *)

  case code of
  1: begin
(*
1003def:
  (similarity on actor variable # def) x transitive triplets
11003def:
  (similarity on dyadic variable # def) x transitive triplets
*)
     if (not dya) and misz(i,p3) then exit;
     if ((dya and (not miszz(i,j,p3))) or (not misz(j,p3))) then
       begin
       it1 := dg.OutTies[i].First;
       it2 := dg.Inties[j].First;
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              Inc(resu);
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       if dya then Result := zz[i,j,p3]*resu
        else
         begin
         if depvar then
           Result := similarity(ya(i,p3),ya(j,p3),p3)*resu
          else Result := similarity(z[i,p3],z[j,p3],p3)*resu;
         end;
       end;
     it1 := dg.OutTies[i].First;
     it2 := dg.OutTies[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if dya then
              begin
              Result := Result + zz[i,it1.Value1,p3];
              end
             else
              begin
              if (not misz(it1.Value1,p3)) then
                begin
                if depvar then
                  Result := Result + similarity(ya(i,p3),ya(it1.Value1,p3),p3)
                 else Result := Result + similarity(z[i,p3],z[it1.Value1,p3],p3);
                end;
              end;
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     end; (* case 1 *)
  2: begin
(*
2003def:
  (similarity on actor variable #2 def) x transitive triplets
12003def:
  (similarity on dyadic variable #2 def) x transitive triplets
*)
       if (not dya) and misz(i,p3) then exit;
     if ((dya and (not miszz(i,j,p3))) or (not misz(j,p3))) then
       begin
       it1 := dg.OutTies[i].First;
       it2 := dg.Outties[j].First;
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              Inc(resu);
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       if dya then Result := zz[i,j,p3]*resu
        else
         begin
         if depvar then
           Result := similarity(ya(i,p3),ya(j,p3),p3)*resu
          else Result := similarity(z[i,p3],z[j,p3],p3)*resu;
         end;
       end;
     it1 := dg.OutTies[i].First;
     it2 := dg.InTies[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if dya then
              begin
              if (not miszz(i,it1.Value1,p3)) then Result := Result + zz[i,it1.Value1,p3];
              end
             else
              begin
              if (not misz(it1.Value1,p3)) then
                begin
                if depvar then
                  Result := Result + similarity(ya(i,p3),ya(it1.Value1,p3),p3)
                 else Result := Result + similarity(z[i,p3],z[it1.Value1,p3],p3);
                end;
              end;
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     end; (* case 2 *)
  3: begin
(*
3003def:
  (similarity on actor variable #3 def) x transitive triplets
13003def:
  (similarity on dyadic variable #3 def) x transitive triplets
*)     if (not dya) and misz(j,p3) then exit;
     it1 := dg.OutTies[i].First;
     it2 := dg.InTies[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if dya then
              begin
              if (not miszz(it1.Value1,j,p3)) then
                  Result := Result + zz[it1.Value1,j,p3];
              end
             else
              begin
              if (not misz(it1.Value1,p3)) then
                begin
                if depvar then
                  Result := Result + similarity(ya(j,p3),ya(it1.Value1,p3),p3)
                 else Result := Result + similarity(z[j,p3],z[it1.Value1,p3],p3);
                end;
              end;
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     it1 := dg.OutTies[i].First;
     it2 := dg.OutTies[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if dya then
              begin
              Result := Result + zz[j,it1.Value1,p3];
              end
             else
              begin
              if (not misz(it1.Value1,p3)) then
                begin
                if depvar then
                  Result := Result + similarity(ya(j,p3),ya(it1.Value1,p3),p3)
                 else Result := Result + similarity(z[j,p3],z[it1.Value1,p3],p3);
                end;
              end;
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     end; (* case 3 *)
  4: begin
(*
4003def:
  (similarity on actor variable #4 def) x transitive triplets
14003def:
  (similarity on dyadic variable #4 def) x transitive triplets
*)     if (not dya) and (misz(j,p3) or misz(i,p3)) then exit;

     if (not (dya and miszz(i,j,p3))) then
     begin
     it1 := dg.OutTies[i].First;
     it2 := dg.OutTies[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if dya then
              begin
              Result := Result + zz[j,it1.Value1,p3];
              end
             else
              begin
              if (not misz(it1.Value1,p3)) then
                begin
                if depvar then
                  Result := Result + similarity(ya(j,p3),ya(it1.Value1,p3),p3)
                 else Result := Result + similarity(z[j,p3],z[it1.Value1,p3],p3);
                end;
              end;
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     if dya then Result := Result * zz[i,j,p3]
      else
       begin
       if depvar then
          Result := Result + similarity(ya(j,p3),ya(i,p3),p3)
         else Result := Result + similarity(z[j,p3],z[i,p3],p3);
       end;
     end; (* if *)

     it1 := dg.OutTies[i].First;
     it2 := dg.InTies[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if dya then
              begin
              if (not (miszz(i,it1.Value1,p3) or miszz(it1.Value1,j,p3))) then
                  Result := Result + zz[i,it1.Value1,p3]*zz[it1.Value1,j,p3];
              end
             else
              begin
              if (not misz(it1.Value1,p3)) then
                begin
                if depvar then
                  Result := Result + similarity(ya(j,p3),ya(it1.Value1,p3),p3)
                                     *similarity(ya(i,p3),ya(it1.Value1,p3),p3)
                 else Result := Result + similarity(z[j,p3],z[it1.Value1,p3],p3)
                                     *similarity(z[i,p3],z[it1.Value1,p3],p3);
                end;
              end;
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     end; (* case 4 *)
  5: begin
(*
8003def:
  transitive triplets
  only within same values of actor variable # def
18003def :
  transitive triplets
  only within values zz[i,j]=zz[j,h]=zz[i,h]=1 of dyadic variable # def.
*)     if dya then
       begin
       if miszz(i,j,p3) or (zz[i,j,p3] <> 1) then exit;
       end
      else
       begin
       if (misz(i,p3) or misz(j,p3)) then exit;
       if depvar then
         begin
         if (abs(ya(i,p3) - ya(j,p3)) >= 1) then exit
          else varval := ya(i,p3);
         end
        else
         begin
         if (abs(z[i,p3] - z[j,p3]) >= 0.0001) then exit
          else varval := round(z[i,p3]);
         end;
       end;

     it1 := dg.OutTies[i].First;
     it2 := dg.Inties[j].First;
     if dya then
       begin
(* twopaths *)
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              if (zz[i,it1.value1,p3] = 1)
                 and (zz[it1.value1,j,p3] = 1) then Inc(resu);
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
(* out-twostars *)
       it1 := dg.OutTies[i].First;
       it2 := dg.OutTies[j].First;
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              if (zz[i,it1.value1,p3] = 1)
                 and (zz[j,it1.value1,p3] = 1) then Inc(resu);
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       result := resu;
       end (* if dya *)
      else
       begin
(* twopaths *)
(* find next it1, it2 with same value on actor variable *)
       s1 := false;
       while ((not s1) and it1.hasnext) do
         begin
         it1 := it1.next;
         s1 := sameval(it1);
         end;
       s2 := false;
       if s1 then while ((not s2) and it2.hasnext) do
         begin
         it2 := it2.next;
         s2 := sameval(it2);
         end;
       while (s1 and s2) do
         case (sign(it1.value1 - it2.value1)) of
          -1: begin
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              end; (* case -1 *)
           0: begin
              inc(resu);
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 0 *)
           1: begin
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 1 *)
         end; (* case (sign(it1.value1 - it2.value1)  *)
       result := resu;
(* out-twostars *)
       it1 := dg.OutTies[i].First;
       it2 := dg.Inties[j].First;
(* find next it1, it2 with same value on actor variable *)
       s1 := false;
       while ((not s1) and it1.hasnext) do
         begin
         it1 := it1.next;
         s1 := sameval(it1);
         end;
       if (not s1) then exit; (* result has been defined above *)
       s2 := false;
       while ((not s2) and it2.hasnext) do
         begin
         it2 := it2.next;
         s2 := sameval(it2);
         end;
       while (s1 and s2) do
         case (sign(it1.value1 - it2.value1)) of
          -1: begin
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              end; (* case -1 *)
           0: begin
              inc(resu);
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 0 *)
           1: begin
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 1 *)
         end; (* case (sign(it1.value1 - it2.value1)  *)
       result := resu;
       end; (* if not dya *)
     end; (* case 5 *)
  6: begin
(*
8005def:
  3-cycles
  only within same values of actor variable # def
18005def:
  3-cycles
  only within values zz[i,j]=zz[j,h]=zz[i,h]=1 of dyadic variable # def.
*)     if dya then
       begin
       if miszz(i,j,p3) or (zz[i,j,p3] <> 1) then exit;
       end
      else
       begin
       if (misz(i,p3) or misz(j,p3)) then exit;
       if depvar then
         begin
         if (abs(ya(i,p3) - ya(j,p3)) >= 1) then exit
          else varval := ya(i,p3);
         end
        else
         begin
         if (abs(z[i,p3] - z[j,p3]) >= 0.0001) then exit
          else varval := round(z[i,p3]);
         end;
       end;

     it1 := dg.OutTies[j].First;
     it2 := dg.Inties[i].First;
     if dya then
       begin
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              if (zz[i,it1.value1,p3] = 1)
                 and (zz[it1.value1,j,p3] = 1) then Inc(resu);
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       result := resu;
       end (* if dya *)
      else
       begin
(* find next it1, it2 with same value on actor variable *)
       s1 := false;
       while ((not s1) and it1.hasnext) do
         begin
         it1 := it1.next;
         s1 := sameval(it1);
         end;
       if (not s1) then exit;
       s2 := false;
       if s1 then while ((not s2) and it2.hasnext) do
         begin
         it2 := it2.next;
         s2 := sameval(it2);
         end;
       while (s1 and s2) do
         case (sign(it1.value1 - it2.value1)) of
          -1: begin
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              end; (* case -1 *)
           0: begin
              inc(resu);
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 0 *)
           1: begin
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 1 *)
         end; (* case (sign(it1.value1 - it2.value1)  *)
       result := resu;
       end; (* if not dya *)
     end; (* case 6 *)
  7: begin
(*
8006def:
  transitive ties
  only within same values of actor variable # def
18006def:
  transitive ties
  only within values zz[i,j]=zz[j,h]=zz[i,h]=1 of dyadic variable # def.
*)

     if dya then
       begin
       if miszz(i,j,p3) or (zz[i,j,p3] <> 1) then exit;
       end
      else
       begin
       if (misz(i,p3) or misz(j,p3)) then exit;
       if depvar then
         begin
         if (abs(ya(i,p3) - ya(j,p3)) >= 1) then exit
          else varval := ya(i,p3);
         end
        else
         begin
         if (abs(z[i,p3] - z[j,p3]) >= 0.0001) then exit
          else varval := round(z[i,p3]);
         end;
       end;

(* First the contribution made by i -> j,
   when there is at least one kk with i -> kk -> j: find such a kk. *)
     it1 := dg.OutTies[i].First;
     it2 := dg.Inties[j].First;
     found := false;
     if dya then
       begin
       while (it1.hasnext and it2.hasnext and (not found)) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              if (zz[i,it1.value1,p3] = 1)
                 and (zz[it1.value1,j,p3] = 1) then found := true;
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       end (* if dya *)
      else
       begin
(* find next it1, it2 with same value on actor variable;
  s1 and s2 indicate that it1 and it2, respectively,
  have the same value on the actor variable *)
       s1 := false;
       while ((not s1) and it1.hasnext) do
         begin
         it1 := it1.next;
         s1 := sameval(it1);
         end;
       s2 := false;
       if s1 then while ((not s2) and it2.hasnext) do
         begin
         it2 := it2.next;
         s2 := sameval(it2);
         end;
       while (s1 and s2 and (not found)) do
         case (sign(it1.value1 - it2.value1)) of
          -1: begin
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              end; (* case -1 *)
           0: begin
              found := true;
              end; (* case 0 *)
           1: begin
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 1 *)
         end; (* case (sign(it1.value1 - it2.value1)  *)
       end; (* if not dya *)
     if found then resu := 1;

(* Second the contribution made by i -> j,
   for those kk with i -> kk, j -> kk for which
   there is no other j' with i -> j' -> kk. *)
     it1 := dg.OutTies[i].First;
     it2 := dg.OutTies[j].First;
     if dya then
       begin
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              if (zz[i,it1.value1,p3] = 1)
                 and (zz[it1.value1,j,p3] = 1) then
                   begin
(* this is kk; is there a j' ? *)
                   kk := it1.value1;
                   found := false;
                   it3 := dg.OutTies[i].First;
                   it4 := dg.InTies[kk].First;
                   while (it3.hasnext and it4.hasnext and (not found)) do
                     case sign(it3.next.value1 - it4.next.value1) of
                      -1: it3 := it3.next;
                       0: begin
                          it3 := it3.next;
                          it4 := it4.next;
                          jj := it3.value1;
                          if (jj <> j) and (zz[i,jj,p3] = 1)
                             and (zz[jj,kk,p3] = 1) then
                               begin
                               found := true;
                               end;
                          end; (* case 0 *)
                       1: it4 := it4.next;
                     end; (* case *)
                   if found then inc(resu);
                   end; (* if *)
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       end (* if dya *)
      else (* not dya *)
       begin
(* find next it1, it2 with same value on actor variable;
  s1 and s2 indicate that it1 and it2, respectively,
  have the same value on the actor variable *)
       s1 := false;
       while ((not s1) and it1.hasnext) do
         begin
         it1 := it1.next;
         s1 := sameval(it1);
         end;
       s2 := false;
       if s1 then while ((not s2) and it2.hasnext) do
         begin
         it2 := it2.next;
         s2 := sameval(it2);
         end;
       while (s1 and s2) do
         case (sign(it1.value1 - it2.value1)) of
          -1: begin
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              end; (* case -1 *)
           0: begin
(* this is kk; is there a j' ? *)
              kk := it1.value1;
              found := false;
              it3 := dg.OutTies[i].First;
              it4 := dg.InTies[kk].First;
              s3 := false;
              while ((not s3) and it3.hasnext) do
                begin
                it3 := it3.next;
                s3 := sameval(it3);
                end;
              s4 := false;
              if s3 then while ((not s4) and it4.hasnext) do
                begin
                it4 := it4.next;
                s4 := sameval(it4);
                end;
              while (s3 and s4 and (not found)) do
                case (sign(it3.value1 - it4.value1)) of
                 -1: begin
                     s3 := false;
                     while ((not s3) and it3.hasnext) do
                       begin
                       it3 := it3.next;
                       s3 := sameval(it3);
                       end;
                     end; (* case -1 *)
                  0: begin
                     if (it3.value1 <> j) then found := true
                      else
                       begin
                       s3 := false;
                       while ((not s3) and it3.hasnext) do
                         begin
                         it3 := it3.next;
                         s3 := sameval(it3);
                         end;
                       s4 := false;
                       while ((not s4) and it4.hasnext) do
                         begin
                         it4 := it4.next;
                         s4 := sameval(it4);
                         end;
                        end;  (* else *)
                     end; (* case 0 *)
                  1: begin
                     s4 := false;
                     while ((not s4) and it4.hasnext) do
                       begin
                       it4 := it4.next;
                       s4 := sameval(it4);
                       end;
                     end; (* case 1 *)
                end; (* case (sign(it3.value1 - it4.value1)  *)
              if (not found) then inc(resu);
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 0 *)
           1: begin
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 1 *)
         end; (* case (sign(it1.value1 - it2.value1)  *)

       end; (* if not dya *)
     result := resu;
     end; (* case 7 *)
  end; (* case *)
end; (* Interactiont_var_trans *)

function DataSpecification.Interaction_var_trans
           (const dg: TValuedDigraph; const i,j,k: integer;
            const par: integer): real;
(* for statistic *)
var p1,p3,code,resu,varval: integer;
    it1, it2: TDoublyLinkedDoubleInt;
    dya, depvar, found, s1, s2: boolean;
function sameval(it: TDoublyLinkedDoubleInt): boolean;
begin
  if misz(it1.value1,p3) then
    sameval := false
   else
    begin
    if depvar then
      sameval := (abs(varval - ya(it.value1,p3)) <= 0)
     else
      sameval := (abs(varval - round(z[it.value1,p3])) <= 0);
    end;
end;
begin
{  Result := Interactiont_var_trans(dg,i,j,k,1,par); moet hetzelfde opleveren }

  Resu := 0;
  Result := 0.0;
  p1 := (par div 1000);
  dya := (p1 >= 10001); (* dya = "variable is dyadic " *)
  if dya then
    begin
    p1 := p1-10000;
    end;
  p3 := (par mod 1000);
  case p1 of
    1003: code := 1;
    2003: code := 2;
    3003: code := 3;
    4003: code := 4;
    8003: code := 5;
    8005: code := 6;
    8006: code := 7;
    else code := 0;
    end; (* case p1 *)
(*1003def, 2003def, 3003def, 4003def:
  (similarity on actor variable # def) x transitive triplets
8003def, 8005def, 8006def:
  transitive triplets, 3-cycles, transitive ties
  only within same values of actor variable # def
11003def, 12003def, 13003def, 14003def:
  (similarity on dyadic variable # def) x transitive triplets
18003def, 18005def, 18006def:     *)

  if (code <= 0) then exit;

  case code of
  1: begin
     if (not dya) and (misz(i,p3) or misz(j,p3))  then exit;
     it1 := dg.OutTies[i].First;
     it2 := dg.Inties[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            Inc(resu);
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     if dya then Result := zz[i,j,p3]*resu
      else Result := similarity(z[i,p3],z[j,p3],p3)*resu;
     end; (* case 1 *)
  2: begin
     if (not dya) and misz(i,p3)  then exit;
     it1 := dg.OutTies[i].First;
     it2 := dg.InTies[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if dya then  Result := Result + zz[i,it1.Value1,p3]
             else
             if not misz(it1.Value1,p3) then
               Result := Result + similarity(z[i,p3],z[it1.Value1,p3],p3);
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     end; (* case 2 *)
  3: begin
     if (not dya) and misz(j,p3)  then exit;
     it1 := dg.OutTies[i].First;
     it2 := dg.InTies[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if dya then  Result := Result + zz[it1.Value1,j,p3]
             else
             if not misz(it1.Value1,p3) then
             Result := Result + similarity(z[j,p3],z[it1.Value1,p3],p3);
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     end; (* case 3 *)
  4: begin
     if (not dya) and (misz(i,p3) or misz(j,p3))  then exit;
     it1 := dg.OutTies[i].First;
     it2 := dg.InTies[j].First;
     while (it1.hasnext and it2.hasnext) do
       case sign(it1.next.value1 - it2.next.value1) of
        -1: it1 := it1.next;
         0: begin
            it1 := it1.next;
            it2 := it2.next;
            if dya then  Result := Result +
              zz[i,it1.Value1,p3]*zz[it1.Value1,j,p3]
             else
             if not misz(it1.Value1,p3) then
             Result := Result +
               similarity(z[i,p3],z[it1.Value1,p3],p3)
               *similarity(z[j,p3],z[it1.Value1,p3],p3);
            end; (* case 0 *)
         1: it2 := it2.next;
       end; (* case *)
     end; (* case 4 *)
  5: begin
(* triplets within variable *)
     if dya then
       begin
       if miszz(i,j,p3) or (zz[i,j,p3] <> 1) then exit;
       end
      else
       begin
       if (misz(i,p3) or misz(j,p3)) then exit;
       if depvar then
         begin
         if (abs(ya(i,p3) - ya(j,p3)) >= 1) then exit
          else varval := ya(i,p3);
         end
        else
         begin
         if (abs(z[i,p3] - z[j,p3]) >= 0.0001) then exit
          else varval := round(z[i,p3]);
         end;
       end;

     it1 := dg.OutTies[i].First;
     it2 := dg.Inties[j].First;
     if dya then
       begin
(* twopaths *)
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              if (zz[i,it1.value1,p3] = 1)
                 and (zz[it1.value1,j,p3] = 1) then Inc(resu);
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       result := resu;
       end (* if dya *)
      else
       begin
(* twopaths *)
(* find next it1, it2 with same value on actor variable *)
       s1 := false;
       while ((not s1) and it1.hasnext) do
         begin
         it1 := it1.next;
         s1 := sameval(it1);
         end;
       s2 := false;
       if s1 then while ((not s2) and it2.hasnext) do
         begin
         it2 := it2.next;
         s2 := sameval(it2);
         end;
       while (s1 and s2) do
         case (sign(it1.value1 - it2.value1)) of
          -1: begin
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              end; (* case -1 *)
           0: begin
              inc(resu);
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 0 *)
           1: begin
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 1 *)
         end; (* case (sign(it1.value1 - it2.value1)  *)
       result := resu;
       end; (* if not dya *)
     end; (* case 5 *)
  6: begin
(* 3-cycles within variable *)
     if dya then
       begin
       if miszz(i,j,p3) or (zz[i,j,p3] <> 1) then exit;
       end
      else
       begin
       if (misz(i,p3) or misz(j,p3)) then exit;
       if depvar then
         begin
         if (abs(ya(i,p3) - ya(j,p3)) >= 1) then exit
          else varval := ya(i,p3);
         end
        else
         begin
         if (abs(z[i,p3] - z[j,p3]) >= 0.0001) then exit
          else varval := round(z[i,p3]);
         end;
       end;

     it1 := dg.OutTies[j].First;
     it2 := dg.Inties[i].First;
     if dya then
       begin
       while (it1.hasnext and it2.hasnext) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              if (zz[i,it1.value1,p3] = 1)
                 and (zz[it1.value1,j,p3] = 1) then Inc(resu);
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       result := resu/3.0;
       end (* if dya *)
      else
       begin
(* find next it1, it2 with same value on actor variable *)
       s1 := false;
       while ((not s1) and it1.hasnext) do
         begin
         it1 := it1.next;
         s1 := sameval(it1);
         end;
       if (not s1) then exit;
       s2 := false;
       if s1 then while ((not s2) and it2.hasnext) do
         begin
         it2 := it2.next;
         s2 := sameval(it2);
         end;
       while (s1 and s2) do
         case (sign(it1.value1 - it2.value1)) of
          -1: begin
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              end; (* case -1 *)
           0: begin
              inc(resu);
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 0 *)
           1: begin
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 1 *)
         end; (* case (sign(it1.value1 - it2.value1)  *)
       result := resu/3.0;
       end; (* if not dya *)
     end; (* case 6 *)
  7: begin
(* transitive ties within variable *)
     if dya then
       begin
       if miszz(i,j,p3) or (zz[i,j,p3] <> 1) then exit;
       end
      else
       begin
       if (misz(i,p3) or misz(j,p3)) then exit;
       if depvar then
         begin
         if (abs(ya(i,p3) - ya(j,p3)) >= 1) then exit
          else varval := ya(i,p3);
         end
        else
         begin
         if (abs(z[i,p3] - z[j,p3]) >= 0.0001) then exit
          else varval := round(z[i,p3]);
         end;
       end;

(* The contribution made by i -> j,
   when there is at least one kk with i -> kk -> j: find such a kk. *)
     it1 := dg.OutTies[i].First;
     it2 := dg.Inties[j].First;
     found := false;
     if dya then
       begin
       while (it1.hasnext and it2.hasnext and (not found)) do
         case sign(it1.next.value1 - it2.next.value1) of
          -1: it1 := it1.next;
           0: begin
              it1 := it1.next;
              it2 := it2.next;
              if (zz[i,it1.value1,p3] = 1)
                 and (zz[it1.value1,j,p3] = 1) then found := true;
              end; (* case 0 *)
           1: it2 := it2.next;
         end; (* case *)
       end (* if dya *)
      else
       begin
(* find next it1, it2 with same value on actor variable;
  s1 and s2 indicate that it1 and it2, respectively,
  have the same value on the actor variable *)
       s1 := false;
       while ((not s1) and it1.hasnext) do
         begin
         it1 := it1.next;
         s1 := sameval(it1);
         end;
       s2 := false;
       if s1 then while ((not s2) and it2.hasnext) do
         begin
         it2 := it2.next;
         s2 := sameval(it2);
         end;
       while (s1 and s2 and (not found)) do
         case (sign(it1.value1 - it2.value1)) of
          -1: begin
              s1 := false;
              while ((not s1) and it1.hasnext) do
                begin
                it1 := it1.next;
                s1 := sameval(it1);
                end;
              end; (* case -1 *)
           0: begin
              found := true;
              end; (* case 0 *)
           1: begin
              s2 := false;
              while ((not s2) and it2.hasnext) do
                begin
                it2 := it2.next;
                s2 := sameval(it2);
                end;
              end; (* case 1 *)
         end; (* case (sign(it1.value1 - it2.value1)  *)
       end; (* if not dya *)
     if found then result := 1.0;
     end; (* case 7 *)
  end; (* case *)
end; (* Interaction_var_trans *)


function DataSpecification.DifMax
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (dg.OutDegree(i) + tv <= par)
    then Result := 1.0 else Result := 0.0;
end;

function DataSpecification.MaxTot
           (const dg: TValuedDigraph; const i,k,par: integer): real;
begin
  if (dg.OutDegree(i) <= par) then Result := dg.OutDegree(i)
      else Result := par;
end;


function DataSpecification.DifRoot
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* square root transformation of outdegree: *)
begin
  if (tv >= 1)
   then Result := root(dg.OutDegree(i))-root(dg.OutDegree(i)-1)
   else Result := root(dg.OutDegree(i)+1)-root(dg.OutDegree(i));
end;

function DataSpecification.RootPlus
           (const dg: TValuedDigraph; const i,k,par: integer): real;
(* square root transformation of outdegree: *)
begin
  result := root(dg.OutDegree(i));
end;

function DataSpecification.DifPow15
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
var od: integer;
begin
  od := dg.outdegree(i);
  if (tv >= 1)
   then Result := od*root(od) - (od-1)*root(od-1)
   else Result := (od+1)*root(od+1) - od*root(od);
end;

function DataSpecification.Pow15Plus
           (const dg: TValuedDigraph; const i,k,par: integer): real;
begin
  result := root(dg.OutDegree(i))*dg.OutDegree(i);
end;


function DataSpecification.DifPow2
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (tv >= 1)
   then Result := 2*dg.outdegree(i) - 1
   else Result := 2*dg.outdegree(i) + 1;
end;

function DataSpecification.Pow2Plus
           (const dg: TValuedDigraph; const i,k,par: integer): real;
begin
  result := sqr(dg.OutDegree(i));
end;

function DataSpecification.DifOutSqrtInt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  result := root(dg.InDegree(i));
end;

function DataSpecification.OutSqrtIn(const dg: TValuedDigraph; const i,k,par: integer): real;
begin
  result := root(dg.InDegree(i))*dg.OutDegree(i);
end;

function DataSpecification.DifOutInt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  result := dg.InDegree(i);
end;

function DataSpecification.OutIn(const dg: TValuedDigraph; const i,k,par: integer): real;
begin
  result := dg.InDegree(i)*dg.OutDegree(i);
end;

function DataSpecification.InSqrtOut(const dg: TValuedDigraph; const i,k,par: integer): real;
begin
  result := root(dg.OutDegree(i))*dg.InDegree(i);
end;



function DataSpecification.Difddt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
var odi: integer;
    ds: real;
    it: TDoublyLinkedDoubleInt;
begin
  case par of
   1: begin
      ds := 0.0;
      it := dg.OutTies[i].First;
      while it.HasNext do begin
        it := it.next;
        ds := ds + dg.outdegree(it.value1);
       end (* while *);
      Result := ((dg.outdegree(j)+1-tv)*(dg.outdegree(i)+1-2*tv)) + ds;
      end;
   2: begin
      ds := 0.0;
      it := dg.OutTies[i].First;
      while it.HasNext do begin
        it := it.next;
        ds := ds + sqrt(dg.outdegree(it.value1));
       end (* while *);
      odi := dg.outdegree(i) + 1 - tv;
      if (tv <= 0) then
        Result := ((ds+sqrt(dg.outdegree(j)+1))*sqrt(odi)) - (ds*sqrt(odi-1))
       else
        Result := (ds*sqrt(odi)) + ((sqrt(dg.outdegree(j)) - ds)*sqrt(odi-1));
      end;
   else Result := 0.0;
   end;  (* case *)
end;

function DataSpecification.Difododt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
var odi: integer;
    ds: real;
    it: TDoublyLinkedDoubleInt;
begin
  case par of
   1: begin
      ds := 0.0;
      it := dg.OutTies[i].First;
      while it.HasNext do begin
        it := it.next;
        ds := ds + dg.outdegree(it.value1);
       end (* while *);
      Result := (dg.outdegree(j)*(dg.outdegree(i)+1-2*tv)) + ds;
      end;
   2: begin
      ds := 0.0;
      it := dg.OutTies[i].First;
      while it.HasNext do begin
        it := it.next;
        ds := ds + sqrt(dg.outdegree(it.value1));
       end (* while *);
      odi := dg.outdegree(i) + 1 - tv;
      if (tv <= 0) then
        Result := ((ds+sqrt(dg.outdegree(j)))*sqrt(odi))- (ds*sqrt(odi-1))
       else
        Result := (ds*sqrt(odi)) + ((sqrt(dg.outdegree(j)) - ds)*sqrt(odi-1));
      end;
   else Result := 0.0;
   end;  (* case *)
end;


function DataSpecification.Prododod(const dg: TValuedDigraph;
              const i,j,k,par: integer): real;
begin
  case par of
    1:  Result := dg.outdegree(i)*dg.outdegree(j);
    2:  Result := sqrt(dg.outdegree(i)*dg.outdegree(j));
   else Result := 0.0;
  end; (* case *)
end;

function DataSpecification.Difodidt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
var odi: integer;
    ds: real;
    it: TDoublyLinkedDoubleInt;
begin
  case par of
   1: begin
      ds := 0.0;
      it := dg.OutTies[i].First;
      while it.HasNext do
        begin
        it := it.next;
        ds := ds + dg.indegree(it.value1);
        end (* while *);
      if (tv <= 0) then
        begin
        Result := (dg.indegree(j)+1)*(dg.outdegree(i)+1) + ds;
        end
       else
        begin
        Result := dg.indegree(j)*(dg.outdegree(i)-1) + ds; // corrected 3.17u
        end;
      end;
   2: begin
      ds := 0.0;
      it := dg.OutTies[i].First;
      while it.HasNext do
        begin
        it := it.next;
        ds := ds + sqrt(dg.indegree(it.value1));
        end (* while *);
      odi := dg.outdegree(i);
      if (tv <= 0) then
        begin
        Result := sqrt((dg.indegree(j)+1)*(odi+1))
                    + ds*(sqrt(odi+1)- sqrt(odi)); // corrected 3.17u
        end
       else
        begin
        Result := sqrt(dg.indegree(j)*(odi-1)) + ds*(sqrt(odi)-sqrt(odi-1));
        end;
      end;
   else Result := 0.0;
   end;  (* case *)
end;

function DataSpecification.Prododid(const dg: TValuedDigraph;
                     const i,j,k,par: integer): real;
begin
  case par of
    1:  Result := dg.outdegree(i)*dg.indegree(j);
    2:  Result := sqrt(dg.outdegree(i)*dg.indegree(j));
   else Result := 0.0;
  end; (* case *)
end;

function DataSpecification.Difidodt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin // corrected 3.17u
  case par of
    1:  Result := dg.indegree(i)*dg.outdegree(j);
    2:  Result := sqrt(dg.indegree(i)*dg.outdegree(j));
   else Result := 0.0;
  end; (* case *)
end;

function DataSpecification.Prodidod(const dg: TValuedDigraph;
                               const i,j,k,par: integer): real;
begin
  case par of
    1:  Result := dg.indegree(i)*dg.outdegree(j);
    2:  Result := sqrt(dg.indegree(i)*dg.outdegree(j));
   else Result := 0.0;
  end; (* case *)
end;

function DataSpecification.Difididt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin // corrected 3.17u
  case par of
    1:  Result := dg.indegree(i)*(dg.indegree(j)+1-tv);
    2:  Result := sqrt(dg.indegree(i)*(dg.indegree(j)+1-tv));
   else Result := 0.0;
  end; (* case *)
end;

function DataSpecification.Prodidid(const dg: TValuedDigraph;
                        const i,j,k,par: integer): real;
begin
  case par of
    1:  Result := dg.indegree(i)*dg.indegree(j);
    2:  Result := sqrt(dg.indegree(i)*dg.indegree(j));
   else Result := 0.0;
  end; (* case *)
end;



function DataSpecification.DifSquare
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* square transformation of outdegree: *)
begin
  Result := 2.0*(dg.OutDegree(i)-par);
end;

function DataSpecification.SquarePlus
           (const dg: TValuedDigraph; const i,k,par: integer): real;
(* square transformation of outdegree: *)
begin
  Result := sqr(dg.OutDegree(i) - par);
end;

function DataSpecification.DifRec
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* reciprocal transformation of outdegree: *)
begin
 Result :=  - 1.0/((dg.OutDegree(i)-tv+1.0+par)*(dg.OutDegree(i)-tv+par));
end;

function DataSpecification.RecTot
              (const dg: TValuedDigraph; const i,k,par: integer): real;
(* reciprocal transformation of outdegree: *)
begin
  Result := 1.0/(dg.OutDegree(i)+par);
end;

function DataSpecification.DifRec2
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
(* squared & reciprocal transformation of outdegree: *)
begin
  Result := - 2.0/((dg.OutDegree(i)-tv+par)*
                   (dg.OutDegree(i)-tv+1.0+par)*(dg.OutDegree(i)-tv+par+2.0));
end;

function DataSpecification.Rec2Tot
           (const dg: TValuedDigraph; const i,k,par: integer): real;
(* squared & reciprocal transformation of outdegree: *)
begin
  Result := 1.0/((dg.OutDegree(i)+par)*(dg.OutDegree(i)+1.0+par));
end;

function DataSpecification.DifDense
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
var n3, n4: Integer;
begin
  if (par <= 4) then begin result := 0; exit; end;
  dg.ThirdCount(i,j,n3,n4);
  if (par >= 6) then begin
   if (dg.TieValue(j,i) >= 1) then Result := n4 else Result := 0;
   end else if (par >= 5) then begin
   if (dg.TieValue(j,i) >= 1) then Result := n3+n4 else Result := n4; end;
end;

function DataSpecification.DenseTot
           (const dg: TValuedDigraph; const i,k,par: integer): real;
var j: integer;
    it: TDoublyLinkedDoubleInt;
    n3, n4 : Integer;
begin
  Result := 0;
  if (par <= 4) then begin
    result := 0.0; exit; end;
  it := dg.OutTies[i].First;
  while it.HasNext do begin
    it := it.next;
    j := it.value1;
    dg.ThirdCount(i,j,n3,n4);
    if (par >= 6) then begin
      if (dg.TieValue(j,i) >= 1) then Result := Result+n4;
      end else if (par >= 5) then begin
      if (dg.TieValue(j,i) >= 1) then Result := Result+n3+n4
        else Result := Result+n4; end;
   end (* while *);
end;

function DataSpecification.DifDyadicCovt
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if miszz(i,j,par) then Result := 0.0 else Result := zz[i,j,par] - zzmean[par];
end;

function DataSpecification.DifDyadicCov
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
  if miszz(i,j,par) then Result := 0.0 else Result := zz[i,j,par] - zzmean[par];
end;


function DataSpecification.DifChangeDyadicCov
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
(* if not missing then *)
  Result := zzc[i,j,par,k] - zzmean[nzz+par];
end;

function DataSpecification.DifChangeDyadicCovt
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
(* if not missing then *)
  Result := zzc[i,j,par,k] - zzmean[nzz+par];
end;

function DataSpecification.DyadicCovTot
           (const dg: TValuedDigraph; const i,k,par: integer): real;
var j: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.OutTies[i].First;
  while it.HasNext do begin
    it := it.next;
    j := it.Value1;
    if not miszz(i,j,par) then  Result := Result + zz[i,j,par] - zzmean[par];
   end; (* while *)
end;

function DataSpecification.DifDyadicCovRect
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (not miszz(i,j,par)) and (symmetric or (dg.TieValue(j,i) >= 1)) then
    Result := zz[i,j,par] - zzmean[par] else Result := 0.0;
end;

function DataSpecification.DifDyadicCovRec
           (const dg: TValuedDigraph; const i,j,k: integer;
            const par: integer): real;
begin
  if ((not miszz(i,j,par)) and (symmetric or (dg.TieValue(j,i) >= 1))) then
    Result := zz[i,j,par] - zzmean[par] else Result := 0.0;
end;

function DataSpecification.DyadicCovRecTot
           (const dg: TValuedDigraph; const i,k,par: integer): real;
var j: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.OutTies[i].First;
  while it.HasNext do begin
    it := it.next;
    j := it.Value1;
    if ((not miszz(i,j,par)) and (dg.TieValue(j,i) >= 1))
      then Result := Result + zz[i,j,par] - zzmean[par];
   end; (* while *)
end;

function DataSpecification.DifChangeDyadicCovRect
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
(* if not missing then *)
  if (symmetric or (dg.TieValue(j,i) >= 1)) then
    Result := zzc[i,j,par,k] - zzmean[nzz+par] else Result := 0.0;
end;

function DataSpecification.DifChangeDyadicCovRec
           (const dg: TValuedDigraph; const i,j,k: integer;
            const par: integer): real;
begin
(* if not missing then *)
  if  (symmetric or (dg.TieValue(j,i) >= 1)) then
    Result := zzc[i,j,par,k] - zzmean[nzz+par] else Result := 0.0;
end;


function DataSpecification.DyadicCovWWX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
var h: integer;
    z: real;
begin
  z := 0.0;
  for h:=n downto 1 do
    if not (miszz(i,h,par) or miszz(h,j,par)) then
      z := z + zz[i,h,par]*zz[h,j,par];
  Result := z;
end;

function DataSpecification.DyadicCovWWXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
var h: integer;
    z: real;
begin
  z := 0.0;
  for h:=n downto 1 do
    if not (miszz(i,h,par) or miszz(h,j,par)) then
      z := z + zz[i,h,par]*zz[h,j,par];
  Result := z;
end;

function DataSpecification.ChangeDyadicCovWWX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
var h: integer;
    z: real;
begin
  z := 0.0;
  for h:=n downto 1 do
//    if not (miszz(i,h,par) or miszz(h,j,par)) then
      z := z + zzc[i,h,par,k]*zzc[h,j,par,k];
  Result := z;
end;

function DataSpecification.ChangeDyadicCovWWXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
var h: integer;
    z: real;
begin
  z := 0.0;
  for h:=n downto 1 do
//    if not (miszz(i,h,par) or miszz(h,j,par)) then
      z := z + zzc[i,h,par,k]*zzc[h,j,par,k];
  Result := z;
end;

function DataSpecification.DyadicCovWXX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
var h: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.InTies[j].First;
  while it.HasNext do begin
    it := it.next;
    h := it.Value1;
    if (h <> i) and (not miszz(i,h,par)) then
      result := result + zz[i,h,par];
   end; (* while *)
end;

function DataSpecification.DyadicCovWXXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
var h: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.InTies[j].First;
  while it.HasNext do begin
    it := it.next;
    h := it.Value1;
    if (h <> i) and (not miszz(i,h,par)) then
      result := result + zz[i,h,par];
  end; (* while *)
end;

function DataSpecification.ChangeDyadicCovWXX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
var h: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.InTies[j].First;
  while it.HasNext do begin
    it := it.next;
    h := it.Value1;
//    if not miszz(i,h,par) then
      if (h <> i) then result := result + zzc[i,h,par,k];
  end; (* while *)
end;

function DataSpecification.ChangeDyadicCovWXXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
var h: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.InTies[j].First;
  while it.HasNext do begin
    it := it.next;
    h := it.Value1;
//    if not miszz(i,h,par) then
      if (h <> i) then result := result + zzc[i,h,par,k];
  end; (* while *)
end;

function DataSpecification.DyadicCovXWX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
var h: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.OutTies[i].First;
  while it.HasNext do begin
    it := it.next;
    h := it.Value1;
    if (h <> j) and (not miszz(h,j,par)) then
      result := result + zz[h,j,par];
   end; (* while *)
end;

function DataSpecification.DyadicCovXWXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
var h: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.OutTies[i].First;
  while it.HasNext do begin
    it := it.next;
    h := it.Value1;
    if (h <> j) then
      begin
      if (not miszz(h,j,par)) then
      result := result + zz[h,j,par];
      if (not miszz(j,h,par)) then
      result := result + zz[j,h,par];
      end;
   end; (* while *)
end;

function DataSpecification.ChangeDyadicCovXWX(const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
var h: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.OutTies[i].First;
  while it.HasNext do begin
    it := it.next;
    h := it.Value1;
    if (h <> j) //and (not miszz(h,j,par))
      then result := result + zzc[h,j,par,k];
   end; (* while *)
end;

function DataSpecification.ChangeDyadicCovXWXt(const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint;  const par: integer): real;
var h: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  it := dg.OutTies[i].First;
  while it.HasNext do begin
    it := it.next;
    h := it.Value1;
    if (h <> j) // and (not missings
      then result := result + zzc[h,j,par,k] + zzc[j,h,par,k];
   end; (* while *)
end;


function DataSpecification.DifCovAltert
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (par > nz) and (par <= nz+nza) then Result := (ya(j,par-nz)-zamean[par-nz])
   else Result := z[j,par];
end;

function DataSpecification.DifCovAlter
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
  if misz(j,par) then Result := 0.0 else
    begin
    if (par > nz) and (par <= nz+nza) then Result := (z[j,par]-zamean[par-nz])
     else Result := z[j,par];
   end;
(* initialise_actors(k) in S_Base.FunctionEvaluations implies
   that z[j,par] = zc[j,thisperiod,par] *)
end;

function DataSpecification.DifSquareCovAltert
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (par > nz) and (par <= nz+nza) then Result := sqr(ya(j,par-nz)-zamean[par-nz])
   else Result := sqr(z[j,par]);
end;

function DataSpecification.DifSquareCovAlter
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
  if misz(j,par) then Result := 0.0 else
    begin
    if (par > nz) and (par <= nz+nza) then Result := sqr(z[j,par]-zamean[par-nz])
     else Result := sqr(z[j,par]);
   end;
(* initialise_actors(k) in S_Base.FunctionEvaluations implies
   that z[j,par] = zc[j,thisperiod,par] *)
end;

function DataSpecification.CovBotht
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (par > nz) and (par <= nz+nza) then
    Result := (ya(j,par-nz)-zamean[par-nz])*(ya(i,par-nz)-zamean[par-nz])
   else Result := z[j,par]*z[i,par];
end;

function DataSpecification.CovBoth
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
  if (misz(j,par) or misz(i,par))
    then Result := 0.0 else
    begin
    if (par > nz) and (par <= nz+nza) then
      Result := (z[j,par]-zamean[par-nz])*(z[i,par]-zamean[par-nz])
     else
      Result := z[j,par]*z[i,par];
    end;
(* initialise_actors(k) in S_Base.FunctionEvaluations implies
   that z[j,par] = zc[j,thisperiod,par] *)
end;

function DataSpecification.CovBothRect
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (symmetric or (dg.TieValue(j,i) >= 1)) then
    begin
    if (par > nz) and (par <= nz+nza) then
    Result := (ya(j,par-nz)-zamean[par-nz])*(ya(i,par-nz)-zamean[par-nz])
     else Result := z[j,par]*z[i,par];
    end
    else Result := 0.0;
(* initialise_actors(k) implies that z[j,par] = zc[j,thisperiod,par] *)
end;

function DataSpecification.CovBothRec
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
  if ((not (misz(i,par) or misz(j,par))) and
      (symmetric or (dg.TieValue(j,i) >= 1))) then
    begin
    if (par > nz) and (par <= nz+nza) then
      Result := (z[j,par]-zamean[par-nz])*(z[i,par]-zamean[par-nz])
     else
      Result := z[j,par]*z[i,par];
    end
   else Result := 0.0;
    (* initialise_actors(k) in S_Base.FunctionEvaluations implies
   that z[j,par] = zc[j,thisperiod,par] *)
end;

// AA start
function DataSpecification.CovIndirectt(const dg: TValuedDigraph; const i,j,k: integer;
                  const tv: shortint; const par: integer): real;
var it1, it2: TDoublyLinkedDoubleInt;
begin
  Result := 0;
  it1 := dg.OutTies[j].First;
  it2 := dg.OutTies[i].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin // h = it1.value1:  a tie j -> h, but not a tie i -> h
        it1 := it1.next;
        if dg.NoTwoPath(i,it1.value1,high(integer)) then Result := Result + z[it1.value1,par];
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
       end; (* case 0 *)
    1: it2 := it2.next;
   end; (* case *)
  while it1.hasnext do
    begin
    it1 := it1.next;
    Result := Result + z[it1.value1,par];
    end;
end; (* CovIndirectt *)


function DataSpecification.CovIndirect(const dg: TValuedDigraph; const i,j,k: integer;
                   const par: integer): real;
(* See Dist2Contr, which was used as the starting point for this procedure. *)
var it1,it2: TDoublyLinkedDoubleInt;
 h: integer;
begin
  Result := 0;
  if (not dg.NoTwoPath(i,j,j)) then Result := -z[j,par];
  it1 := dg.outties[i].first;
  it2 := dg.outties[j].first;
  h := 0;
  while (it1.HasNext and it2.HasNext and (h < j)) do
    begin
    case sign(it1.Next.Value1 - it2.Next.Value1) of
     -1: it1 := it1.Next;
      0:  begin
          it1 := it1.Next;
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h and from i to h *)
          if (j <= h) and dg.NoTwoPath(i,h,j) then Result := Result+z[h,par];
          end;
      1:  begin
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h but not from i to h *)
          if (i <> h) and dg.NoTwoPath(i,h,j) then Result := Result+z[h,par];
          end;
      end; (* case *)
    end;
  while it2.HasNext do
    begin
    it2 := it2.Next;
    h := it2.Value1; (* there is a tie from j to h but not from i to h *)
    if (i <> h) and dg.NoTwoPath(i,h,j) then Result := Result+z[h,par];
    end;
end; (* CovIndirect *)


function DataSpecification.CorCovIndirect(const dg: TValuedDigraph;
                                     const i,k,par: integer): real;
(* This must be redefined here, it seems. *)
begin
 result := CorrectionVarDist2(dg,i,k,par);
end;


// AA end


function DataSpecification.DifCovEgot
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (par > nz) and (par <= nz+nza) then Result := (ya(i,par-nz)-zamean[par-nz])
   else Result := z[i,par];
end;

function DataSpecification.DifCovEgo
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
  if misz(i,par) then Result := 0.0 else
    begin
    if (par > nz) and (par <= nz+nza) then Result := (z[i,par]-zamean[par-nz])
     else Result := z[i,par];
   end;  
(* initialise_actors(k) implies that z[j,par] = zc[j,thisperiod,par] *)
end;

function DataSpecification.DifCovSimt
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (par > nz) and (par <= nz+nza) then
    Result := similarity(ya(i,par-nz),ya(j,par-nz),par)
   else Result := similarity(z[i,par],z[j,par],par);
end;

function DataSpecification.DifCovSim
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
// the calculation of ya in SimStats/ FunctionEvaluations implies
// that this is functionally equivalent to DifCovSimt.
  if (misz(j,par) or misz(i,par)) then Result := 0.0
    else
   Result := similarity(z[i,par],z[j,par],par); // 3.17w
end;

function DataSpecification.DifCovSimRect
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (symmetric or (dg.TieValue(j,i) >= 1)) then
    begin
    if (par > nz) and (par <= nz+nza) then
      Result := similarity(ya(i,par-nz),ya(j,par-nz),par)
     else Result := similarity(z[i,par],z[j,par],par);
    end
   else Result := 0.0;
end;

function DataSpecification.DifCovSimRec
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
  if ((not (misz(i,par) or misz(j,par))) and
      (symmetric or (dg.TieValue(j,i) >= 1))) then
        Result := similarity(z[i,par],z[j,par],par)
    else Result := 0.0;
end;

function DataSpecification.DifCovSamet
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (par > nz) and (par <= nz+nza) then
    begin
    if (abs(ya(i,par-nz) - ya(j,par-nz)) < epsi) then result := 1.0 else result := 0.0;
    end
   else
    begin
    if (abs(z[i,par] - z[j,par]) <  epsi) then result := 1.0 else result := 0.0;
    end;
end;

function DataSpecification.DifCovSame
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
// the calculation of ya in SimStats/ FunctionEvaluations implies
// that this is functionally equivalent to DifCovSamet.
  if (misz(j,par) or misz(i,par)) then Result := 0.0 else
    begin
    if (abs(z[i,par] - z[j,par]) <  epsi) then result := 1.0 else result := 0.0;
    end;
end;


function DataSpecification.DifCovSameRect
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  if (symmetric or (dg.TieValue(j,i) >= 1)) then
    begin
      if (par > nz) and (par <= nz+nza) then
      begin
      if (abs(ya(i,par-nz) - ya(j,par-nz)) <  epsi) then result := 1.0 else result := 0.0;
      end
     else
      begin
      if (abs(z[i,par] - z[j,par]) <  epsi) then result := 1.0 else result := 0.0;
      end;
     end
   else Result := 0.0;
end;

function DataSpecification.DifCovSameRec
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const par: integer): real;
begin
  if ((not (misz(i,par) or misz(j,par))) and
      (symmetric or (dg.TieValue(j,i) >= 1))) then
      begin
      if (abs(z[i,par] - z[j,par]) <  epsi) then result := 1.0 else result := 0.0;
      end
    else Result := 0.0;
end;

function DataSpecification.CovSimTot
           (const dg: TValuedDigraph; const i,k,par: integer): real;
var j: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  if misz(i,par) then exit;
  it := dg.OutTies[i].First;
  while it.HasNext do begin
    it := it.next;
    j := it.Value1;
    if (not misz(j,par)) then begin
      if (par > nz) and (par <= nz+nza) then
        Result := Result + similarity(ya(i,par-nz),ya(j,par-nz),par)
       else Result := Result + similarity(z[i,par],z[j,par],par);
     end;
   end; (* while *)
end;


function DataSpecification.CovSimRecTot
           (const dg: TValuedDigraph; const i,k,par: integer): real;
var j: integer;
    it: TDoublyLinkedDoubleInt;
begin
  result := 0;
  if misz(i,par) then exit;
  it := dg.OutTies[i].First;
  while it.HasNext do begin
    it := it.next;
    j := it.Value1;
    if (not (misz(j,par))) and (dg.TieValue(j,i) >= 1) then begin
        if (par > nz) and (par <= nz+nza) then
          Result := Result + similarity(ya(i,par-nz),ya(j,par-nz),par)
         else Result := Result + similarity(z[i,par],z[j,par],par);
        end;
     end; (* while *)
end;


{function DataSpecification.Interactiont
           (const dg: TValuedDigraph; const i,j,k: integer;
                           const tv: shortint; const par: integer): real;
begin
  Result := Effect[(par mod 1000) - 1].f1ijc(dg,i,j,k,tv)*
              Effect[((par div 1000) mod 1000) - 1].f1ijc(dg,i,j,k,tv);
end;

function DataSpecification.Interaction
           (const dg: TValuedDigraph; const i,k,par: integer): real;
begin
end;
}


procedure DataSpecification.DefineSignedNetworkEffects(var Effs: TEffects);
const  ZeroContriW: TConfigWeight0 = (0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0);
      ZeroConfigW: TConfigWeight1 = (0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0);
var i: integer;
    ContrW: TConfigWeight0;
    ConfgW: TConfigWeight1;

begin
  if (Effs.NumberEffects > 0) then exit;
  (* This procedure must not be executed more than once! *)
  Effs.AddEffect1('outdegree positive ','Number of positive ties',
                 ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('outdegree positive ','Number of positive ties',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('outdegree negative ','Number of negative ties',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('reciprocity P/P',  'Number of P/P dyads',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('reciprocity 0/0', 'Number of 0/0 dyads',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('reciprocity N/N','Number of N/N dyads',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('balance PPP', 'Number of PPP triples',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('balance PNN', 'Number of PNN triples',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('balance NPN', 'Number of NPN triples',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('balance NNP', 'Number of NNP triples',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('anti-balance PPN', 'Number of PPN triples',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('anti-balance PNP', 'Number of PNP triples',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('anti-balance NPP', 'Number of NPP triples',
                  ZeroContriW, ZeroConfigW,0,0,0);
  Effs.Addeffect1('anti-balance NNN',  'Number of NNN triples',
                  ZeroContriW, ZeroConfigW,0,0,0);
  for i:=1 to nzz do begin
    Effs.Addeffect1(ZZnames[i] + ' pos. ties',
                   'Sum of pos. ties  '+ZZNames[i],
                    ZeroContriW, ZeroConfigW,0,0,0);
    Effs.Addeffect1(ZZnames[i] + ' neg. ties', 'Sum of neg. ties  '+ZZNames[i],
                 ZeroContriW, ZeroConfigW,0,0,0);
    Effs.Addeffect1(ZZnames[i] + ' P/P dyads', 'Sum of P/P dyads  '+ZZNames[i],
                  ZeroContriW, ZeroConfigW,0,0,0);
    Effs.Addeffect1(ZZnames[i] + ' 0/0 dyads', 'Sum of 0/0 dyads  '+ZZNames[i],
                  ZeroContriW, ZeroConfigW,0,0,0);
    Effs.Addeffect1(ZZnames[i] + ' N/N dyads', 'Sum of N/N dyads  '+ZZNames[i],
                  ZeroContriW, ZeroConfigW,0,0,0);
   end; (* for i *)
  for i:=1 to (nz+nzc) do begin
    Effs.Addeffect1(ZNames[i]+' pos. ties ego',
                   'Sum of pos. outdegrees  '+ZNames[i],
                  ZeroContriW, ZeroConfigW,0,0,0);
    Effs.Addeffect1(ZNames[i]+' pos. ties alter',
                   'Sum of pos. indegrees   '+ZNames[i],
                  ZeroContriW, ZeroConfigW,0,0,0);
    Effs.Addeffect1(ZNames[i]+' pos. ties similarity',
                   'Sum of pos. ties  '+ZNames[i]+' similarity',
                  ZeroContriW, ZeroConfigW,0,0,0);
    Effs.Addeffect1(ZNames[i]+' neg. ties ego',
                   'Sum of neg. outdegrees  '+ZNames[i],
                  ZeroContriW, ZeroConfigW,0,0,0);
    Effs.Addeffect1(ZNames[i]+' neg. ties alter',
                   'Sum of neg. indegrees   '+ZNames[i],
                    ZeroContriW, ZeroConfigW,0,0,0);
    Effs.Addeffect1(ZNames[i]+' neg. ties similarity',
                   'Sum of neg. ties  '+ZNames[i]+' similarity',
                    ZeroContriW, ZeroConfigW,0,0,0);
   end; (* for i *)

// er stond ook nog @@@     param[24] := 7;
end; (* DefineSignedNetworkEffects *)

procedure DataSpecification.DefineNetworkEffects(var Effs: TEffects);
const  ZeroContriW: TConfigWeight0 = (0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0);
      ZeroConfigW: TConfigWeight1 = (0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0);
var i,parindex,pardum: integer;
    ContrW: TConfigWeight0;
    ConfgW: TConfigWeight1;
    fijrnil: TRealfunctionijr;
    fnil: TRealfunction;
function nextpar(const pv: integer): integer;
(* The use of nextpar is a trick to obtain the correspondence with
   param_effect, and start with default values defined in this procedure,
   when param_effect has been defined initially in S_Start by high(integer).
   This requires that for each call of AddEffects, nextpar is called once. *)
begin
  inc(parindex);
  if (param_effect[parindex] = high(integer)) then param_effect[parindex] := pv;
  nextpar := param_effect[parindex];
end; (* Nextpar *)
begin (* DefineNetworkEffects*)
// @@@ alle symmetric moeten nog gecontroleerd
  fnil := nil;
  fijrnil := nil;
  parindex := 0;

  if (Effs.NumberEffects > 0) then exit;
  (* This procedure must not be executed more than once! *)
  (* For defining the reciprocity model, it is required that the first effect
     is the outdegree effect and the second the reciprocity effect. *)

  if (observations <= 1) then
(* Note that for the symmetric case, the change statistic(i,j) is the change
   for toggling (i,j) as well as (j,i). *)
    begin
    ContrW := ZeroContriW;
    ContrW[1] := 1.0;
    if symmetric then Effs.AddERGMEffect1('degree (density)', 'Number of edges',
                    ContrW, nextpar(0),0,0, fijrnil,0)
     else Effs.AddERGMEffect1('degree (density)', 'Number of ties',
                    ContrW, nextpar(0),0,0, fijrnil,0);

    ContrW := ZeroContriW;
    ContrW[2] := 1.0;
    if (not symmetric) then
        Effs.AddERGMEffect1('reciprocity', 'Number of mutuals',
                    ContrW, nextpar(0),0,0, fijrnil,0);

    ContrW := ZeroContriW;
    ContrW[7] := 1.0;
    if symmetric then
      Effs.AddERGMeffect1('transitive triads' ,  'Number of transitive triads',
                         ContrW, nextpar(0),0,0,fijrnil,0)
     else
      begin
      ContrW[8] := 1.0;
      ContrW[9] := 1.0;
      Effs.AddERGMEffect1('transitive triplets', 'Number of transitive triplets',
                         ContrW, nextpar(0),0,0,fijrnil,0)
      end;

    ContrW := ZeroContriW;
    ContrW[10] := 1.0;
    if (not symmetric) then
      Effs.AddERGMEffect1('3-cycles','3-cycles',ContrW,nextpar(0),0,0,fijrnil,0);

    ContrW := ZeroContriW;
    if symmetric then
      begin
      ContrW[3] := 1;
      ContrW[6] := 1;
      ContrW[0] := -2;
      Effs.AddERGMEffect1('2-stars','2-stars',ContrW,nextpar(0),0,0,fijrnil,0);
      end
     else
      begin
      ContrW[0] := -1;
      ContrW[3] := 1;
      Effs.AddERGMEffect1('out-2-stars','out-2-stars',ContrW,nextpar(0),0,0,fijrnil,0);
      ContrW[3] := 0;
      ContrW[6] := 1;
      Effs.AddERGMEffect1('in-2-stars','in-2-stars',ContrW,nextpar(0),0,0,fijrnil,0);
      ContrW[0] := 0;
      ContrW[2] := -2;
      ContrW[6] := 0;
      ContrW[4] := 1;
      ContrW[5] := 1;
      Effs.AddERGMEffect1('2-paths','2-paths',ContrW,nextpar(0),0,0,fijrnil,0);
      end;

    index_koutstars := 0;
    index_kinstars := 0;
    index_ktri := 0;
    index_ktwop := 0;

    if symmetric then
      begin
      Effs.AddERGMEffect1('#-stars','#-stars/#!',ZeroContriW,nextpar(3),2,6,ChooseBoth,0);
      Effs.AddERGMEffect1('#-stars','#-stars/#!',ZeroContriW,nextpar(4),2,6,ChooseBoth,0);
      Effs.AddERGMEffect1('#-stars','#-stars/#!',ZeroContriW,nextpar(5),2,6,ChooseBoth,0);
      Effs.AddERGMEffect1('#-stars','#-stars/#!',ZeroContriW,nextpar(6),2,10,ChooseBoth,0);
      Effs.AddERGMEffect1('#-stars','#-stars/#!',ZeroContriW,nextpar(7),2,10,ChooseBoth,0);
(* If the maximum is increased beyond 10, then change facmax in S_Dat. *)
      Effs.AddERGMEffect2('Sum 1/(degrees + #)','Sum 1/(degrees + #)',
                          ZeroContriW,nextpar(1),1,10,InvDEmpty,InvDegrees,0);
      Effs.AddERGMEffect2('Sum 1/(degrees + #)(degrees + #+1)','Sum 1/(degrees + #)(degrees + #+1)',
                          ZeroContriW,nextpar(1),1,10,InvDEmpty2,InvDegrees2,0);
      Effs.AddERGMEffect1('alternating k-stars, par. #','alternating k-stars, par. #',
                             ZeroContriW,nextpar(2),1,100,AltStars,0);
      index_koutstars := parindex;

      end
     else
      begin
      Effs.AddERGMEffect1('in-#-stars','in-#-stars/#!',ZeroContriW,nextpar(3),2,6,ChooseIn,0);
      Effs.AddERGMEffect1('in-#-stars','in-#-stars/#!',ZeroContriW,nextpar(4),2,6,ChooseIn,0);
      Effs.AddERGMEffect1('in-#-stars','in-#-stars/#!',ZeroContriW,nextpar(5),2,6,ChooseIn,0);
      Effs.AddERGMEffect1('in-#-stars','in-#-stars/#!',ZeroContriW,nextpar(6),2,10,ChooseIn,0);
      Effs.AddERGMEffect1('in-#-stars','in-#-stars/#!',ZeroContriW,nextpar(6),2,10,ChooseIn,0);
      Effs.AddERGMEffect1('out-#-stars','out-#-stars/#!',ZeroContriW,nextpar(3),2,6,ChooseOut,0);
      Effs.AddERGMEffect1('out-#-stars','out-#-stars/#!',ZeroContriW,nextpar(4),2,6,ChooseOut,0);
      Effs.AddERGMEffect1('out-#-stars','out-#-stars/#!',ZeroContriW,nextpar(5),2,6,ChooseOut,0);
      Effs.AddERGMEffect1('out-#-stars','out-#-stars/#!',ZeroContriW,nextpar(6),2,10,ChooseOut,0);
      Effs.AddERGMEffect1('out-#-stars','out-#-stars/#!',ZeroContriW,nextpar(6),2,10,ChooseOut,0);
(* If the maximum is increased beyond 10, then change facmax in S_Dat. *)
      Effs.AddERGMEffect2('Sum 1/(outdegrees + #)','Sum 1/(outdegrees + #)',
                          ZeroContriW,nextpar(1),1,10,InvDEmpty,InvDegree,0);
      Effs.AddERGMEffect2('Sum 1/(o.deg + #)(o.deg + #+1)','Sum 1/(o.deg + #)(o.deg + #+1)',
                          ZeroContriW,nextpar(1),1,10,InvDEmpty2,InvDegree2,0);
      Effs.AddERGMEffect1('alternating out-k-stars, par. #','alternating out-k-stars, par. #',
                          ZeroContriW,nextpar(2),1,100,AltOutStars,0);
      index_koutstars := parindex;
      Effs.AddERGMEffect1('alternating in-k-stars, par. #','alternating in-k-stars, par. #',
                          ZeroContriW,nextpar(2),1,100,AltInStars,0);
      index_kinstars := parindex;
      end;

    if symmetric then
      begin
      Effs.AddERGMEffect1('indirect connections','number pairs indirectly connected',
                         ZeroContriW,nextpar(0),0,0,Indirect,0);
      Effs.AddERGMEffect1('direct + indirect connections','number pairs direct + indirectly connected',
                         ZeroContriW,nextpar(0),0,0,DirectIndirect,0);
      end
     else
      begin
      Effs.AddERGMEffect1('indirect connections','number ordered pairs indirectly connected',
                         ZeroContriW,nextpar(0),0,0,Indirect,0);
      Effs.AddERGMEffect1('direct + indirect connections','number ordered pairs direct + indirectly connected',
                         ZeroContriW,nextpar(0),0,0,DirectIndirect,0);
      end;

    Effs.AddERGMEffect1('4-cycles','4-cycles',
                       ZeroContriW,nextpar(0),0,0,ThreePaths,0);
(* Here the name of the change statistic refers to itself and not to the effect. *)

    Effs.AddERGMEffect1('3 par. twopaths','3 par. twopaths',
                       ZeroContriW,nextpar(0),0,0,ThreeTwoPaths,0);

    Effs.AddERGMEffect1('alternating independent twopaths, par. #','alternating independent twopaths, par. #',
                       ZeroContriW,nextpar(2),1,10,AltParTwoPaths,0);
    index_ktwop := parindex;

    Effs.AddERGMEffect1('2-triangles','2-triangles',
                       ZeroContriW,nextpar(0),0,0,TwoTriangles,0);

    Effs.AddERGMEffect1('3-triangles','3-triangles',
                       ZeroContriW,nextpar(0),0,0,ThreeTriangles,0);

    Effs.AddERGMEffect1('alternating k-triangles, par. #','alternating k-triangles, par. #',
                       ZeroContriW,nextpar(2),1,10,AltTriangles,0);
    index_ktri := parindex;

    if symmetric then Effs.AddERGMEffect1('coathangers','coathangers',
                       ZeroContriW,nextpar(0),0,0,Coathanger,0);

    Effs.AddERGMEffect1('3-paths','3-paths',
                       ZeroContriW,nextpar(0),0,0,ThreeLine,0);


    if symmetric then
    begin

    (* the arrays for clique intermediate results have length 20, corresponding to maximum 21; see Digraph. *)
    Effs.AddERGMEffect1('cliques of order #','number of cliques of order #',
                       ZeroContriW,nextpar(4),3,20,CliqueChange,0);
    index_cliq1 := parindex;

    Effs.AddERGMEffect1('cliques of order #','number of cliques of order #',
                       ZeroContriW,nextpar(5),3,20,CliqueChange,0);
    index_cliq2 := parindex;

    Effs.AddERGMEffect1('cliques of order #','number of cliques of order #',
                       ZeroContriW,nextpar(6),3,20,CliqueChange,0);
    index_cliq3 := parindex;

    Effs.AddERGMEffect1('cliques of order #','number of cliques of order #',
                       ZeroContriW,nextpar(7),3,20,CliqueChange,0);
    index_cliq4 := parindex;

    //    Effs.AddERGMEffect1('near-cliques of order #','number of near-cliques of order #',
//                       ZeroContriW,nextpar(4),3,20,Red1_CliqueChange,0);
   index_rcliq1 := parindex;

//    Effs.AddERGMEffect1('near-cliques of order #','number of near-cliques of order #',
//                       ZeroContriW,nextpar(5),3,20,Red1_CliqueChange,0);
    index_rcliq2 := parindex;
  end; (* if symmetric *)  

  for i:=1 to nz do
(* AA begin *)
    if symmetric then
      begin
      if compzvar[i] then
        Effs.AddERGMEffect1(ZNames[i],'Sum of degrees times '+Znames[i],
                         ZeroContriW,nextpar(i),i,i,Covari,3)
       else
        begin
        Effs.AddERGMEffect1(ZNames[i],'Sum of degrees times '+Znames[i],
                           ZeroContriW,nextpar(i),i,i,Covari,1);
        Effs.AddERGMEffect1(ZNames[i]+' similarity','Similarity on '+Znames[i],
                           ZeroContriW,nextpar(i),i,i,SimCovar,1);
        Effs.AddERGMEffect1('same '+ZNames[i],'Same values on '+Znames[i],
                           ZeroContriW,nextpar(i),i,i,SameCovar,1);
        end;
      end
     else
      begin
      if compzvar[i] then
        Effs.AddERGMEffect1(ZNames[i],'Sum of degrees times '+Znames[i],
                           ZeroContriW,nextpar(i),i,i,Covari,3)
       else
        begin
        Effs.AddERGMEffect1(ZNames[i]+' ego','Sum of outdegrees times '+Znames[i],
                           ZeroContriW,nextpar(i),i,i,Covari,2);
        Effs.AddERGMEffect1(ZNames[i]+' alter','Sum of indegrees times '+Znames[i],
                           ZeroContriW,nextpar(i),i,i,Covarj,2);
        Effs.AddERGMEffect1(ZNames[i]+' similarity','Similarity on '+Znames[i],
                           ZeroContriW,nextpar(i),i,i,SimCovar,1);
        Effs.AddERGMEffect1('same '+ZNames[i],'Same values on '+Znames[i],
                           ZeroContriW,nextpar(i),i,i,SameCovar,1);
        end;
      end;
(* AA end *)

  for i:=1 to nzz do
      if symmetric then
      Effs.AddERGMEffect1(ZZNames[i]+' (centered)','Sum of ties times '+ZZnames[i],
                         ZeroContriW,nextpar(i),i,i,DyCovar,1)
       else
      Effs.AddERGMEffect1(ZZNames[i]+' (centered)','Sum of ties times '+ZZnames[i],
                         ZeroContriW,nextpar(i),i,i,DyCovar,2);

  if (nz + nzz > 0) then for i:=1 to nintn do
      Effs.AddERGMEffect1('unspecified interaction effect',
                      'unspecified interaction statistic',ZeroContriW,
                    nextpar(0),low(integer),high(integer),nil,4);
(* The specification of this effect is done in procedure SpecifyInteractionEffect
   which is defined in Digraph and called in S_Base *)

  end (* observations <= 1 *)
 else (* observations >=  2 *)
  begin

  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;
  ContrW[1] := 1.0;
  if symmetric then ConfgW[1] := 0.5 else ConfgW[1] := 1.0;
  if symmetric then Effs.Addeffect1('degree (density)', 'Number of edges',
                    ContrW, ConfgW,nextpar(0),0,0)
   else Effs.Addeffect1('outdegree (density)', 'Number of ties',
                    ContrW, ConfgW,nextpar(0),0,0);

  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;
  ContrW[2] := 1.0;
  ConfgW[2] := 1.0;
  if (not symmetric) then
        Effs.Addeffect1('reciprocity', 'Number of reciprocated ties',
                    ContrW, ConfgW,nextpar(0),0,0);

  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;
  ContrW[7] := 1.0;
(* pure tie-oriented model: ContrW[8] = 1, else = 0 *)
  if symmetric then
    ConfgW[7] := 1/6
   else
    begin
(*     ContrW[8] := 1.0;
  This has been erroneously present until version 3.17p.
  Corrected 3.17q.
  Error discovered by Ruth and Krists. *)
    ContrW[9] := 1.0;
    ConfgW[7] := 1.0;
    end;
  if symmetric then
   Effs.Addeffect1('transitive triads' ,  'Number of transitive triads',
                    ContrW, ConfgW,nextpar(0),0,0)
  else Effs.Addeffect1('transitive triplets', 'Number of transitive triplets',
                    ContrW, ConfgW,nextpar(0),0,0);

  if (not symmetric) then
    begin
    ContrW := ZeroContriW;
    ConfgW := ZeroConfigW;
    ContrW[8] := 1.0;
    ConfgW[8] := 1.0;
    Effs.Addeffect1('transitive mediated triplets', 'Number of transitive triplets',
                    ContrW, ConfgW,nextpar(0),0,0);
    end;

  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;
  ContrW[10] := 1.0;
  ConfgW[10] := 1/3;
  if (not symmetric) then
    Effs.Addeffect1('3-cycles','3-cycles',ContrW, ConfgW,nextpar(0),0,0);


  Effs.Addeffect5('transitive ties', 'Number of ties with transitive closure ',
                    nextpar(0),0,0,DifIndirectt,DifIndirect,false);

  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;
  ContrW[2] := -1.0;
  ContrW[4] := 1.0;
  ContrW[8] := -1.0;
  ConfgW[2] := -1.0;
  ConfgW[4] := 1.0;
  ConfgW[8] := -1.0;
(* Statistic is different from that in Siena 2 *)
// AA change: condition omitted
  Effs.Addeffect2('betweenness','betweenness count',
                ContrW, ConfgW,nextpar(0),0,0,CorBetween);

  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;

(*    Contr_fn := -((2*outd_yn(i) + Outd_yn(j) - 2.0*one(yn(i,j)) - one(yn(j,i))
                     - 2.0*(1.0*path2^[i,j] + 1.0*in2star^[i,j]) )/(n-2) - balmean);
  *)
(* Multiplied by (n-2), version 3.17w *)  
  ContrW[0] := 2.0;
  ContrW[1] := (n-2.0)*balmean;
  ContrW[2] := 1.0;
  ContrW[3] := -2.0;
  ContrW[5] := -1.0;
  ContrW[7] := 2.0;
  ContrW[9] := 2.0;
  Effs.Addeffect4('balance', 'Amount of balance',
                 ContrW, nextpar(0),0,0,BalanceSimilarity);

  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;
  if anynetworkmissings then
    begin
    if symmetric then
       Effs.Addeffect3('number of actor pairs at distance 2', 'Number of distances equal to 2',
                ContrW, nextpar(0),0,0, Dist2ActorContr, HalfDist2Contr, HalfCorNdist2)
       else Effs.Addeffect3('number of actors at distance 2',
               'Number of directed distances equal to 2',
               ContrW, nextpar(0),0,0, Dist2ActorContr, Dist2Contr, CorNdist2);
    end
   else
    begin
    if symmetric then
       Effs.Addeffect5('number of actor pairs at distance 2', 'Number of distances equal to 2',
                nextpar(0),0,0, Dist2ActorContr, HalfDist2Contr,false)
       else Effs.Addeffect5('number of actors at distance 2',
               'Number of directed distances equal to 2',
                nextpar(0),0,0, Dist2ActorContr, Dist2Contr, false);
    end;

  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;
  if anynetworkmissings then
    begin
    if symmetric then
       Effs.Addeffect3('number pairs at doubly achieved distance 2',
                 'Number of doubly achieved distances 2',
    ContrW, nextpar(0),0,0, TwoDist2ActorContr, HalfTwoDist2Contr, HalfTwoCorNdist2)
       else Effs.Addeffect3('number pairs at doubly achieved distance 2',
               'Number of doubly achieved distances 2',
               ContrW, nextpar(0),0,0, TwoDist2ActorContr, TwoDist2Contr, TwoCorNdist2);
    end
   else
    begin
    if symmetric then
       Effs.Addeffect5('number pairs at doubly achieved distance 2',
                 'Number of doubly achieved distances 2',
            nextpar(0),0,0, TwoDist2ActorContr, HalfTwoDist2Contr, false)
       else Effs.Addeffect5('number pairs at doubly achieved distance 2',
               'Number of doubly achieved distances 2',
             nextpar(0),0,0, TwoDist2ActorContr, TwoDist2Contr, false);
    end;

  if (not symmetric) then Effs.Addeffect6('dense triads' ,
                    'Sum of triads with # or more ties' ,
                    nextpar(5),5,6,DifDense,DenseTot);
  index_cliq1 := parindex;

  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;
(* The ContrW[0], ContrW[1] elements were forgotten in Siena 1 and 2:
   but a tie from i to j will increase j's indegree!
   The following effect was multiplied by na in version 3.17q,
   to bring it in line with the activity effects.
 *)
(* until 3.17p:
  ContrW[0] := -1/n;
  ContrW[1] := 1/n;
  ContrW[6] := 1/n;
  ConfgW[6] := 1.0; *)
(* from 3.17t: *)
  ContrW[0] := -1.0;
  ContrW[1] := 1.0;
  ContrW[6] := 2.0;
  ConfgW[6] := 1.0;

  if symmetric then
    Effs.Addeffect1('degree of alter', 'Sum of squared degrees',
                    ContrW, ConfgW,nextpar(0),0,0)
   else Effs.Addeffect1('indegree - popularity', // earlier called 'popularity of alter',
                         'Sum of squared indegrees',
                    ContrW, ConfgW,nextpar(0),0,0);
(*  if symmetric then Effs.Addeffect6('sqrt degree of alter',
                   'Sum of degrees',
                   nextpar(0),0,0,DifSqrtIn,SqrtInTot)
   else Effs.Addeffect6('popularity of alter (sqrt measure)',
                   'Sum of indegrees  sqrt(indegree)',
                    nextpar(0),0,0,DifSqrtIn,SqrtInTot);
 the next is functionally equivalent for the evaluation function,
 but implies also an endowment function:
                     *)

  if symmetric then Effs.Addeffect5('degree popularity (sqrt)',  // earlier called 'sqrt degree of alter'
                   'Sum of degrees ^ 1.5',
                   nextpar(0),0,0,DifSqrtInt,DifSqrtIn,false)
   else Effs.Addeffect5('indegree - popularity (sqrt)', // earlier called 'popularity of alter (sqrt measure)'
                   'Sum of indegrees  sqrt(indegree)',
                    nextpar(0),0,0,DifSqrtInt,DifSqrtIn,false);


  ContrW := ZeroContriW;
  ConfgW := ZeroConfigW;
(*   The following effect was multiplied by n in version 3.17q
   to bring it in line with the activity effects.
  until 3.17p:  ContrW[5] := 1/n;
                ConfgW[5] := 1.0; *)
(* from 3.17t: *)
  ContrW[5] := 1.0;
  ConfgW[5] := 1.0;

  if (not symmetric) then Effs.Addeffect1('outdegree - popularity', // earlier called 'activity of alter',
                 'Sum of crossproducts indegree  outdegree',
                    ContrW, ConfgW,nextpar(0),0,0);
  if (not symmetric) then Effs.Addeffect5('outdegree - popularity (sqrt)', // earlier called 'activity of alter (sqrt measure)',
                   'Sum of indegrees  sqrt(outdegree)',
                    nextpar(0),0,0,DifSqrtOutt,DifSqrtOut,false);

(*
  if symmetric then Effs.Addeffect6('degree up to #',
                   'Sum of degrees up to #',
                   nextpar(3),1,high(integer),DifMax,MaxTot)
   else Effs.Addeffect6('outdegree up to #',
                   'Sum of outdegrees up to #',
                    nextpar(3),1,high(integer),DifMax,MaxTot);
*)

  if not symmetric then
    begin
    Effs.Addeffect6('indegree - activity',
                    'Sum of crossproducts indegree  outdegree' ,
                    nextpar(0),0,0,DifOutInt,OutIn);
    Effs.Addeffect6('indegree - activity (sqrt)','Sum of outdegrees  sqrt(indegree)' ,
                    nextpar(0),0,0,DifOutSqrtInt,OutSqrtIn);
    Effs.Addeffect6('outdegree - activity', 'Sum of squared outdegrees' ,
                    nextpar(0),0,0,DifPow2,Pow2Plus);
    Effs.Addeffect6('outdegree - activity (sqrt)',   // earlier called 'outdegree^(1.5)' ,
                    'Sum of outdegrees^(1.5)' ,
                    nextpar(0),0,0,DifPow15,Pow15Plus);
    end;

  if symmetric then Effs.Addeffect6('1/(degree + #)','Sum 1/(degrees + #)',
                    nextpar(1),1,high(integer),DifRec,RecTot)
   else Effs.Addeffect6('1/(outdegree + #)','Sum 1/(outdegrees + #)',
                    nextpar(1),1,high(integer),DifRec,RecTot);
  if symmetric then Effs.Addeffect6('1/(degree+#)(degree+1+#)',
                       'Sum 1/(degrees+#)(degrees+1+#)',
                    nextpar(1),1,high(integer),DifRec2,Rec2Tot)
   else Effs.Addeffect6('1/(outdegree+#)(outdegree+1+#)',
                      'Sum 1/(outdegrees+#)(outdegrees+1+#)',
                    nextpar(1),1,high(integer),DifRec2,Rec2Tot);


  (* assortativity effects *)

   if symmetric then Effs.Addeffect5('degree^(1/#) assortativity',
                   'Sum of degree^(1/#) products',
                   nextpar(2),1,2,Difddt,Prododod,false)
   else
    begin
    Effs.Addeffect5('out-out degree^(1/#) assortativity',
                   'Sum of out-out degree^(1/#) products',
                   nextpar(2),1,2,Difododt,Prododod,false);
    Effs.Addeffect5('out-in degree^(1/#) assortativity',
                   'Sum of out-in degree^(1/#) products',
                   nextpar(2),1,2,Difodidt,Prododid,false);
    Effs.Addeffect5('in-out degree^(1/#) assortativity',
                   'Sum of in-out degree^(1/#) products',
                   nextpar(2),1,2,Difidodt,Prodidod,false);
    Effs.Addeffect5('in-in degree^(1/#) assortativity',
                   'Sum of in-in degree^(1/#) products',
                   nextpar(2),1,2,Difididt,Prodidid,false);
    end;

  (* next: effects of dyadic covariates on network objective function *)
  (* The index number of the covariates is used as parameters,
     superseding param_effect values, and without calling nextpar;
     therefore, nextpar must be called outside of AddEffect.
     --- It seems that the last two lines are superseded.
     *)
  for i:=1 to nzz do
    begin
    Effs.Addeffect7(ZZNames[i]+' (centered)',  'Sum of ties  '+ZZNames[i],
                    nextpar(i),i,i,DifDyadicCovt,DifDyadicCov,true,1);
    if (not symmetric) then
      begin
      Effs.Addeffect5(ZZNames[i]+' (centrd)  reciprocity',  'Sum reciprocated ties  '+ZZNames[i],
                    nextpar(i),i,i,DifDyadicCovRect,DifDyadicCovRec,true);
      end;
      Effs.Addeffect5('WW=>X closure of '+ZZNames[i],  'WW=>X closure of '+ZZNames[i],
                    nextpar(i),i,i,DyadicCovWWXt,DyadicCovWWX,true);
      Effs.Addeffect5('WX=>X closure of '+ZZNames[i],  'WX=>X closure of '+ZZNames[i],
                    nextpar(i),i,i,DyadicCovWXXt,DyadicCovWXX,true);
      Effs.Addeffect5('XW=>X closure of '+ZZNames[i],  'XW=>X closure of '+ZZNames[i],
                    nextpar(i),i,i,DyadicCovXWXt,DyadicCovXWX,true);
    end;

  (* changing dyadic covariates *)
  for i:=1 to nzzc do
    begin
    Effs.Addeffect7(ZZNames[nzz+i]+' (centered)',  'Sum of ties  '+ZZNames[nzz+i],
                    nextpar(i),i,i,DifChangeDyadicCovt,DifChangeDyadicCov,true,1);
    if (not symmetric) then
      begin
      Effs.Addeffect5(ZZNames[nzz+i]+' (centrd)  reciprocity',  'Sum reciprocated ties  '+ZZNames[nzz+i],
                    nextpar(i),i,i,DifChangeDyadicCovRect,DifChangeDyadicCovRec,true);
      end;
      Effs.Addeffect5('WW=>X closure of '+ZZNames[nzz+i],
                      'WW=>X closure of '+ZZNames[nzz+i],
                    nextpar(i),i,i,ChangeDyadicCovWWXt,ChangeDyadicCovWWX,true);
      Effs.Addeffect5('WX=>X closure of '+ZZNames[nzz+i],
                      'WX=>X closure of '+ZZNames[nzz+i],
                    nextpar(i),i,i,ChangeDyadicCovWXXt,ChangeDyadicCovWXX,true);
      Effs.Addeffect5('XW=>X closure of '+ZZNames[nzz+i],
                      'XW=>X closure of '+ZZNames[nzz+i],
                    nextpar(i),i,i,ChangeDyadicCovXWXt,ChangeDyadicCovXWX,true);
    end;

   (* individual covariates;
       for nz+1 <= i <= nz+nza, the dependent action variables. *)
  for i:=1 to nz+nzc do
    begin
    if symmetric then
      begin
      Effs.Addeffect7(ZNames[i], 'Sum of degrees  '+ZNames[i],
                    nextpar(i),i,i,DifCovAltert,DifCovAlter,true,1);
      if (range[i] >= 1.5) then
        Effs.Addeffect7(ZNames[i]+' squared ',  'Sum of degrees  squared '+ZNames[i],
                    nextpar(i),i,i,DifSquareCovAltert,DifSquareCovAlter,true,2);
      end
     else
      begin
(* AA begin *)
      if poszvar[i] and (not compzvar[i]) then
        begin
        Effs.Addeffect7(ZNames[i]+' alter',  'Sum of indegrees  '+ZNames[i],
                    nextpar(i),i,i,DifCovAltert,DifCovAlter,true,2);

        if (range[i] >= 1.5) then
        Effs.Addeffect7(ZNames[i]+' squared alter',  'Sum of indegrees  squared '+ZNames[i],
                    nextpar(i),i,i,DifSquareCovAltert,DifSquareCovAlter,true,2);
        end;

      Effs.Addeffect7(ZNames[i]+' ego',  'Sum of outdegrees  '+ZNames[i],
                    nextpar(i),i,i,DifCovEgot,DifCovEgo,true,3);
      end; (* not symmetric *)

    if poszvar[i] and (not compzvar[i]) then
      begin
      Effs.Addeffect7(ZNames[i]+' similarity',  'Similarity on '+ZNames[i],
                    nextpar(i),i,i,DifCovSimt,DifCovSim,true,1);          // 3.17w
      if (not symmetric) then
      Effs.Addeffect5(ZNames[i]+' similarity  reciprocity',  'Similarity  reciprocity on '+ZNames[i],
                    nextpar(i),i,i,DifCovSimRect,DifCovSimRec,true);
      Effs.Addeffect5('same '+ZNames[i],  'Same values on '+ZNames[i],
                    nextpar(i),i,i,DifCovSamet,DifCovSame,true);
      if (not symmetric) then
      Effs.Addeffect5('same '+ZNames[i]+'  reciprocity',  'Same values  reciprocity on '+ZNames[i],
                    nextpar(i),i,i,DifCovSameRect,DifCovSameRec,true);
      Effs.Addeffect5(ZNames[i]+' ego  '+ZNames[i]+' alter',
                      'Sum '+ZNames[i]+' ego  '+ZNames[i]+' alter',
                    nextpar(i),i,i,CovBotht,CovBoth,true);
      if (not symmetric) then
      Effs.Addeffect5(ZNames[i]+' ego  '+ZNames[i]+' alter  recipr.',
                      'Sum '+ZNames[i]+' ego  '+ZNames[i]+' rec. alter',
                    nextpar(i),i,i,CovBothRect,CovBothRec,true);

      ContrW := ZeroContriW;
      Effs.Addeffect3(ZNames[i]+' of indirect ties',
                      'Sum '+ZNames[i]+' of indirect ties', ContrW,
                    nextpar(i),i,i,CovIndirectt,CovIndirect,CorCovIndirect);
      end;
// AA end
    end; (* i *)

(* Interaction effects *)
  if (nz + nzc + nzz + nzzc >= 1) then for i:=1 to nintn do
    begin
    Effs.Addeffect7('unspecified interaction effect',
                      'unspecified interaction statistic',
                    nextpar(0),low(integer),high(integer),Interactiont_var_trans,Interaction_var_trans,true,4);
(* The specification of this effect is done in procedure SpecifyInteractionEffect
   which is defined in Digraph and called in S_Base.
   The default functions are Interactiont_var_trans and Interaction_var_trans,
   because these depend not only on the digraph but also on covariates,
   and they are not products of primary effects - therefore they cannot
   be defined in Digraph.
    *)
    end;

  end; (* if observations >= 2 *)
end; (* DefineNetworkEffects *)


(* End of include file S_Effects.pas *)
