unit s_ergm_est;

(* This unit contains some material connected to
   S_est that is used only for the observations = 1 case. *)

interface

uses classes, slib, s_constants, digraph, s_dat;

procedure HysteresisAnalysis(var lf, outf: TFileStream; p, pp: integer;
               theta, Targets: pvec; sx: pMatrix);


implementation

uses Forms, Controls, SysUtils, Graphics, eight, s_est, s_hysteresis, siena_7;


procedure HysteresisAnalysis(var lf, outf: TFileStream; p, pp: integer;
               theta, Targets: pvec; sx: pMatrix);
var  thb: textfile;
  h: byte;
  i, varhyst, ni, maxni, sed: integer;
  step, theta0, theta1, thetavh, ix, iy, se, xxx, abscissa, ylo, yhi: real;
  stats: pvec;
  erstr, s: string;
  feasible, sear, err: boolean;
  mtime: array of real;
  cov0: pmatrix;
  ordgs: intvec;

begin
{   if FileExists('thetabounds.txt') then
   with ContinueDigraph do begin
    SetLength(mtime,3);
    assignfile(thb,'thetabounds.txt');
    reset(thb);
    read(thb,varhyst,theta0,theta1);
    closefile(thb);
(* Hysteresis analysis *)
(* Results of the hysteresis analysis are given in 3 places:
   numbers in the .out file,
   graphically on the screen, and in Latex/Pictex format in the .log file. *)
    writeln(outf);
    writeln(outf,'Hysteresis analysis requested for parameter ',
                 varhyst,' (file thetabounds.txt).');
    writeln(lf,'Hysteresis analysis requested for parameter ',varhyst,'.');
    EstForm.PLtext.visible := true;
    EstForm.PLtext.caption := 'preparations for hysteresis analysis... please wait';
    if (varhyst > pp) then begin
      writeln(outf,'This is larger than the number of parameters in the model.');
      writeln(outf,'Changed to parameter 1.');
      varhyst := 1;
     end;
    writeln(outf,'This is the parameter for ',description[varhyst],'.');
    writeln(outf,'Results given are values for this parameter,');
    writeln(outf,'and the ',p,' generated statistics.');
    writeln(outf,'(PicTex data in logfile.');
    if (theta1 < theta0) then begin
      if (theta1 < -10.0) then theta1 := -10.0;
      if (theta0 > 10.0) then theta0 := 10.0;
      thetavh := theta0;
      theta0 := theta1;
      theta1 := thetavh;
     end;
    if (theta0 < -10.0) then theta0 := -10.0;
    if (theta1 > 10.0) then theta1 := 10.0;
    if (theta1 <= theta0 + 0.1) then theta1 := theta0 + 0.1;
    thetavh := theta[varhyst];
    writeln(outf);
    ni := 0;
    maxni := 100;
    step := (theta1-theta0)/maxni;
    EstForm.EIteration.hint := 'from 1 to '+inttostr(2*maxni);
    EstForm.Label3.Caption := 'Generated statistics';
    independentstart := true;
    with EstForm do begin
      ParameterValues.Height := round(PixelsPerInch div 5);
      QAutoCor.Hide;
      Label1.Caption := 'hysteresis analysis';
      Background.Hide;
      LPhase.Hide;
      EPhase.Hide;
      ProgressBar.Hide;
      ParameterValues.Hint := 'variable values for parameter '+inttostr(varhyst);
      Statistics.Hint := 'generated statistics';
      Graphic1.show;
      StartPaint;
      application.processmessages;
     end;
    (* variable varhyst on vertical axis : *)
    (* First find lowest and highest values *)
    if Cconditional then DefineStartMat(outf,1)  (* observed graph *)
      else DefineStartMat(outf,3);  (* empty graph *)
    initialise_y(1);
    if Continuous then rstep := rstep * 20.0;
    if Cconditional then begin
      theta[varhyst] := -100.0;
      FRAN(pp,theta,true,false,sed,stats,mtime,cov0,
           true,feasible,sear,err,erstr,outf,lf,h);
(* cov0 here used as dummy argument *)
     end else statistics(stats,ordgs,1);
    EstForm.ymin := stats[varhyst];
    if Cconditional then DefineStartMat(outf,1)  (* observed graph *)
       else DefineStartMat(outf,4);  (* complete graph *)
    initialise_y(1);
    if Cconditional then begin
      theta[varhyst] := 100.0;
      FRAN(pp,theta,true,false,sed,stats,mtime,cov0,
           true,feasible,sear,err,erstr,outf,lf,h);
     end else statistics(stats,ordgs,1);
    EstForm.ymax := stats[varhyst];
    abscissa := Targets[varhyst];

    with EstForm do begin
      if (ymin > abscissa - 3.0*root(sx[varhyst,varhyst])) then
        ymin := abscissa - 3.0*root(sx[varhyst,varhyst]) else
        ymin := ymin - root(sx[varhyst,varhyst]);
      if (ymax < abscissa + 3.0*root(sx[varhyst,varhyst])) then
        ymax := abscissa + 3.0*root(sx[varhyst,varhyst]) else
        ymin := ymin + root(sx[varhyst,varhyst]);

      if Cconditional then
        DefineStartMat(outf,1)  (* observed graph *)
       else DefineStartMat(outf,3);  (* empty graph *)
      initialise_y(1);
      independentstart := false;

      theta[varhyst] := theta0;
      sed := readstate;
      FRAN(pp,theta,true,false,sed,stats,mtime,cov0,
           true,feasible,sear,err,erstr,outf,lf,h);
      if Continuous then rstep := rstep/20.0;

      EstForm.PLtext.visible := false;

      xmin := 0;
      xmax := maxni;
      xxx := maxni*(thetavh-theta0)/(theta1-theta0);
      Drawxaxis(theta0,theta1);
      Drawyaxis(ymin,ymax);
      VertLine(xxx);
      HorLine(abscissa);
      Graphic1.Canvas.Pen.Color := clBlue;
      application.processmessages;
     end;
    writeln(lf,'Data for PicTex : ');
    writeln(lf,'\multiput{\gast}{ at                              ');
    ylo := abscissa;
    yhi := abscissa;
    repeat
      theta[varhyst] := theta0 + ni*step;
      NiceString(theta[varhyst],s);
      if (varhyst <= 9) then s := ' '+s;
      if (varhyst <= 30) then
        EstForm.ParameterValues.lines[0] := st(varhyst)+'. '+s;
      sed := readstate;
      FRAN(pp,theta,true,false,sed,stats,mtime,cov0,
           true,feasible,sear,err,erstr,outf,lf,h);
      write(outf,theta[varhyst]:10:4,' ; ');
      for i:=1 to p do write(outf,stats[i]:12:4,' ');
      EstForm.EIteration.Text:=inttostr(ni);
      for i:=1 to p do begin
        NiceString(stats[i],s);
        if (i <= 9) then EstForm.Statistics.lines[i-1] := ' '+st(i)+'. '+s;
        if ((i >= 10) and (i <= 30))
         then EstForm.Statistics.lines[i-1] := st(i)+'. '+s;
       end;

      with EstForm do with Graphic1 do begin
        ix := ni;
        iy := stats[varhyst];
        draw1(ix,iy);
       end;
      writeln(lf,theta[varhyst]:10:4,' ',iy:10:4);
      if (iy < ylo) then ylo := iy;
      if (iy > yhi) then yhi := iy;

      Application.ProcessMessages;
      if stopo then begin
         writeln(outf);
         writeln(outf,'Hysteresis iterations interrupted by user after ',ni,' simulations.');
         ni := maxni;
       end;
      inc(ni);
      writeln(outf);
     until (ni > maxni);

    with EstForm do begin
      Graphic1.Canvas.Pen.Color := clRed;
      application.processmessages;
     end;
    writeln(lf,'  /                                               ');
    writeln(lf,'\multiput{\rast}{ at                              ');
    EstForm.EIteration.Text:=inttostr(maxni);
    ni := 0;
    repeat
      theta[varhyst] := theta0 + (maxni-ni)*step;
      NiceString(theta[varhyst],s);
      if (varhyst <= 9) then s := ' '+s;
      if (varhyst <= 30) then
        EstForm.ParameterValues.lines[0] := st(varhyst)+'. '+s;
      sed := readstate;
      FRAN(pp,theta,true,false,sed,stats,mtime,cov0,
           true,feasible,sear,err,erstr,outf,lf,h);
      write(outf,theta[varhyst]:10:4,' ; ');
      for i:=1 to p do write(outf,stats[i]:12:4,' ');
       EstForm.EIteration.Text:=inttostr(maxni+ni);
      for i:=1 to p do begin
        NiceString(stats[i],s);
        if (i <= 9) then EstForm.Statistics.lines[i-1] := ' '+st(i)+'. '+s;
        if ((i >= 10) and (i <= 30))
         then EstForm.Statistics.lines[i-1] := st(i)+'. '+s;
       end;
      with EstForm do with Graphic1 do begin
        ix := maxni-ni;
        iy := stats[varhyst];
        draw1(ix,iy);
       end;
      writeln(lf,theta[varhyst]:10:4,' ',iy:10:4);
      if (iy < ylo) then ylo := iy;
      if (iy > yhi) then yhi := iy;

      Application.ProcessMessages;
      if stopo then begin
         writeln(outf,'Hysteresis iterations interrupted by user after ',ni,' simulations.');
         ni := maxni;
       end;
      inc(ni);
      writeln(outf);
     until (ni > maxni);

    with HysteresisForm do begin
      showmodal;
      if modalresult = mrOK then ;
     end;

    writeln(lf,'  /                                               ');
    writeln(lf,'Headings for PicTex : ');
    with HysteresisForm do StartPictexPlot(theta0,theta1,ylo,yhi,lf);
    writeln(lf,'\axis bottom shiftedto y=',
           FloatToStrF(abscissa,ffFixed,10,3),' /        ');
    writeln(lf,'\axis left shiftedto x=',
           FloatToStrF(thetavh,ffFixed,10,3),' /             ');
    writeln(lf,'..... data .....');
    EndPictexPlot(lf);
    theta[varhyst] := thetavh;
   end; (* if observations = 1 and FileExists *)
}end; (* HysteresisAnalysis *)

end.
