(*[[s]*)
(* This is file S_signed.pas, which is included in S_base.pas *)
(* This file contains some material connected to
   S_base that is used only for the signed case. *)

function ModelSpecification.SignedContr_fn(i,j         : integer; dif: Shortint;
                                     h           : byte;
                                     var halterr : boolean;
                                     var outf,lf : tFileStream) : Real;
(* The contribution to f [effect h]
   of changing ynv(i,j) to ynv(i,j)+dif;
   only the statistic part, not the parameter. *)
var poss, negg: shortint;
    h1, h2, count: integer;
    zi, zj: real;
    dependent: boolean;
begin
exit; // @@@
{  SignedContr_fn := 0.0; (* this saves many "else" statements below *)
  if ((i = j) or (dif = 0)) then exit;
  poss := 0; negg := 0;
  if ((ynv(i,j)+dif >= 0) and (ynv(i,j) >= 0)) then
    poss := dif
   else begin if ((ynv(i,j)+dif <= 0) and (ynv(i,j) <= 0)) then
    negg := -dif;
   end;
(* poss = after-before difference in number of positive ties;
   negg = after-before difference in number of negative ties;
   if dif <> 0, exactly one of these is <> 0. *)
  case h of
   1: begin
        SignedContr_fn := poss;
      end; (* case 1 *)
   2: begin
        SignedContr_fn := negg;
      end; (* case 2 *)
   3: begin  (* PP *)
        if (ynv(j,i) >= 1) then SignedContr_fn := poss;
      end; (* case 3 *)
   4: begin (* 00 *)
        if (ynv(j,i) = 0) then SignedContr_fn := -(negg + poss);
      end; (* case 4 *)
   5: begin (* NN *)
        if (ynv(j,i) <= -1) then SignedContr_fn := negg;
      end; (* case 5 *)
   6: begin (* PPP *)
         if symmetric then SignedContr_fn := poss*in2star^[i,j]
           else SignedContr_fn := poss*(1.0*in2star^[i,j] + 1.0*path2^[i,j]);
      end; (* case 6 *)
   7: begin (* PNN *)
         SignedContr_fn := 1.0*poss*in2starNN^[i,j] + 1.0*negg*path2PN^[i,j];
      end; (* case 7 *)
   8: begin (* NPN *)
         SignedContr_fn := negg*(1.0*in2starPN^[j,i] + 1.0*path2NP^[i,j]);
      end; (* case 8 *)
   9: begin (* NNP *)
         SignedContr_fn := 1.0*negg*in2starPN^[i,j] + 1.0*poss*path2NN^[i,j];
      end; (* case 9 *)
   10: begin (* PPN *)
         SignedContr_fn := 1.0*poss*in2starPN^[j,i] + 1.0*negg*path2^[i,j];
      end; (* case 10 *)
   11: begin (* PNP *)
         SignedContr_fn := poss*(1.0*in2starPN^[i,j] + 1.0*path2PN^[i,j]);
      end; (* case 11 *)
   12: begin (* NPP *)
         SignedContr_fn := 1.0*negg*in2star^[i,j] + 1.0*poss*path2NP^[i,j];
      end; (* case 12 *)
   13: begin (* NNN *)
         if symmetric then SignedContr_fn := negg*in2starNN^[i,j]
           else SignedContr_fn := negg*(1.0*in2starNN^[i,j] + 1.0*path2NN^[i,j]);
      end; (* case 13 *)
   end; (* case h *)

   count := 13;
   if (h <= count) then exit;
   if ((h >= count+1) and (h <= count+5*nzz)) then begin
     h1 := (4+h-count) div 5;     (* values 1...nzz; number of dyadic covariate       *)
     h2 := 1+((4+h-count) mod 5); (* values 1..5; number of dyadic covariate effect *)
     case h2 of
      1 : SignedContr_fn := poss*(zz[i,j,h1] - zzmean[h1]); (* positive ties          *)
      2 : SignedContr_fn := negg*(zz[i,j,h1] - zzmean[h1]); (* negative ties          *)
      3 : begin
          if (ynv(j,i) >= 1) then SignedContr_fn := poss*(zz[i,j,h1] - zzmean[h1]);
           end; (* case 3 *)
      4 : begin
          if (ynv(j,i) = 0) then SignedContr_fn := -(negg + poss)*(zz[i,j,h1] - zzmean[h1]);
           end; (* case 4 *)
      5 : begin
          if (ynv(j,i) <= -1) then SignedContr_fn := negg*(zz[i,j,h1] - zzmean[h1]);
           end; (* case 5 *)
     end; (* case h2 *)
   (* update number of checked effects: *)
    end; (* h *)
   count := count+5*nzz;
   if (h <= count) then exit;

   (* actor covariates: *)
   if ((h >= count+1) and (h <= count+6*(nz+nzc))) then begin
     h1 := (5+h-count) div 6;     (* values 1...(nz+nzc); number of actor covariate   *)
     h2 := 1+((5+h-count) mod 6); (* values 1...6; number of actor covariate effect  *)
     dependent := ((h1 > nz) and (h1 <= nz+nza)); (* flag: is chg.cov. h1 a dependent variable? *)
     if dependent then begin
       zi := ya(i,h1-nz);
       zj := ya(j,h1-nz)
      end else begin
       zi := z[i,h1];
       zj := z[j,h1]
     end;
    end; (* if h *)
   case h2 of
     1 : SignedContr_fn := poss*zi;  (* ego *)
     2 : SignedContr_fn := poss*zj;  (* alter *)
     3 : SignedContr_fn := poss*similarity(zi,zj,h1); (* similarity *)
     4 : SignedContr_fn := negg*zi;  (* ego *)
     5 : SignedContr_fn := negg*zj;  (* alter *)
     6 : SignedContr_fn := negg*similarity(zi,zj,h1); (* similarity *)
    end; (* case h2 *)
   count := count + 6*(nz+nzc);
   if (h > count) then begin
     Halterrorr('Procedure SignedContr_fn (S_Signed.pas included in S_Base) called with h = '+st(h)+').',outf,lf);
     halterr := true;
     exit;
    end;
    }
end; (* ModelSpecification.SignedContr_fn *)


function ModelSpecification.SignedContr_gn(i,j: integer; h: byte; dif: Shortint;
                        var halterr : boolean;var outf,lf : tFileStream): Real;
  (* The contribution of a change in yn(i,j), effect h, to gn, for the signed case *)
begin
  SignedContr_gn := 0.0;
end; (* ModelSpecification.SignedContr_gn *)


function ModelSpecification.SignedUtilityPlusGratificationGain(i,j: integer; dif: shortint; var halterr: Boolean;
                         var outf,lf: TFileStream): real;
(*  networkeffects only *)
(*  f(x(i \leadsto j)) - f(x); networkeffects only *)
var contribution, u, v: real;
    h, k: integer;
begin
  if ((i = j) or (dif = 0))
  then begin
    SignedUtilityPlusGratificationGain := 0.0;
    exit
  end;

  u := 0.0;
  for h := NetworkEffects_f downto 1 do begin
    if Effects_f[h] then begin
      contribution := SignedContr_fn(i,j,dif,h,halterr,outf,lf);
// geen changestat[2,h,j] voor MoM: derivative estimation, andere formules nodig
      if randomeffects then (*MS: actor-dependent random effects.*)
        begin
          k := res[1,h];
          if (k <> 0) then v := RE[i,k] (*<v> is the random effect of actor <i> corresponding to fixed effect <h> in the gratification function.*)
           else v := 0;
         end else v := 0;
        u := u + (alpa_f[h] + v)*contribution;
      end; (* if Effects_f[h] *)
    if halterr then begin
      SignedUtilityPlusGratificationGain := 0.0;
      exit
     end
   end; (* for h *)
  if (homogeneous in [11..19]) (*MS: actor dependent parameter in phase 3, indicated by <derivatives>, or generation of data in procedure <SimulateData>.*)
   then begin (*Note: for most other effects you would have to multiply <addgain> by <contr_fn>!*)
    if ((derivatives and (alternative <> 999999999)) or altern) then begin
      if outlier[i] then u := u + addgain(i,j) (*Monte Carlo generation of data.*)
      end
    end;
(* No multiplication by the direction of the change is necessary here,
   because of the way how SignedContr_fn is defined. *)

  if halterr then exit;
  for h := NetworkEffects_g downto 1 do
    begin
    if Effects_g[h] then
      begin
      contribution := SignedContr_gn(i,j,dif,h,halterr,outf,lf);
// weer geen changestat[3,h,j]
      if randomeffects then (*MS: actor-dependent random effects.*)
        begin
        k := res[2,h];
        if (k <> 0) then v := RE[i,k] (*<v> is the random effect of actor <i> corresponding to fixed effect <h> in the gratification function.*)
        else v := 0;
        end
      else v := 0;
      u := u + (alpa_g[h] + v)*contribution;
      if halterr then exit
      end;
    end; (* for h *)
  SignedUtilityPlusGratificationGain := u;
end;  (* ModelSpecification.SignedUtilityPlusGratificationGain *)



procedure ModelSpecification.SignedStatistics(var fun: pVec; var ordeg: intvec; k: integer);
  (* calculates the functions indicated in array Functions,
      for period from observation k to k+1 *)
var i,j,h,dif,count,count1: integer;
  np, nm, npp, nmm, n00: cardinal;
  nppp,nppn,npnp,npnn,nnpp,nnpn,nnnp,nnnn: cardinal;
  stats          : array[1..ccmax] of real;
  syzzm, syzzp, syzzmm, syzzpp, syzz00: array[1..nzzmax] of real;
  syzp_e, syzp_a, syzp_s, syzm_e, syzm_a, syzm_s: array[1..nzmax] of real;
  out1,in1,rec1, outt : IntVec;    (* outdegree, indegree and reciprocity counts at beginning of the periods *)
  FuncInd       : BBarray;
  calctriads    : boolean;

begin  (* SignedStatistics *)
(* This procedure is called ONLY if signed, replacing Statistics in S_Base. *)

  for i:=1 to ccmax do stats[i] := 0.0;
  for i:=1 to pmax do fun[i] := 0.0;
  for i:=1 to n do begin
    out1[i] := 0; in1[i] := 0; rec1[i] := 0; outt[i] := 0;
   end;

  (* transform 0-1 code of the effect-inclusion variable from the .MO-file
     to a false-true code in variable FuncInd *)
  FunctionsToBooleans(FuncInd);
  calctriads := false;
  for h:=5 to 12 do if FuncInd[observations+h] then calctriads := true;

  count1 := observations + 12 + 5*nzz + 6*(nz+nzc);

  (* compute indegree, outdegree and reciprocity counts at beginning of
     current period (variables ending with '1'): *)
  for i:=1 to n do for j:=1 to n do if (i <>j) then begin
    if (ynvm(i,j,k) <> 0) (* tie present at beginning of the current period k, of sign - or + *)
    then begin
      Inc(out1[i]); Inc(in1[j]);
      if (ynvm(j,i,k) <> 0) then Inc(rec1[j]);
     end;
   end; (* if i <> j *)

  (* change ynv by setting missings to 0.
     This is done in order not to have to check mis(i,j,.) so often.
     The changes are reset at the end of this procedure 'SignedStatistics'. *)
  for i:=1 to n do for j:=1 to n do
   if (mis(i,j,k) or mis(i,j,k+1)) then begin
     DefineSaveMat(i,j,ynv(i,j));
     setynv(i,j,0);
    end else DefineSaveMat(i,j,0);

  np := 0; nm := 0;
  npp := 0; nmm := 0; n00 := 0;
  nppp := 0; nppn := 0; npnp := 0; npnn := 0;
  nnpp := 0; nnpn := 0; nnnp := 0; nnnn := 0;
  for h :=1 to nzz do begin
    syzzm[h]  := 0;
    syzzp[h]  := 0;
    syzzmm[h] := 0;
    syzz00[h] := 0;
    syzzpp[h] := 0;
   end; (* h *)
  for h:=1 to (nz+nzc) do begin
    syzp_e[h] := 0;
    syzp_a[h] := 0;
    syzp_s[h] := 0;
    syzm_e[h] := 0;
    syzm_a[h] := 0;
    syzm_s[h] := 0;
   end; (* for h *)

(* Long i-j loop for all calculations *)
  for i:=1 to n do begin
    for j:=1 to n do begin
      if (ynv(i,j) <> 0) then begin
        inc(outt[i]);

        if (i < j) then begin
(* ties and dyads *)
          if (ynv(i,j) >= 1) then begin
            inc(np);
            if (ynv(j,i) >= 1) then inc(npp);
           end;
          if (ynv(i,j) <= -1) then begin
            inc(nm);
            if (ynv(j,i) <= -1) then inc(nmm);
           end;
         end (* if i < j *) else begin
          if (ynv(i,j) >= 1) then inc(np);
          if (ynv(i,j) <= -1) then inc(nm);
         end; (* if i > j *)

        if calctriads then for h:=n downto 1 do begin
(* triplets *)
          if (ynv(i,j) > 0) then begin
            if (ynv(j,h) > 0) then begin
              if (ynv(i,h) > 0) then inc(nppp) else if (ynv(i,h) < 0) then inc(nppn);
             end else if (ynv(j,h) < 0) then begin
              if (ynv(i,h) > 0) then inc(npnp) else if (ynv(i,h) < 0) then inc(npnn);
             end;
           end else (* ynv(i,j) < 0; note that there earlier was the condition ynv(i,j) <> 0 *) begin
            if (ynv(j,h) > 0) then begin
              if (ynv(i,h) > 0) then inc(nnpp) else if (ynv(i,h) < 0) then inc(nnpn);
             end else if (ynv(j,h) < 0) then begin
              if (ynv(i,h) > 0) then inc(nnnp) else if (ynv(i,h) < 0) then inc(nnnn);
             end;
           end; (*  if ynv(i,j) < 0 *)
          end; (* if calctriads for h *)

(* statistics pertaining to fixed actor covariates *)
       for h:= nz downto 1 do begin
         if not misz(i,h) (* only count non-missing dyadic covariate information *)
           then begin if (ynv(i,j) > 0) then begin
            syzp_e[h] := syzp_e[h] + z[i,h]; (* ego *)
            if (not misz(j,h)) then
              syzp_s[h] := syzp_s[h] + similarity(z[i,h],z[j,h],h); (* similarity *)
           end else (* ynv(i,j) < 0 *) begin
            syzm_e[h] := syzm_e[h] + z[i,h]; (* ego *)
            if (not misz(j,h)) then
             syzm_s[h] := syzm_s[h] + similarity(z[i,h],z[j,h],h); (* similarity *)
           end;
          end; (* if not misz(i,h) *)
         if not misz(j,h) then begin
           if (ynv(i,j) > 0) then
            syzp_a[h] := syzp_a[h] + z[j,h] (* alter *)
           else (* ynv(i,j) < 0 *)
            syzm_a[h] := syzm_a[h] + z[j,h]; (* alter *)
          end; (* if not misz(j,h) *)
        end; (* for h *)

(* statistics pertaining to changing actor covariates *)
       for h:= nzc downto 1 do begin
         if (not miszc(i,h,k)) (* only count non-missing dyadic covariate information *)
           then begin if (ynv(i,j) > 0) then begin
            syzp_e[nz+h] := syzp_e[nz+h] + zc[i,h,k]; (* ego *)
            if (not miszc(j,h,k)) then
              syzp_s[nz+h] := syzp_s[nz+h] + similarity(zc[i,h,k],zc[j,h,k],nz+h); (* similarity *)
           end else (* ynv(i,j) < 0 *) begin
            syzm_e[nz+h] := syzm_e[nz+h] + zc[i,h,k]; (* ego *)
            if (not miszc(j,h,k)) then
             syzm_s[nz+h] := syzm_s[nz+h] + similarity(zc[i,h,k],zc[j,h,k],nz+h); (* similarity *)
           end;
          end; (* if not miszc *)
         if (not miszc(j,h,k)) then begin
           if (ynv(i,j) > 0) then
            syzp_a[nz+h] := syzp_a[nz+h] + zc[j,h,k] (* alter *)
           else (* ynv(i,j) < 0 *)
            syzm_a[nz+h] := syzm_a[nz+h] + zc[j,h,k]; (* alter *)
          end; (* if not miszc  *)
        end; (* for h *)
       end; (* if ynv(i,j) <> 0 *)

       if ((i < j) and (not (mis(i,j,k) or mis(i,j,k+1)))) then begin
         if ((ynv(i,j) = 0) and (ynv(j,i) = 0)) then inc(n00);
        end; (* if *)

(* statistics pertaining to dyadic covariates *)
       for h:= nzz downto 1 do
        if not miszz(i,j,h) (* only count non-missing dyadic covariate information *)
         then case sgn(ynv(i,j)) of
          -1: begin
              syzzm[h] := syzzm[h] + (zz[i,j,h] - zzmean[h]);
              if (ynv(j,i) < 0) then
                syzzmm[h] := syzzmm[h] + (zz[i,j,h] - zzmean[h]);
              end; (* case -1 *)
          0: if ((not (mis(i,j,k) or mis(i,j,k+1) or mis(j,i,k) or mis(j,i,k+1)))
                 and (ynv(j,i) = 0)) then
                  syzz00[h] := syzz00[h] + (zz[i,j,h] - zzmean[h]);
          1: begin
              syzzp[h] := syzzp[h] + (zz[i,j,h] - zzmean[h]);
              if (ynv(j,i) > 0) then
                syzzpp[h] := syzzpp[h] + (zz[i,j,h] - zzmean[h]);
              end; (* case 1 *)
          end; (* case *)

     (* calculate the following only when dyad ij is observed both
        at the beginning *and* at the end of the current period  *)
      if (not (mis(i,j,k) or mis(i,j,k+1)))
      then begin
        dif := abs(ynv(i,j) - ynvm(i,j,k));
        if (dif >= 1) then begin
          stats[k]        := stats[k] + dif;
          stats[count1+1] := stats[count1+1] + dif*out1[i];
          stats[count1+2] := stats[count1+2] + dif*in1[i];
          stats[count1+3] := stats[count1+3] + dif*rec1[i];
          stats[count1+4] := stats[count1+4] + dif/(out1[i]+1);
          for h:=1 to nz do
            if (not misz(i,h)) then stats[count1+4+h] := stats[count1+4+h] + dif*z[i,h];
          for h:=1 to nzc do
            if (not miszc(i,h,k))
              then stats[count1+4+nz+h] := stats[count1+4+nz+h] + dif*zc[i,h,k];
         end; (* if dif >= 1 *)
      end; (* not (mis(i,j,k) or mis(i,j,k+1)) *)
    end; (* for j *)
  end; (* for i *)

  stats[observations]   := np;
  stats[observations+1] := nm;
  stats[observations+2] := npp;
  stats[observations+3] := n00;
  stats[observations+4] := nmm;
  stats[observations+5] := nppp;
  stats[observations+6] := npnn;
  stats[observations+7] := nnpn;
  stats[observations+8] := nnnp;
  stats[observations+9] := nppn;
  stats[observations+10] := npnp;
  stats[observations+11] := nnpp;
  stats[observations+12] := nnnn;
  count := observations + 12;
  for h :=1 to nzz do begin
    stats[count+5*(h-1)+1]:= syzzp[h];
    stats[count+5*(h-1)+2]:= syzzm[h];
    stats[count+5*(h-1)+3]:= syzzpp[h];
    stats[count+5*(h-1)+4]:= syzz00[h];
    stats[count+5*(h-1)+5]:= syzzmm[h];
   end; (* h *)
  count := count + 5*nzz;

  for h:=1 to (nz+nzc) do begin
    stats[count+6*(h-1)+1] := syzp_e[h];
    stats[count+6*(h-1)+2] := syzp_a[h];
    stats[count+6*(h-1)+3] := syzp_s[h];
    stats[count+6*(h-1)+4] := syzm_e[h];
    stats[count+6*(h-1)+5] := syzm_a[h];
    stats[count+6*(h-1)+6] := syzm_s[h];
   end; (* for h *)
  count := count + 6*(nz+nzc);
(* This must be equal to count1 *)

  (* copy statistics into shorter array fun[.] *)
  h := 1;
  repeat
    fun[h] := stats[Functions[h]];
    inc(h);
  until ((h > pmax) or (Functions[h] = 0));

  (*MS: test homogeneity assumptions.*)
  if ((homogeneous in [11..19]) and derivatives)
  then StatisticsSpecial(fun); (*Evaluate statistic for the outlier candidates using simulated digraph.*)
  (*MS*)

(* For the checks using the ordered outdegrees : *)
  ordeg := outt;
  quicksort(ordeg,1,n);

  (* Reset values ynv changed at the start of SignedStatistics: *)
  for i:=1 to n do for j:=1 to n do
   if (SaveMat(i,j) <> 0) then setynv(i,j,SaveMat(i,j));

end; (* ModelSpecification.SignedStatistics *)

procedure ModelSpecification.PlusTie(const i,j: integer);
var h: integer;
  oldval, newval: shortint;
begin
  if (i = j) then exit;
  if Struct(i,j,thisperiod)
  then begin
    ChangeTieStructError(i,j,1);
    exit;
  end; (* if *)
  if (ynv(i,j) >= valplus)
  then begin
    ChangeTieStructError(i,j,2);
    exit;
  end; (* if *)

  (* Put the change into effect *)
  oldval := ynv(i,j);
  newval := ynv(i,j)+1;
  setynv(i,j,newval);

  (* updates of running statistics *)

  if ((observations >= 2) and (ConditioningNetworkVariable >= 1))
    then begin
      if (not (mis(i,j,thisperiod) or mis(i,j,thisperiod+1)))
      then begin
        if (oldval < ynvm(i,j,thisperiod))
        then Dec(Distance)
        else Inc(Distance)
      end
    end;


    {for h:=nza downto 1
    do begin
      if (ya(j,h) > ya(i,h))
      then inc(zaplus[i,h])
      else begin
        if (ya(j,h) < ya(i,h))
        then inc(zamin[i,h])
        else inc(zaeq[i,h])
      end
    end;} (* for h *)

end; (* PlusTie *)

procedure ModelSpecification.MinTie(const i,j: integer);
var h: integer;
  oldval, newval: shortint;
begin
  if (i = j) then exit;
  if Struct(i,j,thisperiod)
  then begin
    ChangeTieStructError(i,j,-1);
    exit;
  end; (* if *)
  if (ynv(i,j) <= valmin)
  then begin
    ChangeTieStructError(i,j,-2);
    exit;
  end; (* if *)

  (* Put the change into effect *)
  oldval := ynv(i,j);
  newval := ynv(i,j)-1;
  setynv(i,j,newval);

  (* updates of running statistics *)

  if ((observations >= 2) and (ConditioningNetworkVariable >= 1))
    then begin
      if (not (mis(i,j,thisperiod) or mis(i,j,thisperiod+1)))
      then begin
        if (oldval > ynvm(i,j,thisperiod))
        then Dec(Distance)
        else Inc(Distance)
      end
    end;


    {for h:=nza downto 1
    do begin
      if (ya(j,h) > ya(i,h))
      then inc(zaplus[i,h])
      else begin
        if (ya(j,h) < ya(i,h))
        then inc(zamin[i,h])
        else inc(zaeq[i,h])
      end
    end;} (* for h *)

end; (* MinTie *)


(*[s]]*)
