{$if not defined(lazarus)}
{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q+,R+,S-,T-,U-,V+,W-,X+,Y+,Z1}
{$MINSTACKSIZE $00004000}
{$MAXSTACKSIZE $00100000}
{$IMAGEBASE $00400000}
{$ifend}
                                 (* S_Base.PAS *)

unit  s_base;

(* This unit contains the object ModelSpecification
   with functions and statistics used in SIENA.
   The fundamental data included are the basic variables for the model
   that can be written to, and recovered from, the project files. *)
(* Main model ingredients:
   lambda indicated by l, rate function;
   objective function f ;
   gratification function g. *)

interface

uses Forms, s_constants, classes, eight, slib, rangen, s_dat, s_matrices,
     digraph;
(* RanGen contains the variable version, which indicates the version
   of the random number generator *)

const stepnomax = 2000;

var
  RE: rrarray; (*MS: actor-dependent random effects in network objective and gratification functions.*)
  d_store, s_store, u_store: rrarray; (*Matrices that keep track of important information when <d_estA> is true.*)
  sc_sum, st_sum: rrarray;
  d_sum: rrrarray; (*Matrices that keep track of important information when <d_estB> is true.*)
  stopo : boolean; (* flag indicating external project termination *)
  changeparameter: boolean;
  extrafile: textfile;
//exf: textfile;

type
 realsteps = array[1..stepnomax] of real;
 ppString = array[1..pmax] of String[60];
 B3array = array[1..3] of Boolean;

 ModelSpecification = object(DataSpecification)
   time: real;
   (* time is the time parameter during the simulations. *)
   Distance: integer;
   (* The distance since the start of the simulations;
      on the dependent variable indicated by ConditioningNetworkVariable and
      ConditioningActionVariable *)
   n_act: integer;
   (* MH: number of active actors;
      TS 18-1-06: changed to number of structurally active actors. *)
   n_actnonmis: array[1..nzmax] of integer;
   (* n_actnonmis[h] = number of structurally active actors
      that have nonmissing values for dependent behavioral variable h
      at the beginning and end of this period *)
   NStructActiveElements: integer;
      (* number of elements in the adjacency matrix that are
         not structurally determined *)
   maxnfree: integer;
      (* maximal number of elements in the rows of the adjacency matrix
        that are not structurally determined; i.e., n-1 if there are
        no structurally determined elements at all;
        used only for observations = 1. calculated in initialise_actors. *)
   xino, nuno: real;
   IndependentStart: Boolean;
   (* For observations = 1, this indicates whether
      at the start of each new runstep, in procedure initialise_y,
      a new independent start is made, or the current value
      of the matrix yn is retained. *)
   CheckChange: Boolean; // @@@@
   (* This is true unless there are no structural zeros and maxdegree >= n
      and (staysametype = 2); if CheckChange = false, which is a nice standard
      situation, then it is superfluous to check if ties may be changed. *)
   Gibbsstring: shortstring;
   ModelTypeString: shortstring;
   (* Indicates the verbal explanations for the Gibbs and ModelType codes. *)
   Description: ppString;
   (* For the effects in the model,
      this indicates the name of the effects. *)
   StatsDescriptions: ppString;
   (* This describes the active statistics used in the model. *)
   active_m: array of array of boolean;
   (* MH: matrix indicating active actors (rows) at observation
            time m (columns) *)
   event_m: array of array of real;
   (* MH: matrix with event times for actors in period m; for
            active actors the event time is 0 *)
   targeta, facta, targetb, factb, avera, averb: Realnvec;
   (* used only for estimation in the VertexPar option *)
   action_cc: array of array of integer;
  (* action_cc: 0=active, 1=join (may leave after),  2=leave (may have joined
     before), 3=first leave and join again later;
     placed here in version 3.17u *)
   thisperiod: byte;
   (* The period; must be between 1 and observations. *)
     LikeCons: real;
   SumRateCons: real;
   theta_c: pvec;
   (*MS: MoM derivative estimation: copy of <theta> in <S_base>.*)
   mp: array of real;
   lambda1 : array of array of real;
   lambda2 : array of array of real;
   lambda4 : array of real;
   (* contribution to rate parameter dependent on actor covariates *)
   lambdar: array of real;
   (* used for reciprocity model with covariates only *)
   Statistics_l: integer;
   NetworkStatistics_l: integer;
   (* number of computed statistics / network statistics for rate of change. *)
   ActiveNetworkEffects_l: integer;
  (* Number of specified rate effects for the dependent network variable *)
   ActiveNetworkEffects_f: integer;
  (* Number of specified evaluation function effects for the dependent network variable *)
   ActiveNetworkEffects_g: integer;
  (* Number of specified endowment function effects for the dependent network variable *)
   ActiveActionEffects_l: integer;
  (* Number of specified rate effects for the dependent action variables *)
   ActiveActionEffects_f: integer;
  (* Number of specified evaluation function effects for the dependent action variables *)
   ActiveActionEffects_g: integer;
  (* Number of specified endowment function effects for the dependent action variables *)
   BeforeActionParameters_f: array of integer;
  (* BeforeActionParameters_f[v] = number of parameters in theta
      before the first evaluation parameter for dependent action variable number v *)
   BeforeActionParameters_g: array of integer;
  (* BeforeActionParameters_g[v] = number of parameters in theta
      before the first endowment parameter for dependent action variable number v *)
   ActionComputedFunction: BBArray;
   (* indicates the index numbers of the functions to be computed by
     ActionStatistics. *)
   BasicRateFunction:  BBArray;
   (* indicates the index numbers of the functions
      for estimating the basic rate parameters. *)
   EndowmentFunction: BBArray;
   (* indicates the index number of statistics defined for the endowment function;
     these get a minus sign for estimation *)
   DependentVariable: array[1..ccmax] of shortint;
   (* indicates the dependent variable to which the parameters refer *)
   AnyNetworkEffects_g: Boolean;
   (* Indicates whether the model includes any network endowment effects *)
   AnyActionEffects_g: array of Boolean;
   (* Indicates whether the model includes any behavior endowment effects for
      a given dependent variable *)
   MadeChangeTieError: boolean;
   ChangeTieErrorText: string;
   (* indicates that an error was made in ChangeTie *)
   (* TS: for calculation of entropies / explained variation : *)
   stepno: integer;
   entrop, entroptime: ^realsteps;
   entropy: array[0..100] of real;
   entropynumber: array[0..100] of integer;
   d_centring: boolean;
   (*MS: moment estimation, derivatives: centring of the statistics.*)
   moments: pvec;
   (*MS: observed moment of statistics.*)
   total_rate: rrarray;
   (*MS: total rate of actor.*)
   rates_na: R_Ar;
   (*MS: rates of network change and behavior change (for each behavioral dimension).*)
   cdscore: rrarray;
   (*MS: moment estimation: cdscore = complete-data score, needed to estimate derivatives.*)
   (*TS: cdscore has got a second index corresponding to the period. *)
   changestat: rrrarray;
   (*MS: moment estimation: stores the change in statistics when estimating derivatives.*)
   last_time: rrarray;
   (*MS: indicates the time point at which the last change happened.*)
    active: array of boolean;
    (* MH: indicates active actors at certain time in period k;
           if exo, this changes dynamically as a consequence of
           actors leaving and joining.      *)
    StructActive, IncludeActor: array of boolean;
    (* Structactive[i] = (Active[i] and (at least one value in row i of the
       adjacency matrix is not structurally determined)) *)
    (* IncludeActor[i] added 3.17u, = structactive[i] and
       not joiner or leaver in this period. *)
(* For ERGM = (observations = 1) case: *)
    SumGibbsChangeFraction: real;

  procedure InitialiseBase1(var outf,lf: TFileStream);
  procedure ChangeDefaults;
  procedure InitialiseBase2(var theta: pvec; var outf,lf: TFileStream);
  (* basic initialisation of this object *)

  procedure SetTheta(var theta: pvec);
  (* Sets theta to alpa *)

  procedure SetAlpha(var theta: pvec);
  (* Sets the alpa parameters to the corresponding thetas,
     and defines intermediate results for efficient calculations. *)

  procedure FunctionEvaluations(var orderdegs: intvec;
                       var halterr: boolean; var outf,lf: TFileStream; var targets: pvec);
  (* Used for calculation of target statistics.
     MS: formerly part of <InitialiseBase>, but not needed when synthetic
     data are generated.*)

  procedure ChangeAltern;
  (*MS: changes the value of global variable <altern>.*)

  procedure InitialiseActiveActors(var halterr: boolean;
                                   var outf, lf: TFileStream);
  (* MH: Determines active actors at start and end of each period
           and sets the values of ynm and mis at proper values *)
  procedure SetLeaversyBack(k: integer);
  (* MH: Changes the values of the yn matrix of actors who left back
           to the last observed values *)
  procedure SetLeaverszBack(k: integer);  // new, 3.17u
  (* Changes the values of the dependent action variables of actors who left.
     back to the last observed values *)
  procedure ReviseDefinition(var halterr: boolean; var outf,lf: TFileStream);
  procedure SetEventLengths;
  function SignedContr_fn(i,j: integer; dif: Shortint; h: byte;
                        var halterr : boolean;var outf,lf : TFileStream): Real;
  (* The contribution of yn(i,j), effect h, to fn, for the signed case *)
  function AddGain(var i, j: integer): real;
  (*Needed to estimate partial derivatives involving specified actor / period: as
  in tests of homogeneity assumptions: <addgain> = <espsilon> in <Polrup>.*)
  function Contr_fa(i: integer; h,k: byte; dif: shortint;
                    thecomponents: array_0_20;
                    var halterr: Boolean; var outf,lf: TFileStream): Real;
  (* The contribution of a change dif in ya(i,k) and Action effect h,
     to the objective function of i.
     First call CalcComponents_fa(i,NetworkEffects_f,ActionEffects_f,
        hstar,k,Effects_f,*** ) ! *)
  function Contr_ga(i           : integer;  (* actor            *)
                    h,                      (* effect number    *)
                    k,                      (* behaviour number *)
                    per         : byte;     (* period number *)
                    dif         : shortint; (* behaviour change *)
                    comp0, comp1: array_0_20;
                    var halterr : Boolean;
                    var outf,lf : TFileStream) : Real;
(* The contribution of a change dif in ya(i,k) and Action effect h,              *)
(* to the endowment function of i.                                               *)
  procedure UtilityComponents_ya(const i,v,h0: integer;
                             var s: rrarray; const perm: B3array;
                             var halterr : Boolean;
                             var outf,lf : TFileStream);
(* Adds the evaluation function contributions (all h) for changing behavior variable
   ya(i,v) by the amount d to array s[d+2,h0+h] , for d = -1, 0, 1;
   perm[1..3] indicate whether changes -1, 0, 1 are possible;
   note that j=i is a formal way of representing no change, hence contribution 0;
   first evaluation function, then endowment function *)
  function SignedContr_gn(i,j: integer; h: byte; dif: Shortint;
                        var halterr : boolean;var outf,lf : TFileStream): Real;
  (* The contribution of a change in yn(i,j), effect h, to gn, for the signed case *)
  function SignedUtilityPlusGratificationGain(i,j: integer; dif: shortint; var halterr: Boolean;
                         var outf,lf: TFileStream): real;
  (*  networkeffects only *)

  procedure DefineFunctions(var outf,lf: TFileStream);
  (* Defines the default choice of Functions, depending on the included effects;
    and the sequence of effects in NewNetworkEffects. *)
  procedure DefineSimFunctions(var outf,lf: TFileStream);
(* Defines the choice of array Functions (= target statistics),
   depending on the specified statistics. *)
  procedure Transform_l;
  (* Transforms alpa_l parameters for more efficient calculations,
     by defining the internal variables eff_l2 and eff_l3,exp_alpha_xxx, and lambda2 *)
  function NetworkLambda(i: integer; var halterr: boolean; var outf,lf: TFileStream): Real;
  (* The change rate for actor i. *)
  function ActionLambda(i,h: integer; var halterr: Boolean; var outf,lf: TFileStream): Real;
  (* The change rate for dependent actor variable j *)
  procedure GetNextevent;
  (* MH: Determines time of next exogenous event *)
//  procedure initialise_y_likel;
  (*MS: called in case of likelihood-based inference (MLE / Bayes):
  can be deleted as soon as likelihood-based inference admits changing composition
  and structural values and hence, <initialise_y> can be called.
  TS: deleted, 260907 *)

  procedure initialiseERGM(const nn, maxn: integer; const d: real; const ccond: boolean);
(* initialises variables for ERGM estimation *)
  procedure initialise_actors(const k: integer);
  (* initialises active, structactive, n_act, n_actnonmis, and changing covariates *)
  procedure initialise_y(const k: integer);
  (* Initialises matrix yn; for observations >= 2: sets yn to ynm[k];
     for observations = 1: depends on IndependentStart.
     Must have been preceded by initialise_actors.
     In the case of MoM inference,
     for longitudinal simulations where exo is possible,
     initialise_y must have been immediately preceded by initialise_actors;
     in S_ML, where exo is impossible so that active and structactive are
     changed only in initialise_actors, other data changes may have intervened.*)
  procedure initialise_y_end(const k: integer); // @@@@@
  procedure ChangeTie(const i,j: integer);
  (* Changes yn(i,j) and carries out all updates *)
  procedure PlusTie(const i,j: integer);
  (* Increases ynv(i,j) by 1 and carries out all updates *)
  procedure MinTie(const i,j: integer);
  (* Decreases ynv(i,j) by 1 and carries out all updates *)
  procedure ChangeTieStructError(i,j: integer; d: shortint);
  (* Reports an error from ChangeTie *)
  procedure ChangeBehavior(const i: integer; const h: byte; const dif: shortint);
  (* Changes ya(i,h) and carries out all updates *)
  procedure ChangeComposition;
  (* MH: Defines new set of active actors at certain time point
           and sets time of next event *)
  procedure ActionStatistics(var fun: pVec; var ordeg: intvec;
                       k: integer; var outf: TFileStream);
  (* calculates the functions indicated in array Functions
     for period from observation  k to k+1 *)
  procedure NetworkStatistics(var fun: pVec;
        var ordeg: intvec; k: integer; var outf: TFileStream);
  (* calculates the functions indicated in array Functions
     for period from observation  k to k+1 *)
  procedure SignedStatistics(var fun: pVec; var ordeg: intvec; k: integer);
  (* The same, for the signed case *)

  function GetRE(const function_i: byte; const ldim, actor: integer): real;
  (*MS: Get the corresponding random effect.*)
  function ComputeCoefficient(const theta: pvec; const actor, dim: integer; const actorRE: rrarray): real;
  (*MS: Compute coefficient, which may be fixed or random.*)
  function Multinomial(const ldim1, ldim2: integer; const theta: pvec; const actorRE: rrarray; const actor, alternative: integer; const statistic: rrarray): real;
  (*MS: Evaluate multinomial choice probability (up to the the normalizing constant) on natural logarithmic scale.*)
  procedure ChangeTheta(var outf: TFileStream; var theta: pvec);
  (*MS: to estimate / test misspecified models, manipulate the value of the (data-generating) parameter
  by utilizing user input.*)
  function Permission: boolean;
  (*MS: By default, grant permission to continue the simulation, unless
  - the estimation method is unconditional moment estimation
  - and time > 1.*)
  function Sign_n(i, j: integer): real;
  (*MS: The sign of the change in statistic which results from one particular change in the network.*)
  function SoLiS: boolean;
  (*MS: "Model specification with Limited Support".*)
  procedure Initialise_d_est(var ascii: TFileStream; const phase: byte; const total_n: cardinal; const dim: byte; const theta: pvec; const override_default: boolean);
  (*MS: method of moments: initialise derivatives estimation.*)
  function d_rate_rho(const dimension: byte): real;
  (*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
  distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
  function d_rate_alpha(var statistic: R_Ar; const dimension: byte): real;
  (*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
  distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
  function d_rate(const rho: boolean; const dimension: byte; var statistic: R_Ar): real;
  (*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
  distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
  function d_loglambda_rho(const dimension: byte): real;
  (*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
  distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
  function d_loglambda_alpha(var statistic: R_Ar; const ego: integer): real;
  (*MS: Evaluates the partial derivative of the rate function with respect to coordinate <k> of the rate parameter.*)
  function d_loglambda(const rho: boolean; const dimension: byte; var statistic: R_Ar; const ego: integer): real;
  (*MS: Evaluates the partial derivative of the rate parameter of the negative exponential
  distribution of the holding times with respect to coordinate <k> of the rate parameter.*)
  function ratio_na: real;
  (*MS: Evaluates, in the case where both the network and some behavioral dimensions change,
  some function of the rates, which is needed in derivative estimation.*)
  procedure StoreRates(lambda_n: real; lambda_a: real; changed_dimension: byte);
  (*MS: Stores the rates of network change and behavior change (for each behavioral dimension).*)
  procedure CompleteStat;
  (*MS: Complete the statistic, needed in derivative estimation.*)
  function Prepare_Score(var rates: array of real): real;
  (*MS: Prepare the evaluation of the complete-data score.*)
  procedure CompleteScore;
  (*MS: unconditional MoM: at the end of each period, complete the complete-data score corresponding
  to the rate parameters.*)
  procedure StoreScoreI(var b: array of real; ego, alter: integer);
  (*MS: Store the complete-data score corresponding to the change made: network objective and endownment function.*)
  procedure StoreScoreII(var b: array of real; var changed: byte; var direction: integer;
                         var minp,plup: boolean);
  (*MS: Store the complete-data score corresponding to the change made: action objective and endowment function.*)
  procedure StoreScoreIII(const variable: byte; const htime: real; const ego: integer);
  (*MS: Evaluate complete-data score corresponding to the network rate function parameters.*)
  function E_Score_Rates(const rho: boolean; const parameter, dimension: byte; const htime: real; const condition: boolean; const ego: integer): real;
  (*MS: Evaluate the complete-data score corresponding to the parameters in the rate functions
  of the behavioral outcome variables.*)
  procedure StoreScoreIV_k(const dimension: byte; var count: byte; const htime: real;
  const condition: boolean; const ego: integer);
  (*MS: Evaluate complete-data score corresponding to the behavior rate function parameters.*)
  procedure StoreScoreIV(const variable: byte; const htime: real; const ego: integer);
  (*MS: Evaluate complete-data score corresponding to the behavior rate function parameters.*)
  procedure Initialise_D(const dim: byte);
  (*MS: Initialise arrays needed to estimate the derivatives of the estimating moment function.*)
  procedure StoreScore(var dim: byte; var statistic: pvec; var d_store: rrarray);
  (*MS: The integrand, needed to estimate the derivatives matrix, is evaluated at the present Monte Carlo sample,
  and stored.*)
  procedure SumScore
          (var dim: byte;  const ob: integer; var statistic: pvec);
  (* TS: analogous to StoreScore for d_estimator = 1 *)
  procedure Estimate_D_EU(var cdim: cardinal; var dim: byte; var d_store: rrarray; var matrix1: pmatrix);
  (*MS: Estimate the derivatives of <E(U | theta)> evaluated at the moment estimate of the parameter.*)
  procedure JacobianMatrix(var dim: byte; var ascii, outf: TFileStream; var matrix1: pmatrix;
                           var warning: integer);
  (*MS: Moment estimation: estimate Jacobian matrix by various score function / likelihood ratio based methods:
  see Schweinberger and Snijders (2005).*)
  (*MS: Comment:
  For the Independent Arcs model, some procedures follow which
  - implement the score function / likelihood ratio method of derivative estimation,
  - evaluate the derivatives analytically.
  It is helpful to look at the Independent Arcs version of the source code in <.../programs/IA>.
  *)
  function Ratio(const maximum_outdegree, outdegree: integer; const theta: pvec): extended;
  (*MS: Compute <(part1 - part2) / (part1 + part2)>,
  where
    <part1 = (maximum_outdegree - outdegree) * exp(theta[2])>
  and
    <part2 = outdegree * exp(- theta[2])>,
  where <outdegree> may represent an arc from <i> to <j>, the outdegree of <i>, or the total outdegree, depending on the input,
  and <maximum_outdegree> represents the maximum the outdegree can attain.
  *)
  procedure CD_Score_IA(const i, j: integer; const q: real; const theta: pvec; const holding_time: real);
  (*MS: Complete-data score: Independent Arcs (IA) model.*)
  procedure CompleteScore_IA(const theta: pvec);
  (*MS: Complete complete-data score: Independent Arcs (IA) model.*)
  procedure Jacobian_IA(const theta: pvec; const M_0, M_1: integer);
  (*MS: Exact (!) Jacobian matrix of estimating moment function: Independent Arcs (IA) model;
  note that the Jacobian matrix is not estimated here, because it can be evaluated analytically;
  see the paper on estimating Jacobian matrices by Schweinberger and Snijders.*)
  procedure Output_d_multiple(var dimension: integer; var estimate: R_Ar);
  (*MS: Multiple sequences output: Jacobian matrix.*)
  procedure StatisticsSpecial(var fun: pvec);
  (*MS: Evaluate the statistic for the outlier candidates using the simulated digraph.*)
  procedure ObservedMomentPeriod(var ystat: pVec; var outf: TFileStream);
  (*MS: computes the observed first moments of specified statistics for period <period>.*)
  procedure PPChecks(const iteration: integer);
  (*MS: posterior predictive checks: store simulated networks and behavior and compute functions of data.*)

  procedure InitialiseAlpha(var outf: TFileStream; var theta: pvec; var ss: integer);
  (* Sets the alpa parameters to the corresponding thetas, prepares
     internal variables, and initialises the random number generator with ss. *)
  function readstate: integer;
  (* gives the current state of the random number generator *)
  function readstate2: integer;
  (* gives the current state of the secondary random number generator *)
  procedure makestate(var k: integer);
  (* sets the state of the random number generator to k *)
(* AA change *)  function ERGMnrun(g: byte): integer ;
  procedure ChoiceProbabilities(var pr: array of real;
       const i: integer; const permitted: array of boolean);
(* Computes logarithms of (change probabilities * constant) *)
  procedure runstep(var error: Boolean; var erstr: shortstring;
                    var halterr : Boolean; var outf,lf: TFileStream);
  procedure ErgmRunStep(var error: Boolean;
                             var erstr: shortstring; var halterr: Boolean;
                             var outf,lf: TFileStream);
  procedure runepoch(t: real; var success: Boolean;
                     updatestat: Boolean; var error: Boolean;
                     var erstr: shortstring; var halterr: Boolean;
                     var outf,lf: TFileStream);
  (* if (observations >= 2) and CConditional: runsteps until
     Distance exceeds t;
     if (observations >= 2) and (not CConditional) runsteps until
     time exceeds t *)

  procedure SimStats(p: Byte; theta: pVec;
                  allstats, varmat: boolean; var fseed: integer;
                  var ystat: pVec; var d: array of real;
                  var dd: pmatrix;
                  updatestat: Boolean;
                  var feasible, error: Boolean;
                  var errstr: shortstring; var outf,lf: TFileStream);
  (* This procedure is used in S_Sim,
     and - through FRAN - in S_Est in POLRUP.
     It simulates one run and stores the outcome in ystat
     and the simulated times in d.
     Secondary simulation results updated if updatestat is true. *)
  procedure SimOnce(kk: integer; extra: boolean; p: Byte; theta: pVec;
                  allstats, varmat: boolean; var fseed: integer;
                  var ystat: pVec; var feasible, error: Boolean;
                  var errstr: shortstring; var outf,lf: TFileStream);
(* Utility for once simulating
   a data set with given parameter values.
   No other procedures utilizing the original data set should follow.*)


  procedure GiveDescriptions;
  (* Defines array Description *)
  procedure DefineStatsNames;
  (* Defines array StatsDescriptions *)
  procedure HaltErrorr(s: string; var f, lf: TFileStream);
  (* Next five procedures used only in VertexPar option *)
  procedure VertexStep(ga: real; var outf: TFileStream);
  procedure InitialiseVertexArrays;
  procedure ReportVertexParameters(var outf: TFileStream);
  procedure InitialiseVertexAverage;
  procedure VertexAverage(nai: Cardinal);

  
  procedure DescribeModeloptions(var outf,lf: TFileStream);
  (* This procedure produces a description of the model options.
     It is not used for running any further calculation or program. *)

  function PseudoDeviance(theta: pvec; var halterr: Boolean;
                          var outf,lf: TFileStream): real;
(* The next two procedures are implemented in S_Ergm *)
  function CheckRunningStatistics(var outf, lf: TFileStream): boolean;
(* Note that in the ERGM case, statistics are calculated by updating,
   based on change statistics.
   The procedure CheckRunningStatistics checks whether these updated values
   are identical to what is produced by ERGMStatistics. *)
  function RandomVertexInComponent
                            (const i: integer; const SameAllowed: boolean): integer;
(* Selects a random vertex in the same component as i;
  if not SameAllowed, it must be but different from i. *)
   function xi(const h: integer): real;
   function nu(const h: integer): real;
    (* contributions to rate parameter for ModelType = 2, not symmetric 
       storage of intermediate parameter functions for ModelType = -1, 1.  *)

  procedure DeActivateModel(var outf: TFileStream);

  private
    {
    zaplus, zaeq, zamin: array of array of integer;
    }
    (* zaplus[i,j] = sum_h Yn_ih Indicator{ya[h,j] > ya[i,j]}
       zaeq[i,j] = sum_h Yn_ih Indicator{ya[h,j] = ya[i,j]}
       zamin[i,j] = sum_h Yn_ih Indicator{ya[h,j] < ya[i,j]} *)
    eff_l3, eff_l2 : array of boolean;
    eff_l4         : boolean;
    (* indicate whether the model includes non-trivial degree effects
       and covariates effects on rate *)
    CalcEntropy: boolean;
    (* indicates whether entropies are to be calculated *)
    exp_alpha_out, exp_alpha_in, exp_alpha_rec: array of array of real;
    (* Transformed alpa_l parameters for degree effects on rate *)
    lambdar_cons, lambdar_in, lambdar_out, lambdar_rec: real;
    (* Transformed alpa_f parameters for the Reciprocity model. *)
    event: array of real;
    (* MH: event times in period k *)
    i_act: integer;
    (* MH: i_act = actor who joins or leaves in period k
             n_act = number of active actors at certain time in period k *)
    nextevent: real;
    (* MH: time of next event in period k *)
    istar, jstar: integer;

(* Some variables used only for the ERGM (observations=1) case *)
    GibbsChange: integer;
     (* current numbers of cell for Gibbs sampling, and numbers of changes
       made in this epoch *)
    GibbsChangeFraction: real;
    probA: real; (* used in S_ergm_est.ERGMRunStep for selecting (i,j) in version A *)
    lograt: array of real;  (* used for proposal probabilities in S_ergm_est.ERGMRunStep *)
    CurrentStats: pvec;
    (* Current values of the sufficient statistics in the ERGM case *)
    ind, outd, pin, pout: array of integer; (* observed indegrees and outdegrees *)

(* For observations >= 2, modeltype = 2: *)
    store_nu, store_xi: array of real;

    procedure writestrings(k: byte; st: array of shortstring; fx: shortstring);
(* writes text for communication with Stocnet *)

    function ErgmUtilityGain(i,j: integer; var halterr: Boolean;
                      var outf,lf: TFileStream): real;
(*  f(x(i \leadsto j)) - f(x) for the observations=1 case *)

end; (* ModelSpecification *)

function FoundError: boolean;


implementation


uses SysUtils;

{$I s_phase3} (*MS: Input file <s_phase3>.*) 

var
 MadeError: boolean;
 CheckStatistics: boolean;

procedure ModelSpecification.writestrings(k: byte; st: array of shortstring; fx: shortstring);
(* writes text for communication with Stocnet *)
var f: textfile;
    i: byte;
    fn : shortstring;
begin
  fn := pname+'.'+fx;
  assign(f,fn);
  rewrite(f);
  if (k >= 10) then
    for i:=0 to k do writeln(f,i:2,'. ',st[i])
   else
    for i:=0 to k do writeln(f,i:1,'. ',st[i]);
  closefile(f);
end;

procedure ModelSpecification.InitialiseBase1(var outf,lf: TFileStream);
(* basic initialisation of this object;
   if target statistics are required,
   this procedure must be followed by FunctionEvaluations;
   note that this is used after reading the data and definition files. *)
var i,j, Modelcode,ns: integer;
   modelstrings0, modelstrings: array of shortstring;
   str: shortstring;
   x: string;
begin
  wwriteln(outf,'Siena version '+Sienaversion);
  {
  SetLength(zaplus,n+1,nza+1);
  SetLength(zaeq,n+1,nza+1);
  SetLength(zamin,n+1,nza+1);
  }
  n := nmax;
  SetLength(exp_alpha_out,nza+1,n+1);
  SetLength(exp_alpha_in,nza+1,n+1);
  SetLength(exp_alpha_rec,nza+1,n+1);
  SetLength(eff_l2,nza+1);
  SetLength(eff_l3,nza+1);
  SetLength(store_xi,n+2);
  SetLength(store_nu,n+2);
  SetLength(lambda1,nza+1,observations);
  SetLength(lambda2,nza+1,n+1);
  SetLength(lambda4,n+1);
  SetLength(active,n+1);
  SetLength(StructActive,n+1);
  SetLength(IncludeActor,n+1); // added 3.17u
  SetLength(mp,n+1);
  time := 0.0;
  distance := 0;
  n_act := n;
  CalcEntropy := false;
  thisperiod := 1;
  istar := 1; jstar := 2;
  IndependentStart := true;
  LikeCons := 0.0;
  SumRateCons := 0.0;
  SumGibbsChangeFraction := 0.0;
  MadeChangeTieError := false;
  ChangeTieErrorText := '';
  Gibbsstring := '';
  ModelTypeString := '';
  param1_clique_s := 0;
  param2_clique_s := 0;
  param3_clique_s := 0;
  param4_clique_s := 0;
  param1_rclique_s := 0;
  param2_rclique_s := 0;
  param1_clique := 0;
  param2_clique := 0;
  param3_clique := 0;
  param4_clique := 0;
  param1_rclique := 0;
  param2_rclique := 0;
  (* by default include basic rate parameters... *)
  for i:=1 to (observations-1)
  do begin
    effects_l[i] := (not skipi[i]);
    for j := 1 to nza
    do effects_l[NetworkEffects_l+(j-1)*ActionEffects_l+i]:=(not skipi[i]);
  end;
  (* ...but change this in case of conditional estimation: *)
  if Cconditional
  then for i:=1 to (observations-1)
  do begin
    if ConditioningNetworkVariable=1
    then effects_l[i] := false
    else effects_l[NetworkEffects_l+(ConditioningActionVariable-1)*ActionEffects_l+i]:=false
  end;
  ActiveNetworkEffects_l := 0;
  for i:=1 to NetworkEffects_l do if effects_l[i] then inc(ActiveNetworkEffects_l);
  ActiveNetworkEffects_f := 0;
  for i:=1 to NetworkEffects_f do if effects_f[i] then inc(ActiveNetworkEffects_f);
  ActiveNetworkEffects_g := 0;
  for i:=1 to NetworkEffects_g do if effects_g[i] then inc(ActiveNetworkEffects_g);
  AnyNetworkEffects_g := (ActiveNetworkeffects_g >= 1);

  ActiveActionEffects_l := 0;
  for i:= NetworkEffects_l + 1 to NetworkEffects_l + nza*ActionEffects_l do
                         if effects_l[i] then inc(ActiveActionEffects_l);
  ActiveActionEffects_f := 0;
  for i:= NetworkEffects_f + 1 to NetworkEffects_f + nza*ActionEffects_f do
                         if effects_f[i] then inc(ActiveActionEffects_f);
  ActiveActionEffects_g := 0;
  SetLength(AnyActionEffects_g,nza+1);
  for j:=1 to nza do
    begin
    AnyActionEffects_g[j] := false;
    for i:=1 to ActionEffects_g do
      if effects_g[NetworkEffects_g + (j-1)*ActionEffects_g + i] then
        begin
        inc(ActiveActionEffects_g);
        AnyActionEffects_g[j] := true;
        end;
    end;

  if (nza <= 0) then setlength(BeforeActionParameters_f,2)
    else setlength(BeforeActionParameters_f,nza+2);
  BeforeActionParameters_f[0] := 0;
  BeforeActionParameters_f[1] := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
               ActiveNetworkEffects_g + ActiveActionEffects_l;
  for j:=2 to nza do
    begin
    BeforeActionParameters_f[j] := BeforeActionParameters_f[j-1];
    for i:= 1 to ActionEffects_f do
            if effects_f[NetworkEffects_f + (j-2)*ActionEffects_f+i]
            then inc(BeforeActionParameters_f[j]);
    end;
  BeforeActionParameters_f[nza+1] := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
               ActiveNetworkEffects_g + ActiveActionEffects_l + ActiveActionEffects_f;

  if (nza <= 0) then setlength(BeforeActionParameters_g,2)
    else setlength(BeforeActionParameters_g,nza+2);
  BeforeActionParameters_g[0] := 0;
  BeforeActionParameters_g[1] := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
               ActiveNetworkEffects_g + ActiveActionEffects_l + ActiveActionEffects_f;
  for j:=2 to nza do
    begin
    BeforeActionParameters_g[j] := BeforeActionParameters_g[j-1];
    for i:= 1 to ActionEffects_f do
            if effects_g[NetworkEffects_g + (j-2)*ActionEffects_g+i]
            then inc(BeforeActionParameters_g[j]);
    end;
  BeforeActionParameters_g[nza+1] := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
               ActiveNetworkEffects_g + ActiveActionEffects_l + ActiveActionEffects_f
                + ActiveActionEffects_g;

(* write text for communication with Stocnet:
   must be called by Siena01 and Siena04.  *)

  if (observations <= 1) then
    begin
      SetLength(modelstrings0,13);
      SetLength(modelstrings,27);
      Modelstrings0[1] := 'Gibbs for single ties';
      Modelstrings0[2] := 'Gibbs for dyads';
      Modelstrings0[3] := 'Gibbs for triplets';
      Modelstrings0[4] := 'M.H. for single ties, A';
      Modelstrings0[5] := 'M.H. for single ties, B';
      Modelstrings0[6] := 'M.H. nondirected';
      Modelstrings0[7] := 'M.H. anti-symmetric';
      Modelstrings0[8] := 'M.H., fixed in- and out-degrees';
      Modelstrings0[9] := 'M.H., fixed out-degrees, A';
      Modelstrings0[10] := 'M.H., fixed out-degrees, B';
      Modelstrings0[11] := 'M.H., fixed in-degrees NOT YET IMPLEMENTED, A';
      Modelstrings0[12] := 'M.H., fixed in-degrees NOT YET IMPLEMENTED, B';
      Modelstrings[0] := 'not implemented';
      for i:=1 to 7 do
        begin
        str := ModelStrings0[i];
//        ModelStrings[i] := str+', independent restarts';
        ModelStrings[i] := str+', independent';
        end;
      for i:=8 to 10 do Modelstrings[i] := 'not implemented';
      for i:=1 to 12 do
        begin
        str := ModelStrings0[i];
        ModelStrings[i+10] := str+', continuous';
        end;
      Modelstrings[8] := 'M.H., fixed in- and out-degrees, continuous';  // deviating from default  for i <= 10
      Modelstrings[26] := 'M.H. nondirected, connected, continuous';
(* Independent option is not available for options >= 11;
   but this is not important in practice. *)
      ns := 22;
      Gibbsstring := ModelStrings[Gibbscode];
(* This section should be consistent with what is done in ReviseDefinition *)
//      SymmetricGibbs := [16];
//      DefaultSymmetricGibbs := 16;
    end
   else
    begin (* observations >= 2 *)
    if symmetric then
      begin
      SetLength(modelstrings,7);
      ModelStrings[0] := 'not implemented';
      Modelstrings[1] := 'Forcing model';
//      Modelstrings[2] := 'Unilateral initiative and reciprocal confirmation';
      Modelstrings[2] := 'Initiative - confirmation';
      Modelstrings[3] := 'Tie-based model';
      Modelstrings[4] := 'Pairwise conjunctive model';
      Modelstrings[5] := 'Pairwise disjunctive model';
      Modelstrings[6] := 'Pairwise compensatory model';
      ns := 6;
      end
     else
      begin
      if signed then
       begin
       SetLength(modelstrings,2);
       ModelStrings[0] := 'not implemented';
       Modelstrings[1] := 'Actor-oriented for signed digraphs';
       ns := 1;
       end
       else
       begin
       SetLength(modelstrings,4);
//       Modelstrings[0] := 'Actor-oriented model; change required at each network micro-step';
       Modelstrings[0] := 'Actor-oriented model; change required';
       Modelstrings[1] := 'Standard actor-oriented model';
//       Modelstrings[2] := '2-step actor-oriented model: degrees - choices (currently not implemented)';
       Modelstrings[2] := '2-step actor-oriented: not implemented';
       Modelstrings[3] := 'Tie-based model';
       ns := 3;
      end;
      end;
    end; (* if observations >= 2 *)
  if (CallingProgram in [1,4]) then writestrings(ns,modelstrings,'cdc');

  if (observations >=2) then
    begin
    ModelCode := ModelType;
    if (abs(ModelCode) >= 10) then
      begin
      OutputType := 2;
      ModelCode := abs(ModelCode) - 10;
      if (ModelType < 0) then ModelCode := - ModelCode;
      end;

    if ((ModelCode = 0) or (StaySameType=1)) then
      begin
      ModelType := 1;
      StaySameType := 1;
      end
     else
      begin
      ModelType := ModelCode;
      StaySameType := 2;
      end;
    if (symmetric) then
      begin
      if not (ModelType in [1..6]) then
        begin
        x :='For non-directed networks, only Model Types 1 to 6 are possible.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%d%s',['Model Type ',ModelType ,' changed to 2.'])+eol;
        outf.Write(pchar(x)^,length(x));
        ModelType := 2;
        end;
      end
     else
      begin (* not symmetric *)
      if (not (ModelType in [0..3])) then
        begin
        x := 'For directed networks, only Model Types 0 to 3 are possible.'+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%d%s',['Model Type ',ModelType ,' changed to 1.'])+eol;
        outf.Write(pchar(x)^,length(x));
        ModelType := 1;
        end;
      end; (* if not symmetric *)
    ModelTypeString := ModelStrings[ModelType];
    end; (* if observations >= 2 *)

(* make only derivative options 0 and 1 available from Stocnet: *)
    if (observations <= 1) then
      begin
      ModelStrings[0] := 'derivative method not relevant for ERGM';
      if (CallingProgram in [1,4]) then writestrings(0,modelstrings,'cdd');
      end
     else
      begin
      ModelStrings[0] := 'crude Monte Carlo';
      ModelStrings[1] := 'score function 1';
      if (CallingProgram in [1,4]) then writestrings(1,modelstrings,'cdd');
      end;

    if (observations <= 1) then
      begin
      ModelStrings[0] := 'unconditional Maximum Likelihood';
      ModelStrings[1] := 'conditional Maximum Likelihood';
      if (CallingProgram in [1,4]) then writestrings(1,modelstrings,'cde');
      end
     else
      begin
      if (nza >= 1) then ns := 20+nxa+nza else ns := 3;
      if high(modelstrings) <= ns then SetLength(modelstrings,ns+1);
      for i:=4 to ns do ModelStrings[i] := '---';
      ModelStrings[0] := 'unconditional Method of Moments';
      ModelStrings[1] := 'conditional Method of Moments';
      ModelStrings[2] := 'Maximum Likelihood';
      ModelStrings[3] := 'Bayesian';
      if (nza >= 1) then
        begin
        ModelStrings[1] := 'conditional Method of Moments (given network distance)';
        for i:=1 to nxa do ModelStrings[20+i] :=
          'Method of Moments, conditional on network distance';
        for i:=1 to nza do ModelStrings[20+nxa+i] :=
          'Method of Moments, conditional on distance for behavior '+inttostr(i);
        end;
      if (CallingProgram in [1,4]) then writestrings(ns,modelstrings,'cde');
      end;
end; (*InitialiseBase1. *)


procedure ModelSpecification.ChangeDefaults;
(* carried out only in Siena01 after Preparations;
   change of default options depending on reading the data *)
  (* Set some defaults for effect inclusion: *)
var nxx, lupo: integer;
begin
  if ((observations = 1) and (symmetric or asymmetric)) then effects_f[16] := true;
  (* This is to ensure that at startup, there is at least one effect in the model. *)
  if (observations = 1) then
    begin
    if symmetric then Gibbscode := 16 else Gibbscode := 14;
    end;
  if (observations>1)
    then effects_f[1] :=  (not (uponly[0] or downonly[0])); (* density *)
  if (not symmetric) then effects_f[2]:=true; (* reciprocity *)
  for lupo := 1 to nza do (* behavioral tendency: *)
      effects_f[Networkeffects_f+(lupo-1)*ActionEffects_f+1] := true;

  (* initialise: *)
{  for lupo :=1 to ccmax do param_effect[lupo]:=0;
  if ((observations >= 1) and (not signed)) then begin (*[s]*)
    (* outdegree up to <param>: *)
    param_effect[8]:= 0;
    (* 1 / (outdegree + <param>): *)
    param_effect[11] := 1;
    (* 1 / ((outdegree + <param>)(outdegree + <param>+1)): *)
    param_effect[12] := 1;
    (* isolate status = indegree up to <param>: *)
    param_effect[14] := 0;
    (* dense triad = <param> or more ties in triad: *)
    param_effect[15] := 6;
    for lupo:=1 to nz+nzc
    do begin
      (* similarity  network effects *)
      param_effect[16+2*nzz+10*(lupo-1) + 1] := 1;
      param_effect[16+2*nzz+10*(lupo-1) + 2] := 2;
      (* similarity  covariate effects *)
      param_effect[16+2*nzz+10*(lupo-1) + 3] := 1;
      param_effect[16+2*nzz+10*(lupo-1) + 4] := lupo;
      (* cov.-ego  network effects *)
      param_effect[16+2*nzz+10*(lupo-1) + 5] := 1;
      param_effect[16+2*nzz+10*(lupo-1) + 6] := 2;
      (* cov.-ego  covariate effects *)
      param_effect[16+2*nzz+10*(lupo-1) + 7] := 1;
      param_effect[16+2*nzz+10*(lupo-1) + 8] := lupo
    end; (* lupo-loop from 1 to nz+nzc *)
  end; (* if *)
}

  (* default is *unconditional* estimation for models with dependent behavioural variable: *)
  if nza>=1 then Cconditional:=false;
end; (* changedefaults *)



procedure ModelSpecification.InitialiseBase2(var theta: pvec; var outf,lf: TFileStream);
(* basic initialisation of this object;
   if target statistics are required,
   this procedure must be followed by FunctionEvaluations;
   note that this is used after reading the data and definition files. *)
var i,j: integer;
  od: array of integer;
  x: string;
begin
  ActiveNetworkEffects_l := 0;
  for i:=1 to NetworkEffects_l do if effects_l[i] then inc(ActiveNetworkEffects_l);
  ActiveNetworkEffects_f := 0;
  for i:=1 to NetworkEffects_f do if effects_f[i] then inc(ActiveNetworkEffects_f);
  ActiveNetworkEffects_g := 0;
  for i:=1 to NetworkEffects_g do if effects_g[i] then inc(ActiveNetworkEffects_g);
  AnyNetworkEffects_g := (ActiveNetworkeffects_g >= 1);
  ActiveActionEffects_l := 0;
  for i:= NetworkEffects_l + 1 to NetworkEffects_l + nza*ActionEffects_l do
                         if effects_l[i] then inc(ActiveActionEffects_l);
  ActiveActionEffects_f := 0;
  for i:= NetworkEffects_f + 1 to NetworkEffects_f + nza*ActionEffects_f do
                         if effects_f[i] then inc(ActiveActionEffects_f);
  ActiveActionEffects_g := 0;
  SetLength(AnyActionEffects_g,nza+1);
  for j:=1 to nza do
    begin
    AnyActionEffects_g[j] := false;
    for i:=1 to ActionEffects_g do
      if effects_g[NetworkEffects_g + (j-1)*ActionEffects_g + i] then
        begin
        inc(ActiveActionEffects_g);
        AnyActionEffects_g[j] := true;
        end;
    end;

  if (nza <= 0) then setlength(BeforeActionParameters_f,2)
    else setlength(BeforeActionParameters_f,nza+2);
  BeforeActionParameters_f[0] := 0;
  BeforeActionParameters_f[1] := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
               ActiveNetworkEffects_g + ActiveActionEffects_l;
  for j:=2 to nza do
    begin
    BeforeActionParameters_f[j] := BeforeActionParameters_f[j-1];
    for i:= 1 to ActionEffects_f do
            if effects_f[NetworkEffects_f + (j-2)*ActionEffects_f+i]
            then inc(BeforeActionParameters_f[j]);
    end;
  BeforeActionParameters_f[nza+1] := ActiveNetworkEffects_l + ActiveNetworkEffects_f +
               ActiveNetworkEffects_g + ActiveActionEffects_l + ActiveActionEffects_f;

  GiveDescriptions;
  DefineFunctionNames(lf);
  DefineStatsNames;
  NewNetworkEffects.DefineTotalContributions(alpa_f,alpa_g);
  SetTheta(theta);
  if changeparameter then ChangeTheta(outf,theta); (*MS: used in simulations to estimate / test misspecified models.*)
  if not assigned(entrop) then new(entrop);
  if not assigned(entroptime) then new(entroptime);
  for i:=0 to 100 do begin
     entropy[i]:= 0;
     entropynumber[i] := 0;
    end;
  for i:=1 to stepnomax do begin
    entrop^[i] := 0.0;
    entroptime^[i] := 0.0;
   end;
   x := booltostr(cconditional,true);

  x := format('%s%s%s%s%s%s%s',['Cconditional = ',BooltoStr(Cconditional,TRUE),
    '; Maxlike = ',BooltoSTr(MaxLike,TRUE),'; Bayesian = ',
  BooltoStr(Bayesian,TRUE),'.'])+eol;
  lf.Write(pchar(x)^,length(x));
  x := format('%s%s%s',['InitialStandard = ',BooltoStr(InitialStandard,TRUE),
    '.'])+eol;
  lf.Write(pchar(x)^,length(x));
  x := format('%s%s',['Continuous = ',BoolToStr(Continuous,TRUE)])+eol;
  lf.Write(pchar(x)^,length(x));
  if (observations = 1)
  then begin
    x:= format('%s%d%s%s%s',['Gibbscode = ',Gibbscode,': ',GibbsString,'.'])+eol;
    lf.Write(pchar(x)^,length(x));
    x := 'VertexPar ='+BooltoStr(VertexPar,TRUE)+'; VertexParNew = '+
    BooltoStr( VertexParNew,TRUE)+'.'+eol;
    lf.Write(pchar(x)^,length(x));
  end else begin
    x := format('%s%d%s',['ConditioningNetworkVariable = ', ConditioningNetworkVariable,','])+eol;
    lf.Write(pchar(x)^,length(x));
    x := format('%s%d%s',['ConditioningActionVariable = ', ConditioningActionVariable,','])+eol;
    lf.Write(pchar(x)^,length(x));
    x := format('%s%d%s%s%s',['ModelType = ',ModelType,': ',ModelTypeString,'.'])+eol;
    lf.Write(pchar(x)^,length(x));
    x := format('%s%d%s%d%s',['OutputType = ',OutputType,'; StaySameType = ',StaySameType,'.'])+eol;
    lf.Write(pchar(x)^,length(x));
  end;
  x := format('%s%d',['MaxDegree = ',MaxDegree])+eol;
  lf.Write(pchar(x)^,length(x));
end; (*InitialiseBase2. *)

procedure ModelSpecification.SetTheta(var theta: pvec);
(* Sets theta to alpa *)
var i,j,h: integer;
begin
  (* initialise *)
  for i:=1 to pmax do theta[i]  := 0.0;
    (* rate parameters *)
  h:=0;
  if (observations >= 2) then
    begin
    if (CConditional and (ConditioningNetworkVariable=1))
(* The first (observations-1) elements of alpa_l are excluded
   from the parameters, because they are the conditioning set.
   Therefore in this condition, the i-loop starts with observations;
   else the i-loop starts with 1. *)
     then for i:=observations to NetworkEffects_l do
      begin
      if Effects_l[i] then
        begin
        inc(h);
        theta[h] := alpa_l[i];
        end
      end (* for i *)
     else for i:=1 to NetworkEffects_l do
      begin
      if Effects_l[i] then
        begin
        inc(h);
        theta[h] := alpa_l[i];
        end
      end; (* for i *)
    end; (* if observations >= 2 *)

    (* network evaluation function parameters *)
  for i:=1 to NetworkEffects_f do if Effects_f[i] then
    begin
    inc(h);
    theta[h] := alpa_f[i];
    end;

    (* network endowment function parameters *)
  if (observations >= 2) then for i:=1 to NetworkEffects_g do
    if Effects_g[i] then
      begin
      inc(h);
      theta[h] := alpa_g[i];
      end;

      (* behaviour rate *)
  if (nza >= 1) then FirstActionRateParameter := h+1;
  for j := 1 to nza do
   if (Cconditional and (j=ConditioningActionVariable)) then
    begin
    for i:=observations to ActionEffects_l do
      begin
      if Effects_l[NetworkEffects_l+(j-1)*ActionEffects_l+i] then
        begin
        inc(h);
        theta[h] := alpa_l[NetworkEffects_l+(j-1)*ActionEffects_l+i];
        end
      end (* for i *)
    end
   else
    begin
    for i:=1 to ActionEffects_l do
      begin
      if Effects_l[NetworkEffects_l+(j-1)*ActionEffects_l+i] then
        begin
        inc(h);
        theta[h] := alpa_l[NetworkEffects_l+(j-1)*ActionEffects_l+i];
        end;
      end; (* for i *)
    end; (* for j *)

    (* objective function parameters *)
    for i:=1 + NetworkEffects_f to MaxEffects_f
    do if Effects_f[i]
    then begin
      inc(h);
      theta[h] := alpa_f[i];
    end;

    (* gratification function parameters *)
    if (observations >= 2)
    then for i:=1 + NetworkEffects_g to MaxEffects_g
    do if Effects_g[i]
    then begin
      inc(h);
      theta[h] := alpa_g[i];
    end;
end;  (* SetTheta *)

procedure ModelSpecification.SetAlpha(var theta: pvec);
(* Sets alpa to theta.
   This is the inverse of SetTheta. *)
(* Changed by TS 05-jan-05:
   alpa_l rate parameters for conditioning variables are not set to 1.0 *)
(* TS: constant factors among alpa_l rate parameters are checked for positivity *)
var i,j,h: integer;
  th1 : pvec;
  sum: real;
begin
  h:=0;
(* First the network effects *)
  if (observations >= 2) then
    begin
    if (CConditional and (ConditioningNetworkVariable=1))
    then for i:=observations to NetworkEffects_l do
      begin
      if Effects_l[i] then
        begin
        inc(h);
        alpa_l[i] := theta[h]
        end;
      end
     else for i:=1 to NetworkEffects_l do
      begin
      if Effects_l[i] then
        begin
        inc(h);
        if (i <= ActivePeriods) and (theta[h] < 1e-4) then theta[h] := 0.0001;    // changed 3.17w
        alpa_l[i] := theta[h]
        end;
      end;
    end;

    for i:=1 to NetworkEffects_f
    do if Effects_f[i]
    then begin
      inc(h);
      alpa_f[i] := theta[h];
    end;

    if (observations >= 2)
    then for i:=1 to NetworkEffects_g
    do if Effects_g[i]
    then begin
      inc(h);
      alpa_g[i] := theta[h];
    end;

(* Now the action effects *)
    for j:=1 to nza do
     if (Cconditional and (j=ConditioningActionVariable)) then
      begin
      for i := observations to ActionEffects_l do
        begin
        if Effects_l[Networkeffects_l+(j-1)*ActionEffects_l+i] then
          begin
          inc(h);
          alpa_l[Networkeffects_l+(j-1)*ActionEffects_l+i] := theta[h];
          end;
        end;
      end
     else
      begin
      for i := 1 to ActionEffects_l do
        begin
        if Effects_l[Networkeffects_l+(j-1)*ActionEffects_l+i] then
          begin
          inc(h);
          if (i <= ActivePeriods) and (theta[h] < 1e-4) then theta[h] := 0.0001;      // changed 3.17w
          alpa_l[Networkeffects_l+(j-1)*ActionEffects_l+i] := theta[h];
          end;
        end;
      end;

    for i:= NetworkEffects_f + 1 to MaxEffects_f
    do if Effects_f[i] then
      begin
      inc(h);
      alpa_f[i] := theta[h];
      end;

    if (observations >= 2)
    then for i:=NetworkEffects_g + 1 to MaxEffects_g
    do if Effects_g[i] then
      begin
      inc(h);
      alpa_g[i] := theta[h];
      end;

    (* TS 29-08-05:
   this procedure, formerly called GetParameter, is extended now by Transform_l
   and everything else that included convenient precalculation of intermediate results. *)
    Transform_l;
    NewNetworkEffects.DefineTotalContributions(alpa_f,alpa_g);
//writeln(outf,' in setaplha ');
//NewNetworkEffects.WriteMod_fSequenceParameters(alpa_f,outf);

    if ExtendedReciprocityModel then begin
      lambdar[0] := 0.0;
      for i:=1 to n do if StructActive[i] then begin
        sum := 0.0;
        for j:=1 to n do
          if ((i <> j) and StructActive[j] and (not struct(i,j,thisperiod)))
            then sum := sum+expo(UtilityGain(i,j,thisperiod,symmetric,true,NewNetworkEffects));
        lambdar[i] := sum/(n-1);  // @@@@@
       end (* if StructActive[i] *) else lambdar[i] := 0.0;
(* lambdar is updated in ChangeTie *)
     end; (* if ExtendedReciprocityModel *)

(* Activate this for checking :
  SetTheta(th1);
  s := 0.0;
  for i:= NumberParameters downto 1 do begin
    s := s + abs(th1[i]-theta[i]);
   end;
  if (s > 0.0001) then begin
    writeln(outf, 'SetAlpha : s = ',s:20:10);
    for i:=1 to NumberParameters do writeln(outf, theta[i]:16:6,' ',th1[i]:16:6);
    writeln(outf, 'This is an error. The program will stop.');
    Halterror('This error occurred in Procedure SetAlpha in S_Dat.',outf,lf);
    halterr := true;
   end; *)
end;  (* SetAlpha *)


procedure ModelSpecification.FunctionEvaluations(var orderdegs: intvec;
                       var halterr: boolean; var outf,lf: TFileStream; var targets: pvec);
(*MS: formerly part of <InitialiseBase>, but not needed when synthetic
      data are generated.
   This is used for calculation of observed functions,
   and should correspond exactly to SimStats in how covariate data
   and networks/simulations are employed
   and how targets/statistics are calculated.*)
(* Initialise_y cannot be used here,
   because in the case of changing actor covariates, the time-lagged definition
   is used:
   network statistics use network of end of the period, behavior of start of the period;
   behavior statistics use network of start of the period, behavior of end of the period.
   *)
var i,j,k,pp: integer;
    statk,statsum: pvec;
    x: string;
begin
  (* following calculation of functions is duplicated in SimStats,
     but then for simulated digraphs *)
(* TS: changed this 20-01-06, to unite the observations=2 case
       and the observations >= 3 case.
       This implies a difference for the case with dependent behavior variables
       if observations = 2,
       because in the earlier version the changing individual covariates
       were used of the contemporaneous observation,
       which now is changed to the previous observation.
       Other change: n_act is calculated in initialise_actors
       so that Networkstatistics and ActionStatistics may depend on n_act. *)
  n := nmax;
  pp:= NumberFunctions;
    x:= format('%s%d',['NumberParameters = ',NumberParameters])+eol;
    lf.Write(pchar(x)^,length(x));
  for i:=1 to pmax do statsum[i] := 0.0;
  for i:=1 to n do orderdegs[i] := 0;
  if (observations <= 1) then
    begin
    thisperiod := 1;
    initialise_actors(1);
    SetDigraph(1);
    ErgmStatistics(CurrentStats,symmetric,orderdegs,false,NewNetworkEffects);
    for i:=1 to NumberFunctions do
      begin
      statsum[i] := CurrentStats[i];
      end;
     cumuldegrees[0] := 0;
     for i:=1 to n do cumuldegrees[i] := cumuldegrees[i-1]+outd_yn(i);
    end
   else
    begin
  (* MH: Only initialise active actors if exogenous events are specified;
     3.17u: put before k-loop, needs to be done only once. *)
    if exo then
      begin
      InitialiseActiveActors(halterr,outf,lf);
      if halterr then exit;
      end; (* if exo *)
    for k := 1 to observations -1 do
      if not skipi[k] then
      begin
      thisperiod := k;
      initialise_actors(k);

      if (nza >= 1) and (k = 1) and (CallingProgram = 5) then
        begin
(* calculate spatial autocorrelation statistics, obs. 1 *)
        SetDigraph(1);
        for i:=1 to n do
          begin
          for j:=1 to nzc do
            begin
          if (zc[i,j,1] < zcmin[j]) then z[i,nz+j] := zcmin[j]
           else
            begin
            if (zc[i,j,1] > zcmax[j]) then z[i,nz+j] := zcmax[j]
             else z[i,nz+j] := zc[i,j,1];
            end;
            setmisz(i,nz+j,miszc(i,j,1))
            end;
          for j:=1 to nza do
            begin
(* For network statistics: put ya equal to zc in start of period *)
            setya(i,j,round(zc[i,j,1]));
            if (ya(i,j) < round(zcmin[j])) then setya(i,j,round(zcmin[j]))
             else
              if (ya(i,j) > round(zcmax[j])) then setya(i,j,round(zcmax[j]));
            end;
          end; (* i *)
(* Now action variables are for time 1, network variables are for time 1 *)
        for j:=1 to nza do
          begin
          SpatialFunctions(StructActive,j,1,mi[j,1],gc[j,1]); // changed 3.17u
          end;
        end; (* if *)


(* Calculate regular statistics *)
(* Following changed TS 13-10-05 *)
(* store changing covariates for current period also in array Z
   -- note that (nzc-nza) must be 0 for observations=2, and nzc=0 for observations=1 *)
    SetDigraph(k+1);                       // cdcd
    if exo then SetLeaversyBack(k); // added 3.17u, mimicking SimStats.
    ForceStructurals(k);
(* preceding line added 3.17r:
   force structurally determined values to be those at the start of the period.
   This is necessary in case that structurally determined values
   change between waves.
   In that case the target statistics use the structurally determined values
   for the start of the period
   - which is accomplished here by the use of ForceStructurals;
   and elements (i,j) for which there is a structurally determined value
   at the end but not at the start of the period
   get their value from the end of the period - which is automatic here.
   Search for ForceStructurals and ForceNewStructurals to see where this happens.
   *)

      for i:=1 to n do
        begin
        for j:=1 to nzc do
          begin
(* TS conditie toegevoegd 21-1-06 : *)
          if (zc[i,j,k] < zcmin[j]) then z[i,nz+j] := zcmin[j]
           else
            begin
            if (zc[i,j,k] > zcmax[j]) then z[i,nz+j] := zcmax[j]
             else z[i,nz+j] := zc[i,j,k];
            end;
          setmisz(i,nz+j,miszc(i,j,k))
          end;
        for j:=1 to nza do
          begin
(* For network statistics: put ya equal to zc in start of period *)
          setya(i,j,round(zc[i,j,k]));
(* TS toegevoegd 4-1-06 : was eerder fout! *)
          if (ya(i,j) < round(zcmin[j])) then setya(i,j,round(zcmin[j]))
           else
            if (ya(i,j) > round(zcmax[j])) then setya(i,j,round(zcmax[j]));
          end;
        end; (* i *)
(* Now action variables are for time k, network variables are for time k+1 *)
      NetworkStatistics(statk,orderdegs,k,outf);   // cdcd

      for i:=1 to n do
        begin
        for j:=1 to nza do
          begin
(* For action statistics: put ya equal to zc in end of period,
   except for structurally inactive actors *)
          if IncludeActor[i] then setya(i,j,round(zc[i,j,k+1]));  // condition added 3.17u
          if (ya(i,j) < round(zcmin[j])) then setya(i,j,round(zcmin[j]))
           else
            if (ya(i,j) > round(zcmax[j])) then setya(i,j,round(zcmax[j]));
          end;
        end;
      if exo then SetLeaverszBack(k); // added 3.17u, mimicking SimStats.
(* Now action variables are for time k+1, network variables are for time k+1,
   except for inactive actors. *)

(* calculate spatial autocorrelation statistics, obs. k+1 *)
      if (CallingProgram = 5) then for j:=1 to nza do
        begin
        SpatialFunctions(StructActive,j,k+1,mi[j,k+1],gc[j,k+1]); // changed 3.17u
        end;

      SetDigraph(k);
(* Now action variables are for time k+1, network variables are for time k *)
      ActionStatistics(statk,orderdegs,k,outf);
      for i:=1 to NumberFunctions do statsum[i] := statsum[i] + statk[i];
      end; (* for k *)
    end; (* if observations >= 2 *)
  for i:=1 to NumberFunctions do targets[i] := statsum[i];
  for i:= NumberFunctions + 1 to pmax do targets[i] := 0.0;
  if (nza >= 1) then outf.write(pchar(eol)^,length(eol));
end; (*FunctionEvaluations*)


procedure ModelSpecification.ChangeAltern;
begin
if altern then altern := false
else altern := true;
end;


procedure ModelSpecification.InitialiseActiveActors(var halterr: boolean;
                             var outf,lf: TFileStream);
(* MH: Determine active actors at start and end of each period
   and set ynm and missing data indicator:
   option 1 - before joining, joiner gets regular observed value or if
              missing the value 0 is imputed; after leaving, leaver gets
              regular observed value or if missing the last observed
              value is imputed.
   option 2 - impute 0's for joiners, treat leavers as missing.
              If first leaving then joining, use last observed values.
   option 3 - joiners and leavers are treated as missing.
   Check whether there are always at least 2 active actors on any time *)
var
  i,j,k,m1,m2,n_m1,n_m2: integer;
  err: string;
(* version 3.17u: local variable action renamed to action_cc and declared
   in ModelSpecification *)
  x : string;
begin
  setlength(action_cc,n+1,observations+1);
  (* action_cc: 0=active, 1=join (may leave after),  2=leave (may have joined
     before), 3=first leave and join again later *)

  (* Specify the active_m, event_m, and action_cc arrays *)
  for i:=1 to nmax do begin
    m1 := trunc(zx[i,1]);
    m2 := trunc(zx[i,2]);
    for k:=1 to observations do begin
      if (skipi[k] and (k < observations)) then
        begin
        HaltError('Joiners-leavers option is active in multi-group project. This is incorrect.',outf,lf);
        halterr := true;
        exit;
        end;
      active_m[i,k] := true;
      event_m[i,k] := 0.0;
      action_cc[i,k] := 0; (* active in all periods *)
    end;
    (* Joining: action_cc=1 *)
    if ((zx[i,1] > 1) and (zx[i,2] > observations)) then begin
      event_m[i,m1] := zx[i,1] - m1;
      for k:=1 to m1 do begin
        active_m[i,k] := false; action_cc[i,k] := 1; end;
    end;
    (* Leaving: action_cc=2 *)
    if ((zx[i,1] < 1) and (zx[i,2] < observations)) then begin
      event_m[i,m2] := zx[i,2] - m2;
      for k:=m2 to observations-1 do begin
        active_m[i,k+1] := false; action_cc[i,k+1] := 2; end;
    end;
    (* Joining and leaving or leaving and joining *)
    if ((zx[i,1] > 1) and (zx[i,2] < observations)) then begin
      event_m[i,m1] := zx[i,1] - m1;
      event_m[i,m2] := zx[i,2] - m2;
      (* First joining then leaving: action_cc=1 and 2 *)
      if (m1 <= m2) then begin
        for k:=1 to m1 do begin
          active_m[i,k] := false; action_cc[i,k] := 1; end;
        for k:=m2 to observations-1 do begin
          active_m[i,k+1] := false; action_cc[i,k+1] := 2; end;
      end
      (* First leaving then joining: action_cc=3 *)
      else for k:=m2 to m1-1 do begin
        active_m[i,k+1] := false; action_cc[i,k+1] := 3; end;
    end; (* if zx > 1 and zx < obs *)
  end; (* for i *)


  (* Set ynm and mis: first handle joiners, second leave-joiners, last leavers *)
  for k:=1 to observations do begin
    for i:=1 to n do begin
      if (action_cc[i,k] = 1) then (* Join *)
        case nzx of
        1,2: for j:=1 to n do
               if (mis(i,j,k) and (i <> j)) then begin
                 setym(i,j,k,false); setym(j,i,k,false);
                 setmis(i,j,k,false); setmis(j,i,k,false);
               end;
          3: for j:=1 to n do begin
               setmis(i,j,k,true); setmis(j,i,k,true);
             end;
        end; (* case nzx *)
    end; (* for i *)

    for i:=1 to n do begin
      if (action_cc[i,k] = 3) then (* Leave then join *)
        case nzx of
        1,2: for j:=1 to n do
               if (mis(i,j,k) and (i <> j)) then begin
                 setyvm(i,j,k,ynvm(i,j,k-1)); setyvm(j,i,k,ynvm(j,i,k-1));
                 setmis(i,j,k,false); setmis(j,i,k+1,false);
               end;
          3: for j:=1 to n do begin
               setmis(i,j,k,true); setmis(j,i,k,true);
             end;
        end; (* case nzx *)
    end; (* for i *)

    for i:=1 to n do begin
      if (action_cc[i,k] = 2) then (* Leave *)
        case nzx of
          1: for j:=1 to n do
               if (mis(i,j,k) and (i <> j)) then begin
                 setyvm(i,j,k,ynvm(i,j,k-1)); setyvm(j,i,k,ynvm(j,i,k-1));
                 setmis(i,j,k,false); setmis(j,i,k+1,false);
               end;
        2,3: for j:=1 to n do begin
               setmis(i,j,k,true); setmis(j,i,k,true);
             end;
        end; (* case nzx *)
    end; (* for i *)
  end; (* for k *)


  (* Check on empty sets of active actors *)
  for j:=1 to observations-1 do begin
    n_m1 := 0; (* # active at observation time j and j+1 *)
    for i:=1 to n do
      if (active_m[i,j] and active_m[i,j+1]) then inc(n_m1);
    if (n_m1 < 2) then begin
       n_m2 := 0; (* # active at start period j *)
       for i:=1 to n do begin
         if active_m[i,j] then inc(n_m2);
         event[i] := event_m[i,j];
       end; (* for i *)
       nextevent := 0.0; i_act := 0;
       while (nextevent < 1) do begin
         if (n_m2 < 2) then begin
           if (n_m2 = 1) then begin
             x := 'Only one active actor left ';
             lf.Write(pchar(x)^,length(x));
           end
           else begin
             x :='No active actors left ';
             lf.Write(pchar(x)^,length(x));
           end;
           if (i_act > 0) then begin
             x := format('%s%4.2f%s%d%s',['after time ',nextevent,
             ' in period ', j, '.'])+eol;
             lf.Write(pchar(x)^,length(x));
             if active_m[i_act,j] then begin
               x := format('%s%d%s',['Last leaver was actor ', i_act, '.'])+eol;
               lf.Write(pchar(x)^,length(x));
             end
             else begin
               x := format('%s%d%s',['Last joiner was actor ', i_act, '.'])+eol;
               lf.Write(pchar(x)^,length(x));
             end;
           end else begin
           x := format('%s%d%s',['at start period ', j, '.'])+eol;
           lf.Write(pchar(x)^,length(x));
           end;
           halterr := true;
         end; (* if n_m2<2 *)
         GetNextevent;
         if active_m[i_act,j] then dec(n_m2) (* leaver *)
           else inc(n_m2); (* joiner *)
       end; (* while nextevent<1 *)
    end; (* if n_m1<2 *)
  end; (* for j *)
  if halterr then begin
    err := 'No/Only one active actor(s) left.';
    Halterrorr(err,outf,lf);
    exit;
  end
end; (* InitialiseActiveActors *)


procedure ModelSpecification.SetLeaversyBack(k: integer);
(* MH: Changes the simulated values of the yn matrix at time k+1 of actors
         who left, back to the last observed values at observation time k,
         which are already correctly specified in ynm[k+1]. *)
var
  i,j: integer;
begin
  for i:=1 to n do if (action_cc[i,k+1] in [2,3])  // changed 3.17u; it was (not active[i])
  then
    begin
    SetDataRowTo(i,k);   // changed from k+1 to k, version 3.17u
    SetDataColumnTo(i,k);   // changed from k+1 to k, version 3.17u
// TS: this replaces:
//    for j:=1 to n
//    do begin
//      setynv(i,j,ynvm(i,j,k));
//      setynv(j,i,ynvm(j,i,k))
//    end;
  end;
end; (* SetLeaversyBack *)

procedure ModelSpecification.SetLeaverszBack(k: integer);   // new, 3.17u; split off from  SetLeaversyBack
(* MH: Changes the simulated values of the yn matrix at time k+1 of actors
         who left, back to the last observed values at observation time k,
         which are already correctly specified in ynm[k+1]. *)
var
  i,j: integer;
begin
  for i:=1 to n do
    if (action_cc[i,k+1] in [2,3]) then
  (* action_cc defined in InitialiseActiveActors: values 2 and 3 refer to leavers
     in the preceding or earlier periods, 3 are those who later will join. *)
      begin
      for j:=1 to nza do setya(i,j,round(zc[i,j,k]));  // changed from k+1 to k, version 3.17u
      end;
end; (* SetLeaverszBack *)


procedure ModelSpecification.Transform_l;
(* THIS PROCEDURE CALCULATES SOME HYBRID VARIABLES EXPRESSING EXPONENTIAL TRANSFORMATIONS
   OF BOTH RATE PARAMETERS AND LOCAL STATISTICS.

   TS 1-aug-05 abolished maxindegree maxoutdegree maxreciprocity

   To avoid frequent calculation of exponential functions, this procedure is called
   once every time that new parameters alpa_l are set.
   CS 29-mrt-05: extension to calculations for behavioural rate functions.
   CS 19-apr-05: old rate effects of outdegree, indegree and reciprocity are
                 replaced by new effects. *)
var i,j,h,
    ModelCode,
    count        : integer;
    xi0,
    contribution,
    alf1,alf2,alf3 : real;
    flag         : boolean;
begin
  if (Observations <= 1) then exit; (* no rate parameters for exponential random graph models *)
  if symmetric then begin
    ModelCode := 1;
//    if (ModelType > 0) then ModelCode := 1 else ModelCode := -1;
   end else ModelCode := ModelType;

  (* Set arrays exp_alpha_out, exp_alpha_in and exp_alpha_rec that contain parameter transformations
     for network effects on the rate functions, array nu[.] for the 1/outdegree parameter on network
     rate (transformations for all possible outdegrees), and/or array xi[.] that for effects of
     outdegree transformations on network rate for ModelType=2. Array eff_l3 contains a flag
     summarizing whether any of these effects was selected in the modelspecification. *)

 (* start with NETWORK RATE parameter transformations: *)

  for i:=1 to (observations-1) do lambda1[0,i]:=alpa_l[i]; (* basic rate parameters *)  // changed TS

  if ReciprocityModel then begin
    if effects_f[1] then alf1 := alpa_f[1] else alf1 := 0.0;
    if effects_f[2] then alf2 := alpa_f[2] else alf2 := 0.0;
    if effects_g[2] then alf3 := alpa_g[2] else alf3 := 0.0;
    lambdar_cons := expo(alf1);
    lambdar_out := (expo(-alf1) - expo(alf1))/(n-1);
    lambdar_in := (expo(alf1 + alf2) - expo(alf1))/(n-1);
    lambdar_rec := (expo(alf1) - expo(-alf1) - expo(alf1+alf2)
                     + expo(-alf1-alf2-alf3))/(n-1);
   end; (* if ReciprocityModel *)

  case ModelCode of
   1: begin
        flag := false;

        (* for outdegree effect on network rate : *)
        if effects_l[observations]
        then begin
          for i := n-1 downto 0
          do exp_alpha_out[0,i] := expo(alpa_l[observations]*i);
          flag := true
        end
        else for i := n-1 downto 0
        do exp_alpha_out[0,i] := 1.0;

        (* for indegree effect on network rate : *)
        if effects_l[observations+1]
        then begin
          for i := n-1 downto 0
          do exp_alpha_in[0,i] := expo(alpa_l[observations+1]*i);
          flag := true
        end
        else for i := n-1 downto 0
        do exp_alpha_in[0,i] := 1.0;

        (* for reciprocity effect on network rate : *)
        if effects_l[observations+2]
        then begin
          for i := n-1 downto 0
          do exp_alpha_rec[0,i] := expo(alpa_l[observations+2]*i);
          flag := true
        end
        else for i := n-1 downto 0
        do exp_alpha_rec[0,i] := 1.0;

        (* for 1/outdegree effect on network rate : *)
        if effects_l[observations+3]
        then begin
          for  i:= n-1 downto 0
          do store_nu[i+1] := expo(alpa_l[observations+3]/(i+1.0));
          flag := true
        end
        else for i := n-1 downto 0 do store_nu[i+1] := 1.0;

        (* remember if any of the above effects is included *)
        eff_l3[0] := flag
      end; (* case 1 *)
   2: begin
        for i := n-2 downto 0 do
        begin
          xi0 := 0.0;
          (* outdegree effect: *)
          if effects_l[observations] then xi0 := xi0 + alpa_l[observations];
          (* logarithmic outdegree effect: *)
          if effects_l[observations+1] then xi0 := xi0 + alpa_l[observations+1]/(i+1);
          (* factorial outdegree effect: *)
          if effects_l[observations+2] then xi0 := xi0 + alpa_l[observations+2]*log(i+1);
          store_xi[i+1] := expo(xi0);
        end;
        store_xi[0] := 1.0;
        for i:= n-1 to n do store_xi[i+1] := 0.0;
        (* effect 1/outdegree: *)
        if effects_l[observations+3]
        then for i:= n-2 downto 0 do store_nu[i+1] := max(0.0, 1.0 + alpa_l[observations+3]/(i+1.0))
        else for i:=n-2 downto 0 do store_nu[i+1] := 1.0;
        nuno := 0.0;
        for i:=n-1 to n do store_nu[i+1] := 0.0;
      end; (* case 2 *)
  end; (* case ModelCode *)

  (* now BEHAVIOUR RATE parameter transformations: *)
  for j := 1 to nza
  do begin
    flag := false;
    (* basic rate parameters *)
    for i:=1 to (observations-1) do lambda1[j,i]:= alpa_l[NetworkEffects_l+(j-1)*ActionEffects_l+i]; // changed TS
    (* for outdegree effect on behaviour rate : *)
    if effects_l[NetworkEffects_l+(j-1)*ActionEffects_l+observations]
    then begin
      for i := n-1 downto 0
      do exp_alpha_out[j,i] := expo(alpa_l[NetworkEffects_l+(j-1)*ActionEffects_l+observations]*i);
      flag := true
    end
    else for i := n-1 downto 0
    do exp_alpha_out[j,i] := 1.0;

    (* for indegree effect on behaviour rate : *)
    if effects_l[NetworkEffects_l+(j-1)*ActionEffects_l+observations+1]
    then begin
      for i := n-1 downto 0
      do exp_alpha_in[j,i] := expo(alpa_l[NetworkEffects_l+(j-1)*ActionEffects_l+observations+1]*i);
      flag := true
    end
    else for i := n-1 downto 0
    do exp_alpha_in[j,i] := 1.0;

    (* for reciprocity effect on behaviour rate : *)
    if effects_l[NetworkEffects_l+(j-1)*ActionEffects_l+observations+2]
    then begin
      for i := n-1 downto 0
      do exp_alpha_rec[j,i] := expo(alpa_l[NetworkEffects_l+(j-1)*ActionEffects_l+observations+2]*i);
      flag := true
    end
    else for i := n-1 downto 0
    do exp_alpha_rec[j,i] := 1.0;

    (* remember if any of the above effects is included *)
    eff_l3[j] := flag
  end; (* for j=1..nza *)

  (* Array eff_l2 contains a flag summarizing whether there are covariate effects on the rate functions.
      - for ModelCode = 2, eff_l2[0] indicates covariate effects on outdegree,
      - for ModelCode = -1,1, eff_l2[0] indicates direct covariate effects on rate,   [s]
      - eff_l2[k] for k>0 indicates direct covariate effects on behaviour k's rate. *)
  for j := 0 to nza do eff_l2[j] := false;
  for h :=1 to nz+nzc
  do begin
    if effects_l[observations+3+h] then eff_l2[0] := true;
    for j := 1 to nza
    do if effects_l[NetworkEffects_l+(j-1)*ActionEffects_l+observations+2+h] then eff_l2[j] := true;
  end;

  (* Set arrays lambda2[k][.] that contain joint transformations of
     sums over covariates of (parameter  actors' score on covariate): *)

  (* first NETWORK RATE: *)
  if eff_l2[0] then
    begin
    for i:=n downto 1 do lambda2[0,i]:=0.0;
    for h:=nz+nzc downto 1 do
      begin
      count := observations + 3 + h;
      if effects_l[count] then
        begin
        for i:=n downto 1 do
          begin
          contribution := z[i,h];
          if d_estAB then changestat[1,count,i] := contribution; (*MS: MoM: derivative estimation.*)
          lambda2[0,i]:=lambda2[0,i] + (alpa_l[count] * contribution);
          end;
        end;
      end;
    for i:=n downto 1 do lambda2[0,i]:=expo(lambda2[0,i]);
    end
  else
    begin
    for i:=n downto 1 do lambda2[0,i]:=1.0;
    end;

  (* now BEHAVIOUR RATE: *)
  for j := 1 to nza do
  if eff_l2[j] then
    begin
    for i:=n downto 1 do lambda2[j,i]:=0.0;
    for h:=nz+nzc downto 1 do
      begin
      count := NetworkEffects_l+(j-1)*ActionEffects_l+observations+2+h;
      if effects_l[count] then
        begin
        for i:=n downto 1 do
          begin
          contribution := z[i,h];
          if d_estAB then changestat[1,count,i] := contribution; (*MS: MoM: derivative estimation.*)
          lambda2[j,i]:=lambda2[j,i] + (alpa_l[count] * contribution);
          end;
        end;
      end;
    for i:=n downto 1 do lambda2[j,i]:=expo(lambda2[j,i]);
    end
  else
    begin
    for i:=n downto 1 do lambda2[j,i]:=1.0;
    end;

  (* for ModelType=2, things are slightly different: *)
  if (ModelCode = 2)
  then begin
    (* eff_l4 indicates whether there are also direct covariate effects on rate *)
    eff_l4 := false;
    for h :=1 to nz+nzc do if effects_l[observations+3+nz+nzc+h] then eff_l4 := true;

    if eff_l4
    then begin
      for i:=n downto 1 do lambda4[i]:=0.0;
      for h:=nz+nzc downto 1 do
        if effects_l[observations+3+nz+nzc+h]
        then for i:=n downto 1 do lambda4[i]:=lambda4[i]+alpa_l[observations+3+nz+nzc+h]*z[i,h];
      for i:=n downto 1 do lambda4[i]:=expo(lambda4[i]);
    end
    else for i:=n downto 1 do lambda4[i]:=1.0;
  end; (* if ModelCode = 2 *)

end; (* procedure Transform_l *)


function ModelSpecification.NetworkLambda(i           : integer;
                                          var halterr : boolean;
                                          var outf,lf : TFileStream) : Real;
(* The change rate for actor i *)
var lambda, xis : real;
    s, ModelCode     : integer;
    x : string;
begin
  halterr := false;
  if (not StructActive[i]) and (not (maxlike or Bayesian))
  then begin
    NetworkLambda := 0.0;
    exit;
  end;
  ModelCode := ModelType;
  if symmetric then begin
    ModelCode := 1;
//    if (ModelType < 0) then ModelCode := -1 else ModelCode := 1;
   end;
  if ReciprocityModel then begin
    if ExtendedReciprocityModel then Result := lambda1[0,thisperiod]*lambdar[i]
     else Result := lambda1[0,thisperiod]*(lambdar_cons + lambdar_out*outd_yn(i)
               + lambdar_in*ind_yn(i) + lambdar_rec*ReciDegree_yn(i));
    exit;
   end; (* if ReciprocityModel *)
  case ModelCode of
   -1, 1: begin
        if eff_l2[0] (* covariate effects on rate *)
        then lambda := lambda1[0,thisperiod]*lambda2[0,i]
        else lambda := lambda1[0,thisperiod];
        if eff_l3[0] (* network effects on rate *)
        then begin
          if effects_l[observations] (* outdegree effect *)
          then
            begin
            lambda := lambda * exp_alpha_out[0,outd_yn(i)];
          if d_estAB then changestat[1,observations,i] := outd_yn(i); (*MS: MoM: derivative estimation.*)
            end;
          if effects_l[observations+1] (* indegree effect *)
          then
            begin
            lambda := lambda * exp_alpha_in[0,ind_yn(i)];
          if d_estAB then changestat[1,observations+1,i] := ind_yn(i); (*MS: MoM: derivative estimation.*)
            end;
          if effects_l[observations+2] (* reciprocity effect *)
          then
            begin
            lambda := lambda * exp_alpha_rec[0,ReciDegree_yn(i)];
          if d_estAB then changestat[1,observations+2,i] := ReciDegree_yn(i); (*MS: MoM: derivative estimation.*)
            end;
          if effects_l[observations+3] (* 1/outdegree effect *)
          then
            begin
            lambda := lambda*nu(outd_yn(i));
          if d_estAB then changestat[1,observations+3,i] := 1 / (1.0 + outd_yn(i)); (*MS: MoM: derivative estimation.*)
            end;
        end; (* eff_l3 *)
        NetworkLambda := lambda
      end; (* case 1 *)
   2: begin
        s := outd_yn(i);
        xis := xi(s)*lambda2[0,i];
        NetworkLambda := alpa_l[thisperiod]*lambda4[i]*
           ( (nu(s)*xis/(1.0+xis)) + (nu(s-1)/(1.0+(xi(s-1)*lambda2[0,i]))) );
      end; (* case 2 *)
   else begin
     NetworkLambda := 0.0;
     halterr := true;
     x := format('%s%d%s',['Procedure Transform in S_Base called for ModelType = ',ModelType,'.'])+eol;
     lf.Write(pchar(x)^,length(x));
     x := 'This is not implemented.'+eol;
     lf.Write(pchar(x)^,length(x));
     x := format('%s%d%s',['NetworkLambda error: Modeltype ',ModelType,' does not exist.'])+eol;
     lf.Write(pchar(x)^,length(x));
     exit;
   end; (* case else *)
  end; (* case *)
end; (* NetworkLambda *)


function ModelSpecification.ActionLambda(i,                     (* actor, between 1 and n *)
                                         h           : integer; (* behaviour, between 1 and nza *)
                                         var halterr : boolean;
                                         var outf,lf : TFileStream): real;
var
  index  : integer;
  lambda : real;
begin
  (* CS 28-mrt-2005: extension of behaviour rate to network and actor covariate influences *)
  halterr := false;
  if (not StructActive[i]) and (not (maxlike or Bayesian))
  then begin
    ActionLambda := 0.0;
    exit;
  end
  else begin
    if eff_l2[h] (* covariate effects on rate *)
    then lambda := lambda1[h,thisperiod]*lambda2[h,i]
    else lambda := lambda1[h,thisperiod];
    if eff_l3[h] (* network effects on rate *)
    then begin
      index:=NetworkEffects_l+(h-1)*ActionEffects_l+observations;
      if effects_l[index] (* outdegree effect *)
      then begin
        lambda := lambda * exp_alpha_out[h,outd_yn(i)];
        if d_estAB then changestat[1,index,i] := outd_yn(i); (*MS: MoM: derivative estimation.*)
      end;
      inc(index);
      if effects_l[index] (* indegree effect *)
      then begin
        lambda := lambda * exp_alpha_in[h,Ind_yn(i)];
        if d_estAB then changestat[1,index,i] := Ind_yn(i); (*MS: MoM: derivative estimation.*)
      end;
      inc(index);
      if effects_l[index] (* reciprocity effect *)
      then begin
        lambda := lambda * exp_alpha_rec[h,ReciDegree_yn(i)];
        if d_estAB then changestat[1,index,i] := ReciDegree_yn(i); (*MS: MoM: derivative estimation.*)
      end;
    end; (* eff_l3 *)
    ActionLambda := lambda

  end
end; (* ActionLambda *)


procedure ModelSpecification.ReviseDefinition(var halterr: boolean;
                                var outf,lf: TFileStream);
var
  i,j,tempi, klad: integer;
  modelnames     : SSarray;
  Correspondence : SBYarray;
  VertexPar_f    : BBarray;
  param          : SSIarray;
  eff_od, eff_l  : boolean;
  someskip,parnum, code
                 : integer;
  atleastone     : Boolean;
  (* Vertexpar_f indicates for in the Vertexpar (observations  = 1) case,
     which parameters must be excluded because they
     would lead to overparametrisation. *)
  ertext: shortstring;
  x : string;
procedure minmax(var mi, ma: integer; m: integer);
begin
  if (m < mi) then mi := m;
  if (m > ma) then ma := m;
end; (* minmax *)
begin
// @@@ Alles met effects_f veranderen moet hier worden bijgesteld.
  halterr := false;

  if (nza >= 1) then
    begin
(* Check that all interactions between actor variables for dependent behavior variables
   are correctly coded: *)
      setlength(ni,nza+1,ninta+1);
      setlength(pv0,nza+1,ninta+1);
      setlength(pv1,nza+1,ninta+1);
      setlength(pv2,nza+1,ninta+1);
      setlength(pv3,nza+1,ninta+1);
      for i:=1 to nza do for j:=1 to ninta do
        begin
        parnum := nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 20 +2*(nz+nzc-1)+j;
        if (param_effect[parnum] <> 0) and (param_effect[parnum] <= 100) then
          begin
          x := format('%s%d%s%d%s%d%s',['For dependent behavior variable ',i,
          ', interaction effect ',j, ' has code ', param_effect[parnum],
          ' which is not meaningful.'])+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'This interaction effect is dropped.'+eol;
          outf.Write(pchar(x)^,length(x));
          param_effect[parnum] := 0;
          end;
        unpack(param_effect[parnum],ni[i,j],pv1[i,j],pv2[i,j],pv3[i,j]);
        if (pv1[i,j] > nz+nzc) or (pv2[i,j] > nz+nzc) or (pv3[i,j] > nz+nzc) then
          begin
          x := format('%s%d%s%d%s',['For dependent behavior variable ',i,', interaction effect ',j,
                       ' is incorrectly coded.'])+eol;
          outf.Write(pchar(x)^,length(x));
          if (pv1[i,j] > nz+nzc) then pv1[i,j] := 0;
          if (pv2[i,j] > nz+nzc) then pv2[i,j] := 0;
          if (pv3[i,j] > nz+nzc) then pv3[i,j] := 0;
          if ((pv1[i,j]*pv2[i,j] + pv2[i,j]*pv3[i,j] + pv1[i,j]*pv3[i,j]) = 0) then
            begin
(* 0 or 1 correct codes remain: no interaction *)
            param_effect[parnum] := 0;
            ni[i,j] := 0;
            x :='This effect is skipped.'+eol;
            outf.Write(pchar(x)^,length(x));
            end
           else
            begin
            if (pv3[i,j] = 0) then
              begin
              param_effect[parnum] :=  pv1[i,j] + 100*pv2[i,j];
              end
             else
              begin
              if (pv1[i,j] = 0) then
                begin
                param_effect[parnum] :=  pv2[i,j] + 100*pv3[i,j];
                pv1[i,j] := pv2[i,j];
                pv2[i,j] := pv3[i,j];
                pv3[i,j] := 0;
                end
               else
                begin
                (* pv2[i,j] must be 0 *)
                param_effect[parnum] :=  pv1[i,j] + 100*pv3[i,j];
                pv2[i,j] := pv3[i,j];
                pv3[i,j] := 0;
                end;
              end;
            x := format('%s%d%s%d%s%d%s',['The coding that will be used is ',param_effect[parnum],
                         ', referring to the interaction between variables ',
                           pv1[i,j],' and ', pv2[i,j],'.'])+eol;
            outf.Write(pchar(x)^,length(x));
            end;
          end (* if incorrectly coded *)
         else
          begin
          if (ni[i,j] > 0) then
            begin
            x := format('%s%d%s%s',['Interaction effect ',j,
                       ' for the dependent behavior variable ',Znames[nz+i]])+eol;
            outf.Write(pchar(x)^,length(x));
            x := format('%s%d%s%s',['refers to variables ',pv1[i,j],': ',Znames[pv1[i,j]]]);
            outf.Write(pchar(x)^,length(x));
            if (ni[i,j] = 2) then begin
              x := format('%s%d%s%s%s',[' and ',pv2[i,j],': ',Znames[pv2[i,j]],'.'])+eol;
              outf.Write(pchar(x)^,length(x));
            end
             else begin
              x := format('%s%d%s%s%s%d%s%s%s',[', ',pv2[i,j],': ',Znames[pv2[i,j]],', and ',
                                pv3[i,j],': ',Znames[pv3[i,j]],'.'])+eol;
             outf.Write(pchar(x)^,length(x));
             end;
            end;

          end;
(* recode the pv numbers in case some are equal to the dependent variable itself;
  then pv0 is the multiplicity of the dependent variable,
  the dependent variable is omitted from pv1, pv2, pv3,
  - i.e., these are recoded to 0 -
  and these are reordered so that the first 1 or 2 are nonzero. *)
        pv0[i,j] := 0;
        if (pv1[i,j] = nz+i) then
          begin
          inc(pv0[i,j]);
          end;
        if (pv2[i,j] = nz+i) then
          begin
          inc(pv0[i,j]);
          end;
        if (pv3[i,j] = nz+i) then
          begin
          inc(pv0[i,j]);
          end;
(* reorder: *)
        if (pv1[i,j] = nz+i) then
          begin
          pv1[i,j] := pv2[i,j];
          if (pv3[i,j] = 0) then pv2[i,j] := nz+i
           else
           begin
           pv2[i,j] := pv3[i,j];
           pv3[i,j] := nz+i;
           end;
          end;
        if (pv1[i,j] = nz+i) then
          begin
          pv1[i,j] := pv2[i,j];
          if (pv3[i,j] = 0) then pv2[i,j] := nz+i
           else
           begin
           pv2[i,j] := pv3[i,j];
           pv3[i,j] := nz+i;
           end;
          end;
        if (pv2[i,j] = nz+i) and (pv3[i,j] > 0) then
          begin
          pv2[i,j] := pv3[i,j];
          pv3[i,j] := nz+i;
          end;
        end; (* for i, j *)
    end;

  DefineModel_fNames(modelnames,Correspondence,VertexPar_f, param);
  if (observations = 1) then
    begin
(* This section should be consistent with InitialiseBase1 *)
    if ExistStruct[1] and (not (Gibbs in [1,5])) then // (Gibbs in [2,3,8]) then
      begin
      if (callingprogram = 1) then
        x := format('%s%d%s%s%s',['The default Gibbs option is ', Gibbscode, ' (',Gibbsstring,')'])+eol
       else
        x := format('%s%d%s%s%s',['The requested Gibbs option is ', Gibbscode, ' (',Gibbsstring,')'])+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'but this is (provisionally) incompatible with the existence of structurally determined values.'+eol;
      outf.Write(pchar(x)^,length(x));
//      Gibbs := 4;
//      Gibbscode := 14;
      Gibbs := 1;
      Gibbscode := 11;

      Continuous := true;
      Gibbsstring := 'Gibbs steps for single tie variables, continuous chain';
      x := format('%s%d%s%s%s',['It is changed to Gibbs option ', Gibbs, ' (',Gibbsstring,')'])+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    if ((Gibbs = 6) and (not symmetric)) then
      begin
      if (callingprogram = 1) then
       x := format('%s%d%s%s%s',['The default Gibbs option is ', Gibbscode, ' (',Gibbsstring,')'])+eol
      else
       x := format('%s%d%s%s%s',['The requested Gibbs option is ', Gibbscode, ' (',Gibbsstring,')'])+eol;
      outf.Write(pchar(x)^,length(x));
      x :='but this is incompatible with a directed network.'+eol;
      outf.Write(pchar(x)^,length(x));
      Gibbs := 4;
      Gibbscode := 14;
      Continuous := true;
      Gibbsstring := 'M.H. steps for single tie variables, version A, continuous chain';
      x := format('%s%d%s%s%s',['It is changed to Gibbs option ', Gibbscode, ' (',Gibbsstring,')'])+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    if ((Gibbs = 7) and (not asymmetric)) then
      begin
      if (callingprogram = 1) then
       x := format('%s%d%s%s%s',['The default Gibbs option is ', Gibbscode, '(',Gibbsstring,')'])+eol
      else
       x := format('%s%d%s%s%s',['The requested Gibbs option is ', Gibbscode, '(',Gibbsstring,')'])+eol;
      outf.Write(pchar(x)^,length(x));
      x :='but this is incompatible with a network that is not antisymmetric.'+eol;
      outf.Write(pchar(x)^,length(x));
      Gibbs := 4;
      Gibbscode := 14;
      Continuous := true;
      Gibbsstring := 'M.H. steps for single tie variables, version A, continuous chain';
      //writeln(outf,'It is changed to Gibbs option ', Gibbscode, ' (',Gibbsstring,')');
      x := format('%s%d%s%s%s',['It is changed to Gibbs option ', Gibbscode, ' (',Gibbsstring,')'])+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    if (symmetric and ((Gibbs <> 6) and (Gibbs <> 16))) then
      begin
(* if this combination is not allowed : *)
wwriteln(outf,'Gibbs '+inttostr(Gibbs));
      x := format('%s%d%s',['The observed graph is symmetric but the Gibbs option is ',Gibbscode,','])+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'designed for directed networks.'+eol;
      outf.Write(pchar(x)^,length(x));
      Gibbscode := 16;
      Gibbs := 6;
      Continuous := true;
      Gibbsstring := 'M.H. nondirected, continuous chain';
      x := format('%s%d%s%s%s',['This is changed to Gibbs option ', Gibbscode,' (',Gibbsstring,').'])+eol;
      outf.Write(pchar(x)^,length(x));
//if this combination is allowed:      writeln(outf,'Note that the observed graph is undirected, but analysed as a directed graph.');
      end;
    if (asymmetric and (Gibbs <> 7) and (Gibbs <> 17)) then
       begin
       x := 'Note that the observed digraph is antisymmetric, but analysed as a general directed graph.'+eol;
       outf.Write(pchar(x)^,length(x));
       x := 'If you wish to analyze this digraph under the model assumption that it is antisymmetric,'+eol;
       outf.Write(pchar(x)^,length(x));
       x := 'use code 17 for the Gibbs/MH steps (in the advanced options).'+eol;
       outf.Write(pchar(x)^,length(x));
       asymmetric := false;
       end;
      if ((Gibbs in [9,10,19,20,21,22]) and symmetric) then
        begin
        if (callingprogram = 1) then
         x := format('%s%d%s%s%s',['The default Gibbs option is ', Gibbscode, ' (',Gibbsstring,').'])+eol
        else
         x := format('%s%d%s%s%s',['The requested Gibbs option is ', Gibbscode, ' (',Gibbsstring,').'])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'but this is incompatible with a non-directed network.'+eol;
        outf.Write(pchar(x)^,length(x));
        Gibbs := 8;
        Gibbscode := 18;
        Continuous := true;
        Gibbsstring := 'M.H. fixed in- and out-degrees, continuous chain';
        x := format('%s%d%s%s%s',['It is changed to Gibbs option ', Gibbscode, ' (',Gibbsstring,')'])+eol;
        outf.Write(pchar(x)^,length(x));
        end;                   

    effects_f[1] := (not Cconditional) and (not VertexPar)
                      and (not (uponly[0] or downonly[0]));
    if (uponly[0] or downonly[0]) then fix_f[1] := true;


{    if (VertexPar  or (Gibbs = 6))
    then begin
      someskip := false;
      for i:=1 to MaxEffects_f do if (VertexPar_f[i] and effects_f[i]) then someskip := true;
      (* These are effects that depend on the degees only, and therefore
         are not estimable if incidental vertex parameters are included,
         or if Gibbs sampling conditional on the degrees is done. *)
      if someskip
      then begin
        writeln(outf,'For the option with incidental vertex parameters, ');
        write(outf,'the following effects are unidentified and therefore are deleted from the model :');
        for i:=1 to MaxEffects_f
        do if (VertexPar_f[i] and effects_f[i])
        then begin
          writeln(outf);
          write(outf,modelnames[i]);
          effects_f[i] := false;
        end;
        writeln(outf,'.');
      end;
    end;  (* if (VertexPar  or (Gibbs = 6)) *)
}

  end  (* if observations = 1 *)
  else continuous := false;

  (* set basic rate parameters *)
  for i := 1 to observations-1
  do begin
    Effects_l[i]:=(not skipi[i]); (* network rate *)
    for j:= 1 to nza
    do Effects_l[NetworkEffects_l+(j-1)*ActionEffects_l+i]:=(not skipi[i]);
  end;

  (* adjust rate parameters for conditional estimation *)
  if Cconditional
  then begin
    for i := 1 to observations-1
    do begin
      if ConditioningNetworkVariable=1
      then Effects_l[i]:=false
      else Effects_l[NetworkEffects_l+(ConditioningActionVariable-1)*ActionEffects_l+i]:=false
    end
  end;


  (* adjust rate parameters for reciprocity model *)
  if ReciprocityModel then begin
    d_estimator := 0;
    someskip := 0;
    for i:= observations to NetworkEffects_l do
      if Effects_l[i] then begin
       someskip := i;
       Effects_l[i] := false;
      end; (* for i / if *)
    if (someskip > 0) then begin
      x := 'For the Reciprocity Model (model code 6),'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'no other rate effects are allowed. ' + eol;
      outf.Write(pchar(x)^,length(x));
      x := 'These effects in the model specification are skipped.' +eol;
      outf.Write(pchar(x)^,length(x));
     end;
    ExtendedReciprocityModel := false;
    for i:= 3 to NetworkEffects_f do
     if (effects_f[i] or effects_g[i]) then ExtendedReciprocityModel := true;
    ExtendedReciprocityModel := true;
    (*  this last statement overrides the earlier assignments and is not necessary, @@@@@
         but it turns out that the ExtendedReciprocityModel is faster than the
         ReciprocityModel. *)
    if ExtendedReciprocityModel then SetLength(lambdar,n+1);
   end;

  (* check and set extra parameters for various statistics and effects *)
  someskip := 0;
  with NewNetworkEffects do
    begin
    for i:=1 to NumberEffects do
      begin
      (* The following statement only operates upon interaction effects. *)
      SpecifyInteractionEffect(i,param_effect[i],code,ertext);
      if (ertext > '') then
        begin
        x := format('%s%d%s',['*** Warning: incorrect attempt to specify interaction effect ',i,' ***'])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'See the index numbers of evaluation function effects in the file '+pname+'.eff . '+eol;
        outf.Write(pchar(x)^,length(x));
        x := ertext+eol;
        outf.Write(pchar(x)^,length(x));
        someskip := i;
        effects_f[i] := false;
        effects_g[i] := false;
        end
       else
(* The following is a hack, not elegant.
  The reason is that Digraph does not know the Siena data and variable names
  and here we wish to have the variable name inserted in the middle of the
  interaction effect name. *)
        if (code > 0) then
          begin
          if (code <= 999) and (code > nz+nzc) then
            begin
            ertext := 'There is no actor variable numbered '+inttostr(code)
                  +' for interacting with transitivity.';
            x := ertext+eol;
            outf.Write(pchar(x)^,length(x));
            DeleteName(i);
            someskip := i;
            param_effect[i] := 0;
            effects_f[i] := false;
            effects_g[i] := false;
            end;
          if (code > 1000+ nzz+nzzc) then
            begin
            ertext := 'There is no dyadic variable numbered '
                   +inttostr(code-1000)
                  +' for interacting with transitivity.';
            param_effect[i] := 0;
            DeleteName(i);
            someskip := i;
            effects_f[i] := false;
            effects_g[i] := false;
            x := ertext+eol;
            outf.Write(pchar(x)^,length(x));
            end;
          if (code <= nz+nzc) then AppendName(i,Znames[code]);
          if (code <= nz+nzc) and ((param_effect[i] div 1000000) in [1..4,11..14])
             then AppendName(i,', '+inttostr((param_effect[i] div 1000000)));
          if (code >= 1001) and (code <= 1000+nzz+nzzc)
//             and ((param_effect[i] div 1000000)-10 in [1..4,11..14])
             then AppendName(i,ZZnames[code-1000]);
          if (code >= 1001) and (code <= 1000+nzz+nzzc)
             and ((param_effect[i] div 1000000)-10 in [1..4])
             then AppendName(i,', '+inttostr((param_effect[i] div 1000000)-10));
          if (code >= 1001) and (code <= 1000+nzz+nzzc)
             and ((param_effect[i] div 1000000)-10 = 8)
             then AppendName(i,' = 1');
          end;
      if (not SetParam(i,param_effect[i])) then
        begin
        someskip := i;
        x := format('%s%d%s%d%s',['Parameter value ',param_effect[i],
          ' is not permitted for effect ',i,'.'])+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      if (kind(i) = 4) and (param_effect[i] = 0) and (effects_f[i] or effects_g[i]) then
        begin
        x := 'Interaction effect with parameter 0 is unspecified and cannot be used.'+eol;
        outf.Write(pchar(x)^,length(x));
        effects_f[i] := false;
        effects_g[i] := false;
        end;
      if (observations >= 2) and (not (maxlike or bayesian)) then
        begin
(* check *)
if endowed(i) <> endowedindex(i) then  begin
x := format('%s%d',['1234 index ??? ',i])+eol;
outf.Write(pchar(x)^,length(x));
end;
        if (effects_g[i] and (not EndowedIndex(i))) then
          begin
//          x := format('%s%s%s',['There is no endowment function for effect ',
//           NameIndex(i),'; it is skipped from the specification.'])+eol;
          x := 'There is no endowment function for effect '+NameIndex(i)+';'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'it is skipped from the specification.'+eol;
          outf.Write(pchar(x)^,length(x));
          effects_g[i] := false;
          thetarandom_g[i] := false;
          fix_g[i] := false;
          test_g[i] := false;
          alpa_g[i] := 0.0;
          end;
        end; (* if observations >= 2 *)
      end; (* for i *)
    end; (* with NewNetworkeffects *)

  if (someskip > 0) then
    begin
    x :='Warning:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%d%s',['You chose some parameter values in effects that are not permitted (code ',someskip,').'])+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'This was restored to permitted values.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Estimation can be resumed now.'+eol;
    outf.Write(pchar(x)^,length(x));
    end;

  if (observations = 1) then
    begin
(* The following is reproduced in DefineSimFunctions, but then for the statistics *)
    if (index_kinstars > 0) then param_kinstars := NewNetworkeffects.Param(index_kinstars)
      else param_kinstars := 0;
    if (index_koutstars > 0) then param_koutstars := NewNetworkeffects.Param(index_koutstars)
      else param_koutstars := 0;
    if (index_ktri > 0) then param_ktri := NewNetworkeffects.Param(index_ktri)
      else param_ktri := 0;
    if (index_ktwop > 0) then param_ktwop := NewNetworkeffects.Param(index_ktwop)
      else param_ktwop := 0;

    pclique_min := 0;
    pclique_max := 0;
    if (index_cliq1 >= 1) then
     begin
     if effects_f[index_cliq1] then
      begin
      param1_clique := NewNetworkeffects.Param(index_cliq1);
      minmax(pclique_min,pclique_max,param1_clique);
     end
      else param1_clique := 0;
     end;
    if (index_cliq2 >= 1) then
     begin
     if effects_f[index_cliq2] then
      begin
      param2_clique := NewNetworkeffects.Param(index_cliq2);
      minmax(pclique_min,pclique_max,param2_clique);
     end
      else param2_clique := 0;
     end;
    if (index_cliq3 >= 1) then
     begin
     if effects_f[index_cliq3] then
      begin
      param3_clique := NewNetworkeffects.Param(index_cliq3);
      minmax(pclique_min,pclique_max,param3_clique);
     end
      else param3_clique := 0;
     end;
    if (index_cliq4 >= 1) then
     begin
     if effects_f[index_cliq4] then
      begin
      param4_clique := NewNetworkeffects.Param(index_cliq4);
      minmax(pclique_min,pclique_max,param4_clique);
     end
      else param4_clique := 0;
     end; 
    param1_clique_s := param1_clique;
    param2_clique_s := param2_clique;
    param3_clique_s := param3_clique;
    param4_clique_s := param4_clique;
    pclique_min_s := pclique_min;
    pclique_max_s := pclique_max;
    end; (* if observations <= 1 *)

    if (nza >= 1) then
      begin
      if effects_f[NetworkEffects_f + 9] or effects_f[NetworkEffects_f + 10] then  begin
       x := format('%s%d%s',['Parameter for dense triads = ',param1_clique,
       ' is also used for behavior dynamics.'])+eol;
       outf.Write(pchar(x)^,length(x));
      end;
      end;
(* This parameter is the parameter for dense triads;
   it is also used for the dense triad statistics in the action statistics. *)

//    if (not symmetric) and (ModelType = 3) and Cconditional
//    then begin
//      writeln(outf,'Conditional option not implemented for Model Type 3.');
//      writeln(outf,'Changed to unconditional simulation.');
//      Cconditional := false;
//    end;
    if ((OutputType = 2) and not ((not symmetric) and (ModelType = 2))) then CalcEntropy := true;

  if ((observations >= 2) and (ModelType = 2) and (not symmetric))
(* TS: This must still be adapted for Siena version 3. @@@ *)
  then begin
    eff_od := (effects_f[1] or effects_f[9] or effects_f[10]
               (* density OR outdeg. up to 3 OR sqrt(outdeg)... *)
               or effects_f[11] or effects_f[12] or effects_f[13]);
               (* ...OR sqr(outdeg) OR inverse(outdeg) OR inv.-sqr.(outdeg)... *)
    for i:=1 to nz+nzc do eff_od := (eff_od or effects_f[16+2*nzz+10*(i-1)+3]);
                                    (* ...OR outdegree  covariate *)
    if eff_od
    then begin
      x := 'For Model Type 2, outdegree effects in the objective function'+eol;
      outf.Write(pchar(x)^,length(x));
      x :='are meaningless. They are deleted from the model specification.'+eol;
      outf.Write(pchar(x)^,length(x));
      effects_f[1] := false;
      for i:= 9 to 13 do effects_f[i] := false;    // @@@ must be changed
      for i:= 1 to nz+nzc do effects_f[16+2*nzz+10*(i-1)+3] := false;
    end;
  end;

  if (observations >= 2) then for i:=1 to nza do
    begin
    for j:= 1 to nz+nzc-1 do
      begin
//writeln(outf,   nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 17 + (2*j),' ',   //abcd
//modelnames[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 17 + (2*j)],': ',
//param_effect[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 17 + (2*j)],' ',
//effects_f[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 17 + (2*j)]);
      if effects_f[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + (2*j)]
        and (not (param_effect[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + (2*j)]
             in [1..3])) then
        begin
        x := format('%s%d%s%s%s',['WARNING: you included the behavior interaction effect ',
              nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + (2*j),
               ', ',modelnames[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + (2*j)],','])+eol;
        outf.Write(pchar(x)^,length(x));
        x := format('%s%d%s',['but specifying the parameter as ',
                param_effect[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + (2*j)],
                ', which is not in the permissible range from 1 to 3.'])+eol;
        outf.Write(pchar(x)^,length(x));
        x := 'This now is a meaningless effect, the statistic is 0.'+eol;
        outf.Write(pchar(x)^,length(x));
        if (param_effect[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + (2*j)] >= 4)
           then
          begin
          x := 'This is reset to 0.'+eol;
          outf.Write(pchar(x)^,length(x));
          param_effect[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + (2*j)] := 0;
          end;
        end;
      end;
    end; (* for i *)

  if ((observations >= 2) and (ModelType = 3) and (not symmetric))
  then begin
    eff_l := false;
    for i:=observations to NetworkEffects_l do if effects_l[i]
    then begin
      eff_l := true;
      effects_l[i]:= false;
      x := format('%s%d',['eff_l ',i])+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    if eff_l
    then begin
      x := 'For Model Type 3, rate effects for network change'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'are not implemented. They are deleted from the model specification.'+eol;
      outf.Write(pchar(x)^,length(x));
    end;
  end;

{  if ((observations >=2) and (effects_f[1] and effects_f[10] and effects_f[2]
                              (* outdegree AND sqr(outdeg.) AND reciprocity... *)
                              and effects_f[3] and effects_f[4] and effects_f[7])
                              and (not signed))  (*[s]*)
                              (* ...AND transitivity AND balance AND activity(alter) *)
  then begin
    writeln(outf,'You selected six effects which together are linearly dependent:');
    writeln(outf,'outdegree, squared outdegree, reciprocity, ');
    writeln(outf,'transitive triplets, balance, and alter`s activity.');
    writeln(outf,'The linear dependence leads to a redundant parameter.');
    writeln(outf,'Therefore, the balance effect is deleted from the model.');
    effects_f[4] := false;
// @@@ must be changed
  end;}

  if ((observations >=2) and (effects_f[1] and effects_g[1]) and (not signed) and not(MaxLike or Bayesian))
                              (* outdegree in evaluation and endowment function *)
  then begin
    x :='You selected two effects which together with the rate function'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'are linearly dependent: outdegree-evaluation and outdegree-endowment.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'The linear dependence leads to a redundant parameter.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Therefore, the outdegree-endowment effect is deleted from the model.'+eol;
    outf.Write(pchar(x)^,length(x));
    effects_g[1] := false;
  end;

  if ((maxlike or Bayesian) and exo) then
    begin
    x := 'Warning: composition change is not implemented for likelihood-based estimation.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Instead, this can be represented by structural zeros.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Composition change is suppressed in the following estimation.'+eol;
    outf.Write(pchar(x)^,length(x));
    exo := false;
    x := 'Change exo (composition change) to false.'+eol;
    lf.Write(pchar(x)^,length(x));
    end;

  if (MaxLike or Bayesian) then (*MS: adapted.*)
    begin
    eff_l := false;
    for i := observations to NetworkEffects_l do
      begin
      if effects_l[i] then
        begin
        eff_l := true;
        effects_l[i] := false;
        end;
      if eff_l then
        begin
        x := 'For maximum likelihood and Bayesian estimation, non-constant rate functions for the network evolution'+eol;
        outf.Write(pchar(x)^,length(x));
        x :='are not implemented and hence replaced by constant rate functions.'+eol;
        outf.Write(pchar(x)^,length(x));
        end;
      end;
    end; (* if MaxLike or Bayesian *)


  if (MaxLike or Bayesian) then (*MS: adapted.*)
    begin
    if (ModelType >= 2) then
      begin
      x := 'Maximum likelihood and Bayesian estimation only implemented for ModelType = 1.'+eol;
      outf.Write(pchar(x)^,length(x));
      wwriteln(outf,'This now is restored.');
      ModelType := 1;
      end;
    end; (* if MaxLike or Bayesian *)

  (* Temporarily, the d_estimator option is not implemented for action endowment effects: *)
  (* This is changed now. *)
  if (d_estimator >= 1) then
    begin
//    for i:=1 to nza do
//      if AnyActionEffects_g[i] then
//        begin
 //       d_estAB := false;
 //       d_estA := false;
 //       d_estB := false;
 // abcdef skipped 3.17d.       d_estimator := 0;
 //       end;
// AA change: condition callingprogram added
    if (d_estimator = 0) and (CallingProgram <> 1) then
      begin
      x := 'In the current version, the option of LR estimation of the derivative matrix'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'is not yet implemented for models with endowment effects for dependent behavior variables.'+eol;
      outf.Write(pchar(x)^,length(x));
      x := 'Therefore this option is changed to the Finite Differences method.'+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    if symmetric then
      begin
        d_estAB := false;
        d_estA := false;
        d_estB := false;
        d_estimator := 0;
// AA change: condition callingprogram added
        if (CallingProgram <> 1) then
          begin
          x := 'In the current version, the option of LR estimation of the derivative matrix'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'is not yet implemented for non-directed networks.'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'Therefore this option is changed to the Finite Differences method.'+eol;
          outf.Write(pchar(x)^,length(x));
          end;
      end;
    if (d_estimator >= 1) and (modeltype = 3) then
      begin
        d_estAB := false;
        d_estA := false;
        d_estB := false;
        d_estimator := 0;
// AA change: condition callingprogram added
        if (CallingProgram <> 1) then
          begin
          x := 'In the current version, the option of LR estimation of the derivative matrix'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'is not yet implemented for tie-based models.'+eol;
          outf.Write(pchar(x)^,length(x));
          x := 'Therefore this option is changed to the Finite Differences method.'+eol;
          outf.Write(pchar(x)^,length(x));
          end;
      end;
    end;

  (* Now determine whether model components are minimally present in the current specification: *)
  (* rate functions: *)
  tempi := 0;                 
  atleastone := false;
  for i:=1 to MaxEffects_l do if effects_l[i]
  then begin
    inc(tempi);
    atleastone := true;
  end;
  effects_l[0] := atleastone;
  (* objective functions: *)
  atleastone := false;
  for i:=1 to MaxEffects_f do if effects_f[i]
  then begin
    inc(tempi);
    atleastone := true;
  end;
  effects_f[0] := atleastone;
  (* gratification functions: *)
  atleastone := false;
  for i:=1 to MaxEffects_g do if effects_g[i]
  then begin
    inc(tempi);
    atleastone := true;
  end;
  effects_g[0] := atleastone;
  NumberParameters := tempi;
  (*MS: <NumberParameters> is the number of fixed effects: in random effects models,
  the (co)variance parameters are not counted.*)
  if exo then SetEventLengths;


  NewNetworkEffects.DefineModSequences(outf,effects_f,effects_g,(MaxLike or Bayesian));


end; (* procedure ReviseDefinition *)

procedure ModelSpecification.SetEventLengths;
begin
  SetLength(event,n+2);
  setlength(active_m,n+1,observations+2);
  setlength(event_m,n+1,observations+2);
end;


function ModelSpecification.Contr_fa(i           : integer;  (* actor            *)
                                     h,                      (* effect number    *)
                                     k           : byte;     (* behaviour number *)
                                     dif         : shortint; (* behaviour change *)
                                     thecomponents: array_0_20;
                                     var halterr : Boolean;
                                     var outf,lf : TFileStream) : Real;
(* The contribution of a change dif in ya(i,k) and Action effect h,              *)
(* to the objective function of i.                                               *)
(* should be called only for observations >= 2                                   *)
(* and for 1 <= h <= ActionEffects_f
   First call CalcComponents_fa !
   If CalcComponents_fa has been called with dif = 0,
   then contr_fa gives the value (not the change value)
   of the behavior evaluation function.for behavior dif.     --- ??? ---         *)

var h2,h3,h4,h5,
    dumin   : integer;
    dumrl,interact,dr   : real;
    ers,x     : string;

  procedure peripheral;
  (* updates peripheral count of actor i, externalized due to transparency of code  *)
  (* triad h3-h4-h5 must be identified as 'dense' before calling this procedure     *)
  begin
    if yn(i,h3) then inc(dumin);
    if yn(i,h4) then inc(dumin);
    if yn(i,h5) then inc(dumin);
  end; (* peripheral *)

  procedure peripheralsimilarity;
  (* updates peripheral  similarity sum of actor i, also externalized for lucidity *)
  (* triad h3-h4-h5 must be identified as 'dense' before calling this procedure     *)
  var addrl : real;
  begin
    if yn(i,h3)
    then begin
      addrl :=    similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                + similarity(ya(i,k)+dif,ya(h5,k),k+nz)
                - similarity(ya(i,k) , ya(h3,k) , k+nz)
                - similarity(ya(i,k) , ya(h4,k) , k+nz)
                - similarity(ya(i,k) , ya(h5,k) , k+nz);
      dumrl := dumrl + addrl;
      if yn(i,h4) then dumrl := dumrl + addrl;
      if yn(i,h5) then dumrl := dumrl + addrl
    end
    else if yn(i,h4)
    then begin
      addrl :=    similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                + similarity(ya(i,k)+dif,ya(h5,k),k+nz)
                - similarity(ya(i,k) , ya(h3,k) , k+nz)
                - similarity(ya(i,k) , ya(h4,k) , k+nz)
                - similarity(ya(i,k) , ya(h5,k) , k+nz);
      dumrl := dumrl + addrl;
      if yn(i,h5) then dumrl := dumrl + addrl
    end
    else if yn(i,h5)
    then dumrl := dumrl + similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                + similarity(ya(i,k)+dif,ya(h5,k),k+nz)
                - similarity(ya(i,k) , ya(h3,k) , k+nz)
                - similarity(ya(i,k) , ya(h4,k) , k+nz)
                - similarity(ya(i,k) , ya(h5,k) , k+nz);
  end; (* peripheralsimilarity *)

function zvar(p,i: integer): real;
begin
 if (p <= 0) then zvar := 1.0
  else
   begin
   if (p > nz) and (p <= nz+nza) then (* one of the dependent behavior variables *)
     zvar := (ya(i,p-nz)-zamean[p-nz])
    else (* actor covariate *) zvar := z[i,p];
   end;
end; (* zvar *)

begin (* Contr_fa *)
(* To understand this procedure, look at CalcComponents_fa
   for the calculation of thecomponents.
   When this is changed, check if  CalcComponents_fa must be changed. *)
(* In ActionStatistics, Contr_fa is used only when the component is linear in dif *)
//wwriteln(outf,'contr_fa  00 :: '+inttostr(h));                                       
  case h of
     1 : Contr_fa := dif;        (* main effect behaviour                       *)
     2 : if (thecomponents[0] <= 0.1)       (* ave. similarity to outgoing neighbours           *)
         then Contr_fa := 0.0
         else
           Contr_fa := (thecomponents[2])/thecomponents[0];
     3 :      (* tot. similarity to outgoing neighbours           *)
         Contr_fa := (thecomponents[2]);
         (*  The original, quicker routine:
         case dif of
           -1:  Contr_fa := (zamin[i,k] - zaeq[i,k] - zaplus[i,k]);
            1:  Contr_fa := (zaplus[i,k] - zaeq[i,k] - zamin[i,k]);
           else Contr_fa := 0.0;
         end;
         *)      //1234
     4 : Contr_fa := Ind_yn(i)*dif;  (* indegree  behaviour                       *)
     5 : Contr_fa := outd_yn(i)*dif; (* outdegree  behaviour                      *)
     6 : if (Ind_yn(i)<=0) (* isolate  behaviour                        *)
           then Contr_fa := dif
           else Contr_fa := 0.0;
     7 :             (* ave. similarity  reciprocity                    *)
         if (thecomponents[1] <= 0.1)
         then Contr_fa := 0.0
         else Contr_fa := thecomponents[7]/thecomponents[1];
     8 : begin            (* tot. similarity  reciprocity                    *)
         Contr_fa := thecomponents[7];
//         begin
//           dumrl:=0.0;
//           dumin:=0;
//           for h3:=n downto 1 do
//           if (yn(i,h3) and yn(h3,i))
//           then begin
//             dumrl:=dumrl + similarity(ya(i,k)+dif,ya(h3,k),k+nz)
//                          - similarity(ya(i,k)    ,ya(h3,k),k+nz);
                      (* again, calculation relies on ya(i,.)+dif remaining between zmin[.] and zmax[.] *)
//             inc(dumin)
//           end;
//           if (dumin <= 0) then Contr_fa := 0.0 else Contr_fa := dumrl/dumin; // @@@@
         end;
     9 : if (thecomponents[0] <= 0.1)       (* ave. similarity  popularity alter               *)
         then Contr_fa := 0.0
         else
           Contr_fa := thecomponents[9]/thecomponents[0];
     10 :        (* pop. alter, or tot. similarity  popularity alter               *)
          if (param_effect[nxa*NetworkEffects_f + (k-1)*ActionEffects_f + h] = 1) then
            begin
            if (outd_yn(i) >= 1) then Contr_fa := dif*thecomponents[10]/outd_yn(i)
             else Contr_fa := 0.0;
            end
           else Contr_fa := thecomponents[9];
//           begin
//           dumrl:=0.0;
//           for h3:=n downto 1 do
//           if yn(i,h3)
//           then dumrl:=dumrl + (  similarity(ya(i,k)+dif,ya(h3,k),k+nz)
//                                - similarity(ya(i,k)    ,ya(h3,k),k+nz) ) * ind_yn(h3);
//                      (* again, calculation relies on ya(i,.)+dif remaining between zmin[.] and zmax[.] *)
//           Contr_fa := dumrl/outd_yn(i)
//         end;
     11 :                  (* ave. similarity  reciprocity  pop.(alter)      *)
         if (thecomponents[1] <= 0.1) then Contr_fa := 0.0
           else Contr_fa := thecomponents[11]/thecomponents[1];
     12 :                  (* tot. similarity  reciprocity  pop.(alter)      *)
           Contr_fa := thecomponents[11];
//         else begin
//           dumrl:=0.0;
//           dumin:=0;
//           for h3:=n downto 1 do
//           if (yn(i,h3) and yn(h3,i))
//           then begin
//             dumrl:=dumrl + (  similarity(ya(i,k)+dif,ya(h3,k),k+nz)
//                          - similarity(ya(i,k)    ,ya(h3,k),k+nz) ) * ind_yn(h3);
                      (* again, calculation relies on ya(i,.)+dif remaining between zmin[.] and zmax[.] *)
//             inc(dumin)
//           end;
//           Contr_fa := dumrl/dumin
//         end;
    13 : (* average alter effect *)
         begin
         if (thecomponents[0] <= 0.1)
            then Contr_fa := 0.0
           else Contr_fa := dif*((thecomponents[13]/thecomponents[0])-zamean[k]); // 1234
//             else Contr_fa := dif*((thecomponents[13]/thecomponents[0])); // 1234
         end;
    14 :  (*  average reciprocated alter *)
         if (thecomponents[1] <= 0.1) then Contr_fa := 0.0
           else Contr_fa := dif*((thecomponents[14]/thecomponents[1])-zamean[k]);

{         begin                   (* peripheral to dense triad  behaviour          *)
           dumin:=0;
           case param_effect[15] of
             6 : begin (* dense triad defined with 6 ties *)
                   for h3:=n downto 1
                   do if not yn(h3,i)
                   then for h4:=h3-1 downto 1
                   do if (yn(h3,h4) and yn(h4,h3) and not yn(h4,i))
                   then for h5:=h4-1 downto 1
                   do if (yn(h3,h5) and yn(h5,h3) and yn(h4,h5) and yn(h5,h4) and not yn(h5,i))
                   then peripheral
                 end; (* dense triad defined with 6 ties *)
             5 : begin (* dense triad defined with 5 ties *)
                   for h3:=n downto 1
                   do begin
                      if not yn(h3,i)
                      then for h4:=h3-1 downto 1
                      do if not yn(h4,i)
                      then begin
                        if yn(h3,h4)
                        then begin (*1*)
                          if yn(h4,h3)
                          then begin (*2*)
                            for h5:=h4-1 downto 1
                            do if not yn(h5,i)
                            then begin
                              if yn(h3,h5)
                              then begin (*3*)
                                if yn(h5,h3)
                                then begin (*4*)
                                  if yn(h4,h5)
                                  then peripheral (*5*)
                                  else if yn(h5,h4)
                                       then peripheral (*5*)
                                end (*4*)
                                else begin (*3*)
                                  if (yn(h4,h5) and yn(h5,h4))
                                  then peripheral (*5*)
                                end(*3*)
                              end (*3*)
                              else begin (*2*)
                                if (yn(h5,h3) and yn(h4,h5) and yn(h5,h4))
                                then peripheral (*5*)
                              end (*2*)
                            end (*loop 3rd actor h5*)
                          end (*2*)
                          else begin (*1*)
                            for h5:=h4-1 downto 1
                            do if not yn(h5,i)
                            then begin
                              if (yn(h3,h5) and yn(h5,h3) and yn(h4,h5) and yn(h5,h4))
                              then peripheral (*5*)
                            end (*loop 3rd actor h5*)
                          end (*1*)
                        end (*1*)
                        else begin (*0*)
                          if yn(h4,h3)
                          then begin (*1*)
                            for h5:=h4-1 downto 1
                            do if not yn(h5,i)
                            then begin
                              if (yn(h3,h5) and yn(h5,h3) and yn(h4,h5) and yn(h5,h4))
                              then peripheral (*5*)
                            end (*loop 3rd actor h5*)
                          end (*1*)
                        end (*0*)
                      end (*loop 2nd actor h4*)
                   end; (*loop 1st actor h3*)
                 end; (* dense triad defined with 5 ties *)
           end; (* case *)
           Contr_fa := dumin*dif
         end;
}
    15 : (* squared effect beh. on beh.; nonlinear, therefore not used in ActionStatistics *)
         Contr_fa := 1.0 + 2.0*(ya(i,k)-zamean[k])*dif;

{         begin                   (* peripheral to dense triad  similarity         *)
           dumrl:=0;
           case param_effect[15] of
             6 : begin (* dense triad defined with 6 ties *)
                   for h3:=n downto 1
                   do if not yn(h3,i)
                   then for h4:=h3-1 downto 1
                   do if (yn(h3,h4) and yn(h4,h3) and not yn(h4,i))
                   then for h5:=h4-1 downto 1
                   do if (yn(h3,h5) and yn(h5,h3) and yn(h4,h5) and yn(h5,h4) and not yn(h5,i))
                   then peripheralsimilarity
                 end; (* dense triad defined with 6 ties *)
             5 : begin (* dense triad defined with 5 ties *)
                   for h3:=n downto 1
                   do begin
                      if not yn(h3,i)
                      then for h4:=h3-1 downto 1
                      do if not yn(h4,i)
                      then begin
                        if yn(h3,h4)
                        then begin (*1*)
                          if yn(h4,h3)
                          then begin (*2*)
                            for h5:=h4-1 downto 1
                            do if not yn(h5,i)
                            then begin
                              if yn(h3,h5)
                              then begin (*3*)
                                if yn(h5,h3)
                                then begin (*4*)
                                  if yn(h4,h5)
                                  then peripheralsimilarity (*5*)
                                  else if yn(h5,h4)
                                       then peripheralsimilarity (*5*)
                                end (*4*)
                                else begin (*3*)
                                  if (yn(h4,h5) and yn(h5,h4))
                                  then peripheralsimilarity (*5*)
                                end(*3*)
                              end (*3*)
                              else begin (*2*)
                                if (yn(h5,h3) and yn(h4,h5) and yn(h5,h4))
                                then peripheralsimilarity (*5*)
                              end (*2*)
                            end (*loop 3rd actor h5*)
                          end (*2*)
                          else begin (*1*)
                            for h5:=h4-1 downto 1
                            do if not yn(h5,i)
                            then begin
                              if (yn(h3,h5) and yn(h5,h3) and yn(h4,h5) and yn(h5,h4))
                              then peripheralsimilarity (*5*)
                            end (*loop 3rd actor h5*)
                          end (*1*)
                        end (*1*)
                        else begin (*0*)
                          if yn(h4,h3)
                          then begin (*1*)
                            for h5:=h4-1 downto 1
                            do if not yn(h5,i)
                            then begin
                              if (yn(h3,h5) and yn(h5,h3) and yn(h4,h5) and yn(h5,h4))
                              then peripheralsimilarity (*5*)
                            end (*loop 3rd actor h5*)
                          end (*1*)
                        end (*0*)
                      end (*loop 2nd actor h4*)
                   end; (*loop 1st actor h3*)
                 end; (* dense triad defined with 5 ties *)
           end; (* case *)
           Contr_fa := dumrl
         end;
}
     16 : begin                   (* group  behaviour                      *)
(* case 16 must be renewed - presumably incorrect now *)
           dumin:=0;
           case param1_clique of
             6 : begin (* dense triad defined with 6 ties *)
                   for h3:=n downto 1
                   do if (yn(i,h3) and yn(h3,i))
                   then for h4:=h3-1 downto 1
                   do if (yn(i,h4) and yn(h4,i) and yn(h3,h4) and yn(h4,h3))
                   then inc(dumin);
                 end; (* dense triad defined with 6 ties *)
             5 : begin (* dense triad defined with 5 ties *)
                   for h3:=n downto 1
                   do if yn(i,h3)
                   then begin (*1*)
                     if yn(h3,i) (*2*)
                     then begin
                       for h4:=h3-1 downto 1
                       do begin
                         if yn(i,h4) (*3*)
                         then begin
                           if yn(h4,i) (*4*)
                           then begin
                             if yn(h3,h4) (*5*)
                             then inc(dumin)
                             else begin (*4*)
                               if yn(h4,h3) (*5*)
                               then inc(dumin)
                             end
                           end (*4*)
                           else begin (*3*)
                             if (yn(h3,h4) and yn(h4,h3)) (*5*)
                             then inc(dumin)
                           end (*3*)
                         end (*3*)
                         else begin (*2*)
                           if (yn(h4,i) and yn(h3,h4) and yn(h4,h3)) (*5*)
                           then inc(dumin)
                         end (*2*)
                       end (* 3rd actor h4 *)
                     end (*2*)
                     else begin (*1*)
                       for h4:=h3-1 downto 1
                       do begin
                         if yn(i,h4) and yn(h4,i) and yn(h3,h4) and yn(h4,h3) (*5*)
                         then inc(dumin)
                       end (* 3rd actor h4 *)
                     end (*1*)
                   end (*1*)
                   else begin (*0*)
                     if yn(h3,i) (*1*)
                     then for h4:=h3-1 downto 1
                          do begin
                            if yn(i,h4) and yn(h4,i) and yn(h3,h4) and yn(h4,h3) (*5*)
                            then inc(dumin)
                          end (* 3rd actor h4 *)
                   end; (*0*)
                 end; (* dense triad defined with 5 ties *)
           end; (* case *)
           Contr_fa := dumin*dif
         end;
    17 : begin                   (* similarity in groups                     *)
(* case 17 must be renewed - presumably incorrect now *)
           dumrl:=0.0;
           case param1_clique of
             6 : begin (* dense triad defined with 6 ties *)
                   for h3:=n downto 1
                   do if (yn(i,h3) and yn(h3,i))
                   then for h4:=h3-1 downto 1
                   do if (yn(i,h4) and yn(h4,i) and yn(h3,h4) and yn(h4,h3))
                   then dumrl:=dumrl + similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                                     + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                                     - similarity(ya(i,k)    ,ya(h3,k),k+nz)
                                     - similarity(ya(i,k)    ,ya(h4,k),k+nz);
                 end; (* dense triad defined with 6 ties *)
             5 : begin (* dense triad defined with 5 ties *)
                   for h3:=n downto 1
                   do if yn(i,h3)
                   then begin (*1*)
                     if yn(h3,i) (*2*)
                     then begin
                       for h4:=h3-1 downto 1
                       do begin
                         if yn(i,h4) (*3*)
                         then begin
                           if yn(h4,i) (*4*)
                           then begin
                             if yn(h3,h4) (*5*)
                             then dumrl:=dumrl + similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                                               + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                                               - similarity(ya(i,k)    ,ya(h3,k),k+nz)
                                               - similarity(ya(i,k)    ,ya(h4,k),k+nz)
                             else begin (*4*)
                               if yn(h4,h3) (*5*)
                               then dumrl:=dumrl + similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                                                 + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                                                 - similarity(ya(i,k)    ,ya(h3,k),k+nz)
                                                 - similarity(ya(i,k)    ,ya(h4,k),k+nz);
                             end
                           end (*4*)
                           else begin (*3*)
                             if (yn(h3,h4) and yn(h4,h3)) (*5*)
                             then dumrl:=dumrl + similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                                               + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                                               - similarity(ya(i,k)    ,ya(h3,k),k+nz)
                                               - similarity(ya(i,k)    ,ya(h4,k),k+nz);
                           end (*3*)
                         end (*3*)
                         else begin (*2*)
                           if (yn(h4,i) and yn(h3,h4) and yn(h4,h3)) (*5*)
                           then dumrl:=dumrl + similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                                             + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                                             - similarity(ya(i,k)    ,ya(h3,k),k+nz)
                                             - similarity(ya(i,k)    ,ya(h4,k),k+nz);
                         end (*2*)
                       end (* 3rd actor h4 *)
                     end (*2*)
                     else begin (*1*)
                       for h4:=h3-1 downto 1
                       do begin
                         if (yn(i,h4) and yn(h4,i) and yn(h3,h4) and yn(h4,h3)) (*5*)
                         then dumrl:=dumrl + similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                                           + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                                           - similarity(ya(i,k)    ,ya(h3,k),k+nz)
                                           - similarity(ya(i,k)    ,ya(h4,k),k+nz);
                       end (* 3rd actor h4 *)
                     end (*1*)
                   end (*1*)
                   else begin (*0*)
                     if yn(h3,i) (*1*)
                     then for h4:=h3-1 downto 1
                          do begin
                            if (yn(i,h4) and yn(h4,i) and yn(h3,h4) and yn(h4,h3)) (*5*)
                            then dumrl:=dumrl + similarity(ya(i,k)+dif,ya(h3,k),k+nz)
                                              + similarity(ya(i,k)+dif,ya(h4,k),k+nz)
                                              - similarity(ya(i,k)    ,ya(h3,k),k+nz)
                                              - similarity(ya(i,k)    ,ya(h4,k),k+nz);
                          end (* 3rd actor h4 *)
                   end; (*0*)
                 end; (* dense triad defined with 5 ties *)
           end; (* case *)
           Contr_fa := dumrl (* note that this is not normalized but summed over all dense triads *)
         end;
     18 : Contr_fa := thecomponents[1]*dif;  (* reciprocated degree  behaviour    *)
     19 : if (thecomponents[0] <= 0.1)       (* popularity ego x
                                    ave. similarity to outgoing neighbours           *)
         then Contr_fa := 0.0
         else
           Contr_fa := Ind_yn(i)*thecomponents[2]/thecomponents[0];
(*
// This was experimental, seems not to give much. T.S. July 2008.
       18 : if (thecomponents[0] <= 0.1)        ave. similarity to outgoing neighbours, down
            then Contr_fa := 0.0
            else
              Contr_fa := (thecomponents[3])/thecomponents[0];
                Contr_fa := (thecomponents[3]);
       19 : if (thecomponents[0] <= 0.1)        ave. similarity to outgoing neighbours, up
            then Contr_fa := 0.0
            else
              Contr_fa := (thecomponents[4])/thecomponents[0];
              Contr_fa := (thecomponents[4]);
*)
    else (* case h *)
      begin (* effect number h>=20 *)
      if (h <= 20+2*(nz+nzc-1)) then
        begin
        (* effect h is effect of covariate or other behavioural dimension *)
        h2 := (h-18) div 2; (* number of actor variable *)
        h3 := h mod 2;
        (* main effects of covariates on action dynamics *)
        if h2 >=(nz+k) then inc(h2);  (* control for the fact that the behavioral variable k itself
                                       does not occur in the list of possible variable effects.
                                       Now h2 is the number of the actor variable in question. *)
//wwriteln(outf,'contr_fa :: '+inttostr(h)+' '+inttostr(h2)+' '+inttostr(h3)+' '
//     +inttostr(nz+nzc));
(* Changed 27/10/8: start *)
        if (h2 <= nz + nzc) then
          begin
          case h3 of
            0: interact := dif;
            1: case param_effect[nxa*NetworkEffects_f + (k-1)*ActionEffects_f + h] of
                1: begin
                   if (thecomponents[0] <= 0.1) then interact := 0.0
                    else interact := (thecomponents[2]+thecomponents[3])/thecomponents[0];
                   end;
                2: interact := (thecomponents[2]+thecomponents[3]);
                3: begin
                   if (thecomponents[0] <= 0.1) then interact := 0.0
                    else interact := dif*((thecomponents[13]/thecomponents[0])-zamean[k]);
                   end;
                end; (* case *)
             end; (* case h3 *)
          if (h2<=nz) or (h2>(nz+nza))
           (* Earlier comment: Erroneously, the last inequality was > instead of >= up to version 3 beta 3u *)
           (* I think now, version 3.12, that the supposed correction to >= was an error.
              After all, h2 now is just the index number of the variable. *)
          then Contr_fa := z[i,h2]*interact (* actor variable h is a covariate *)
          else Contr_fa := (ya(i,h2-nz)-zamean[h2-nz])*interact (* variable is another dependent variable *)
          end
         else   //          (h = 20+2*(nz+nzc-1))
          begin
(* one-sided influence: only by out-ties with higher behavior *)
          case param_effect[nxa*NetworkEffects_f + (k-1)*ActionEffects_f + h] of
              1: begin
                 if (thecomponents[0] <= 0.1) then Contr_fa := 0.0
                  else Contr_fa := thecomponents[15]/thecomponents[0];
                 end;
              2: Contr_fa := thecomponents[15];
              3: begin
                 if (thecomponents[0] <= 0.1) then Contr_fa := 0.0
                  else Contr_fa := ((thecomponents[16]/thecomponents[0])-zamean[k]);
                 end;
              end; (* case *)
          end;
        end (* if (h <= 20+2*(nz+nzc-1)) *)
(* Changed 27/10/8: end *)
      else
        begin
        if (h <= 18 + 2*(nz+nzc) + 2*ninta) then
          begin
          h2 := h-(18 + 2*(nz+nzc));
          case pv0[k,h2] of
(* pv0[h,k2] is the number of times that the dependent variable k itself
   occurs among the interacting variables *)
          0: Contr_fa := dif*zvar(pv1[k,h2],i)*zvar(pv2[k,h2],i)*zvar(pv3[k,h2],i);
          1: case ni[k,h2] of
               2: Contr_fa := zvar(pv1[k,h2],i)*
                             (1.0 + 2.0*dif*(ya(i,k)-zamean[k]));
               3: Contr_fa := zvar(pv1[k,h2],i)*zvar(pv2[k,h2],i)*
                             (1.0 + 2.0*dif*(ya(i,k)-zamean[k]));
              end; (* case *)
          2: begin
              dr := (ya(i,k)-zamean[k]);
              case ni[k,h2] of
               2: Contr_fa := (3.0*dr*(dr*dif + 1.0) + dif);
               3: Contr_fa := zvar(pv1[k,h2],i)*(3.0*dr*(dr*dif + 1.0) + dif);
                end; (* case *)
              end;
          3: begin
              dr := (ya(i,k)-zamean[k]);
              Contr_fa := dr*(dr*(4.0*dr*dif+ 6.0) + 4.0*dif) + 1.0;
              end;
           end; (* case *)
          end
         else
          begin (* effect number h is too high to make sense *)
          str(h,ers);
          Contr_fa := 0.0;
          ers := 'Contr_fa called for h = '+ers+'.';
          x := ers;
          outf.Write(pchar(x)^,length(x));
          Halterrorr(ers,outf,lf);
          halterr := true;
          end;
        end; (* else clause for if (h <= 19+2*(nz+nzc-1)) *)
    end; (* else clause for case h of else *)
  end; (* case h *)
end; (* Contr_fa *)

function ModelSpecification.Contr_ga(i           : integer;  (* actor            *)
                                     h,                      (* effect number    *)
                                     k,                      (* behaviour number *)
                                     per         : byte;     (* period number *)
                                     dif         : shortint; (* behaviour change *)
                                     comp0, comp1: array_0_20;
                                     var halterr : Boolean;
                                     var outf,lf : TFileStream) : Real;
(* The contribution of a change dif in ya(i,k) and Action effect h,              *)
(* to the endowment function of i.                                               *)
(* Completely ad hoc, programmed as an adaptation of Contr_fa *)
(* should be called only for observations >= 2                                   *)
(* and for 1 <= h <= ActionEffects_f
   First call CalcComponents_fa !
   If CalcComponents_fa has been called with dif = 0,
   then contr_ga gives the value (not the change value)
   of the behavior endowment function.for behavior dif.                         *)

var h1,h2,h3,h4,h5,
    dumin   : integer;
    dumrl,interact   : real;
    ers,x     : string;


begin (* Contr_ga *)
(* To understand this procedure, look at CalcComponents_fa
   for the calculation of thecomponents *)
  case h of
     1 : Contr_ga := dif;        (* main effect behaviour                       *)
     2 :     (* ave. similarity to outgoing neighbours           *)
         if (comp0[0] <= 0.1) then
           begin
           if (comp1[0] <= 0.1) then
             begin
             Contr_ga := 0.0;
             end
            else
              begin
              Contr_ga := (comp1[2]+comp1[3])/comp1[0];
              end;
           end
          else
           begin
           if (comp1[0] <= 0.1) then
             begin
             Contr_ga := -(comp0[2]+comp0[3])/comp0[0];
             end
            else
             begin
             Contr_ga := ((comp1[2]+comp1[3])/comp1[0]) - ((comp0[2]+comp0[3])/comp0[0]);
             end;
           end;


     3 :     (* tot. similarity to outgoing neighbours           *)
         Contr_ga := (comp1[2]+comp1[3]) - (comp0[2]+comp0[3]);
     4 : Contr_ga := Indm(i,per)*dif;  (* indegree  behaviour                       *)
     5 : Contr_ga := Outdm(i,per)*dif; (* outdegree  behaviour                      *)
     6 : if (Indm(i,per) <= 0) (* isolate  behaviour                        *)
           then Contr_ga := dif
           else Contr_ga := 0.0;
     7 :            (* ave. similarity  reciprocity                    *)
         if (comp0[1] <= 0.1) then
           begin
           if (comp1[1] <= 0.1) then
           Contr_ga := 0.0
          else
           Contr_ga := comp1[6]/comp1[1];
           end
          else
           begin
           if (comp1[1] <= 0.1) then
           Contr_ga := -comp0[7]/comp0[1]
           else
           Contr_ga := comp1[7]/comp1[1] - comp0[7]/comp0[1];
           end;
     8 :            (* tot. similarity  reciprocity                    *)
           Contr_ga := comp1[7] - comp0[7];
     9 : (* ave. similarity  popularity alter               *)
         if (comp0[0] <= 0.1) then
           begin
           if (comp1[0] <= 0.1) then
           Contr_ga := 0.0
          else
           Contr_ga := comp1[9]/comp1[0];
           end
          else
           begin
           if (comp1[0] <= 0.1) then
           Contr_ga := -comp0[9]/comp0[0]
           else
           Contr_ga := comp1[9]/comp1[0] - comp0[9]/comp0[0];
           end;
     10 :        (* pop. alter, or tot. similarity  popularity alter               *)
          if (param_effect[nxa*NetworkEffects_f + (k-1)*ActionEffects_f + h] = 1) then
            begin
            if (outd_yn(i) >= 1) then Contr_ga := dif*comp1[10]/outd_yn(i)
             else Contr_ga := 0.0;
            end
           else Contr_ga := comp1[9] - comp0[9];
     11 :                  (* ave. similarity  reciprocity  pop.(alter)      *)
         if (comp0[1] <= 0.1) then
           begin
           if (comp1[1] <= 0.1) then
           Contr_ga := 0.0
          else
           Contr_ga := comp1[11]/comp1[1];
           end
          else
           begin
           if (comp1[1] <= 0.1) then
           Contr_ga := -comp0[11]/comp0[1]
           else
           Contr_ga := (comp1[11]/comp1[1]) - (comp0[11]/comp0[1]);
           end;
     12 :                  (* tot. similarity  reciprocity  pop.(alter)      *)
           Contr_ga := comp1[11] - comp0[11];
     13 : (* average alter effect *)
         if (comp0[0] <= 0.1) then
           begin
           if (comp1[0] <= 0.1) then
           Contr_ga := 0.0
          else
           Contr_ga := (ya(i,k)-zamean[k])*((comp1[13]/comp1[0])-zamean[k]);
           end
          else
           begin
           if (comp1[0] <= 0.1) then
           Contr_ga := -(zc[i,k,per]-zamean[k])*((comp0[13]/comp0[0])-zamean[k])
           else
           Contr_ga := ((ya(i,k)-zamean[k])*((comp1[13]/comp1[0])-zamean[k]))
                 - ((zc[i,k,per]-zamean[k])*((comp0[13]/comp0[0])-zamean[k]));
           end;
    14 :  (*  average reciprocated alter *)
         if (comp0[1] <= 0.1) then
           begin
           if (comp1[1] <= 0.1) then
           Contr_ga := 0.0
          else
           Contr_ga := (ya(i,k)-zamean[k])*((comp1[14]/comp1[1])-zamean[k]);
           end
          else
           begin
           if (comp1[1] <= 0.1) then
           Contr_ga := -(zc[i,k,per]-zamean[k])*((comp0[14]/comp0[1])-zamean[k])
           else
           Contr_ga := ((ya(i,k)-zamean[k])*((comp1[14]/comp1[1])-zamean[k]))
                 - ((zc[i,k,per]-zamean[k])*((comp0[14]/comp0[1])-zamean[k]));
           end;
     15 : (* squared effect beh. on beh.; nonlinear, therefore not used in ActionStatistics *)
           Contr_ga := 1.0 + 2.0*(ya(i,k)-zamean[k])*dif;
     16 : begin                   (* group  behaviour                      *)
           Contr_ga := 0.0;
          end;
     17 : begin                   (* similarity in groups                     *)
           Contr_ga := 0.0;
          end;
(* Changed 27/10/8: start *)
     18 : Contr_ga := Recidm(i,per)*dif;  (* reciprocated degree  behaviour    *)
     19 : Contr_ga := 0.0;
    else (* case h *)
      begin (* effect number h>=20 *)
      if (h <= 20+2*(nz+nzc-1)) then
        begin
        (* effect h is effect of covariate or other behavioural dimension *)
        h2 := (h-18) div 2; (* number of actor variable *)
        h3 := h mod 2;
        (* main effects of covariates on action dynamics *)
        if h2 >=(nz+k) then inc(h2);  (* control for the fact that the behavioral variable k itself
                                       does not occur in the list of possible variable effects.
                                       Now h2 is the number of the actor variable in question. *)
           case h3 of
             0: interact := dif;
             1: case param_effect[nxa*NetworkEffects_f + (k-1)*ActionEffects_f + h] of
                 1: begin
                    if (comp0[0] <= 0.1) then
                      begin
                      if (comp1[0] <= 0.1) then
                      interact := 0.0
                     else
                      interact := (comp1[2]+comp1[3])/comp1[0];
                      end
                     else
                      begin
                      if (comp1[0] <= 0.1) then
                      interact := -(comp0[2]+comp0[3])/comp0[0]
                      else
                      interact := ((comp1[2]+comp1[3])/comp1[0]) - ((comp1[2]+comp1[3])/comp0[0]);
                      end;
                    end;
                 2: interact := (comp1[2]+comp1[3]) - (comp0[2]+comp0[3]);
                 3: begin
                    if (comp0[0] <= 0.1) then
                      begin
                      if (comp1[0] <= 0.1) then
                      interact := 0.0
                     else
                      interact := (ya(i,k)-zamean[k])*((comp1[13]/comp1[0])-zamean[k]);
                      end
                     else
                      begin
                      if (comp1[0] <= 0.1) then
                      interact := -(zc[i,k,per]-zamean[k])*((comp0[13]/comp0[0])-zamean[k])
                      else
                      interact := ((ya(i,k)-zamean[k])*((comp1[13]/comp1[0])-zamean[k]))
                            - ((zc[i,k,per]-zamean[k])*((comp0[13]/comp0[0])-zamean[k]));
                      end;
                    end;
                 end; (* case *)
             end; (* case h3 *)


        if (h2 <= nz + nzc) then
          begin
           if (h2 <= nz) or (h2 > (nz+nza))  (* See comment in Contr_fa about > or >=  *)
             then Contr_ga := z[i,h2]*interact (* actor variable h is a covariate *)
             else Contr_ga := (ya(i,h2-nz)-zamean[h2-nz])*interact; (* variable is another dependent variable *)
          end;
(* Changed 27/10/8: end *)
           end (* if (h <= 19+2*(nz+nzc-1))*)
      else begin (* effect number h is too high to make sense *)
            str(h,ers);
           Contr_ga := 0.0;
           ers := 'Contr_ga called for h = '+ers+'.';
           x := ers;
           outf.Write(pchar(x)^,length(x));
           Halterrorr(ers,outf,lf);
           halterr := true;
           end;
         end; (* case else: h >= 20 *)
       end; (* case h *)
end; (* Contr_ga *)

procedure ModelSpecification.UtilityComponents_ya(const i,v,h0: integer;
                             var s: rrarray; const perm: B3array;
                             var halterr : Boolean;
                             var outf,lf : TFileStream);
(* Adds the objective function contributions (all h) for changing behavior variable
   ya(i,v) by the amount d to array s[d+2,h0+h] , for d = -1, 1;
   perm[1..3] indicate whether changes -1, 0, 1 are possible;
   note that d=0 represents no change, hence contribution 0;
   first evaluation function, then endowment function *)
var j,jj: integer;
   thecomponents: array_0_20;
   x : string;
begin
  if perm[1] then
    begin
    CalcComponents_fa(i,NetworkEffects_f,ActionEffects_f,nz,v,thisperiod,-1,false,Effects_s,thecomponents,outf);
    jj := h0;
    for j:= 1 to ActionEffects_f do
      if Effects_f[NetworkEffects_f + (v-1)*ActionEffects_f + j] then
        begin
        inc(jj);
        s[1,jj] := s[1,jj] + Contr_fa(i,j,v,-1,thecomponents,halterr,outf,lf);
        if halterr then x := format('%s%d',['Contr_fa 1 ',jj]);
        end;
    for j:= 1 to ActionEffects_g do
      if Effects_g[NetworkEffects_g + (v-1)*ActionEffects_g + j] then
        begin
        inc(jj);
        s[1,jj] := s[1,jj] + Contr_fa(i,j,v,-1,thecomponents,halterr,outf,lf);
        if halterr then x := format('%s%d',['Contr_fa -1 ',jj]);
        end;
    end;
  if perm[3] then
    begin
    CalcComponents_fa(i,NetworkEffects_f,ActionEffects_f,nz,v,thisperiod,1,false,Effects_f,thecomponents,outf);
    jj := h0;
    for j:= 1 to ActionEffects_f do
      if Effects_f[NetworkEffects_f + (v-1)*ActionEffects_f + j] then
        begin
        inc(jj);
        s[3,jj] := s[3,jj] + Contr_fa(i,j,v,1,thecomponents,halterr,outf,lf);
        if halterr then x := format('%s%d',['Contr_fa 3 ',jj]);
        end;
    end;

  if halterr then
    begin
    x := 'error in UtilityComponents_ya'+x+eol;
    outf.Write(pchar(x)^,length(x));
    exit;
    end; (* if halterr *)
end; (* UtilityComponents_ya*)


procedure ModelSpecification.DefineFunctions(var outf,lf: TFileStream);
(* Defines the default choice, and ordering,
   of array Functions (= target statistics),
   depending on the included effects;
   and the sequence of effects in NewNetworkEffects. *)
var i,h,k: integer;
  modelnames: SSarray;
  Correspondence: SBYarray;
  VertPar: BBarray;
  param: SSIarray;
  x: string; // abcdef
begin
  (* initialise array *)
  for i:=1 to pmax do
    begin
    Functions[i] := 0;
    end;
  for i:=0 to ccmax do
    begin
    ActionComputedFunction[i] := false;
    BasicRateFunction[i] := false;
    EndowmentFunction[i] := false;
    if (i >= 1) then DependentVariable[i] := 0;
    end;
  Statistics_l := 0;
  NetworkStatistics_l := 0;
  h := 0;
  (* Network statistics *)
  (* rate function statistics *)
  if (observations >= 2) then
    begin
    DefineModel_lNames(modelnames,Correspondence,param);
    for i:=1 to NetworkEffects_l
    do if Effects_l[i] then
      begin
      inc(h);
      inc(Statistics_l);
      Functions[h] := Correspondence[i];
      if (i <= (observations-1)) then BasicRateFunction[h] := (not skipi[i]);
      inc(NetworkStatistics_l);
      end;
    end; (* if obs >= 2 *)
  (* evaluation function statistics *)

  DefineModel_fNames(modelnames,Correspondence,VertPar,param);
  for i:=1 to NetworkEffects_f do if (Effects_f[i] and (h < pmax)) then
    begin
    inc(h);
    Functions[h] := Correspondence[i];
    end;
  (* endowment function statistics *)
  if (observations >= 2) then
    begin
    DefineModel_gNames(modelnames,Correspondence,param);
    for i:=1 to NetworkEffects_g do if (Effects_g[i] and (h < pmax)) then
      begin
      inc(h);
      Functions[h]:= Correspondence[i];
      EndowmentFunction[h] := true;
      end;
    end;

  (* Action statistics *)
  (* rate function statistics *)
    DefineModel_lNames(modelnames,Correspondence,param);
    for i:=1 to nza*ActionEffects_l
    do if (Effects_l[NetworkEffects_l + i] and (h < pmax)) then
      begin
      inc(h);
      inc(Statistics_l);
      Functions[h] := Correspondence[NetworkEffects_l + i];
      ActionComputedFunction[h] := true;
      DependentVariable[h] := (i-1) div ActionEffects_l + 1;
      if ((i mod ActionEffects_l <= (observations-1))
        and (i mod ActionEffects_l >= 1)) then BasicRateFunction[h] :=
             (not skipi[i mod ActionEffects_l]); 
      end;

  (* evaluation function statistics *)
    DefineModel_fNames(modelnames,Correspondence,VertPar,param);
    for i:=1 to nza*ActionEffects_f do
     if (Effects_f[NetworkEffects_f + i] and (h < pmax)) then
      begin
      inc(h);
      Functions[h] := Correspondence[NetworkEffects_f + i];
      ActionComputedFunction[h] := true;
      DependentVariable[h] := (i-1) div ActionEffects_f + 1;
      end;
  (* endowment function statistics *)
    DefineModel_gNames(modelnames,Correspondence,param);
    for i:=1 to nza*ActionEffects_g do
     if (Effects_g[NetworkEffects_g + i] and (h < pmax)) then
      begin
      inc(h);
      Functions[h]:= Correspondence[NetworkEffects_g + i];
      EndowmentFunction[h] := true;
      ActionComputedFunction[h] := true;
      DependentVariable[h] := (i-1) div ActionEffects_f + 1;
      end;

  Numberfunctions := h;
  if (h > pmax)
    then HaltErrorr('The number of statistics is '+st(h)+', greater than constant pmax = '+st(pmax)+'.',outf,lf);
  if estimation then
    begin
(* This condition is at this moment superfluous.
   It is mentioned here as a reminder that the Stat_fSequence and Stat_gSequence
   are defined differently for simulation. *)
    NewNetworkEffects.DefineStat_fSequence(effects_f);
    NewNetworkEffects.DefineStat_gSequence(effects_g,(MaxLike or Bayesian));
    end;

  for i:=1 to ccmax do Effects_s[i] := (Effects_f[i] or Effects_g[i]);  // abcdef
  for i:=1 to ccmax do Effects_sg[i] := Effects_g[i];   // abcdef

(* And for interactions between behavior effects: *)
  for i:=1 to nza do
    begin
    for h:=1 to nz + nzc - 1 do
      begin
      if Effects_s[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + 2*h] then
      case param_effect[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + 2*h] of
        1: Effects_s[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 2] := true;
        2: Effects_s[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 3] := true;
        3: Effects_s[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 13] := true;
        end; (* case *)
      end;
    end;

  for i:=1 to nza do
    begin
    for h:=1 to nz + nzc - 1 do
      begin
      if Effects_sg[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + 2*h] then
      case param_effect[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + 2*h] of
        1: Effects_sg[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 2] := true;
        2: Effects_sg[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 3] := true;
        3: Effects_sg[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 13] := true;
        end; (* case *)
      end;
    end;

  Effects_s[0] := (Effects_f[0] and Effects_g[0]);

(*x :='  effects_s        effects_sg'+eol;
outf.Write(pchar(x)^,length(x));
for i := 1 to ccmax do
  begin
  if effects_s[i] then x := inttostr(i)+'    +    ' else x := inttostr(i)+'    -    ';
  if effects_s[i] then x := x+'    +    '+eol else x := x+'    -    '+eol;
  outf.Write(pchar(x)^,length(x));
  end;
  *)
end; (* DefineFunctions *)

procedure ModelSpecification.DefineSimFunctions(var outf,lf: TFileStream);
(* Defines the choice of arrays Functions and Effects_s (= target statistics),
   depending on the specified statistics.  *)
var i,j,h,co,co1,coa,cob: integer;
  ordegs: intvec;
  effects_ff: BBArray;
  dummodelnames: SSarray;
  Correspondence: SBYarray;
  param: SSIarray;
  functind, dumpar_f: BBarray;
  dumparam: SSIarray;
  klad: shortint;
procedure minmax(var mi, ma: integer; m: integer);
begin
  if (m < mi) then mi := m;
  if (m > ma) then ma := m;
end; (* minmax *)
begin
(* Order of statistic types is as follows.
   this order will eventually supersede the arrays "contribution".
   (observations - 1)                    for network base rate of change;
   networkeffects_l - (observations - 1) for network change, in addition to base rates;
   networkeffects_f                      for network evaluation;
   networkeffects_f                      for network endowment;

   nza*actioneffects_l                   for action rate of change;
   nza*actioneffects_f                   for action evaluation;
   nza*actioneffects_f                   for action endowment;           *)

  FunctionsToBooleans(FunctInd);
  for i:=0 to ccmax do begin
    ActionComputedFunction[i] := false;
    EndowmentFunction[i] := false;
   end;
  Statistics_l := 0;
  NetworkStatistics_l := 0;
  h := 0;

(* for network change: *)
  co1 := networkeffects_l;
  for i:=1 to co1 do if FunctInd[i] then
    begin
    inc(h);
    inc(Statistics_l);
    inc(NetworkStatistics_l);
    end;
  co := co1;

(*   for network evaluation: *)
  co1 := co+networkeffects_f;
  for i:=co+1 to co1 do if FunctInd[i] then inc(h);
  co := co1;

(*   for network endowment: *)
  co1 := co+networkeffects_g;
  for i:=co+1 to co1 do if FunctInd[i] then
    begin
    inc(h);
    EndowmentFunction[h] := true;
    end;
  co := co1;

(*   for action dependent variables: *)
(*   for action change rates: *)
    co1 := co + nza*actioneffects_l;
    for i:=co+1 to co1 do if FunctInd[i] then
      begin
      inc(h);
      inc(Statistics_l);
      ActionComputedFunction[h] := true;
      DependentVariable[h] := (i-co-1) div ActionEffects_l + 1;
      end;
    co := co1;

(*   for action evaluation: *)
    co1 := co + nza*actioneffects_f;
    for i:=co+1 to co1 do if FunctInd[i] then
      begin
      inc(h);
      ActionComputedFunction[h] := true;
      DependentVariable[h] := (i-co-1) div ActionEffects_f + 1;
      end;
    co := co1;

(*   for action endowment: *)
    co1 := co + nza*actioneffects_g;
    for i:=co+1 to co1 do if FunctInd[i] then
      begin
      inc(h);
      EndowmentFunction[h] := true;
      ActionComputedFunction[h] := true;
      DependentVariable[h] := (i-co-1) div ActionEffects_g + 1;
      end;
    co := co1;

  for j:=0 to ccmax do
    begin
    effects_ff[j] := false;
    effects_s[j] := false;
    end;
  for j:= NetworkEffects_f downto 1 do if FunctInd[NetworkEffects_l+j] then
    begin
    effects_ff[j] := true;
    effects_s[j] := true;
    end;
  NewNetworkEffects.DefineStat_fSequence(effects_ff);

  for j:=0 to ccmax do effects_ff[j] := false;
    for j:= NetworkEffects_f downto 1 do
    begin
      if FunctInd[NetworkEffects_l+NetworkEffects_f+j] then
      begin
      effects_ff[j] := true;
      effects_s[j] := true;
      effects_sg[j] := true;
      end;
    end;

  NewNetworkEffects.DefineStat_gSequence(effects_ff,(MaxLike or Bayesian));
//exit;

  coa := NetworkFunctions+nza*actioneffects_l;
  cob := NetworkFunctions+nza*(actioneffects_l+actioneffects_f);
  for j:= nza*ActionEffects_f downto 1 do
    if FunctInd[coa+j] or FunctInd[cob+j] then
      begin
      effects_s[nxa*Networkeffects_f+j] := true;
      effects_sg[nxa*Networkeffects_f+j] := true;
      end;

  for i:=1 to nza do
    begin
    for j := 1 to Actioneffects_f do
      if FunctInd[cob + (i-1)*actioneffects_f + j] then AnyActionEffects_g[i] := true;
    end;

(* And for interactions between behavior effects: *)
  for i:=1 to nza do
    begin
    for h:=1 to nz + nzc - 1 do
      begin
      if Effects_s[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + 2*h] then
      case param_effect[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + 2*h] of
        1: Effects_s[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 2] := true;
        2: Effects_s[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 3] := true;
        3: Effects_s[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 13] := true;
        end; (* case *)
      end;
    end;

  for i:=1 to nza do
    begin
    for h:=1 to nz + nzc - 1 do
      begin
      if Effects_sg[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + 2*h] then
      case param_effect[nxa*NetworkEffects_f + (i-1)*ActionEffects_f + 19 + 2*h] of
        1: Effects_sg[NetworkEffects_f + (i-1)*ActionEffects_f + 2] := true;
        2: Effects_sg[NetworkEffects_f + (i-1)*ActionEffects_f + 3] := true;
        3: Effects_sg[NetworkEffects_f + (i-1)*ActionEffects_f + 13] := true;
        end; (* case *)
      end;
    end;

  if (observations <= 1) then
    begin

    pclique_min_s := 0;
    pclique_max_s := 0;
    if (index_cliq1 >= 1) then
     begin
     if effects_f[index_cliq1] then
      begin
      param1_clique_s := NewNetworkeffects.Param(index_cliq1);
      minmax(pclique_min_s,pclique_max_s,param1_clique_s);
     end
      else param1_clique_s := 0;
     end;
    if (index_cliq2 >= 1) then
     begin
     if effects_f[index_cliq2] then
      begin
      param2_clique_s := NewNetworkeffects.Param(index_cliq2);
      minmax(pclique_min_s,pclique_max_s,param2_clique_s);
     end
      else param2_clique_s := 0;
     end;
    if (index_cliq3 >= 1) then
     begin
     if effects_f[index_cliq3] then
      begin
      param3_clique_s := NewNetworkeffects.Param(index_cliq3);
      minmax(pclique_min_s,pclique_max_s,param3_clique_s);
     end
      else param3_clique_s := 0;
     end;
    if (index_cliq4 >= 1) then
     begin
     if effects_f[index_cliq4] then
      begin
      param4_clique_s := NewNetworkeffects.Param(index_cliq4);
      minmax(pclique_min_s,pclique_max_s,param4_clique_s);
     end
      else param4_clique_s := 0;
     end;

    if (index_rcliq1 >= 1) then
     begin
     if FunctInd[index_rcliq1] then param1_rclique_s := NewNetworkeffects.Param(index_rcliq1)
       else param1_rclique_s := 0;
     end;
    if (index_rcliq2 >= 1) then
     begin
     if FunctInd[index_rcliq2] then param2_rclique_s := NewNetworkeffects.Param(index_rcliq2)
      else param2_rclique_s := param1_rclique_s;
     end; 
    if (param1_rclique_s > param2_rclique_s) then
      begin
      klad := param1_rclique_s ;
      param1_rclique_s := param2_rclique_s ;
      param2_rclique_s := klad;
      end;
    end;
end; (* DefineSimFunctions *)

procedure ModelSpecification.DefineStatsNames;
var h, j: integer;
begin
  for h:=1 to pmax do StatsDescriptions[h] := 'name '+st(h)+' undefined.';
  h := 1;
  repeat
    StatsDescriptions[h] := FDescriptions[Functions[h]];
    (* This means that Functions[h] has meaning described by Statsdescriptions[h]. *)
    inc(h);
   until ((h > pmax) or (Functions[h] = 0));
  for h:=1 to pmax do
    for j:= 60-length(StatsDescriptions[h]) downto 1 do
      StatsDescriptions[h] :=StatsDescriptions[h] + ' ';
end; (* DefineStatsNames *)


procedure ModelSpecification.ActionStatistics(var fun: pVec;
                  var ordeg: intvec;
                  k: integer; var outf: TFileStream);
  (* calculates the functions indicated in array Functions,
      for period from observation k to k+1 *)

  (* Order of the functions:
   A. Statistics for Network dynamics
      1. amount of change for current period
      2. other statistics for rate function
      3. statistics for evaluation function
      4. statistics for gratification function
   B. Statistics for Action dynamics
      1. amounts of change for current period
      2. other statistics for rate function
      3. statistics for evaluation function
      4. statistics for gratification function  *)

  (* calculates the ordered degrees for the last observation and puts them in ordeg *)
  (* only not-missing data are used for the calculations *)
  (* all statistics are calculated and put in array stats.
     In the end, only the "active" statistics are transferred to fun. *)

var i,j,h,hh,
    count,
    count1,count2  : integer;
    stats          : array[1..ccmax] of real;
    effect,
    s,scin, scout,
    screc,
    ntranss,
    ncycls,
    ncliqs,
    nperiph,
    nbal, nbaltot,
    stot,srec,spop,sact,siso,
    npaths2, ndist2,
    npairpaths2, ntrij,
    n0t1, n1t0,
    n00t1, n10t0,
    n01t1, n11t0   : integer;
    (* n00t1 = number of changes from 00 to (10 or 11), etc. *)
    n0, n1,
    n00, n01,
    n10, n11       : integer;
    outt,inn,recc,              (* outdegree, indegree and reciprocity counts in current network *)
    out1, in1,rec1 : IntVec;    (* outdegree, indegree and reciprocity counts at beginning of the periods *)
    (*p00t1, p10t0, p01t1, p11t0, *)
    dumreal, rdelta,
    r01, rs, rt, ru,
    rv, rw, rx, ry,
    rz, rzz,
    rktriads,
    rkparallel     : real;
    ruz            : array[1..nzmax] of real;
    netbeh         : array of real; (* temporary variables for calculating behaviour  network statistics for objective function *)
    gratbeh        : array of real; (* same, but used for gratification function statistics *)
    twopath        : Boolean;
    halter,
    calcgroups,
    calcperipherals: boolean;
    FunctInd: BBarray;
    compos, compos_init: array_0_20;
    firstef, delta, mindelta: integer; // abcdef
(* for interaction effects between actor variables: *)


    procedure groupcount(i,j,h: integer);
    (* CS: keeps track of various positional count statistics *)
    var hh, numconnect : integer;
    begin
      Inc(ncliqs); (* count group *)
      if calcperipherals
      then for hh:= n downto 1 (* count peripherals to this group *)
      do if ((hh<>i)and(hh<>j)and(hh<>h))
      then begin
(* added by TS 23-jan-05 *)
        if symmetric
        then begin
          numconnect := one(yn(hh,i))+ one(yn(hh,j))+ one(yn(hh,h));
          if (numconnect = 1)
          then inc(nperiph);
        end
        else begin (* not symmetric *)
          if not (yn(i,hh) or yn(j,hh) or yn(h,hh))
          then begin
            if yn(hh,i)
            then begin
              inc(nperiph);
            end;
            if yn(hh,j)
            then begin
              inc(nperiph);
            end;
            if yn(hh,h)
            then begin
              inc(nperiph);
            end;
          end;
         end; (* if not symmetric *)
      end; (* if ((hh<>i)and(hh<>j)and(hh<>h)) *)
    end; (* groupcount *)

    procedure groupcountcovego(i,j,hh,h: integer);
    var hhh, numconnect : integer;
    begin
      if not misz(hh,h)
      then for hhh:= n downto 1
      do if ((hhh<>hh) and (hhh<>j) and (hhh<>i)  and (not misz(hhh,h)))
      then begin
        if symmetric
        then begin
          numconnect := one(yn(hhh,i))+ one(yn(hhh,j))+ one(yn(hhh,hh));
          (* if this is 1 then hhh is peripheral to dense triad i,j,hh in beginning of period k *)
          if (numconnect = 1)
          then rs := rs + z[hhh,h] (* count peripheral to dense triad  covariate *)
        end
        else begin (* not symmetric *)
          if (not (yn(i,hhh) or yn(j,hhh) or yn(hh,hhh))) (* hhh gets no tie from this triad *)
          then begin
            if yn(hhh,i)
            then rs := rs + z[hhh,h]; (* count peripheral tie of hhh to i-j-hh via i *)
            if yn(hhh,j)
            then rs := rs + z[hhh,h]; (* count peripheral tie of hhh to i-j-hh via j *)
            if yn(hhh,hh)
            then rs := rs + z[hhh,h]; (* count peripheral tie of hhh to i-j-hh via hh *)
          end (* if no tie from the triad *)
        end (* if not symmetric *)
      end (* hhh *)
    end;

    procedure groupcountcovsim(i,j,hh,h: integer);
    var hhh, numconnect : integer;
    begin
      if not misz(hh,h)
      then for hhh:= n downto 1
      do if ((hhh<>hh) and (hhh<>j) and (hhh<>i)  and (not misz(hhh,h)))
      then begin
        if symmetric
        then begin
          numconnect := one(yn(hhh,i))+ one(yn(hhh,j))+ one(yn(hhh,hh));
          (* if this is 1 then hhh is peripheral to dense triad i,j,hh in beginning of period k *)
          if (numconnect = 1)
          then rs := rs + similarity(z[hhh,h],z[i,h],h)
                        + similarity(z[hhh,h],z[j,h],h)
                        + similarity(z[hhh,h],z[hh,h],h)
          else begin (* not symmetric *)
            if (not (yn(i,hhh) or yn(j,hhh) or yn(hh,hhh))) (* hhh gets no tie from this triad *)
            then begin
              if yn(hhh,i)
              then rs := rs + similarity(z[hhh,h],z[i,h],h)
                            + similarity(z[hhh,h],z[j,h],h)
                            + similarity(z[hhh,h],z[hh,h],h); (* count peripheral tie of hhh to i-j-hh via i *)
              if yn(hhh,j)
              then rs := rs + similarity(z[hhh,h],z[i,h],h)
                            + similarity(z[hhh,h],z[j,h],h)
                            + similarity(z[hhh,h],z[hh,h],h); (* count peripheral tie of hhh to i-j-hh via j *)
              if yn(hhh,hh)
              then rs := rs + similarity(z[hhh,h],z[i,h],h)
                            + similarity(z[hhh,h],z[j,h],h)
                            + similarity(z[hhh,h],z[hh,h],h)
           end; (* count peripheral tie of hhh to i-j-hh via hh *)
          end; (* if no tie from the triad *)
         end; (* if not symmetric *)
      end (* hhh *)
    end;

    procedure groupcountvariable(i,j,hh,h: integer);
    var hhh, numconnect    : integer;
        rl,rr  : real;
    begin (* i, j and hh constitute a dense triad at beginning of period k *)
      if not misz(hh,h)
      then begin
        netbeh[ 9] := netbeh[ 9] + ya(i,h); (* behaviour  dense triads *)
        netbeh[10] := netbeh[10] + rs + similarity(ya(i,h),ya(hh,h),h+nz); (* behavioural homogeneity in dense triad *)
        if calcperipherals
        then for hhh:=n downto 1 (* hhh is 'candidate' for peripheral status to triad i,j,hh *)
        do if ((hhh<>hh) and (hhh<>j) and (hhh<>i) and (not misz(hhh,h)))
         then begin
(* symmetric added by TS 23-jan-05 *)
          if symmetric then begin
           numconnect := one(ynm(hhh,i,k))+ one(ynm(hhh,j,k))+ one(ynm(hhh,hh,k));
(* if this is 1 then hhh is peripheral to dense triad i,j,hh in beginning of period k *)
           if (numconnect = 1) then begin
                netbeh[11] := netbeh[11] + ya(hhh,h);
                netbeh[12] := netbeh[12] + (  similarity(ya(hhh,h),ya(hh,h),h+nz)
                   + similarity(ya(hhh,h),ya(j ,h),h+nz)
                   + similarity(ya(hhh,h),ya(i ,h),h+nz));
               end;
           end else (* not symmetric *) begin
          if  (not (ynm(i,hhh,k) or ynm(j,hhh,k) or ynm(hh,hhh,k))) then begin
            rr := ya(hhh,h); (* behaviour  peripheral to dense triad *)
            rl := (  similarity(ya(hhh,h),ya(hh,h),h+nz)
                   + similarity(ya(hhh,h),ya(j ,h),h+nz)
                   + similarity(ya(hhh,h),ya(i ,h),h+nz)); (* similarity  peripheral to dense triad *)
            if (ynm(hhh,i,k)) (* tie to i *)
            then begin  (* hhh peripheral to dense triad i,j,hh in beginning of period k *)
              netbeh[11] := netbeh[11] + rr;
              netbeh[12] := netbeh[12] + rl;
            end;
            if (ynm(hhh,j,k)) (* tie to j *)
            then begin  (* hhh peripheral to dense triad i,j,hh in beginning of period k *)
              netbeh[11] := netbeh[11] + rr;
              netbeh[12] := netbeh[12] + rl;
            end;
            if (ynm(hhh,hh,k)) (* tie to hh *)
            then begin  (* hhh peripheral to dense triad i,j,hh in beginning of period k *)
              netbeh[11] := netbeh[11] + rr;
              netbeh[12] := netbeh[12] + rl;
            end;
           end;
          end; (* if not symmetric *)
        end; (* for hhh - potential 4th peripheral actor *)
      end
    end;

    procedure peripheralcountnetwork(i,j,hh,hhh,behavioral,number: integer);
    var rl,rr  : real;
    begin (* j,hh,hhh constitute a dense triad when procedure is called, and i is peripheral to it *)
      if not miszc(i,behavioral,k)
      then begin
        rl := zc[i,behavioral,k]; (* behaviour h of actor i at beginning of period k *)
        netbeh[12+ 9] := netbeh[12+ 9] + number*rl; (* prev.behaviour  peripheral to dense triad *)
        if not(miszc(j,behavioral,k) or miszc(hh,behavioral,k) or miszc(hhh,behavioral,k))
        then begin
          rr := (  similarity(rl,zc[j,behavioral,k],behavioral+nz)
                 + similarity(rl,zc[hh,behavioral,k],behavioral+nz)
                 + similarity(rl,zc[hhh,behavioral,k],behavioral+nz) ); (* similarity-sum of actor i to dense triad *)
          netbeh[12+10] := netbeh[12+10] + number*rr; (* prev.similarity  peripheral to dense triad *)
        end
      end
    end;

function zvar(p,i: integer): real;
begin
  if (p <= nz) then
    begin
    if misz(i,p) then zvar := 0.0 else zvar := z[i,p];
    end
   else
   begin
     if (p <= nz+nza) then
       begin
       if miszc(i,p,k) then zvar := 0.0 else zvar := (ya(i,p)-zamean[p]);
       end
      else
       begin
       if miszc(i,p,k) then zvar := 0.0 else zvar := z[i,p];
       end;
   end;
end; (* zvar *)

begin (* ActionStatistics *)
  (* initialize variables *)
  setlength(netbeh,ActionEffects_f+1);
  setlength(gratbeh,ActionEffects_f+1);
  setlength(out1,n+1);
  setlength(in1,n+1);
  setlength(rec1,n+1);
  setlength(outt,n+1);
  setlength(inn,n+1);
  setlength(recc,n+1);
  ntranss := 0; ncycls := 0;
  ncliqs := 0; nperiph:=0;
  nbal := 0; nbaltot := 0;
  npaths2 := 0; npairpaths2 := 0; ndist2 := 0;
  n0t1 := 0; n1t0 := 0;
  n00t1:=0; n10t0 := 0; n01t1 := 0; n11t0 := 0;
  n0 := 0; n1 := 0;
  n00 := 0; n01 := 0; n10 := 0; n11 := 0;
  rktriads:= 0.0; rkparallel := 0.0;
  for i:=1 to ccmax do stats[i] := 0.0;

  FunctionsToBooleans(FunctInd);

  (* change yn by setting missings to false.
     This is done in order not to have to check mis(i,j,.) so often.
     The changes are reset at the end of this procedure 'statistics'. *)
  MissingsToSaveMat(k);

  (* calculations necessary for network objective function statistics *)
  (* indegree, outdegree and reciprocity counts at beginning of
         current period (variables ending with '1'): *)
  for i:=n downto 1 do
    begin
    out1[i] := Outdm(i,k);
    in1[i]  := Indm(i,k);
    rec1[i] := Recidm(i,k);
    inn[i] := ind_yn(i);
    outt[i] := outd_yn(i);
    recc[i] := recidegree_yn(i);
    end;

(* BEHAVIOUR RATE FUNCTION  *)
(* Raw rates *)

(* count is in the following not the cumulative count of computed functions,
   but the index of the place where the next functions will be put. *)

  for h:=1 to nza do  (* a very long loop, almost to the end of the procedure *)
    begin
  (* Rates interacting with degrees and other variables *)
    rs:=0; rt:=0; ru := 0; rv:=0;
    for j:=1 to nzmax do ruz[j] := 0.0;
    for i:=n downto 1 do
     if IncludeActor[i] and (not (miszc(i,h,k) or miszc(i,h,k+1))) then // IncludeActor added 3.17u
      begin
      rdelta := round(abs(ya(i,h)-zc[i,h,k]));
      rs := rs + rdelta;
      rt := rt + outt[i]*rdelta;
      ru := ru + inn[i]*rdelta;
      rv := rv + recc[i]*rdelta;
      for j:= 1 to nz do
        if (not misz(i,j)) then ruz[j] := ruz[j] + z[i,j]*rdelta;
      for j:= (nz+1) to (nz+nza) do
        if (not miszc(i,j-nz,k)) then ruz[j] := ruz[j] + ya(i,j-nz)*rdelta;
      for j:= (nz+nza+1) to (nz+nzc) do
        if (not miszc(i,j-nz,k)) then ruz[j] := ruz[j] + zc[i,j-nz,k]*rdelta;
      end;

    count := NetworkFunctions + (h-1)*Actioneffects_l;
    stats[count+k] := rs;
    count := count + observations - 1;
    stats[count+1] := rt;
    stats[count+2] := ru;
    stats[count+3] := rv;
    for j:=nz+nzc downto 1 do stats[count+3+j]:=ruz[j];

(* STATISTICS FOR BEHAVIOUR OBJECTIVE FUNCTION
   first some calculations / subgraph counts.
    NOTE that for effects behaviour ==>> network, the variables zc and yn are used,
      while for effects network ==>> behaviour, it is ya and ynm !! *)

    (* initialise *)
    count := NetworkFunctions + nza*Actioneffects_l + (h-1)*ActionEffects_f;

    (* Calculation of statistics of network variables >> BEHAVIOUR dynamics: *)
{CS: temporally disabled: }
    calcperipherals := false;//FunctInd[count+{22*(h-1)+}11] or     (* main effect peripheral   *)
                       // FunctInd[count+{22*(h-1)+}12]; (* similarity of peripheral *)
    calcgroups := calcperipherals
                  or FunctInd[count +  9]       (* main effect group     *)
                  or FunctInd[count + 10];      (* similarity in group   *)

(* If the number of effects is changed, then also change ActionEffects_f *)
    for j :=ActionEffects_f downto 1 do netbeh[j] := 0.0;
    firstef := NetworkEffects_f + (h-1)*ActionEffects_f;
    for i:=n downto 1 do
     if IncludeActor[i] and (not (miszc(i,h,k) or miszc(i,h,k+1))) then // IncludeActor added 3.17u
      begin
      CalcComponents_fa(i,NetworkEffects_f,ActionEffects_f,nz,h,k,0,false,
                        Effects_s,compos,outf);  // abcd was Effects_f
(* for effects 1..16, which are linear in ya(i,h) given compos, Contr_fa can be used *)
      delta := round(ya(i,h) - zamean[h]);
      for j :=1 to 14 do
       if effects_s[firstef+j] then  // abcd was Effects_f
        begin
        netbeh[j] := netbeh[j] + Contr_fa(i,j,h,delta,compos,halter,outf,outf);
        end;
      if effects_s[firstef+15] then  // abcd was Effects_f
        netbeh[15] := netbeh[15] + sqr(ya(i,h) - zamean[h]);  (* nonlinear *)
      for j :=16 to 19 do
       if effects_s[firstef+j] then  // abcd was Effects_f
        begin
        netbeh[j] := netbeh[j] + Contr_fa(i,j,h,delta,compos,halter,outf,outf);
        end;


  (* actor variables => dependent behavior variables, evaluation function *)
  (* This duplicates contr_fa, but it has an extra missing condition *)
  (* for each actor variable there are two effects:
      main effect & interaction with influence, depending on parameter *)

      count := nxa*NetworkEffects_f  + (h-1)*ActionEffects_f + 19;

      for j:=1 to nz do
        begin
        if (not misz(i,j)) then
          begin  // logic for the index is firstef+19+(2*j)-1
          if effects_s[count-1+(2*j)] then
            netbeh[18 + 2*j] := netbeh[18 + 2*j] + delta*z[i,j];
          if effects_s[count+(2*j)] then
(* The next calculation depends on the fact that Effects_s,
   used in CalcComponents_fa, also is true for the main influence effects
   of which these are the interactions; see DefineFunctions. *)
            case param_effect[count+(2*j)] of
              1: netbeh[19 + (2*j)] :=
                 netbeh[19 + (2*j)] + (z[i,j]*Contr_fa(i,2,h,delta,compos,halter,outf,outf));
              2: netbeh[19 + (2*j)] :=
                 netbeh[19 + (2*j)] + (z[i,j]*Contr_fa(i,3,h,delta,compos,halter,outf,outf));
              3: netbeh[19 + (2*j)] :=
                 netbeh[19 + (2*j)] + (z[i,j]*Contr_fa(i,13,h,delta,compos,halter,outf,outf));
              end; (* case *)
          end;
        end;
      for j:=1 to h-1 do
        begin
        if (not (miszc(i,j,k) or miszc(i,j,k+1))) then
          begin
          if effects_s[count-1+2*(nz+j)] then
              netbeh[18 + 2*(nz+j)] :=
                netbeh[18 + 2*(nz+j)] + delta*(zc[i,j,k]-zamean[j]);
          if effects_s[count+2*(nz+j)] then
            case param_effect[count+2*(nz+j)] of
              1: netbeh[18 + 2*(nz+j)] :=
                 netbeh[18 + 2*(nz+j)] + ((zc[i,j,k]-zamean[j])*Contr_fa(i,2,h,delta,compos,halter,outf,outf));
              2: netbeh[18 + 2*(nz+j)] :=
                 netbeh[18 + 2*(nz+j)] + ((zc[i,j,k]-zamean[j])*Contr_fa(i,3,h,delta,compos,halter,outf,outf));
              3: netbeh[18 + 2*(nz+j)] :=
                 netbeh[18 + 2*(nz+j)] + ((zc[i,j,k]-zamean[j])*Contr_fa(i,13,h,delta,compos,halter,outf,outf));
              end; (* case *)
           end;
        end;
    (* j = h is excluded because that is the current action variable under consideration *)
      for j:=h+1 to nza do
        begin
        if (not (miszc(i,j,k) or miszc(i,j,k+1))) then
          begin
          if effects_s[count-3+2*(nz+j)] then
              netbeh[16 + 2*(nz+j)] :=
                netbeh[16 + 2*(nz+j)] + delta*(zc[i,j,k]-zamean[j]);
          if effects_s[count-2+2*(nz+j)] then
            case param_effect[count-2+2*(nz+j)] of
              1: netbeh[17 + 2*(nz+j)] :=
                 netbeh[17 + 2*(nz+j)] + ((zc[i,j,k]-zamean[j])*Contr_fa(i,2,h,delta,compos,halter,outf,outf));
              2: netbeh[17 + 2*(nz+j)] :=
                 netbeh[17 + 2*(nz+j)] + ((zc[i,j,k]-zamean[j])*Contr_fa(i,3,h,delta,compos,halter,outf,outf));
              3: netbeh[17 + 2*(nz+j)] :=
                 netbeh[17 + 2*(nz+j)] + ((zc[i,j,k]-zamean[j])*Contr_fa(i,13,h,delta,compos,halter,outf,outf));
              end; (* case *)
           end;
        end;

      for j:=nza+1 to nzc do
        begin
        if (not (miszc(i,j,k) or miszc(i,j,k+1))) then
          begin
          if effects_s[count-3+2*(nz+j)] then
              netbeh[16 + 2*(nz+j)] := netbeh[16 + 2*(nz+j)] + delta*zc[i,j,k]; // 12345
          if effects_s[count-2+2*(nz+j)] then
            case param_effect[count-2+2*(nz+j)] of
              1: netbeh[17 + 2*(nz+j)] :=
                 netbeh[17 + 2*(nz+j)] + (zc[i,j,k]*Contr_fa(i,2,h,delta,compos,halter,outf,outf));
              2: netbeh[17 + 2*(nz+j)] :=
                 netbeh[17 + 2*(nz+j)] + (zc[i,j,k]*Contr_fa(i,3,h,delta,compos,halter,outf,outf));
              3: netbeh[17 + 2*(nz+j)] :=
                 netbeh[17 + 2*(nz+j)] + (zc[i,j,k]*Contr_fa(i,13,h,delta,compos,halter,outf,outf));
              end; (* case *)
           end;
        end; (* j *)
      count1 := 17 + 2*(nz+nzc);
(* Added 27/10/8: *)
      if (not (miszc(i,j,k) or miszc(i,j,k+1))) then
        begin
        if effects_s[count-1+2*(nz+nzc)] then
        netbeh[count1+1] :=
               netbeh[count1+1] + Contr_fa(i,count1+1,h,delta,compos,halter,outf,outf);
        end;
      count1 := count1+1;
(* end Added 27/10/8 *)

      for j:=1 to ninta do
        begin
        case ni[h,j] of
        2: netbeh[count1 + j] :=
                     netbeh[count1 + j] + delta*zvar(pv1[h,j],i)*zvar(pv2[h,j],i);
        3: netbeh[count1 + j] :=
                     netbeh[count1 + j] + delta*zvar(pv1[h,j],i)*zvar(pv2[h,j],i)*zvar(pv3[h,j],i);
        end; (* case *)
        end; (* j *)
    end; (* i - focal actor *)

    (* Now calculate and store the statistics; *)
    count := NetworkFunctions + nza*ActionEffects_l + (h-1)*ActionEffects_f;
    for i:=ActionEffects_f downto 1 do stats[count+i] := netbeh[i];

(* BEHAVIOURAL GRATIFICATION FUNCTION STATISTICS; *)

    if AnyActionEffects_g[h] then
      begin
      for i:=ActionEffects_f downto 1 do gratbeh[i] := 0.0;

      for i:=n downto 1 do
       if IncludeActor[i] and (not (miszc(i,h,k) or miszc(i,h,k+1))) then // IncludeActor added 3.17u
        begin
        delta := round(ya(i,h) - zc[i,h,k]);
        mindelta := -delta; //abcdef
        if (delta <0) (* decrease in behaviour h ? *) then
          begin
          CalcComponents_fa(i,NetworkEffects_f,ActionEffects_f,nz,h,k,mindelta,false,
                        Effects_sg,compos_init,outf);  // abcdef
// This yields compos_init for the situation of the behavior at the start of period k.
          CalcComponents_fa(i,NetworkEffects_f,ActionEffects_f,nz,h,k,0,false,
                        Effects_sg,compos,outf);
// This yields compos for the situation of the current behavior.
// 3.17d abcdef 0 vervangen door delta
          for j :=1 to 14 do if effects_sg[firstef+j] then
            gratbeh[j] := gratbeh[j] + Contr_ga(i,j,h,k,delta,
                                 compos_init,compos,halter,outf,outf);
          if effects_sg[firstef+15] then
            gratbeh[15] := gratbeh[15] + sqr(ya(i,h) - zamean[h])
                                       - sqr(zc[i,h,k] - zamean[h]);
          for j :=16 to 19 do if effects_sg[firstef+j] then
            gratbeh[j] := gratbeh[j] + Contr_ga(i,j,h,k,delta,
                                 compos_init,compos,halter,outf,outf);

          (* main effects and interactions of other actor variables *)
           count := nxa*NetworkEffects_f  + (h-1)*ActionEffects_f + 17;
           for j:=1 to nz do
            begin
            if (not misz(i,j)) then
              begin  // logic for the index is firstef+17+(2*j)-1
              if effects_sg[count-1+(2*j)] then
                gratbeh[18 + 2*j] := gratbeh[18 + 2*j] + delta*z[i,j];
              if effects_sg[count+(2*j)] then
(* The next calculation depends on the fact that effects_sg,
   used in CalcComponents_fa, also is true for the main influence effects
   of which these are the interactions; see DefineFunctions. *)
                case param_effect[count+(2*j)] of    // abcdef
                  1: gratbeh[19 + (2*j)] :=
                     gratbeh[19 + (2*j)] +
                     (z[i,j]*Contr_ga(i,2,h,k,delta,compos_init,compos,halter,outf,outf));
                  2: gratbeh[19 + (2*j)] :=
                     gratbeh[19 + (2*j)] +
                     (z[i,j]*Contr_ga(i,3,h,k,delta,compos_init,compos,halter,outf,outf));
                  3: gratbeh[19 + (2*j)] :=
                     gratbeh[19 + (2*j)] +
                     (z[i,j]*Contr_ga(i,13,h,k,delta,compos_init,compos,halter,outf,outf));
                  end; (* case *)
              end;
            end;
          for j:=1 to h-1 do
            begin
            if (not (miszc(i,j,k) or miszc(i,j,k+1))) then
              begin
              if effects_sg[count-1+2*(nz+j)] then
                  gratbeh[18 + 2*(nz+j)] :=
                    gratbeh[18 + 2*(nz+j)] + delta*(zc[i,j,k]-zamean[j]);
              if effects_sg[count+2*(nz+j)] then
                case param_effect[count+2*(nz+j)] of           // abcdef
                  1: gratbeh[19 + 2*(nz+j)] :=
                     gratbeh[19 + 2*(nz+j)] +
                     ((zc[i,j,k]-zamean[j])*Contr_ga(i,2,h,k,delta,compos_init,compos,halter,outf,outf));
                  2: gratbeh[19 + 2*(nz+j)] :=
                     gratbeh[19 + 2*(nz+j)] +
                     ((zc[i,j,k]-zamean[j])*Contr_ga(i,3,h,k,delta,compos_init,compos,halter,outf,outf));
                  3: gratbeh[19 + 2*(nz+j)] :=
                     gratbeh[19 + 2*(nz+j)] +
                     ((zc[i,j,k]-zamean[j])*Contr_ga(i,13,h,k,delta,compos_init,compos,halter,outf,outf));
                  end; (* case *)
               end;
            end;
    (* j = h is excluded because that is the current action variable under consideration *)
          for j:=h+1 to nza do
            begin
            if (not (miszc(i,j,k) or miszc(i,j,k+1))) then
              begin
              if effects_sg[count-3+2*(nz+j)] then
                  gratbeh[16 + 2*(nz+j)] :=
                    gratbeh[16 + 2*(nz+j)]  + delta*(zc[i,j,k]-zamean[j]);
              if effects_sg[count-2+2*(nz+j)] then
                case param_effect[count-2+2*(nz+j)] of         // abcdef
                  1: gratbeh[17 + 2*(nz+j)] :=
                     gratbeh[17 + 2*(nz+j)] +
                     ((zc[i,j,k]-zamean[j])*Contr_ga(i,2,h,k,delta,compos_init,compos,halter,outf,outf));
                  2: gratbeh[17 + 2*(nz+j)] :=
                     gratbeh[17 + 2*(nz+j)] +
                     ((zc[i,j,k]-zamean[j])*Contr_ga(i,3,h,k,delta,compos_init,compos,halter,outf,outf));
                  3: gratbeh[17 + 2*(nz+j)] :=
                     gratbeh[17 + 2*(nz+j)] +
                     ((zc[i,j,k]-zamean[j])*Contr_ga(i,13,h,k,delta,compos_init,compos,halter,outf,outf));
                  end; (* case *)
               end;
            end;

          for j:=nza+1 to nzc do
            begin
            if (not (miszc(i,j,k) or miszc(i,j,k+1))) then
              begin
// 1234 writeln(outf,':a: ',count-3+2*(nz+j):3,' ',effects_sg[count-3+2*(nz+j)]);
              if effects_sg[count-3+2*(nz+j)] then
                  gratbeh[16 + 2*(nz+j)] :=
                    gratbeh[16 + 2*(nz+j)] + delta*zc[i,j,k];
              if effects_sg[count-2+2*(nz+j)] then
                case param_effect[count-2+2*(nz+j)] of              // abcdef
                  1: gratbeh[17 + 2*(nz+j)] :=
                     gratbeh[17 + 2*(nz+j)] +
                     (zc[i,j,k]*Contr_ga(i,2,h,k,delta,compos_init,compos,halter,outf,outf));
                  2: gratbeh[17 + 2*(nz+j)] :=
                     gratbeh[17 + 2*(nz+j)] +
                     (zc[i,j,k]*Contr_ga(i,3,h,k,delta,compos_init,compos,halter,outf,outf));
                  3: gratbeh[17 + 2*(nz+j)] :=
                     gratbeh[17 + 2*(nz+j)] +
                     (zc[i,j,k]*Contr_ga(i,13,h,k,delta,compos_init,compos,halter,outf,outf));
                  end; (* case *)
               end;
            end; (* j *)
          count1 := 17 + 2*(nz+nzc);

(* Added 27/10/8: *)
      if (not (miszc(i,j,k) or miszc(i,j,k+1))) then
        begin
        if effects_sg[count-1+2*(nz+nzc)] then
        gratbeh[count1+1] :=
               gratbeh[count1+1] + Contr_ga(i,count1+1,h,k,delta,compos_init,compos,halter,outf,outf);
        end;
      count1 := count1+1;
(* end Added 27/10/8 *)

          for j:=1 to ninta do
            begin
            case ni[h,j] of
              2: gratbeh[count1 + j] :=
                     gratbeh[count1 + j] + delta*zvar(pv1[h,j],i)*zvar(pv2[h,j],i);
              3: gratbeh[count1 + j] :=
                     gratbeh[count1 + j] +
                           delta*zvar(pv1[h,j],i)*zvar(pv2[h,j],i)*zvar(pv3[h,j],i);
              end; (* case *)
            end; (* j *)
          end; (* decrease of behaviour *)
        end; (* i as 1st actor in counting loops *)

      count := NetworkFunctions + nza*(ActionEffects_l + ActionEffects_f) +
             (h-1)*ActionEffects_g;
      for j:=ActionEffects_f downto 1 do stats[count+j] := gratbeh[j];

  (* the BEHAVIOUR GRATIFICATION FUNCTION STATISTICS now are all calculated. *)
    end; (* if AnyActionEffects_g[h] *)

  end; (* for h *)

  (* copy statistics into shorter array fun[.] *)

  h := 1;
  repeat
    if ActionComputedFunction[h] then
      begin
      fun[h] := stats[Functions[h]];
      end;
// writeln(outf,h:3,Functions[h]:4,' ',ActionComputedFunction[h]);
(* Later - i.e. in another procedure - the endowment effects get a minus sign. *)
    inc(h);
  until ((h > pmax) or (Functions[h] = 0));
(* this can lead to the calculation of Functions[pmax+1], which is legal
   because of the definition of pWordVec in S_Cons. *)

  (*MS: test homogeneity assumptions.*)
  if ((homogeneous in [11..19]) and derivatives)
  then StatisticsSpecial(fun); (*Evaluate statistic for the outlier candidates using simulated digraph.*)
  (*MS*)

(* reset the values of yn that were changed earlier in this procedure: *)
  RestoreSaveMat;
end; (* ActionStatistics *)

procedure ModelSpecification.NetworkStatistics
        (var fun: pVec;
         var ordeg: intvec; k: integer; var outf: TFileStream);
  (* calculates the functions indicated in array Functions
     for period from observation  k to k+1 *)
var dgdif: TValuedDigraph;
    i, j, h, s, sums, scin, scout, screc, fts, funh0,
      numberstats,numberrates,nzplus: integer;
    changezh : array[1..nzmax] of real;
    rs: real;
    updatestat: boolean;
    funk: pVec;
    FunctInd: BBarray;
    calc_l: boolean; (* Indicates whether any rate statistics must be calculated. *)
begin
  numberrates := 0;
  for i:=1 to observations-1 do if not(skipi[i]) then inc(numberrates);
(* numberrates is the number of basic rate parameters *)
  for i:=1 to pmax do fun[i] := 0.0;
  calc_l := false;
  numberstats := 0;
  (*  numberstats = number of change rate statistics in addition to the
      raw total amounts of change *)
  FunctionsToBooleans(FunctInd);

  for i:=observations-1 downto 1 do
    if (estimation and not cconditional) or FunctInd[i] then begin
      calc_l := true;
     end;
  for i:= NetworkEffects_l downto observations do if FunctInd[i] then begin
      calc_l := true;
      inc(numberstats);
     end;
(* Create new data matrices, without ties for missings :
   dgdif is the difference graph. *)

 if calc_l then
  begin
  CreateDifMat(k,dgdif);
  (* Calculate rate statistics : *)
  scout := 0; scin := 0; screc := 0; rs := 0.0; sums := 0;
  for i:=1 to nzmax do changezh[i] := 0.0;
  for i:=n downto 1 do
  begin
    s := dgdif.OutDegree(i); (* number of changed ties outgoing from i *)
    Inc(sums,s);
    if FunctInd[observations] then
      Inc(scout, s*Outdm(i,k)); (* changing tie  previous outdegree *)
    if FunctInd[observations+1] then
    Inc(scin,  s*Indm(i,k)); (* changing tie  previous indegree *)
    if FunctInd[observations+2] then
    Inc(screc, s*Recidm(i,k)); (* changing tie  previous reciprocated degree *)
    if FunctInd[observations+3] then
    rs := rs + s/(Outdm(i,k)+1); (* changing tie  1/outdegree statistic *)
    for h:=nz+nzc downto 1 do if FunctInd[observations+3+h]
      then changezh[h] := changezh[h] + s*z[i,h]; (* changing tie  actor covariates *)
   end; (* for i *)

// ModelType = 2 is not yet implemented here. @@@
  if not (cconditional  and (ConditioningNetworkVariable = 1) and estimation) then
    begin
    fts := 0;
    for i:=1 to k do if (not skipi[i]) then inc(fts); // abcdefg
    fun[fts] := sums;
    end;

  (* change rate statistics other than total raw change start at fts;
  objective function statistics start at funh0 *)
  if (estimation and cconditional and (ConditioningNetworkVariable = 1)) then
    begin
    fts := 1;
    funh0 := numberstats+1;
    end
   else
    begin
    fts := 1;
    for i:=1 to observations-1 do if not(skipi[i]) then inc(fts);
    funh0 := numberstats + fts;
    end;

  h := fts-1;
  if FunctInd[observations] then begin
    inc(h);
    fun[h] := scout;
   end;
  if FunctInd[observations+1] then begin
    inc(h);
    fun[h] := scin;
   end;
  if FunctInd[observations+2] then begin
    inc(h);
    fun[h] := screc;
   end;
  if FunctInd[observations+3] then begin
    inc(h);
    fun[h] := rs;
   end;
  nzplus := nz+nzc;
  for j:=1 to nzplus do if FunctInd[observations+3+j] then begin
    inc(h);
    fun[h] := changezh[j];
   end;

  dgdif.Destroy(outf,'dgdif in NetworkStatistics');
 end; (* if calc_l *)

(* Statistics corresponding to the evaluation function: *)
  updatestat := (OutputType >= 2);

  CalcFunctions_yn(k,funk,ordeg,NewNetworkEffects,updatestat,outf);
  for h:=1 to NewNetworkEffects.NumberStats_f
    do begin
       fun[funh0 + h - 1] := funk[h];
      end;

  for h:=NewNetworkEffects.NumberStats_f+1
    to NewNetworkEffects.NumberStats_f + NewNetworkEffects.NumberStats_g
    do begin
       if estimation then fun[funh0 + h - 1] := -funk[h] else
(* former line changed compared to 3bet2 *)
         fun[funh0 + h - 1] := funk[h];
        end;
end; (* NetworkStatistics *)

procedure ModelSpecification.GetNextevent;
(* MH: specifies event time and joining/leaving actor in period k,
   thereby corrupting array event during 1 simulation run *)
var i, ii: integer;
    nt: real;
begin
  ii := 0;
  nt := 1.0;
  for i:=1 to n do begin
    if (event[i] > 0.0) then begin
      if ((event[i] >= nextevent) and (event[i] < nt)) then begin
         nt := event[i];
         ii := i;
      end; (* if event > nextevent *)
    end; (* if event > 0*)
  end; (* for i *)
  nextevent := nt;
  i_act := ii;
  if (i_act > 0) then event[i_act] := 0.0;
end; (* GetNextevent *)

(* procedure ModelSpecification.initialise_y_likel;
  MS: called in case of likelihood-based inference (MLE / Bayes):
can be deleted as soon as likelihood-based inference admits changing composition
and structural values and hence, <initialise_y> can be called.
TS: deleted 260907
var i: integer;
begin
for i := 1 to n do StructActive[i] := true;
Setlength(ComponentNumber,n+1,n+1);
ComponentNumber[0,0] := 1;
end;
*)

procedure ModelSpecification.initialise_actors(const k: integer);
(* Initialises Structurally active actors at begin period k.
   MH: If observations >= 2, determines active actors at
         begin period k and sets nextevent *)
var i,j: integer;
    rpar: real;
begin
  n := nni[k];
  NStructActiveElements := 0;
  n_act := n;
  maxnfree := 0;

  for i:=1 to n do if (StructRowElements(i,k) >= n-1) then
    begin
    StructActive[i] := false;
    IncludeActor[i] := false; // added 3.17u
    dec(n_act);
    end
   else
    begin
    StructActive[i] := true;
    IncludeActor[i] := true; // added 3.17u
    if (n-1-StructRowElements(i,k) > maxnfree) then maxnfree :=  n-1-StructRowElements(i,k);
    Inc(NStructActiveElements,n-1-StructRowElements(i,k));
    end; (* if *)

  if (observations <= 1) then
    begin
    initialiseERGM(n, maxnfree, densty, cconditional);
    setlength(arraykoutstars,n+1);
    setlength(arraykinstars,n+1);
    setlength(arrayktri,n+1);
    setlength(arrayktwop,n+1);
    for i:=0 to n do
      begin
      if (param_koutstars > 1) then rpar := log(1.0 - (1.0/param_koutstars)) else rpar := 0;
      arraykoutstars[i] := param_koutstars*(1.0-expo(i*rpar));
      if (param_kinstars > 1) then rpar := log(1.0 - (1.0/param_kinstars)) else rpar := 0;
      arraykinstars[i] := param_kinstars*(1.0-expo(i*rpar));
      if (param_ktri > 1) then rpar := log(1.0 - (1.0/param_ktri)) else rpar := 0;
      arrayktri[i]   := expo(i*rpar);
      if (param_ktwop > 1) then rpar := log(1.0 - (1.0/param_ktwop)) else rpar := 0;
      arrayktwop[i]   := expo(i*rpar);
      end;
    end;

if (param_koutstars = 1) then
  begin
  arraykoutstars[0] := 1;
  for i:=1 to n do arraykoutstars[i] := 0;
  end;

  if (observations >= 2) then
    begin
    (* MH: initialisation active actors -> all actors active.
       This array is always initialised, because it is needed to
       set the yn[i,j] values and to simulate correctly. *)
    for i:=1 to n do active[i] := true;
    if exo then
      begin
      (* MH: initialize variables that handle exogenous events *)
      nextevent := 0.0;   (* time of next event *)
      i_act := 0;         (* actor who joins/leaves at nextevent *)

      (* MH: set proper values for event variables *)
      for i:=1 to n do event[i] := event_m[i,k];  (* event times in period k*)
      GetNextevent;                (* set value of nextevent and i_act *)
      (* MH: determine set of active actors at begin period k *)
      for i:=1 to n do
        begin
        if (not active_m[i,k]) then
          begin (* if not active at begin k *)
          active[i] := false;
          StructActive[i] := false;
          IncludeActor[i] := false; // added 3.17u
          end (* if not active_m[i,k] *)
         else
          if (not active_m[i,k+1]) then IncludeActor[i] := false; // added 3.17u
        end; (* for i *)
      end; (* if exo *)
    for j:=1 to nza do
      begin
      n_actnonmis[j] := n_act;
      for i:=n downto 1 do
        if (Structactive[i] and (miszc(i,j,k) or miszc(i,j,k+1)))
          then dec(n_actnonmis[j]);
      end;
    end; (* observations >= 2 *)

(* changing covariates : *)
  for j:=1 to nzc do for i:=n downto 1 do
    begin
    z[i,nz+j] := zc[i,j,k];
    setmisz(i,nz+j,miszc(i,j,k))
    end;

  for j:=1 to nza do czamean[j] := pzamean[j,k];

  if ExistStruct[k] then Component(j,false,ComponentIndicator,k);
  (* j is a dummy here *)
  MoreComponents := (ComponentNumber[0,0] >= 2);

  CheckChange := (StaySameType <= 1) or ExistStruct[k] or (maxdegree >= n-1); // @@@@@
end; (* initialise_actors *)

procedure ModelSpecification.initialise_y(const k: integer);
(* if observations = 1:
   Initialises yn:
     if independentstart,
        then sets yn to a random matrix with expected density 0.5,
        satisfying symmetry / asymmetry constraints, if relevant;
     if (not independentstart), then leaves yn as it is.
   if observations >= 2 :
   Initialises yn to ynm[k].
   Note that this is used in procedure SimStats and S_ML.SetData,
   and its functionality for getting the data
   should be replicated in procedure FunctionEvaluations.
   Initialises Structurally active actors at begin period k.
   MH: If observations >= 2, determines active actors at
         begin period k and sets nextevent *)
var i,j,jmax,h,nn: integer;
  od: array of integer;
begin
  n := nni[k];
  time := 0.0;
  if (observations <= 1) then
    begin
    Gibbschange := 0;
    if IndependentStart then
      begin
      SetStructOne(1);
(* Sets current valued digraph to only the structural ones for observation 1. *)
      if asymmetric then
        begin
        for i:=2 to n do
          for j:=1 to i-1 do
             if (rando(version) < 0.5) then setyn(i,j,true) else setyn(j,i,true);
        end
       else
        begin (* not asymmetric *)
        h := 0;
        for i:=1 to n do inc(h,outdm(i,1) - outd_yn(i));
        (* h new ties to be formed *)
        repeat
         i := randomd(version,n)+1;
         repeat j := randomd(version,n)+1 until (i <> j);
         if (not (struct(i,j,1) or yn(i,j) or (outd_yn(i) >= maxdegree)
                       or (symmetric and (outd_yn(j) >= maxdegree)))) then
           begin
           setyn(i,j,true);
           dec(h);
           if symmetric then
             begin
             setyn(j,i,true);
             dec(h);
             end;
           end;
         until (h <= 0);
        end;
      SetLength(od,n+1);
      ErgmStatistics(CurrentStats,symmetric,od,true,NewNetworkEffects); // in initialise_y
      cumuldegrees[0] := 0;
      for i:=1 to n do cumuldegrees[i] := cumuldegrees[i-1]+outd_yn(i);
    end (* if independentstart *)
   else exit; (* nothing to be done; retain existing matyn *)
  end (* observations <=1 *)

  else begin(* observations >= 2 *)
    (* MH: initialisation active actors -> all actors active.
       This array is always initialised, because it is needed to
       set the yn[i,j] values and to simulate correctly. *)

    (* MH: set yn; if i or j not active set yn false;
       CS: set ya; if i not active, set ya=round(global mean);
       TS: the first was already dropped, I now (21-01-06) drop the second also
           because, as described in the manual, it is more flexible and natural
           to leave it up to the user whether to specify a missing
           or a real value. *)
// if network is not to be set to observation, e.g.,
// for calculating spatial autocorrelations without setting back:
// add    if (k = 1) then
    SetDigraph(k);
    for i:=n downto 1
    do begin
        for j:=nza downto 1 do
          begin
          setya(i,j,round(zc[i,j,k]));
(* TS toegevoegd 4-1-06 : !!! @@@;
   mooier om deze truncering al in de Siena datafile te doen.  *)
          if (ya(i,j) < round(zcmin[j])) then setya(i,j,round(zcmin[j]));
          if (ya(i,j) > round(zcmax[j])) then setya(i,j,round(zcmax[j]));
          end;
    end; (* for i *)

  end; (* observations >= 2 *)

  Distance := 0;
  thisperiod := k;
end; (* initialise_y *)

procedure ModelSpecification.initialise_y_end(const k: integer);
var i,j: integer;
begin
    SetDigraph(k+1);
    for i:=1 to n
    do begin
        for j:= nza downto 1 do
          begin
          setya(i,j,round(zc[i,j,k+1]));
          if (ya(i,j) < round(zcmin[j])) then setya(i,j,round(zcmin[j]));
          if (ya(i,j) > round(zcmax[j])) then setya(i,j,round(zcmax[j]));
          end;
    end; (* for i *)
end; (* initialise_y_end *)


procedure ModelSpecification.ChangeTie(const i,j: integer);
var h: integer;
(* For the signed case, this procedure is replaced by PlusTie and MinTie [s] *)
begin
  if (i = j) then exit;
  if Struct(i,j,thisperiod)
  then begin
    ChangeTieStructError(i,j,0);
    exit;
  end; (* if *)

(* first part of update in case of  ExtendeReciprocityModel *)
  if ExtendedReciprocityModel then begin
    lambdar[i] :=
      lambdar[i] - expo(UtilityGain(i,j,thisperiod,symmetric,true,NewNetworkEffects))/(n-1);  // @@@@@
    lambdar[j] :=
      lambdar[j] - expo(UtilityGain(j,i,thisperiod,symmetric,true,NewNetworkEffects))/(n-1);  // @@@@@
   end;

  (* Put the change into effect *)
  if yn(i,j)
  then begin
    (* change 1 to 0 *)
    setyn(i,j,false);

    (* updates of running statistics *)
    if ((observations >= 2) and (ConditioningNetworkVariable >= 1))
    then begin
      if (not (mis(i,j,thisperiod) or mis(i,j,thisperiod+1)))
      then begin
        if ynm(i,j,thisperiod)
        then Inc(Distance)
        else Dec(Distance)
      end
    end;

    {for h:=nza downto 1
    do begin
      if (ya(j,h) > ya(i,h))
      then dec(zaplus[i,h])
      else begin
        if (ya(j,h) < ya(i,h))
        then dec(zamin[i,h])
        else dec(zaeq[i,h])
      end
    end;} (* for h *)

  end
  else begin (* not yn(i,j) *)
    (* change 0 to 1 *)
    setyn(i,j,true);

    (* updates of running statistics *)
    if ((observations >= 2) and (ConditioningNetworkVariable >= 1))
    then begin
      if (not (mis(i,j,thisperiod) or mis(i,j,thisperiod+1)))
      then begin
        if ynm(i,j,thisperiod)
        then Dec(Distance)
        else Inc(Distance)
      end
    end;

    {for h:=nza downto 1
    do begin
      if (ya(j,h) > ya(i,h))
      then inc(zaplus[i,h])
      else begin
        if (ya(j,h) < ya(i,h))
        then inc(zamin[i,h])
        else inc(zaeq[i,h])
      end
    end;} (* for h *)

  end; (* if yn(i,j) *)

(* second part of update in case of  ExtendeReciprocityModel *)
    if ExtendedReciprocityModel then begin
      lambdar[i] := lambdar[i] +
                       expo(UtilityGain(i,j,thisperiod,symmetric,true,NewNetworkEffects))/(n-1);  // @@@@@
      lambdar[j] := lambdar[j] +
                       expo(UtilityGain(j,i,thisperiod,symmetric,true,NewNetworkEffects))/(n-1);  // @@@@@
     end;
end; (* ChangeTie *)


procedure ModelSpecification.ChangeTieStructError(i,j: integer; d: shortint);
  (* Reports an error from ChangeTie *)
begin
  MadeChangeTieError := true;
  case d of
   -2: ChangeTieErrorText := 'MinTie called for position ('
                        +inttostr(i)+','+inttostr(j)+') with value '+inttostr(ynv(i,j));
   -1: ChangeTieErrorText := 'MinTie called for structurally determined position ('
                        +inttostr(i)+','+inttostr(j)+')';
    0:  ChangeTieErrorText := 'ChangeTie called for structurally determined position ('
                        +inttostr(i)+','+inttostr(j)+')';
    1: ChangeTieErrorText := 'PlusTie called for structurally determined position ('
                        +inttostr(i)+','+inttostr(j)+')';
    2: ChangeTieErrorText := 'PlusTie called for position ('
                        +inttostr(i)+','+inttostr(j)+') with value '+inttostr(ynv(i,j));
   else
      ChangeTieErrorText := 'ChangeTie called for code '+inttostr(d);
   end; (* case *)
end; (* ChangeTieStructError *)


procedure ModelSpecification.ChangeBehavior(const i: integer;const h: byte; const dif: shortint);
  (* Changes ya(i,h) and carries out all updates *)
var j: integer;
begin
  if (dif = 0) then exit;
  (* put the change into effect *)
  case dif of
   -1: Minya(i,h);
    1: Plusya(i,h);
  end; (* case *)

 (* updates of running statistics *)

 czamean[h] := czamean[h] + dif/n_act;

   {for j :=n downto 1 do begin
     if yn(i,j) then
       case (ya(j,h) - ya(i,h)) of
         -1: if (dif >= 1) then
               begin inc(zamin[i,h]); dec(zaeq[i,h]); end;
          0:  case dif of
              -1: begin inc(zaeq[i,h]); dec(zamin[i,h]);  end;
               1: begin inc(zaeq[i,h]); dec(zaplus[i,h]); end;
              end; (* case *)
          1:  if (dif <= -1) then
               begin inc(zaplus[i,h]); dec(zaeq[i,h]); end;
        end; (* case *)
     if yn(j,i) then
       case (ya(j,h) - ya(i,h)) of
         -1: if (dif >= 1) then
               begin inc(zaplus[j,h]); dec(zaeq[j,h]); end;
          0:  case dif of
              -1: begin inc(zaeq[j,h]); dec(zaplus[j,h]);  end;
               1: begin inc(zaeq[j,h]); dec(zamin[j,h]); end;
              end; (* case *)
          1:  if (dif <= -1) then
               begin inc(zamin[j,h]); dec(zaeq[j,h]); end;
        end; (* case *)
    end;} (* for j *)
    (* zaplus[i,j] = sum_h Yn_ih Indicator{ya[h,j] > ya[i,j]}
       zaeq[i,j] = sum_h Yn_ih Indicator{ya[h,j] = ya[i,j]}
       zamin[i,j] = sum_h Yn_ih Indicator{ya[h,j] < ya[i,j]} *)

    (* update Distance *)
    if ((ConditioningActionVariable = h) and (not (miszc(i,h,thisperiod) or miszc(i,h,thisperiod+1))))
    then case dif of
      -1 : if (zc[i,h,thisperiod] <= ya(i,h))
           then Dec(Distance)
           else Inc(Distance);
       1 : if (zc[i,h,thisperiod] >= ya(i,h))
           then Dec(Distance)
           else Inc(Distance);
    end; (* case *)
end; (* ChangeBehavior *)


procedure ModelSpecification.ChangeComposition;
(* MH: Update active and nextevent, and change matrix yn *)
(* TS: lines for StructActive added *)
var
  i,j: integer;
  sum: real;
begin
  if (active[i_act]) then begin (* leaver *)
    active[i_act] := false;
    dec(n_act);
    NStructActiveElements := 0;
    for j:=1 to n do begin
      (* if there are ties at time of change, remove them *)
      if (yn(i_act,j)) then ChangeTie(i_act,j);
      if (yn(j,i_act)) then ChangeTie(j,i_act);
      StructActive[j] := false;
      if active[j] then for i := n downto 1 do begin
       if ((not struct(i,j,thisperiod)) and active[i]) then begin
          StructActive[j] := true;
          Inc(NStructActiveElements);
         end; (* if *)
       end; (* for i *)
     end; (* for j *)
  end (* if leaver *) else begin (* joiner *)
    active[i_act] := true;
    inc(n_act);
    StructActive[i_act] := false;
    for j:=1 to n do begin
      (* if there are ties before time of change, restore them *)
      if (ynm(i_act,j,thisperiod)  and (not yn(i_act,j)))  // changed 3.17u
          then ChangeTie(i_act,j);
      if (ynm(j,i_act,thisperiod)  and (not yn(j,i_act)))  // changed 3.17u
         then ChangeTie(j,i_act);
      if ((not struct(i_act,j,thisperiod)) and active[j])
        then StructActive[i_act] := true;
      if ((not struct(j,i_act,thisperiod)) and active[j])
        then StructActive[j] := true;
    end;
  end; (* if joiner *)
(* update running statistics *)
  for j:=nza downto 1 do
    begin
    sum := 0.0;
    for i:=n downto 1 do if StructActive[i] then sum := sum + ya(i,j);
    czamean[j] := sum/n_act;
    end;
  time := nextevent; (* first set time equal to time of event *)
  GetNextevent;      (* then determine time of next event *)
end; (* ChangeComposition *)


procedure ModelSpecification.InitialiseAlpha(var outf: TFileStream; var theta: pvec; var ss: integer);
(* Sets alpa to theta
   and initialises the random number generator with ss. *)
begin
  SetAlpha(theta);
  makestate(ss);
end;  (* InitialiseAlpha *)


function ModelSpecification.readstate: integer;
begin
  readstate := takeseed(version);
end;

function ModelSpecification.readstate2: integer;
begin
  readstate2 := takeseed(version2);
end;


procedure ModelSpecification.makestate(var k: integer);
begin
  if ((modelseed<>0) and not seedsown)
  then begin        (* CS: here, the random state initialisation *)
    k := modelseed; (*     according to the .MO-file occurs      *)
    seedsown := true;
    init(version,k,false);
    init(version2,k,false);
  end
  else begin
    init(version,k,true);
    init(version2,k,true);
  end;
end;




(* AA change *)  function ModelSpecification.ERGMnrun(g: byte): integer;
var dr: real;
begin
  dr := densty;
  if (dr < 0.05) then dr := 0.05;
  if (dr > 0.95) then dr := 0.95;
  if (g in [4,6,7,9,10,11,12]) then (* version A *)
    result := trunc(rstep*((n*n) - nstruct(1))*densty)
   else
    result := trunc(rstep*((n*n) - nstruct(1))/(2.0*dr*(1.0-dr)));
  if (result <= 10) then result := 10;
end;



function ModelSpecification.AddGain(var i, j: integer): real;
(*Needed in two situations:
(a) to estimate actor / period dependent partial derivatives: <addgain> = <espsilon> in <Polrup>
(b) to simulate data and estimate / test misspecified models: the true model is the model with
<alternative> and the misspecified model is the one being specified and estimated / tested.
In both cases <alternative> is set outside this procedure.*)
begin
addgain := 0; (*Default*)
(*The magnitude by which the parameter value of the specified actor / period deviates from the null hypothesis.*)
if (homogeneous < 20) then (*Actors as candidates for outlier test.*)
  begin
  if yn(i,j) then
    begin
    if (homogeneous = 11) then addgain := (alternative - alpa_f[1]) (*outdegree*)
    else if (homogeneous = 12) then
      begin
      if yn(j,i) then addgain := (alternative - alpa_f[2]); (*Reciprocity*)
      end;
    end;
  end;
end;


{$I s_ergm}
(* Includes the file S_ergm.pas, as if it were part of this file. *)

{$I s_signed }
(* Includes the file S_signed.pas, as if it were part of this file. *)


function ModelSpecification.Permission: boolean;
(*MS: By default, grant permission to continue the simulation, unless
- the estimation method is unconditional moment estimation
- and time > 1.*)
var condition: boolean;
begin
condition := true; (*Default: continue the simulation.*)
if not(cconditional or (observations = 1)) then (*Unconditional moment estimation and no ERGM model.*)
  begin
  if (time > 1.0) then (*End point of time interval reached, stop simulation right now.*)
    begin
    time := 1.0;
    condition := false;
    end;
  end;
Permission := condition;
end; (*MS: Permission.*)


procedure ModelSpecification.ChoiceProbabilities(var pr: array of real;
       const i: integer; const permitted: array of boolean);
(* Computes logarithms of (change probabilities * constant);
   does not change pr[j] if (not permitted[j]). *)
var count,h,j,var_n_tot: integer;
    sp,pup: real;
    alf: array of real;
    statistic: rrarray;
begin
{(* Compute contributions to objective functions
   for increasing and decreasing tie variables: *)
  SetLength(prup,n+1);
  for j:=n downto 0 do prup[j] := 0.0;
  AddNoTies_yn(i,prup,alpa_f,permitted,NewNetworkEffects);
  SetLength(prdown,n+1);
  for j:=n downto 0 do prdown[j] := 0.0;
  SubtractTies_yn(i,prdown,alpa_f,permitted,NewNetworkEffects);

(* Transform to proportional to cumulative probabilities: *)
  for j:=1 to n do begin
    if permitted[j] then prup[j] := prup[j-1] + expo(prup[j])
     else prup[j] := prup[j-1];
    if permitted[j] then prdown[j] := prdown[j-1] + expo(prdown[j])
     else prdown[j] := prdown[j-1];
   end;
  if (prup[n]+prdown[n] < 1e-40) then begin
    erstr := 'Too small probabilities in ChooseTieVariable';
    error := true;
    exit;
   end; (* if  *)
  pup := prup[n]/(prup[n] + prdown[n]);

  if (rando(version) < pup) then begin
(* increase a tie variable *)
    Result := Newranp(version,n,prup);
   end else begin
(* decrease a tie variable *)
    Result := Newranp(version,n,prdown);
   end;
}

(* voor correspondentie met eerdere versies eerst nog anders:  *)
  for j:=n downto 0 do pr[j] := 0.0;

  if (randomeffects or d_estAB or (homogeneous in [11..19])) then
    begin
    if d_estAB then
      begin
      NewNetworkEffects.SetChasZero(0,n,changestat); // added 3.17q
      var_n_tot := ActiveNetworkEffects_l + ActiveNetworkEffects_f + ActiveNetworkEffects_g;
      SetArray(statistic,n,NumberParameters);
      UtilityComponents_yn(i,thisperiod,ActiveNetworkEffects_l,n,statistic,permitted,NewNetworkEffects);
      sp := 0;
      for j := 1 to n do
        begin
        mp[j] := 0;
        if permitted[j] then
          begin
          pr[j] := Multinomial(ActiveNetworkEffects_l+1,var_n_tot,theta_c,RE,i,j,statistic); (*Multinomial choice probability up to the the normalizing constant on logarithmic scale.*)
          if d_estAB then
            begin
            count := ActiveNetworkEffects_l;
            for h := 1 to NetworkEffects_f do if Effects_f[h] then
              begin
              Inc(count);
              changestat[2,h,j] := statistic[j,count];
              end;
            for h := 1 to NetworkEffects_g do if Effects_g[h] then
              begin
              Inc(count);
              changestat[3,h,j] := statistic[j,count];
              end;
            end;
          mp[j] := expo(pr[j]);
          sp := sp + mp[j];
          end;
        end;
      for j := 1 to n do mp[j] := mp[j] / sp;
      end
    else (*not d_estAB *)
      begin
      SetLength(alf,NetworkEffects_f);
      for j := 1 to NetworkEffects_f do alf[j-1] := alpa_f[j] + GetRE(1,j,i); (*MS: coefficient: objective function; <j>: effect; <i>: actor.*)
      if ((homogeneous in [11..19]) and ((derivatives and (alternative <> 999999999)) or altern) and outlier[i])
      then alf[homogeneous-10-1] := alternative;
      if (not downonly[thisperiod]) then AddNoTies_yn(i,thisperiod,pr,alf,permitted,d_estAB,changestat,NewNetworkEffects);
      for j := 1 to NetworkEffects_f do alf[j-1] := alf[j-1] + alpa_g[j] + GetRE(2,j,i); (*MS: coefficient: gratification function; <j>: effect; <i>: actor.*)
      if (not uponly[thisperiod]) then SubtractTies_yn(i,thisperiod,pr,alf,permitted,d_estAB,changestat,NewNetworkEffects);
      end;
    end
  else (* no randomeffects or .... *)
    begin
    if (not downonly[thisperiod]) then AddNoTies_yn(i,thisperiod,pr,permitted,NewNetworkEffects);
    if (not uponly[thisperiod]) then SubtractTies_yn(i,thisperiod,pr,permitted,NewNetworkEffects);
    end;

  pr[i] := 0.0;  (* this is valid for StaySameType = 2;
           for StaySameType = 1, it must be -infinity, must be reset later
           or handled by a "permitted"-type array. *)
end; (* ChoiceProbabilities *)


procedure ModelSpecification.runstep(var error: Boolean;
                             var erstr: shortstring; var halterr: Boolean;
                             var outf,lf: TFileStream);
(* one change *)
(* The system random number state must have been set
   before entering this procedure. *)
(* the step consists of a stochastic change of the adjacency matrix *)
var i,j,h,hh,
    hvar,s,sum,
    a1,a2,
    b1,b2,
    c1,c2,iii,
    ModelCode,
    effectnumber      : integer;
    tie1,tie2,tie3, ynij,
    cc,change,yih,
    plus,minpossible,
    SomePermitted,
    continue,
    ChangeToPlus, ChangeToMin,
    pluspossible      : Boolean;
    sp,sp1,sp2,
    lxi,lama, taumin,
    eu,eu1,eu2,lnp,lnp2,
    lami,lamsum2,lamsumrev,ran,
    pro,pro1,pro2,entr,
    p2,p3,p4,p5,
    p6,p7,p8,
    coefficient, yai,
    contribution      : real;
    hstar             : byte;
    stats0, stats1    : pvec;
    thecomponents     : array_0_20;
    PlusPermitted, MinPermitted,
    permitted         : array of boolean;
    ords              : intvec;
    booleandummy      : boolean;
    prr,
    rs_lamsum, rs_tau : array of real;
    rs_p              : array of array of real;
    x : string;

begin (* ModelSpecification.Runstep *)
  (* The ModelCode variable gives the possibility of a double use of the
     variables ModelCode and ModelType, which enables a symmetry between the
     treatment of non-directed and directed networks,
     without this symmetry having to be reflected in the code
     numbers used for ModelType.   *)
     (* Changes in runstep should be copied to S_ML.CalcRatelProb;
        but since in S_ML.CalcRatelProb only one probability needs to be calculated,
        for model specifications such as type 3 it is more efficient
        not to use a common procedure to calculate the rate and
        full probability vectors. *)

  if signed then setlength(rs_p,1+nza,2*n+1) else
    setlength(rs_p,1+nza,n+1); (* rs_p contains vectors of actor-rates for each dep. dimension;             *)
  setlength(rs_lamsum,1+nza); (* similarly, rs_lamsum contains the sums of these rates, and           *)
  setlength(rs_tau,1+nza); (* rs_tau contains random draw from the corresp. exponential distribution. *)
                          (* for all 3 dynamic arrays: entry 0 = network, entry k>0 = behaviour k *)
  setlength(permitted,nmax+1);
//  setlength(ords,n+1);

  if signed then
    begin
    setlength(pluspermitted,n+1);
    setlength(minpermitted,n+1);
    end;
  error := false; erstr := '';
  halterr := false;
  ModelCode := ModelType;
  if symmetric then begin
    if (ModelType = 2) then ModelCode := 1;
   end;
  if (observations = 1)
  then ErgmRunStep(error,erstr,halterr,outf,lf)
  else begin (* observations >= 2 *)
    for hvar := 0 to nza do rs_p[hvar][0] := 0.0; (* If this is forgotten, ranp runs into unpredictable errors! *)
    inc(stepno);

    if (ModelCode = 3)
    then begin
     rs_lamsum[0] := alpa_l[thisperiod]*NStructActiveElements;
    end
    else begin
      (* Choose actor i who will be allowed to change one of his tie variables.
         MH: Only active actors are allowed to make changes.
             For nonactive actors all probabilities are set to 0. *)

      (* First calculate the network change rates for each actor: *)
      rs_lamsum[0] := 0.0; lamsum2 := 0.0;
      for i:= n downto 1
      do begin
        if (StructActive[i])
        then begin
          (* MH: only active actors; TS: StructActive implies Active *)
          lami := NetworkLambda(i,halterr,outf,lf); (* network change rate for actor i *)
          if halterr then exit;
          if (lami < 0.0)
          then begin
            str(lami:20:10,erstr);
            erstr := 'Negative network change rate('+inttostr(i)+'); lami = '+erstr+' in RunStep.';
            error := true;
            exit;
          end;
        end (* if StructActive i *)
        else lami := 0.0; (* MH: nonactive actors *)
        rs_lamsum[0] := rs_lamsum[0] + lami; (* sum over structurally active actors *)
        if symmetric then lamsum2 := lamsum2 + sqr(lami);
        rs_p[0][i] := lami; (* recall rate of network change for actor i *)
      end; (* for i *)
    end; (* ModelCode *)

    if (symmetric and (ModelCode >= 4))
    then lamsumrev := sqr(rs_lamsum[0]) - lamsum2
    else lamsumrev := rs_lamsum[0];
    if (lamsumrev <= 0.0)
    then begin
      str(lamsumrev:20:10,erstr);
      erstr := 'Nonpositive network change rate in RunStep: '+erstr;
      error := true;
      exit;
    end;

    (* Draw time increment rs_tau[0] until next network event: *)
    rs_tau[0] := -log(rando(version))/lamsumrev;
    taumin := rs_tau[0]; (* earliest network event = so far earliest event overall *)
    hstar := 0;

    (* Now take a look at behavioural events: *)
    for hvar:=nza downto 1 do
      begin
      rs_lamsum[hvar] := 0.0;
      for i:= n downto 1 do
        begin
        if (StructActive[i]) then
          begin
          lami := ActionLambda(i,hvar,halterr,outf,lf);
          if halterr then exit;
          if (lami <= 0.0) then
            begin
            str(lami:20:10,erstr);
            erstr := 'Nonpositive action change rate in RunStep: '+erstr;
            erstr:=erstr+' - actor '+st(i)+' in period '+st(thisperiod)+' actor outdegree='+st(outd_yn(i));
            error := true;
            exit;
            end;
          end (* if StructActive i *)
         else lami := 0.0; (* nonactive actors *);

        rs_lamsum[hvar] := rs_lamsum[hvar] + lami; (* sum over structurally active actors *)
        rs_p[hvar][i] := lami; (* recall rate of action change for actor i *)
        end; (* for i *)

      (* Draw time increment rs_tau[hvar] until next event on behaviour hvar: *)
      rs_tau[hvar] := -log(rando(version))/rs_lamsum[hvar]; (* earliest change in behaviour hvar *)
      if (rs_tau[hvar] < taumin) then
        begin (* Only the smallest time increment has an effect -- so, update taumin here: *)
        hstar := hvar;
        taumin := rs_tau[hvar];
        end;
      end; (* for hvar *)

    (* Now taumin is the earliest time of change,
       and hstar is the variable to change, where hstar = 0 means a network change. *)
    time := time+taumin;

    continue := Permission; (*MS: By default, grant permission to continue the simulation, unless moment estimation is unconditional and time > 1.*)
    if d_estAB then (*MS: MoM: derivative estimation.*)
      begin
      for h := 0 to nza do
        begin
        total_rate[h,0] := rs_lamsum[h]; (*Sum of actor-dependent rate functions for dependent variable <h>.*)
        for i := 1 to n do total_rate[h,i] := rs_p[h,i]; (*Actor dependent rate function for dependent variable <h>.*)
        end;
      if continue then last_time[hstar,thisperiod] := time;
      end;

    (* MH: If exogenous events are to occur, check whether time >= nextevent and change composition. *)
    cc := false;
    if (exo and (time >= nextevent) and (time < 1.0))
    then cc := true;

    (* MH: if exo then unconditional estimation and simulation run always ends when time>t=1.0 *)
    if cc then ChangeComposition (* MH: Actor joining or leaving the network. *)
   else if continue then
      begin (* this begin ... end runs on until almost the end of runstep *)

      if (hstar <= 0) then
        begin

      (* CHANGE NETWORK: *)
      (* =============== *)

        if (ModelCode <> 3) then (* not tie-based*)  (* this is the start of AAB: *)
          begin (* choose active actor to change something *)
          for i:=1 to n do rs_p[0][i] := rs_p[0][i] + rs_p[0][i-1];
          repeat
            i := newranp(version,n,rs_p[0]);
          (* The following check has probability 1 under correct specifications. *)
           until ((i >= 1) and (i <= n) and StructActive[i]);

(* At this point, for actor-based models, actor i has been determined.
   Now choose the change to be made by this actor:
   i.e. choose the actor j with whom the tie variable changes. *)

          if (symmetric and (ModelCode >= 4)) then
            begin
            (* In this case, the choice of the potential j
               is a matter of the rate function;
               the same probability vector rs_p is used,
               but the extra condition i <> j will be applied. *)
            repeat
              j := newranp(version,n,rs_p[0]);
            until ((j >= 1) and (j <= n) and (j <> i) and
                   (not Struct(i,j,thisperiod)) and StructActive[j]);
            (* And now it depends on the ModelCode how the choice is made *)

            (* In the following case statement, the probability lnp
               of changing the tie variable is calculated. *)
            ynij := yn(i,j);
            if ynij and uponly[thisperiod] then exit;
            if (not ynij) and downonly[thisperiod] then exit;
            lnp := UtilityGain(ynij,i,j,thisperiod,true,false,NewNetworkEffects);
            lnp2 := UtilityGain(ynij,j,i,thisperiod,true,false,NewNetworkEffects);
            case ModelCode of
              4: begin
                 if ynij then
                   begin  (* both must agree to sever the tie *)
                   lnp := expit(lnp);
                   lnp2 := expit(lnp2);
                   lnp := lnp*lnp2;
                   end
                  else
                   begin  (* at least one must wish to create the tie *)
                   lnp := expit(lnp);
                   lnp2 := expit(lnp2);
                   lnp := lnp + lnp2 - lnp*lnp2;
                   end;
                 end;
              5: begin
                 if ynij then
                   begin  (* at least one must wish to sever the tie *)
                   lnp := expit(lnp);
                   lnp2 := expit(lnp2);
                   lnp := lnp + lnp2 - lnp*lnp2;
                   end
                  else
                   begin  (* both must agree to create the tie *)
                   lnp := expit(lnp);
                   lnp2 := expit(lnp2);
                   lnp := lnp*lnp2;
                   end;
                 end;
              6: begin
                 lnp := expit(lnp+lnp2);
                 end;
             end; (* case *)
            if CalcEntropy
            then entr := (- lnp*log(lnp) - (1.0-lnp)*log(1.0-lnp))/log(2.0);
            if (rando(version) < lnp)
            then
              begin
              ChangeTie(i,j);
              ChangeTie(j,i);
              end;
            end
           else
            begin (* not (symmetric and ModelCode >= 4) *)
            (* Now starts the part for nonsymmetric model types -1,1,2.
               actor i chooses the optimal other actor j *)
            case ModelCode of
             -1 : begin
                    if downonly[thisperiod] then
                      for j:=n downto 1 do PlusPermitted[j] := false else
                      PermitPlus(n,i,thisperiod,valplus,PlusPermitted);
                    if uponly[thisperiod] then
                      for j:=n downto 1 do MinPermitted[j] := false else
                      PermitMin(n,i,thisperiod,valmin,MinPermitted);
                    if exo then for j:=n downto 1 do
                     if (not StructActive[j]) then
                      begin
                      PlusPermitted[j] := false;
                      MinPermitted[j] := false;
                      end;
//TS: replaces:                  for j:=n downto 1 do begin
//                      if ((not Struct(i,j,thisperiod)) and StructActive[j]
//                           and (ynv(i,j) < valplus)) then Pluspermitted[j] := true
//                           else PlusPermitted[j] := false;
//                      if ((not Struct(i,j,thisperiod)) and StructActive[j]
//                           and (ynv(i,j) > valmin)) then Minpermitted[j] := true
//                           else MinPermitted[j] := false;
//                     end;
                  end; (* case -1 *)
              1 : begin
                  if checkchange then PermitChange(n,i,thisperiod,maxdegree,permitted)
                    else for j :=n downto 1 do permitted[j] := true;
                  end; (* case 1 *)
              2 : begin
                  s := outd_yn(i);
                  if downonly[thisperiod] then plus := false
                   else if uponly[thisperiod] then plus := true
                   else
                    begin
                    plus :=  (s <= 0);
                    if (not (plus or (s >= (n-1))))
                    then
                      begin
                      lxi :=  lambda2[0,i]*xi(s);
                      sp := nu(s)*lxi*(1.0+lambda2[0,i]*xi(s-1))/(nu(s-1)*(1.0+lxi));
                      plus := (rando(version) <= sp);
                      end; (* if *)
                    end;
// TS: following still must be adapted to PermitChange etc.
                  if plus
                  then for j:=n downto 1
                  do permitted[j] := ((not yn(i,j)) and (outd_yn(i) < maxdegree)
                        and (not Struct(i,j,thisperiod)) and StructActive[j])
                  else for j:=n downto 1
                  do permitted[j] := (yn(i,j) and (not Struct(i,j,thisperiod)) and StructActive[j]);
                end; (*  case 2 *)
              end; (* case ModelCode *)

            if (ModelCode = -1) then
              begin
              PlusPermitted[i] := (StaySameType = 2) and (not downonly[thisperiod]);  // ????
              MinPermitted[i] := false;
              end
             else Permitted[i] := (StaySameType = 2);

            (* TS 15-jan-05 added the following check.
               It is not defined as an error check because for a different i
               there might exist a permitted j . *)
            if checkchange or uponly[thisperiod] or downonly[thisperiod] then                  // @@@@@
              begin                                      // @@@@@
              somepermitted := (StaySameType = 2);
              j := n;
              repeat
                if (ModelCode = -1) then begin
                  if (MinPermitted[j] or PlusPermitted[j]) then somepermitted := true;
                 end else begin
                  if permitted[j] then somepermitted := true;
                 end;
                dec(j);
              until ((j <= 0) or SomePermitted);
              if (not SomePermitted) then exit;
              end                                                // @@@@@
             else somepermitted := true; // @@@@@

            sp := 0.0;
(* For the signed option, rs_p[0][1..n] contains the probabilities of increasing ynv(i,j),
   and rs_p[0][(n+1)..(2*n)] the probabilities of decreasing ynv(i,j). *)
            if (ModelCode = -1) then
              begin
              for j:=1 to n do
                begin
                if PlusPermitted[j] then
                  begin
                  if (i = j) then lnp := 0.0
                 else
                    begin
                    lnp := SignedUtilityPlusGratificationGain(i,j,+1,halterr,outf,lf);
                    if halterr then
                      begin
                      error := true;
                      erstr := 'error in SignedUtilityPlusGratificationGain';
                      exit;
                      end; (* if halterr *)
                    end; (* if i <> j *)
                  rs_p[0][j] := expo(lnp);
                  sp:=sp+rs_p[0][j];
                  end (* if PlusPermitted[j] *)
                 else rs_p[0][j] := 0.0;
                if (MinPermitted[j] and (i <> j)) then
                  begin
                  lnp := SignedUtilityPlusGratificationGain(i,j,-1,halterr,outf,lf);
                  if halterr then
                    begin
                    error := true;
                    erstr := 'error in SignedUtilityPlusGratificationGain';
                    exit;
                    end; (* if halterr *)
                  rs_p[0][n+j] := expo(lnp);
                  sp:=sp+rs_p[0][n+j];
                  end (* if MinPermitted[j] *)
                 else rs_p[0][n+j] := 0.0;
                end; (* for j *)
              if (sp < 1e-40) then
                begin
                erstr := 'Too small probabilities';
                error := true;
                exit;
                end; (* if sp *)
              for j:=2*n downto 1 do rs_p[0][j] := rs_p[0][j]/sp;
              repeat
               j:=ranp(version,2*n,rs_p[0]);
              until (((j<= n) and PlusPermitted[j]) or ((j > n) and MinPermitted[j-n]));
              (* For exact precision the last condition is satisfied with probability 1 *)
              if (i=j) then
                begin
                ChangeToPlus :=false;
                ChangeToMin := false;
                end
               else
                begin
                if (j <= n) then
                  begin
                  ChangeToPlus := true;
                  ChangeToMin := false;
                  end
                 else
                  begin
                  ChangeToPlus := false;
                  ChangeToMin := true;
                  j := j-n;
                  end;
                end;
              end
             else
              begin (* ModelCode <> -1 *)
              SetLength(prr,n+1);
              ChoiceProbabilities(prr,i,permitted);
(* MS: note: present random effects <RE> are used, nested in ChoiceProbabilities.*)
(* Transform to proportional to cumulative probabilities: *)
              for j:=1 to n do
                begin
                if permitted[j] then prr[j] := prr[j-1] + expo(prr[j])
                  else prr[j] := prr[j-1];
                end;
              if (prr[n] < 1e-40) then
                begin
                erstr := 'Too small probabilities in RunStep';
                error := true;
                exit;
                end; (* if  *)

              j := Newranp(version,n,prr);

              if (j <= 0) or (not permitted[j]) then
                begin
                erstr := 'Error : not permitted['+inttostr(j)+'] in RunStep('+inttostr(i)+').';
                error := true;
                exit;
                end;
              end; (* if ModelCode <> -1 *)

(* Now j has been chosen *)

              (* TS: calculation of entropy : *)
            if CalcEntropy
            then begin
              entr := 0.0;
              if signed then begin
                for h:=n downto 1 do begin
                  if Pluspermitted[h] then entr := entr - rs_p[0][h]*log(rs_p[0][h]);
                  if Minpermitted[h] then entr := entr - rs_p[0][n+h]*log(rs_p[0][n+h]);
                 end;
                entr := entr/log(2*n);
               end else begin
                for h:=n downto 1
                do if permitted[h] then entr := entr - rs_p[0][h]*log(rs_p[0][h]);
                entr := entr/log(n);
               end;
             end; (* if CalcEntropy *)

            if (symmetric and (ModelType = 2)) then
              begin
              ynij := yn(i,j);   (* = yn(j,i) *)
              (* reciprocal confirmation required of new tie *)
              if (not ynij) then
                begin
                lnp := UtilityGain(false,j,i,thisperiod,true,false,NewNetworkEffects);
                lnp := expit(lnp);
                if CalcEntropy
                then entr := entr - lnp*log(lnp) - (1.0-lnp)*log(1.0-lnp);
                if (rando(version) < lnp) then
                  begin
                  ChangeTie(i,j);
                  ChangeTie(j,i);
                  end;
                end (* not ynij *)
               else
                begin (* yn(i,j): no confirmation required for canceling a tie *)
                ChangeTie(i,j);
                ChangeTie(j,i);
                end;
              end (* symmetric and ModelType = 2 *)
             else
              begin
              if signed then
                begin
                if ChangeToPlus then PlusTie(i,j) else
                  begin
                  if ChangeToMin then MinTie(i,j)
                  end;
                if symmetric then
                  begin
                  error := true;
                  erstr := 'Signed symmetric graphs not implemented (runstep).';
                  exit;
                  end;
                end (* if signed *)
               else
                begin
                ChangeTie(i,j);
                if symmetric then ChangeTie(j,i);
                end; (* if not signed  *)
              end;  (* if not (symmetric and (ModelType = 2)) *)
            end; (* if not (symmetric and ModelCode >= 3) *)
          end (* this is the end of AAB: if ModelCode <> 3 *)
         else
          begin (* ModelCode = 3: tie-based *)
           (* choose a random pair (i,j) only under the restriction that
              (i,j) is not structurally determined *)
            repeat
             i := randomd(version,n)+1;
             j := randomd(version,n)+1;
             until ((i <> j) and (not Struct(i,j,thisperiod)));
            if symmetric then
              begin
              ynij := yn(i,j);
              lnp := UtilityGain(ynij,i,j,thisperiod,true,false,NewNetworkEffects);
              lnp2 := UtilityGain(ynij,j,i,thisperiod,true,false,NewNetworkEffects);
              lnp := 0.5*(lnp + lnp2);
              end
             else
              lnp := utilitygain(i,j,thisperiod,symmetric,false,NewNetworkEffects);
            lnp := expit(lnp);
            if CalcEntropy then
              entr := - lnp*log(lnp) - (1.0-lnp)*log(1.0-lnp);
            if (rando(version) < lnp)
            then begin
              ChangeTie(i,j);
              if symmetric then ChangeTie(j,i);
            end;
         end; (* if ModelCode = 3 *)

        if (CalcEntropy and (stepno >= 1) and (stepno <= stepnomax))
        then begin
          entrop^[stepno] := entr;
          entroptime^[stepno] := time;
        end;
      end (* if hstar = 0 *)

      else begin (* hstar > 0 *)

      (* CHANGE BEHAVIOUR hstar: *)
      (* ======================= *)

        (* Draw actor i who may change his behaviour: *)
        rs_p[hstar][0] := 0.0;
        for i:=1 to n do rs_p[hstar][i] := rs_p[hstar][i]/rs_lamsum[hstar];
        repeat
          i:=ranp(version,n,rs_p[hstar]);
        until ((i >= 1) and (i <= n) and active[i]);

        (* Now see which change is made *)
        (*MS: the code below is extended to random effects models;
        important for simulation by program Siena 5 and procedure SimulateData!*)

        yai := ya(i,hstar);   // RR: in view of a compiler error in Linux fpc.
        minpossible := (yai > zcmin[hstar]);
        pluspossible := (yai < zcmax[hstar]);

(* added 3.17q: *)
        if d_estAB then for h:= 1 to ActionEffects_f do
          begin
          changestat[5,h,0] := 0.0;
          changestat[6,h,0] := 0.0;
          changestat[5,h,2] := 0.0;
          changestat[6,h,2] := 0.0;
          end;

        if minpossible
        then begin
          lnp := 0.0;
          CalcComponents_fa(i,NetworkEffects_f,ActionEffects_f,nz,
                            hstar,thisperiod,-1,false,Effects_s,
                            thecomponents,outf);
          for h:= 1 to ActionEffects_f do
            begin
            effectnumber := NetworkEffects_f + (hstar-1)*ActionEffects_f + h;
(* use Actioneffects_f = ActionEffects_g, NetworkEffects_f = Networkeffects_g *)
            if (Effects_f[effectnumber] or Effects_g[effectnumber]) then
              begin
//wwriteln(outf,' rstep :: '+inttostr(effectnumber));
              contribution := Contr_fa(i,h,hstar,-1,thecomponents,halterr,outf,lf);
              if Effects_f[effectnumber] then
                begin
                if d_estAB then changestat[5,h,0] := contribution; (*MS: MoM: derivative estimation.*)
                coefficient := alpa_f[effectnumber] + GetRE(1,effectnumber,i); (*MS: compute coefficient: objective function.*)
                lnp := lnp + coefficient*contribution;
                end;
              if Effects_g[effectnumber] then
                begin
                if d_estAB then changestat[6,h,0] := contribution; (*MS: MoM: derivative estimation.*)
                coefficient := alpa_g[effectnumber] + GetRE(2,effectnumber,i); (*MS: compute coefficient: gratification function.*)
                lnp := lnp + coefficient*contribution;
                end;
              end;
            end; (* for h *)


          if halterr
          then begin
            erstr := 'error in Contr_fa- '+st(hstar);
            error := true;
            exit;
          end; (* if halterr *)

          rs_p[hstar][1] := expo(lnp);
        end (* if minpossible*)
        else rs_p[hstar][1] := 0.0;

        rs_p[hstar][2] := 1.0; (* this is expo(0),
                        corresponding to zero utility change when not changing behaviour hstar *)

        if pluspossible
        then begin
          lnp := 0.0;
          CalcComponents_fa(i,NetworkEffects_f,ActionEffects_f,nz,
                          hstar,thisperiod,1,false,
                          Effects_f,thecomponents,outf);

          for h:= 1 to ActionEffects_f  do
            begin
            effectnumber := NetworkEffects_f + (hstar-1)*ActionEffects_f + h;
            if Effects_f[effectnumber] then
              begin
              contribution := Contr_fa(i,h,hstar,1,thecomponents,halterr,outf,lf);
              if d_estAB then changestat[5,h,2] := contribution; (*MS: MoM: derivative estimation.*)
              coefficient := alpa_f[effectnumber] + GetRE(1,effectnumber,i); (*MS: compute coefficient: objective function.*)
              lnp := lnp + coefficient*contribution;
              end;
            end;


          if halterr
          then begin
            erstr := 'error in Contr_fa+ '+st(hstar);
            error := true;
            exit;
          end; (* if halterr *)

          rs_p[hstar][3] := expo(lnp);
        end (* if pluspossible*)
        else rs_p[hstar][3] := 0.0;

        (* Calculate choice probabilities and determine action *)
        sp := rs_p[hstar][1] + rs_p[hstar][2] + rs_p[hstar][3];
        for h:=1 to 3 do rs_p[hstar][h] := rs_p[hstar][h]/sp;
        repeat
          h:=ranp(version,3,rs_p[hstar]);
        until ((h >= 1) and (h <= 3));
        if ((h <= 1) and (not minpossible))
        then begin
          x := format('%s%d%s',['*** error: step below minimum behavior variable ',h,' ***'])+eol;
          lf.Write(pchar(x)^,length(x));
          h := 2;
        end;
        if ((h >= 3) and (not pluspossible))
        then begin
          x := format('%s%d%s',['*** error: step above maximum behavior variable ',h,' ***'])+eol;
          lf.Write(pchar(x)^,length(x));
          h := 2;
        end;
        case h of
          1 : ChangeBehavior(i,hstar,-1);
          3 : ChangeBehavior(i,hstar,1);
        end (* case *)
      end; (* if hstar > 0 *)

      if d_estAB then
        begin (*MS: MoM: derivative estimation.*)
        if (hstar = 0) then StoreScoreI(mp,i,j) (*Network objective / endowment function parameters.*)
        else StoreScoreII(rs_p[hstar],hstar,h,minpossible,pluspossible); (*Behavior objective / endowment function parameters.*)
        StoreScoreIII(hstar,taumin,i); (*Network rate function parameters.*)
        StoreScoreIV(hstar,taumin,i); (*Behavior rate function parameters.*)
        end; (*MS: MoM: must be placed inside the <cc> loop.*)

    end; (* if not cc; time >= nextevent *)
  end; (* if observations >= 2 *)

  if checkstatistics then begin
    if signed then signedstatistics(stats0,ords,thisperiod) else begin
      NetworkStatistics(stats0,ords,thisperiod,outf);
      ActionStatistics(stats0,ords,thisperiod,outf);
     end;

    if (observations <= 1) then begin
     end else begin (* observations >= 2 *)

       x := format('%s%16d%s',['seed ',takeseed(version),'; ']);
       outf.Write(pchar(x)^,length(x));
       x := format('%s%4d%s%7.4f%s%3d%s',['  ',distance,' ',time,' ',hstar,' ']);
       outf.Write(pchar(x)^,length(x));
       if (hstar = 0) then begin
       x := format('%3d%s%3d%s',[i,' ',j,' ']);
       outf.Write(pchar(x)^,length(x));
       end
       else outf.write(pchar('   *   *')^,8);
       if signed then begin
         sum := 0;
         if changetoplus then sum := sum+1;
         if changetomin then sum := sum-2;
         case sum of
          -2: outf.write(pchar(string(' -1 '))^,4);
           0: outf.write(pchar(string('  0 '))^,4);
           1: outf.write(pchar(string('  1 '))^,4);
          else  outf.write(pchar(string(' ?? '))^,4);
         end; (* case *)
        end; (* if signed *)
       outf.write(pchar(string(' | '))^,3);
       for h:=1 to Numberfunctions do begin
         x := format('%8.2f',[stats0[h]]);
         outf.Write(pchar(x)^,length(x));
        end; (* for h *)
      // writeln(outf);
       outf.Write(pchar(eol)^,length(eol));
     end; (* if observations >= 2 *)
   end; (* if checkstatistics *)
end; (* runstep *)


procedure ModelSpecification.runepoch(t: real; var success: Boolean;
                             updatestat: Boolean;
                             var error: Boolean; var erstr: shortstring;
                             var halterr: Boolean; var outf,lf: TFileStream);
(* if observations >= 2 and CConditional: runsteps until Distance exceeds t
   else runsteps until time exceeds t. *)
(* It is possible that Distance reaches an equilibrium state lower than t.
   Then success := false, and miss is the amount by which distance falls
   short of t. *)
var pastruns, MaxDis, nrun1, i, j, h: integer;
    (* pastruns is the number of runs since the first occurrence
         of this maximum distance;
         if pastruns is too large, then it is inferred that an
         equilibrium has been reached for this parameter vector
         and the `equilibrium' network is taken as the outcome
         of the simulation;
       MaxDis is the maximal distance yet observed *)
    stats0: pvec;
    ords: intvec;

od: array of integer;
x: string;

begin (* runepoch *)
  error := false;
  erstr := '';
  halterr := false;
  pastruns := 0;
  stepno := 0;
//  setlength(ords,n+1);
  for i:=1 to stepnomax
  do begin
    entrop^[i] := 0.0;
    entroptime^[i] := 0.0;
  end;
  updatestat := (OutputType >= 2);

  if checkstatistics then begin
    outf.Write(pchar(eol)^,length(eol));
    x := format('%s%d',['Start period ',thisperiod])+eol;
    outf.Write(pchar(x)^,length(x));
    x := format('%s%d',['seed ',takeseed(version)])+eol;
    outf.Write(pchar(x)^,length(x));
//    for i:=1 to n do begin
//      write(outf,i:3,'. ');p
//      for j:=1 to n do write(outf,ynvm(i,j,thisperiod):3);
//      writeln(outf);
//      write(outf,i:3,'. ');
//      for j:=1 to n do write(outf,ynv(i,j):3);
//      writeln(outf);
//     end; (* for i *)
//    writeln(outf);
    NetworkStatistics(stats0,ords,thisperiod,outf);
    ActionStatistics(stats0,ords,thisperiod,outf);
    if (observations <= 1) then begin
     end else begin (* observations >= 2 *)
       x := format('%s%4d%s%7.4f%s',['start period: ',distance,' ',time,'      | ']);
       outf.Write(pchar(x)^,length(x));
       for h:=1 to Numberfunctions do begin
         x := format('%8.2f',[stats0[h]]);
         outf.Write(pchar(x)^,length(x));
        end; (* for h *)
       outf.Write(pchar(eol)^,length(eol));
     end; (* if observations >= 2 *)
   end; (* if checkstatistics *)


  if (observations <= 1)
  then begin
(* AA change *)    nrun1 := ERGMnrun(gibbs);
    istar := 1; jstar := 1;
    GibbsChange := 0;
    repeat
      inc(pastruns);
      ErgmRunStep(error,erstr,halterr,outf,lf);
    until ((pastruns >= nrun1) or halterr);
    if halterr then exit;
    GibbsChangeFraction := GibbsChange/nrun1;
    SumGibbsChangeFraction := SumGibbsChangeFraction + GibbsChangeFraction;
    success := true;
  end

  else begin (* observations >= 2 *)
    MaxDis := 0;
    if CConditional
    then begin
      while ((Distance < t) and (not (error or stopo)) and (pastruns < MaxRuns))
      do begin
        runstep(error,erstr,halterr,outf,lf);
        if halterr then exit;
        if (Distance > MaxDis)
        then begin
          MaxDis := Distance;
          pastruns := 0;
        end
        else inc(pastruns);
        success := (Distance >= t);
      end
    end
    else begin (* not conditional, MH: standard for exogeneous events *)
      while ((time < t) and (not (error or stopo)))
      do begin
        runstep(error,erstr,halterr,outf,lf);
        if halterr then exit;
        inc(pastruns);
        success := (time >= t);
      end;
    end; (* if-then-else Cconditional *)


      (* added 3.17s:
   Force new structurally determined values at the end of this period
   to be equal to those in the observed data.
   This is necessary in case that structurally determined values
   change between waves.
   In that case the statistics use the structurally determined values
   for the start of the period  - which is automatic here in runepoch;
   and elements (i,j) for which there is a structurally determined value
   at the end but not at the start of the period
   get their value from the end of the period
   - which is accomplished here by the use of ForceNewStructurals.
   Search for ForceStructurals and ForceNewStructurals to see where this happens.  *)

   ForceNewStructurals(thisperiod+1);

   (* distribute entroptime evenly *)
    if updatestat
    then begin
      for i:=1 to stepno
      do if (i <= stepnomax)
      then begin
        h := trunc(100.0*(thisperiod-1 + (entroptime^[i]/(time+1e-20)))/(observations-1));
        if (h >= 0) and (h <= 100)
        then begin
          entropy[h] := entropy[h] + entrop^[i];
          inc(entropynumber[h])
        end (* if *)
      end (* for i do if *)
    end (* if updatestat *)
  end; (* observations >= 2 *)

  if checkstatistics then begin
  x:= format('%s%d',['seed ',takeseed(version)])+eol;
  outf.Write(pchar(x)^,length(x));
  end;
  Application.ProcessMessages
end; (* runepoch *)

procedure ModelSpecification.SimStats(p: Byte; theta: pVec;
                  allstats, varmat: boolean; var fseed: integer;
                  var ystat: pVec; var d: array of real;
                  var dd: pmatrix;
                  updatestat: Boolean;
                  var feasible, error: Boolean;
                  var errstr: shortstring; var outf,lf: TFileStream);
(* This procedure is used in S_Sim, and in S_Est in POLRUP.
   It simulates one run and stores the outcome in ystat
   and (for observations >= 2) the simulated times in d;
   Secondary simulation results are updated if updatestat.
   If allstats, then the distances are included in ystat,
   otherwise they are left out  - changed now, can be omitted. *)
(* The arguments varmat, dd, are for compatibility with Scores in S_ML
   and my plan is to use this later for some simplifications in POLRUP (TS). *)
var dt,gc,mi,
    statperiod,
    original : real;
    i,j,k,activek,
    sed      : integer;
    pp       : byte;
    success,
    halterr  : boolean;
    statsum,
    statk    : pvec;
    ordes    : intvec;
    dummy    : extended;
    alpa1    : CRarray;
    yaa      : array of array of byte;
    x: string;
begin (* SimStats *)
(* calculation of functions duplicated in FunctionEvaluations,
   but then for observed digraphs *)
(* TS changed 25-01-06, as in FunctionEvaluations *)
  Error := false; errstr := ''; halterr := false;
  updatestat := (OutputType >= 2);
  setlength(yaa,nmax+1,nza+1); // abcdefg
  setlength(ordes,nmax+1);

  InitialiseAlpha(outf,theta,fseed);
//writeln(outf,' checkk: ');
//NewNetworkEffects.WriteMod_fSequenceParameters(alpa_f,outf);

(* TS added 05-jan-05 *)
  alpa1 := alpa_l;
  if CConditional
  then for i:=1 to observations-1
  do if (ConditioningNetworkvariable=1)
  then alpa_l[i] := 1.0 (* CS: used to be 1.0 for old rate parametrisation; TS teruggezet *)
  else alpa_l[NetworkEffects_l+(ConditioningActionVariable-1)*ActionEffects_l+i] := 1.0;

  pp := NumberFunctions;
  if d_estAB then SetZero(cdscore,pp,observations-1); (*MS: method of moments, phase 3: derivative estimation.*)
  if checkstatistics then
    begin
    x := 'Option checkstatistics: '+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Parameters for rate function: '+eol;
    outf.Write(pchar(x)^,length(x));
    for i:=1 to NetworkEffects_l do if effects_l[i] then begin
      x := format('%8.4f',[alpa_l[i]]);
      outf.Write(pchar(x)^,length(x));
    end;
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := 'Parameters for objective function: '+eol;
    outf.Write(pchar(x)^,length(x));
    for i:=1 to NetworkEffects_f do if effects_f[i] then begin
      x := format('%8.4f',[alpa_f[i]]);
      outf.Write(pchar(x)^,length(x));
    end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    x := 'write intermediate results to output file at the end of each runstep:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Distance; time; hstar; i; j; ';
    outf.Write(pchar(x)^,length(x));
    if signed then begin
      x := ' dif; | ' +eol;
      outf.Write(pchar(x)^,length(x));
    end
     else  begin
      x := ' | '+eol;
      outf.Write(pchar(x)^,length(x));
     end;
    for i:=1 to NumberFunctions do begin
    x := StatsDescriptions[i]+eol;
    outf.Write(pchar(x)^,length(x));
    end;
    for i:=1 to 2 do begin
    x := labels_f[i]+eol;
    outf.Write(pchar(x)^,length(x));
    end;
   // writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    end; (* if checkstatistics *)


//for i:=1 to pp do write(outf,'duz: ', currentstats[i]:6:0); writeln(outf); //@@@

  if (observations <= 1) then
    begin
    initialise_y(1);
//r i:=1 to n do begin
//for k:=1 to n do if yn(i,k) then write(outf,'1') else write(outf,'0');
//writeln(outf);
//end;
//wwrite(outf,'dus: ');
//for i:=1 to pp do wwrite(outf,' '+ floattostr(currentstats[i])); wwriteln(outf,''); //@@@

(* This sets GibbsChange to 0,
   and sets an independent starting matrix if IndependentStart. *)
    dt := 1.0;
    runepoch(dt,success,false,error,errstr,halterr,outf,lf);

    if GibbsConnected then
      begin
      if (not ConnectedNetwork) then
        wwriteln(outf,'*** Warning: simulated network was not connected (but it was requested to be so).');
      end;

    if halterr then error := true;
    success := true;
    feasible := true;
    if (not (gibbs in [2,3,8])) and estimation then  //@@@@
      for i:=1 to pp do ystat[i] := CurrentStats[i]
(* procedure was not (yet?) optimized for other gibbs values; see ergmrunstep *)
      else ErgmStatistics(ystat,symmetric,ordes,estimation,NewNetworkEffects);  // @@@ gaat dit goed voor andere gibbs?
    if (pp < pmax) then ystat[pp+1] := GibbsChangeFraction;

    if updatestat then
      begin
      for i:=1 to n do ordes[i] := outd_yn(i);
      quicksort(ordes,1,n);
      end;
    end
   else
    begin   (* observations >= 2 *)
    for i:=1 to pmax do statsum[i] := 0.0;
    feasible := true;
    activek := 0;
    for k:=1 to observations-1 do
      if not skipi[k] then
      begin
      inc(activek);
      initialise_actors(k);
      initialise_y(k);
        (* CS: following line added 25-apr-2005:
               for having correct rate parameters, some running statistics must be accurate,
               so after initialising data, also these need to be updated *)
      InitialiseAlpha(outf,theta,fseed);
(* Three next statements were added by TS 02-jan-05,
   in order to make a tighter use of common random numbers *)
      dummy := rando(version2);  (* only needed to make one step of the
                                    version2 random number generator *)
      sed := readstate2;
      makestate(sed);  (* use the version2 random number generator
           to reinitialise the version <=basic> random number generator;
           since the version2 random number generator IS USED FOR NOTHING ELSE,
           this implies for each observation period,
           the same fseed used as argument in SimStats
           implies the same starting value of the random number stream
           for this observation period. *)

      if CConditional then
        begin
        if (ConditioningNetworkVariable >= 1)
        then dt := ObservedNetworkDistance[1,k]
        else dt := ObservedActionDistance[ConditioningActionVariable,k];
        end
       else dt := 1.0; (* not Cconditional *)
        (*MS*)
      if (observations >= 3) and (homogeneous > 20) then
        begin
        if ((derivatives and (alternative <> 999999999)) or altern) then
          begin
          if (k = period) then
            begin
            original := theta[coordinate];
            theta[coordinate] := alternative;
            SetAlpha(theta); (*Set to be estimated elements in <alpa_f>, <alpa_g> etc. equal to the corresponding elements in <theta>.*)
            end;
          end;
        end;
        (*MS*)
      runepoch(dt,success,updatestat,error,errstr,halterr,outf,lf);
        (*MS*)

      if (observations >= 3) and (homogeneous > 20) then
        begin
        if ((derivatives and (alternative <> 999999999)) or altern) then
          begin
          if (k = period) then
            begin
            theta[coordinate] := original;  (*MS: "reset" parameter.*)
            SetAlpha(theta);
            end;
          end;
        end;
        (*MS*)
      if halterr then error := true;

      if ((Callingprogram = 5) and (nza >= 1)) then
(* Compute spatial autocorrelations *)
        begin
        if (k <= (pmax - pp) div (2*nza)) then for j:=1 to nza do
(* for greater k there is no space in the ystat array *)
          begin
          SpatialFunctions(StructActive,j,k+1,mi,gc); // changed 3.17u
          statsum[pp+(2*nza*(k-1))+(2*j)-1] := mi;
          statsum[pp+(2*nza*(k-1))+(2*j)] := gc;
          end;
        end;

      if exo then
        begin
        SetLeaversyBack(k);  //  put after SpatialFunctions, 3.17u
        (* MH: change simulated values leavers (all false) back to last
           observed ones (ynm[k]), as was done for observed data, in
           order not to influence estimation *)
        end;

  (* For network statistics: put ya equal to zc in start of period *)
      for j:=1 to nza do for i:=n downto 1 do
        begin
        yaa[i,j] := ya(i,j);  // store
        setya(i,j,round(zc[i,j,k]));
        if (ya(i,j) < round(zcmin[j])) then setya(i,j,round(zcmin[j]))
         else
          if (ya(i,j) > round(zcmax[j])) then setya(i,j,round(zcmax[j]));
        end;
      NetworkStatistics(statk,ordes,k,outf);

      if (nza >= 1) then
        begin
        for i:=n downto 1 do
          if IncludeActor[i] then   // condition added 3.17u
            for j:=nza downto 1 do setya(i,j,yaa[i,j]);
        SetDigraph(k);
        if exo then SetLeaverszBack(k);  // added 3.17u
        ActionStatistics(statk,ordes,k,outf);
        end;

      if not Cconditional then success := true;
      feasible := (feasible and success);
      for i:=1 to pp do statsum[i] := statsum[i] + statk[i];
        (*MS*)
      if ((observations >= 3) and (homogeneous > 20) and derivatives and (k = period))
        then statperiod := statk[coordinate];
      if (d_estAB and not(cconditional)) then CompleteScore; (*MS: unconditional MoM: complete score for the rate parameters.*)
        (*MS*)
      d[activek] := time;
      if d_estB then SumScore(pp,k,statk); (*TS: period-dependent analog of StoreScore *)
      end; (* for k *)
    for i:=1 to pmax do ystat[i] := statsum[i];
    if d_estA then StoreScore(pp,ystat,d_store); (*MS: ingredients to derivative estimation.*)
    if ((homogeneous > 20) and derivatives) then ystat[pp] := statperiod (*MS*)
    end; (* if observations >= 2 *)
  if updatestat then for i:=n downto 1 do Inc(SimDegrees[i,ordes[i]]);
(* This is outside the k-loop, so the ordered degrees are used only
   for the last time point.*)
  fseed := readstate;
  (* TS added 05-jan-05 *)
  if Cconditional then alpa_l := alpa1;
end; (* SimStats *)

procedure ModelSpecification.SimOnce(kk: integer;  extra: boolean;
                  p: Byte; theta: pVec;
                  allstats, varmat: boolean; var fseed: integer;
                  var ystat: pVec; var feasible, error: Boolean;
                  var errstr: shortstring; var outf,lf: TFileStream);
(* Utility for once simulating
   a data set with given parameter values.
   No other procedures utilizing the original data set should follow.
   It is copied with some changes from SimStats.
   The main difference is that at the start of
   observations >= 2, the data set is not reset
   to the observed data (which happens in initialise_y) .
   Changes in structurally determined values or in missing values
   are not taken into account.
   It also writes to files with project name sisim,
   for potential use in simulation studies.
   *)
var dt, statperiod: real;
    i,h: integer;
    success, halterr, writepf, DacExists: boolean;
    k, fin, repeats: integer;
    simdatf,f,datf,pf: textfile;
    simdatfn, ppname, fname, pfname: shortstring;
    alpa1: CRarray;
    statsum, statk: pvec;
    ordes: intvec;
    od: array of integer;
    x: string;
    deg,zdist: iarray;
    ef: textfile;
    z2: boarray;

label    repeatagain;

procedure writenetwork;
var i,j,sum: integer;
begin
  if (thisperiod < 9)
     then fname := ppname +'.d0'+st(thisperiod+1)
     else fname := ppname +'.d'+st(thisperiod+1);
  if FileExists(fname) then DacExists := true;
  writethesedata(fname);
  if (observations >= 2) then writeln(simdatf,'Observation ', thisperiod+1,'.');
  writesimulateddata(false,simdatf);
//  if (thisperiod >= 2) then writeln(outf,'Distance = ',...);  // if required, use Digraph.CountDifNoNo
  if writepf then
    begin
    if (observations >= 2) then
      writeln(pf,'*Network ',pname,' T',k+1);
    writeln(pf,'*Vertices ',n);
    for i:=1 to n do writeln(pf,i:3,' ',char(39),i,char(39));
    if symmetric then writeln(pf,'*Edges ') else writeln(pf,'*Arcs ');
//    if (observations >= 2) then writeln(pf,pname,' T',k+1);
    writeSimulatedData(symmetric,pf);
    if (observations >= 2) then writeln(pf);
    end;
end; (* writenetwork *)
begin (* SimOnce *)
(* writing to ppname is ad hoc, for batch simulation purposes *)
  Error := false; errstr := ''; halterr := false;
  repeats := 0;
  SetLength(ordes,n+1);
  DacExists := false;
  InitialiseAlpha(outf,theta,fseed);
  alpa1 := alpa_l;
(* Facility for writing to files in Siena format: taken from S_Dat.WritePData *)
  ppname := 'sisim'+inttostr(kk);
  if CConditional then begin
    for i:=1 to observations-1 do
     if (ConditioningNetworkvariable=1) then
      alpa_l[i] := 1.0
     else
      alpa_l[NetworkEffects_l+(ConditioningActionVariable-1)*ActionEffects_l+i] := 1.0;
     end; (* if Cconditional *)

repeatagain:

  simdatfn :=  pname+'.sd0';
  assign(simdatf,simdatfn);
  rewrite(simdatf);

(* prepare filename and header for pajek file *)
  if (observations <= 1) then pfname := pname+'_r1.net'
    else pfname := pname+'_r1.paj';
  fin := 1;
  repeat
   if FileExists(pfname) then
     begin
     inc(fin);
     if (observations <= 1) then pfname := pname+'_r'+char(fin+48)+'.net'
      else pfname := pname+'_r'+char(fin+48)+'.paj';
     end;
   until (not FileExists(pfname)) or (fin >= 9);
  if  FileExists(pfname) then
    begin
    x := 'No Pajek file is written, because the intended file names'+eol;
    outf.Write(pchar(x)^,length(x));
    if (observations <= 1) then begin
      x := pname+'_r1.net';
      outf.Write(pchar(x)^,length(x));
      end
      else begin
        x := pname+'_r1.paj';
        outf.Write(pchar(x)^,length(x));
     end;
    x := ' to '+pfname+' all are in use.'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'If you wish a Pajek output file, delete one of these.'+eol;
    outf.Write(pchar(x)^,length(x));
    writepf := false;
    end
   else
    begin
    x := format('%s%s%s',['Pajek output written to file ',fname,' .'])+eol;
    lf.Write(pchar(x)^,length(x));
    assignfile(pf,pfname);
    rewrite(pf);
//    if (observations >= 2) then writeln(pf,'*Network ',pname,' simulation');
    writepf := true;
    end;

  if checkstatistics then
    begin
    x :='Option checkstatistics: '+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'write intermediate results to output file at the end of each runstep:'+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'Distance; time; hstar; i; j; '+eol;
    outf.Write(pchar(x)^,length(x));
    for i:=1 to NumberFunctions do begin
      x := StatsDescriptions[i]+eol;
      outf.Write(pchar(x)^,length(x));
    end;
    //writeln(outf);
    outf.Write(pchar(eol)^,length(eol));
    end; (* if checkstatistics *)

  if (observations <= 1) then IndependentStart := false;
    (* The first matrix was already generated in FunctionEvaluations *)
  initialise_actors(1);
  initialise_y(1);
  k := 0;
  if (observations >= 2) then
    begin
    thisperiod := 0;
    writenetwork;
    end;
  for i:=1 to pmax do statsum[i] := 0.0;

(* Note that the networkstatistics and actionstatistics
   depend on the global means of covariates and they
   also can depend on the previous observation;
   therefore the data set must be changed to calculate these statistics
   and no other procedures utilizing the original data set should follow. *)

  case observations of
    1: begin
         thisperiod := 1;
         dt := 1.0;
         runepoch(dt,success,false,error,errstr,halterr,outf,lf);
         if halterr then error := true;
         writenetwork;
       end;
    else
      begin(* case observations >= 2 *)
      feasible := true;
      for k:=1 to observations-1 do
        begin
        thisperiod := k;
        time := 0.0;
        Distance := 0;
        for h:=1 to nzc do for i:=1 to n do z[i,nz+h] := zc[i,h,k];
        if CConditional then
          begin
          if (ConditioningNetworkVariable >= 1)
          then dt := ObservedNetworkDistance[ConditioningNetworkVariable,k]
           else dt := ObservedActionDistance[ConditioningActionVariable,k];
          end
         else dt := 1.0; (* not Cconditional *)
        runepoch(dt,success,false,error,errstr,halterr,outf,lf);
        writenetwork;
        MakeDigraph(k+1);
        for h:=1 to nza do for i:=1 to n do zc[i,h,k+1] := ya(i,h);
//        NetworkStatistics(statk,ordes,k,outf);
//        if (nza >= 1) then ActionStatistics(statk,ordes,k,outf);
//        for i:=1 to NumberFunctions do statsum[i] := statsum[i] + statk[i];
        if (not success) then writeln(simdatf,'period ',k,' (no success...)');
        end; (* for k *)
      end (* case-else, i.e. observations >= 3 *)
  end; (* case *)


  if (nza > 0) then
    begin
  (* covariate data *)
    fname := ppname + '.dac';
    if FileExists(fname) then DacExists := true;
    assignfile(datf,fname);
    rewrite(datf);

    for h:=1 to nza do
      begin
      writeln(simdatf,'Dependent behavior variable ',h,'.');
      for i:=1 to n do
        begin
        for k:=1 to observations do write(simdatf,' ',zc[i,h,k]:2:0);
        writeln(simdatf);
        end;
      writeln(simdatf);
      end;

    for i:=1 to n do
      begin
      for h:=1 to nzc do
        begin
        if (h <= nza) then fin := observations
         else fin := observations-1;
        for k:=1 to fin do write(datf,zc[i,h,k]:8:4,' ');
        end;
      writeln(datf);
      end; (* i *)
    closefile(datf);
    end; (* if nza > 0 *)

(* finalize *)
  closefile(simdatf);
  if Cconditional then alpa_l := alpa1;

(* output of statistics *)
  Calculations(false,outf); // to use the correct covariate means
  if (observations <= 1) then
     ErgmStatistics(statsum,symmetric,ordes,false,NewNetworkEffects)  // @@@ gaat dit goed voor andere gibbs?
    else FunctionEvaluations(ordes,halterr,outf,outf, statsum);

  finalize(zc);     // this is to preclude further use of the changed data set.

  Heading(2,outf,'One simulated run: statistics and output.');
  for i:=1 to NumberFunctions do begin
    x := format('%s%11.4f',[StatsDescriptions[i],statsum[i]])+eol;
  outf.Write(pchar(x)^,length(x));
  end;

  if extra then
    begin
    inc(repeats);
    DistancesFromHighDegree(deg,zdist);
    if (2*zdist[0] < n) and (repeats <= 10) then
      begin
      wwriteln(outf,'Component of highest degree node has size '+inttostr(zdist[0])+'.');
      wwriteln(outf,'(try number '+inttostr(repeats)+')');
      wwriteln(outf,'Start again.');
      goto repeatagain;
      end;

    SetLength(z2,n+1);
    for i:=1 to n do z2[i] := (zdist[i] < n);
(* For checking - information : *)
(*  for i:=1 to n do
    begin
    wwrite(outf,RightAlignString(st(i),3));
    end;
  wwriteln(outf,'');
  for i:=1 to n do
    begin
    wwrite(outf,RightAlignString(st(Outd_yn(i)),3));
    end;
  wwriteln(outf,'');
  for i:=1 to n do
    begin
    wwrite(outf,RightAlignString(st(deg[i]),3));
    end;
  wwriteln(outf,'');
  for i:=1 to n do
    begin
    wwrite(outf,RightAlignString(st(zdist[i]),3));
    end;
  wwriteln(outf,'');
*)
    x := format('%s%7d',['Component size                            ',
                          zdist[0]])+eol;
    outf.Write(pchar(x)^,length(x));

    writeln(extrafile,statsum[1]:3:0,' ',statsum[2]:3:0,' ',
                      statsum[3]:3:0,' ',zdist[0]:3);

    WriteErGMInfile(zdist[0],'sim'+st(kk),'netw'+st(kk)+'.net');
    wwriteln(outf,'');
    assign(ef,'netw'+st(kk)+'.net');
    rewrite(ef);
    writeln(ef,'*Vertices ',zdist[0]);
    for i:=1 to zdist[0] do writeln(ef,i:3,' ',char(39),i,char(39));
    writeln(ef,'*Arcs ');
    writeSubsetSimulatedData(deg,z2,ef);
    closefile(ef);
   
  end; (* if extra *)

  outf.Write(pchar(eol)^,length(eol));
  x := 'One simulated data set was written to file '+simdatfn;
  outf.Write(pchar(x)^,length(x));
  if writepf then
    begin
    x := ','+eol;
    outf.Write(pchar(x)^,length(x));
    x := 'and in Pajek format to file '+ pfname+'.'+eol;
    outf.Write(pchar(x)^,length(x));
    closefile(pf);
    end
   else  begin
    x := '.'+eol;
    outf.Write(pchar(x)^,length(x));
   end;
  x := 'The simulated data set was also written in internal SIENA format'+eol;
  outf.Write(pchar(x)^,length(x));
  x := 'to project files with the project name <<'+ppname+'>>';
  outf.Write(pchar(x)^,length(x));
  if DacExists then
    begin
    x := ','+eol;
    outf.Write(pchar(x)^,length(x));
    x :=' overwriting the existing files.'+eol;
    outf.Write(pchar(x)^,length(x));
    end
   else begin
    x := '.'+eol;
    outf.Write(pchar(x)^,length(x));
   end;
   if extra then
     begin
     wwriteln(outf,'The large connected component was written to '
                    +'netw'+st(kk)+'.net'+' with .in file sim'+st(kk)+'.in .');
     end;
  outf.Write(pchar(eol)^,length(eol));
end; (* SimOnce *)

function ModelSpecification.nu(const h: integer): real;
begin
  nu := store_nu[h+1];
end;

function ModelSpecification.xi(const h: integer): real;
begin
  xi := store_xi[h+1];
end;

procedure ModelSpecification.GiveDescriptions;
(* Defines array with labels for the included effects *)
var i,h,count: integer;
  modelnames: SSarray;
  Correspondence: SBYarray;
  VertPar: BBarray;
  param: SSIarray;
begin
  h := 0;
  (* initialise arrays *)
  for i:=1 to pmax
  do Description[i] := '';

  (* network rate function parameters *)
  DefineModel_lNames(modelnames,Correspondence,param);
  for i:=1 to NetworkEffects_l do if Effects_l[i] then
    begin
    inc(h);
    Description[h] := 'rate:  '+modelnames[i]
    end;
  (* network evaluation function parameters *)
  DefineModel_fNames(modelnames,Correspondence,Vertpar,param);
  for i:=1 to NetworkEffects_f do if Effects_f[i] then
    begin
    inc(h);
    if (observations >= 2) then Description[h] := 'eval:  '+modelnames[i]
     else Description[h] := modelnames[i];
    end;
  (* network endowment function parameters *)
  DefineModel_gNames(modelnames,Correspondence,param);
  for i:=1 to NetworkEffects_g do if Effects_g[i] then
    begin
    inc(h);
    if NewNetworkEffects.Endowed(i) then Description[h] := 'endow: '+modelnames[i]
     else Description[h] := 'not implemented';
    end;
  (* action rate function parameters *)
  DefineModel_lNames(modelnames,Correspondence,param);
  for i:=1 + NetworkEffects_l to MaxEffects_l do if Effects_l[i] then
    begin
    inc(h);
    Description[h] := 'rate:  '+modelnames[i]
    end;
  (* action evaluation function parameters *)
  DefineModel_fNames(modelnames,Correspondence,Vertpar,param);
  for i:=1 + NetworkEffects_f to MaxEffects_f do if Effects_f[i] then
    begin
    inc(h);
    Description[h] := 'eval:  '+modelnames[i]
    end;
  (* action endowment function parameters *)
  DefineModel_gNames(modelnames,Correspondence,param);
  for i:=1 + NetworkEffects_g to MaxEffects_g do if Effects_g[i] then
    begin
    inc(h);
    Description[h] := 'endow: '+modelnames[i];
    end;
  (* bring effect labels to same length: *)
  count := h;
  for i:=1 to count do Alignstring(Description[i],60);
end; (* GiveDescriptions *)


function va(b: boolean): integer;
begin
 if b then va := 1 else va := 0;
end;

procedure ModelSpecification.DescribeModeloptions(var outf, lf: TFileStream);
(* This procedure produces a description of the model options.
   It is not used for running any further calculation or program. *)
   var effile: TFileStream;
   x: string;
  procedure DescribeThem(var ouf: TFileStream; option: byte);
  var
    i,j,mt,n0,
    ModelMax,
    Type0,
    digit          : integer;
    modelnames     : SSarray;
    Correspondence : SBYarray;
    VertPar        : BBarray;
    param          : SSIArray;
    x : string;
  begin
  if (option = 1) then
    begin
    Heading(2,ouf,'Model descriptions.');
    x := 'Parameters and target statistics for the current project.'+eol;
    ouf.Write(pchar(x)^,length(x));
    //writeln(ouf);
    ouf.Write(pchar(eol)^,length(eol));
    end else
    begin
    if (observations >= 2) then begin
      x :='List of objective function effects for this data set.'+eol;
      ouf.Write(pchar(x)^,length(x));
      end
     else begin
     x := 'List of statistics available for this data set.'+eol ;
     ouf.Write(pchar(x)^,length(x));
     end;
    //writeln(ouf);
    ouf.Write(pchar(eol)^,length(eol));
    end;

    digit:=1; (* digit is used for nice layout in the output file *)
    if (MaxFunctions>=10) then inc(digit);
    if (MaxFunctions>=100) then inc(digit);

    if ((observations = 1) or signed) then ModelMax := 1 else ModelMax := 2;
    (* van hier tot hier: zolang modeltype 2 nog niet klaar in Siena 3:
    *)
    ModelMax := 1;
    Type0 := ModelType;
    for mt := 1 to ModelMax
    do begin
      ModelType := mt;
      if (ModelMax >= 2)
      then begin
        Heading(2,ouf,'Descriptions for Model Type = '+MakeString(ModelType)+':');
      end;

      if (observations >= 2) and (option = 1) then
        begin
        DefineModel_lNames(modelnames,Correspondence,Param);
        for i:=1 to MaxEffects_l
        do for j:= 48-length(Modelnames[i]) downto 1 do Modelnames[i] := Modelnames[i] + ' ';
        x := 'Parameters for defining the rate function'+eol;
        ouf.Write(pchar(x)^,length(x));
        x := '(with number of corresponding default statistic):'+eol;
        ouf.Write(pchar(x)^,length(x));
        for i:=1 to MaxEffects_l
        do begin
           x := format('%*d%s%48s%s%*d%s',[digit,i,'. ',Modelnames[i],'  (',
             digit,Correspondence[i],')'])+eol;
         ouf.Write(pchar(x)^,length(x));
        end;
        //writeln(ouf);
        ouf.Write(pchar(eol)^,length(eol));
        end;

      DefineModel_fNames(modelnames,Correspondence,VertPar,param);     // 123456
      for i:=1 to MaxEffects_f
      do for j:= 48-length(Modelnames[i]) downto 1
      do Modelnames[i] := Modelnames[i] + ' ';
      if (option = 1) then
        begin
        x := 'Parameters for defining the evaluation function'+eol;
        ouf.Write(pchar(x)^,length(x));
        x := '(with number of corresponding default statistic):'+eol;
        ouf.Write(pchar(x)^,length(x));
        end;
      for i:=1 to MaxEffects_f do
        begin
        if (i <= 9) then n0 := 2 else n0 := 1;
        if (i >= 100) then n0 := 0;
        if (option = 2) then for j := 1 to n0 do ouf.write(pchar(string('0'))^,1);
        if (option = 1) then begin
          x := format('%*d%s%48s%s%*d%s',[digit,i,'. ',Modelnames[i],'  (',  digit,
           Correspondence[i],')'])+eol;
           ouf.Write(pchar(x)^,length(x));
           end
         else begin
          x := format('%d%s%48s',[i,'. ',Modelnames[i]])+eol;
          ouf.Write(pchar(x)^,length(x));
         end;
        end;
      //writeln(ouf);
      ouf.Write(pchar(eol)^,length(eol));

      if (observations >= 2) and (option = 1) then
        begin
        DefineModel_gNames(modelnames,Correspondence,Param);
        for i:=1 to MaxEffects_g
        do for j:= 48-length(Modelnames[i]) downto 1
        do Modelnames[i] := Modelnames[i] + ' ';
        x := 'Parameters for defining the endowment function'+eol;
        ouf.Write(pchar(x)^,length(x));
        if (MaxEffects_g = 0) then begin
          x := '(none)'+eol;
          ouf.Write(pchar(x)^,length(x));
         end
          else begin
          x := '(with number of corresponding default statistic):'+eol;
           ouf.Write(pchar(x)^,length(x));
          end;
        for i:=1 to MaxEffects_g
        do begin
          x:= format('%*d%s%48s%s%*d%s',[digit,i,'. ',Modelnames[i],'  (',  digit,
            Correspondence[i],')'])+eol;
          ouf.Write(pchar(x)^,length(x));
         end;
        //writeln(ouf);
         ouf.Write(pchar(eol)^,length(eol));
        end;

    if (option = 1) then
      begin
      DefineFunctionNames(lf);
      x := 'Statistics:'+eol;
       ouf.Write(pchar(x)^,length(x));
      for i:=1 to MaxFunctions do begin
      x := format('%*d%s%s',[digit,i,'. ',FDescriptions[i]])+eol;
       ouf.Write(pchar(x)^,length(x));
      end;
      //writeln(ouf);
       ouf.Write(pchar(eol)^,length(eol));
      end;
    end; (* for mt, i.e., ModelType *)
    ModelType := Type0;
  end; (* DescribeThem *)
begin  (* DescribeModeloptions *)
  if ((observations = 1) or signed) then DescribeThem(outf,1) else begin
      x := 'Model descriptions are written to file '+pname+'.log.'+eol;
      outf.Write(pchar(x)^,length(x));
    //writeln(outf);
      outf.Write(pchar(eol)^,length(eol));
      DescribeThem(lf,1);
   end;

  //assign(effile,pname+'.eff');
  //rewrite(effile);
   if (not FileExists(pname+'.eff')) then begin
    effile := TFileStream.Create(pname+'.eff', fmCreate);
    effile.Free;
    effile := TFileStream.Create(pname+'.eff', fmOpenWrite or fmShareDenyWrite);
  end
  else begin
    effile := TFileStream.Create(pname+'.eff',fmOpenReadWrite or fmShareDenyWrite);
    effile.size := 0;
  end;
  DescribeThem(effile,2);
  //closefile(effile);
  effile.Free;
  x := 'A list of objective function effects is given in file '+pname+'.eff.'+eol;
  outf.Write(pchar(x)^,length(x));
  x := 'The numbers in this list can be used for specifying interaction effects (see the SIENA manual).'+eol;
  outf.Write(pchar(x)^,length(x));
  //writeln(outf);
  outf.Write(pchar(eol)^,length(eol));
end; (* DescribeModeloptions *)


procedure ModelSpecification.Halterrorr(s: string; var f, lf: TFileStream);
begin
  Halterror(s,f,lf);
  DeActivateModel(f);
end; (* Halterrorr *)


procedure ModelSpecification.DeActivateModel(var outf: TFileStream);
begin
  FinishDat(outf);
  closefile(extrafile);
end;


function FoundError: Boolean;
begin
  FoundError := MadeError;
end;


begin (* S_Base *)
  MadeError := false;
  checkstatistics := false;
  (* TS: checkstatistics is meant to be used
     only in the simulation (Siena05) case with niter = 2.  *)

   assign(extrafile,'extrafile.txt');
   rewrite(extrafile);
//   assign(exf,'extraf.txt');
//   rewrite(exf);
end.

(* formerly:
   xi, nu: array[-1..nmax] of real;
   xino, nuno: real;
   lambda2 : array of array[-1..nmax] of real;
   lambda4 : array[1..nmax] of real;
    active: array[1..nmax] of boolean;
    StructActive: array[1..nmax] of boolean;

    *)
